package com.tetaj.hvmhelper;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.LinkedList;

import com.jopdesign.wcet.uppaal.model.Location;
import com.jopdesign.wcet.uppaal.model.NTASystem;
import com.jopdesign.wcet.uppaal.model.Template;
import com.jopdesign.wcet.uppaal.model.Transition;
import com.tetaj.ircfg.ControlFlowGraph;
import com.tetaj.modelgenerator.uppaalmodelgeneratorircfg.FQNameHelper;

public class HvmUppaalHelper {

	public static void modifyInvokeTemplate(Template template, String[] nativeMethods, String methodIdentifier, HashMap<String, ControlFlowGraph> cfgs) throws Exception{
		
		LinkedList<Transition> copiedTransListQueue = new LinkedList<Transition>();
				
		for(Transition trans : template.getTransitions())
			copiedTransListQueue.add(trans);
		
		boolean foundEiCall =false;
		
		for(Transition trans : copiedTransListQueue){
			for(String update : trans.getAttrs().getUpdates()){
				if(update.toLowerCase().contains("eicall")){
					foundEiCall = true;
					Location extraLocation = new Location("nativeCall");
					extraLocation.setCommited();
					template.addLocation(extraLocation);
					
					Location oldTargetLocaction = trans.getTarget();
					
					trans.setTarget(extraLocation);
					
					for(String nativeMethod : nativeMethods){
						if(cfgs.containsKey(nativeMethod)){
							Location nativeMethodInvokeLocation = new Location("Native_" + nativeMethod);
							
							Transition nativeTransition = new Transition(extraLocation, nativeMethodInvokeLocation);
							nativeTransition.getAttrs().appendGuard("NATIVE_CALL=="+ nativeMethod);
							nativeTransition.getAttrs().setSync("invoke_" +methodIdentifier+"_"+ FQNameHelper.getUppaalFriendlyName(nativeMethod) + "!");
							
							Transition nativeWaitTransition = new Transition(nativeMethodInvokeLocation,oldTargetLocaction);
							nativeWaitTransition.getAttrs().appendUpdate("NATIVE_CALL=NO_NATIVE");
							nativeWaitTransition.getAttrs().setSync("return_" +methodIdentifier+"_"+ FQNameHelper.getUppaalFriendlyName(nativeMethod) + "?");
													
							template.addLocation(nativeMethodInvokeLocation);
							template.addTransition(nativeTransition);
							template.addTransition(nativeWaitTransition);	
						} else {
							
							System.out.println("Existing methods");
							for (String existingMethod : cfgs.keySet()) {
								System.out.println(existingMethod);
							}
							
							throw new Exception("Could not find CFG for native method " + nativeMethod);
						}
					}
					
					//prevents deadlocking
					Transition noNativeTrans = new Transition(extraLocation, oldTargetLocaction);
					noNativeTrans.getAttrs().appendGuard("NATIVE_CALL==NO_NATIVE");
					template.addTransition(noNativeTrans);
					
				}
			}
		}
		if(foundEiCall==false){
			throw new Exception("Did not find any EICALL");
		}
		
	}

	public static void appendNativeDeclarations(NTASystem uppaal, String[] nativeMethods) {
		
		int i=1;
		for(String nativeMethod : nativeMethods){
			uppaal.appendDeclaration("const int " + nativeMethod + " = " + i++ + ";");
		}
		uppaal.appendDeclaration("const int NO_NATIVE = 0;");
		uppaal.appendDeclaration("int[0,"+i+"] " + "NATIVE_CALL = NO_NATIVE;");
		
	}
	
}
