package com.tetaj.ircfg;

import java.util.ArrayList;
import java.util.HashMap;

public class BasicBlock implements Comparable<BasicBlock> {
	
	private int id;
	private ArrayList<Instruction> instructions;
	
	private ArrayList<Edge> incomingEdges;
	private ArrayList<Edge> outgoingEdges;
	
	private Edge fallthroughEdge;
	private ArrayList<Edge> branchEdges;
	
	private HashMap<Integer, ControlFlowGraph> methodInvokeCfg;
	private HashMap<String, Object> propertyMap;
	
	private String nativeCallName;
	
	public BasicBlock(int id) {
		this.id = id;
		
		this.instructions = new ArrayList<Instruction>();
		
		this.incomingEdges = new ArrayList<Edge>();
		this.outgoingEdges = new ArrayList<Edge>();
		this.branchEdges = new ArrayList<Edge>();
		this.fallthroughEdge = null;
		
		this.methodInvokeCfg = new HashMap<Integer, ControlFlowGraph>();
		
		this.propertyMap = new HashMap<String, Object>();
		
	}

	public int getId() {
		return this.id;
	}
	
	public void setId(int value) {
		this.id = value;
	}

	public void addInstruction(Instruction instruction) {
		instruction.setId(this.instructions.size());
		this.instructions.add(instruction);
	}

	public ArrayList<Instruction> getInstructions() {
		return instructions;
	}

	public ArrayList<Edge> getIncomingEdges() {
		return this.incomingEdges;
	}
		
	public ArrayList<Edge> getOutgoingEdges() {
		return this.outgoingEdges;
	}
	
	public Edge getFallthroughEdge() {
		return this.fallthroughEdge;
	}
	
	public ArrayList<Edge> getBranchEdges() {
		return this.branchEdges;
	}
	
	public void addBranch(BasicBlock target) {
		Edge edge = new Edge(this, target, Edge.Type.Branch);
		
		this.branchEdges.add(edge);
		this.outgoingEdges.add(edge);
		target.addIncomingEdge(edge);
	}

	public void setFallthrough(BasicBlock target) throws ControlFlowGraphException {
		if (this.fallthroughEdge != null) {
			throw new ControlFlowGraphException("Cannot overwrite existing registered fallthrough basic block");
		}
		
		Edge edge = new Edge(this, target, Edge.Type.Fallthrough);
		
		this.fallthroughEdge = edge;
		this.outgoingEdges.add(edge);
		target.addIncomingEdge(edge);
	}
	
	public void removeEdge(Edge edge) {	
		if (edge.getType() == Edge.Type.Branch) {
			this.branchEdges.remove(edge);
		} else {
			this.fallthroughEdge = null;
		}
		
		this.outgoingEdges.remove(edge);
		edge.getTarget().removeIncomingEdge(edge);
	}
	
	public boolean hasBranchEdge() {
		return (this.branchEdges.size() > 0);
	}
	
	public boolean hasFallthroughEdge(){
		return (this.fallthroughEdge != null);
	}
	
	public void addMethodInvokeCfg(ControlFlowGraph methodInvokeCfg, int instrId) {
		this.methodInvokeCfg.put(instrId,methodInvokeCfg);
	}

	public HashMap<Integer, ControlFlowGraph> getMethodInvokeCfg() {
		return methodInvokeCfg;
	}
	
	public boolean isMethodInvoke() {
		return (this.methodInvokeCfg.size() != 0);
	}
	
	public void setProperty(String key, Object value) {
		this.propertyMap.put(key, value);
	}
	
	public Object getProperty(String key) {
		return this.propertyMap.get(key);
	}
	
	public boolean hasProperty(String key) {
		return this.propertyMap.containsKey(key);
	}

	@Override
	public int compareTo(BasicBlock bb) {
		int bb1Id = this.getId();        
        int bb2Id = ((BasicBlock)bb).getId();
       
        if(bb1Id > bb2Id)
            return 1;
        else if(bb1Id < bb2Id)
            return -1;
        else
            return 0;    
	}
	
	private void addIncomingEdge(Edge edge) {
		this.incomingEdges.add(edge);
	}

	private void removeIncomingEdge(Edge edge) {
		this.incomingEdges.remove(edge);
	}
	
	public void removeInstruction(Instruction instr) {
		this.instructions.remove(instr);
	}

	public void setNativeCallName(String nativeCallName) {
		this.nativeCallName = nativeCallName;
	}

	public String getNativeCallName() {
		return nativeCallName;
	}
}
