package com.tetaj.ircfg;

import java.util.HashMap;
import java.util.ArrayList;

public class Edge {

	public enum Type {
		Fallthrough, Branch,
	}
	
	private BasicBlock source;
	private BasicBlock target;
	private Type type;
	
	private LoopInfo loopIterationInfo;
	private LoopInfo loopEntryInfo;
	private ArrayList<LoopInfo> loopExitInfo;
	
	private HashMap<String, Object> propertyMap;
	
	public Edge(BasicBlock source, BasicBlock target, Type type) {
		
		this.source = source;
		this.target = target;
		this.type = type;
		
		this.loopIterationInfo = null;
		this.loopExitInfo = new ArrayList<LoopInfo>();
		
		this.propertyMap = new HashMap<String, Object>();
		
	}
	
	public BasicBlock getSource() {
		return this.source;
	}
	
	public BasicBlock getTarget() {
		return this.target;
	}
	
	public Type getType() {
		return this.type;
	}
	
	public void setAsLoopIteration(LoopInfo loopInfo) throws ControlFlowGraphException {
		if (this.loopIterationInfo != null) {
			throw new ControlFlowGraphException("This edge is already set as loop iteration - edge from basic block " + this.source.getId() + " to " + this.target.getId() + " overwriting loop " + this.loopIterationInfo.getId() + " with " + loopInfo.getId());
		}
		
		this.loopIterationInfo = loopInfo;
	}
	
	public LoopInfo getLoopIterationInfo() {
		return this.loopIterationInfo;
	}
	
	public boolean isLoopIteration() {
		return (this.loopIterationInfo != null);
	}
	
	public void setAsLoopEntry(LoopInfo loopInfo) throws ControlFlowGraphException {
		if (this.loopEntryInfo != null) {
			throw new ControlFlowGraphException("This edge is already set as loop entry - edge from basic block " + this.source.getId() + " to " + this.target.getId() + " overwriting loop " + this.loopIterationInfo.getId() + " with " + loopInfo.getId());
		}
		
		this.loopEntryInfo = loopInfo;
	}
	
	public LoopInfo getLoopEntryInfo() {
		return this.loopEntryInfo;
	}
	
	public boolean isLoopEntry() {
		return (this.loopEntryInfo != null);
	}
	
	public void setAsLoopExit(LoopInfo loopInfo) {
		this.loopExitInfo.add(loopInfo);
	}
	
	public ArrayList<LoopInfo> getLoopExitInfo() {
		return this.loopExitInfo;
	}
	
	public boolean isLoopExit() {
		return (this.loopExitInfo.size() > 0);
	}
	
	public void setProperty(String key, Object value) {
		this.propertyMap.put(key, value);
	}
	
	public Object getProperty(String key) {
		return this.propertyMap.get(key);
	}
	
	public boolean hasProperty(String key) {
		return this.propertyMap.containsKey(key);
	}
}
