package com.tetaj.modelgenerator;

import java.util.HashMap;
import java.util.LinkedList;

import javax.xml.transform.Result;
import javax.xml.transform.Source;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamResult;

import jintgen.gen.PrettyPrinter;

import com.tetaj.cfggenerators.AVRToIRCFG;
import com.tetaj.cfggenerators.JBCToIRCFG;
import com.tetaj.cfggenerators.constants.Constants;

import com.tetaj.hvmhelper.HvmCfgHelper;
import com.tetaj.hvmhelper.HvmUppaalHelper;

import com.tetaj.ircfg.BasicBlock;
import com.tetaj.ircfg.Edge;
import com.tetaj.ircfg.Instruction;

import com.tetaj.ircfg.ControlFlowGraph;
import com.tetaj.prettyprinter.AppConfig;
import com.tetaj.prettyprinter.IRCFGPrettyPrinter;
import com.tetaj.modelgenerator.cfganalysis.CfgAnalysis;
import com.tetaj.modelgenerator.staticmodels.StaticModels;
import com.tetaj.modelgenerator.uppaalmodelgeneratorircfg.UppaalModelGenerator;

import com.jopdesign.wcet.uppaal.model.NTASystem;
import com.jopdesign.wcet.uppaal.model.Template;

public class HvmBuilder {

	/**
	 * [program] classPath sourcePath fqMethodName [targetName]
	 * 
	 * @param args
	 */
	public static void main(String[] args) {

		if (args.length < 2) {
			System.out.println("Usage: java -jar JvmModelGenerator.jar <elfPath> <sourcePath> [targetName]");
			System.exit(1);
		}
		
		String targetFileName = "model.xml";
		
		String sourcePath = args[1];
		AppConfig.appendSourcePaths(sourcePath);
		targetFileName = (args.length > 2) ? args[2] : targetFileName;
		
		HashMap<String, ControlFlowGraph> cfgtable = null;
		
		try {
			String elfPath = args[0];
			String fqMethodName = "methodInterpreter";
			
			cfgtable = HvmCfgHelper.generateCfgsForHvm(fqMethodName, elfPath, Constants.HVM_JBC_OPCODE_MNEMONICS, "handle");
			
			
		//	IRCFGPrettyPrinter printer = new IRCFGPrettyPrinter("/home/luckow/universitetet/phd/papers/JTRES2012/experiments/models/optimizedHVM_prettyprint/");
		//	printer.createVisualRepresentation(cfgtable);
			
			
			//cfgtable = HvmCfgHelper.generateCfgsForHvm(fqMethodName, elfPath, Constants.HVM_JBC_OPCODE_MNEMONICS_TETASARTS, "handle");
			//Handle Native Methods
			HvmCfgHelper.generateNativeCfgs(elfPath, Constants.NATIVE_METHODS, cfgtable);
			
			cfgtable.remove(fqMethodName);
			
			// HACK!
			cfgtable.remove("hack");
			cfgtable.remove("hack0");
			cfgtable.remove("hack1");
			
			CfgAnalysis analysis = new CfgAnalysis(cfgtable);
			analysis.conductCfgAnalyses();
			
			//IRCFGPrettyPrinter pp = new IRCFGPrettyPrinter("/tmp/cfgs/");
			//pp.createVisualRepresentation(cfgtable);
			
			HvmCfgHelper.handleInterpretationLoopCfg(cfgtable);
			
			UppaalModelGenerator modelgenerator;
			NTASystem uppaal;
			
			modelgenerator = new UppaalModelGenerator("methodInterpreter", cfgtable, "assembly", Constants.AVR_OPCODE_MNEMONICS);				
			uppaal = modelgenerator.getUppaalSystem();
			Template metamocInterface = StaticModels.getMetamocInterfaceModel("assembly");
			uppaal.addTemplate(metamocInterface);
			uppaal.setSystem(uppaal.getSystem().substring(0, uppaal.getSystem().length()-1)+ ",\n\t" +metamocInterface.getId() +";");
							
			//Handle Native calls
			HvmUppaalHelper.modifyInvokeTemplate(uppaal.getTemplate("handleINVOKEVIRTUAL_OPCODE_T"), Constants.NATIVE_METHODS, "assembly", cfgtable);
			HvmUppaalHelper.modifyInvokeTemplate(uppaal.getTemplate("handleINVOKESPECIAL_OPCODE_T"), Constants.NATIVE_METHODS, "assembly", cfgtable);
			HvmUppaalHelper.modifyInvokeTemplate(uppaal.getTemplate("handleINVOKESTATIC_OPCODE_T"), Constants.NATIVE_METHODS, "assembly", cfgtable);
			
			HvmUppaalHelper.appendNativeDeclarations(uppaal, Constants.NATIVE_METHODS);				
			
			Source xml = new DOMSource(uppaal.toXML());
			
			TransformerFactory tranFactory = TransformerFactory.newInstance();
		    Transformer transformer = tranFactory.newTransformer();
		  
		    Result targetFile = new StreamResult(targetFileName);
		    transformer.transform(xml, targetFile);
			
		    // statistics
		    HashMap<Integer, Integer> edgeStatistics = new HashMap<Integer, Integer>();
		    LinkedList<ControlFlowGraph> visited = new LinkedList<ControlFlowGraph>();
		    LinkedList<ControlFlowGraph> queue = new LinkedList<ControlFlowGraph>();
		    
		    for (ControlFlowGraph cfg : cfgtable.values()) {
		    	queue.push(cfg);
		    }

		    while (queue.size() != 0) {
		    	ControlFlowGraph cfg = queue.pop();
		    	
		    	for (BasicBlock basicBlock : cfg.getBasicBlocks()) {
		    		int edgeCount = basicBlock.getOutgoingEdges().size();
		    		if (edgeStatistics.containsKey(edgeCount)) {
		    			edgeStatistics.put(edgeCount, edgeStatistics.get(edgeCount) + 1);
		    		} else {
		    			edgeStatistics.put(edgeCount, 1);
		    		}
		    		
		    		if (edgeCount > 1) {
		    			System.out.println("Edge count " + edgeCount + " for " + cfg.getMethodName() + " at line " + basicBlock.getInstructions().get(basicBlock.getInstructions().size()-1).getLinenumber());
		    			for (Edge edge : basicBlock.getOutgoingEdges()) {
		    				System.out.println("Target > line " + edge.getTarget().getInstructions().get(0).getLinenumber());
		    			}
		    		}
		    		
		    		if (basicBlock.isMethodInvoke()) {
		    			for(ControlFlowGraph invoked : basicBlock.getMethodInvokeCfg().values()){
		    			
			    			if (!visited.contains(cfg)) {
			    				queue.push(invoked);
			    				visited.push(invoked);
			    			}
		    			}
		    		}
		    	}
		    }
		    
		    System.out.println("Statistics");
		    System.out.println("================");
		    System.out.println("Branches : Count");
		    
		    for (Integer edgeCount : edgeStatistics.keySet()) {
		    	System.out.println(edgeCount.toString() + " : " + edgeStatistics.get(edgeCount));
		    }
		    
		    
		} catch (Exception e) {
			System.out.println("An error occured when generating the Model.\n");
			e.printStackTrace();
			System.exit(1);
		}
		
		System.out.println("done");
	}

}
