package com.tetaj.modelgenerator.cfganalysis;

import java.util.HashMap;
import java.util.LinkedList;

import com.tetaj.cfggenerators.constants.Constants;
import com.tetaj.ircfg.BasicBlock;
import com.tetaj.ircfg.ControlFlowGraph;
import com.tetaj.ircfg.Edge;

public class CollapseIfConditionAnalysis implements IAnalysis{

	private HashMap<String, ControlFlowGraph> CFGtable;

	public CollapseIfConditionAnalysis(HashMap<String, com.tetaj.ircfg.ControlFlowGraph> CFGTable) {
		this.CFGtable = CFGTable;
	}
	
	private BasicBlock getRealTarget(Edge edge, ControlFlowGraph cfg) {
		// A bit of a hack
		// Some if statements are build such that one branch goes to the body of the if, and one
		// goes to a basic block with a single rjump pointing to the basic block connecting the
		// two branches of the condition. This method ignores the stated target and points directly 
		// to the "real" target. 
		
		// Please note that this i potentially unsafe in some cases. Thus a warning is thrown. Please
		// ensure that this is always applied in a safe manner
		
		BasicBlock basicBlock = edge.getTarget();
		
		if (basicBlock.getOutgoingEdges().size() == 1 &&
			basicBlock.getInstructions().size() == 1) {

			String opcode = basicBlock.getInstructions().get(0).getMnemonic();	

			if (opcode.equals("rjmp")) {
				System.out.println("Warning: if statement found with rjmp, please ensure a safe if condition analysis at line " + basicBlock.getInstructions().get(0).getLinenumber() + " in file " + cfg.getSourceFilePath() + " in method " + cfg.getMethodName());
				basicBlock = basicBlock.getOutgoingEdges().get(0).getTarget();
			}
		}
		
		return basicBlock;
	}
	
	@Override
	public void conductAnalysis() throws CfgAnalysisException {
		
		for (ControlFlowGraph cfg : CFGtable.values()) {
			
			for (BasicBlock basicBlock : cfg.getBasicBlocks()) {
				
				if (basicBlock.getOutgoingEdges().size() == 2) {
					BasicBlock basicBlock0 = this.getRealTarget(basicBlock.getOutgoingEdges().get(0), cfg);
					BasicBlock basicBlock1 = this.getRealTarget(basicBlock.getOutgoingEdges().get(1), cfg);
					
					Edge candidateEdge;
					Edge searchFromEdge;
					BasicBlock joinBasicBlock;
					
					if ((Integer)basicBlock0.getProperty("dfstvalue") < (Integer)basicBlock1.getProperty("dfstvalue")) {
						candidateEdge = basicBlock.getOutgoingEdges().get(1);
						searchFromEdge = basicBlock.getOutgoingEdges().get(0);
						joinBasicBlock = basicBlock1;
					} else {
						candidateEdge = basicBlock.getOutgoingEdges().get(0);
						searchFromEdge = basicBlock.getOutgoingEdges().get(1);
						joinBasicBlock = basicBlock0;
					}
					
					if (!candidateEdge.hasProperty("backedge") && !searchFromEdge.hasProperty("backedge")) {
						LinkedList<BasicBlock> queue = new LinkedList<BasicBlock>();
						queue.push(searchFromEdge.getTarget());
						
						BasicBlock tempBasicBlock;
						while (queue.size() != 0) {
							tempBasicBlock = queue.pop();
							
							if (tempBasicBlock == joinBasicBlock){
								
								candidateEdge.getSource().removeEdge(candidateEdge);
								break;
							} else {
								for (Edge outEdge : tempBasicBlock.getOutgoingEdges()) {
									if (!outEdge.hasProperty("backedge") && !(outEdge.isLoopIteration() && outEdge.getLoopIterationInfo().getLoopBound()<=0)) {
										queue.push(outEdge.getTarget());
									}
								}
									
							}
						}
					}
					
				}
			}
		}
		
		
		
	}

}
