package com.tetaj.modelgenerator.uppaalmodelgeneratorircfg;

import java.util.ArrayList;

import com.jopdesign.wcet.uppaal.model.Location;
import com.jopdesign.wcet.uppaal.model.Transition;
import com.tetaj.ircfg.BasicBlock;

public class BasicBlockModel {

	private Location ingoingLocation;
	private Location outgoingLocation;
	private ArrayList<Location> locations;
	private ArrayList<Transition> transitions;
	private String uniqueIdentifier;
	private String nativeCallName;
	private boolean nativeCall;
	
	public BasicBlockModel(com.tetaj.ircfg.BasicBlock basicBlock, String uniqueIdentifier) {
		this.locations = new ArrayList<Location>();
		this.transitions = new ArrayList<Transition>();
		this.uniqueIdentifier = uniqueIdentifier;
		
		ingoingLocation = new Location("BasicBlockBegin" + "_ID" + basicBlock.getId());
		this.locations.add(ingoingLocation);
		
		Location currentLocation = ingoingLocation;		
		
		int opcode = 0;
		for(int i= 0; i<basicBlock.getInstructions().size();i++){
			
			opcode = basicBlock.getInstructions().get(i).getOpcode();
			if(opcode>=0){
					Location bytecodeExecutingLocation = new Location(this.uniqueIdentifier.toUpperCase()+"_" + basicBlock.getInstructions().get(i).getMnemonic() +"_ID" + basicBlock.getId() + "_" +i);
					Transition bytecodeExecutingTransition = new Transition(currentLocation, bytecodeExecutingLocation);
					bytecodeExecutingTransition.getAttrs().setSync(this.uniqueIdentifier + "_execute!");
					bytecodeExecutingTransition.getAttrs().appendUpdate(this.uniqueIdentifier + "_instruction = " + this.uniqueIdentifier.toUpperCase() +"_"+ basicBlock.getInstructions().get(i).getMnemonic().toUpperCase());
					this.locations.add(bytecodeExecutingLocation);
					this.transitions.add(bytecodeExecutingTransition);
					
					currentLocation = bytecodeExecutingLocation;

					if(basicBlock.getInstructions().get(i).isMethodinvoke() && !basicBlock.getInstructions().get(i).isNativecall()){
						currentLocation = generateMethodInvokeLocations(basicBlock, currentLocation, i);		
					}else if(basicBlock.getInstructions().get(i).isNativecall()){
						this.nativeCall = true;
						this.nativeCallName = FQNameHelper.getUppaalFriendlyName(basicBlock.getNativeCallName());
						bytecodeExecutingTransition.getAttrs().appendUpdate("NATIVE_CALL = " + FQNameHelper.getUppaalFriendlyName(basicBlock.getNativeCallName().toLowerCase()));
					}
					
			}
		}
		
		currentLocation.setCommited();
		outgoingLocation = new Location("BasicBlockEnd" + "_ID" + basicBlock.getId());
		
		//outgoingLocation.setUrgent();
		outgoingLocation.setCommited();
		this.locations.add(outgoingLocation);
		
		Transition outgoingTransition = new Transition(currentLocation, outgoingLocation);
		this.transitions.add(outgoingTransition);		
	}
	
	private Location generateMethodInvokeLocations(com.tetaj.ircfg.BasicBlock basicBlock, Location currentLocation, int instructionId) {
			
		Location methodInvokeLocation = new Location("Wait_for_method_invoke_ID" + basicBlock.getId() + "_" + instructionId);
		Transition methodInvokeTransition = new Transition(currentLocation, methodInvokeLocation);
		methodInvokeTransition.getAttrs().setSync(FQNameHelper.getUppaalFriendlyName("invoke_"  + this.uniqueIdentifier + "_"+ basicBlock.getMethodInvokeCfg().get(instructionId).getMethodName()) + "!");
		this.locations.add(methodInvokeLocation);
		this.transitions.add(methodInvokeTransition);
		
		Location methodReturnLocation = new Location("Method_returned_ID" + basicBlock.getId() + "_" + instructionId);
		Transition methodReturnTransition = new Transition(methodInvokeLocation, methodReturnLocation);
		methodReturnTransition.getAttrs().setSync(FQNameHelper.getUppaalFriendlyName("return_"  + this.uniqueIdentifier + "_"+ basicBlock.getMethodInvokeCfg().get(instructionId).getMethodName()) + "?");
		this.locations.add(methodReturnLocation);
		this.transitions.add(methodReturnTransition);
		
		return methodReturnLocation;
	}


	public Location getIngoingLocation() {
		return ingoingLocation;
	}

	public Location getOutgoingLocation() {
		return outgoingLocation;
	}

	public ArrayList<Location> getLocations() {
		return locations;
	}

	public ArrayList<Transition> getTransitions() {
		return transitions;
	}

	public void setNativeCallName(String nativeCallName) {
		this.nativeCallName = nativeCallName;
	}

	public String getNativeCallName() {
		return nativeCallName;
	}

	public void setNativeCall(boolean nativeCall) {
		this.nativeCall = nativeCall;
	}

	public boolean isNativeCall() {
		return nativeCall;
	}

}
