package com.tetaj.modelgenerator.annotationextractor.extractor;

import java.io.IOException;
import java.util.ArrayList;
import java.util.LinkedList;

import org.junit.Before;
import org.junit.Test;

import com.tetaj.prettyprinter.AppConfig;
import com.tetaj.modelgenerator.annotationextractor.parser.Annotation;
import com.tetaj.modelgenerator.annotationextractor.parser.Annotation.AnnotationType;

import static org.junit.Assert.*;

public class JavaAnnotationExtractorTest {

	CStyleAnnotationExtractor annoExtractor;
	
	@Before
	public void setUp() {
		this.annoExtractor = new CStyleAnnotationExtractor();
		ArrayList<String> sourcePaths = AppConfig.getSourcePaths();
		sourcePaths.add("./resources/testappsrc/");
	}
	
	@Test
	public void testSimpleAnnoExtractor() throws IOException {
		LinkedList<Annotation> annoList = annoExtractor.retrieveAnnotations("org/testapp/AnnotationTestClass.java", 10);
		
		assertNotNull(annoList);
		assertTrue(annoList.size() == 1);

		Annotation theAnno = annoList.pop();
		
		assertEquals(AnnotationType.LOOPBOUND, theAnno.getAnnotationType());
		assertEquals("5", theAnno.getAnnotationValue());
	}
	
	@Test
	public void testDoubleAnnoExtractor() throws IOException {
		LinkedList<Annotation> annoList = annoExtractor.retrieveAnnotations("org/testapp/AnnotationTestClass.java", 16);
		
		assertNotNull(annoList);
		assertTrue(annoList.size() == 2);

		Annotation theAnno = annoList.pop();
		
		assertEquals(AnnotationType.LOOPBOUND, theAnno.getAnnotationType());
		assertEquals("66", theAnno.getAnnotationValue());
		
		theAnno = annoList.pop();
		
		assertEquals(AnnotationType.LOOPBOUND, theAnno.getAnnotationType());
		assertEquals("66", theAnno.getAnnotationValue());
	}
	
	@Test
	public void testSingleAndNoAnnoExtractor() throws IOException {
		LinkedList<Annotation> annoList = annoExtractor.retrieveAnnotations("org/testapp/AnnotationTestClass.java", 23);
		
		assertNotNull(annoList);
		assertTrue(annoList.size() == 1);

		Annotation theAnno = annoList.pop();
		
		assertEquals(AnnotationType.LOOPBOUND, theAnno.getAnnotationType());
		assertEquals("7", theAnno.getAnnotationValue());
	}
	
}
