package com.tetaj.modelgenerator.integration;

import static org.junit.Assert.*;

import java.io.IOException;

import javax.xml.transform.Result;
import javax.xml.transform.Source;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerException;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamResult;

import org.jdom.JDOMException;
import org.junit.Test;

import com.jopdesign.wcet.uppaal.model.DuplicateKeyException;
import com.jopdesign.wcet.uppaal.model.NTASystem;
import com.jopdesign.wcet.uppaal.model.XmlSerializationException;
import com.tetaj.cfggenerators.constants.Constants;
import com.tetaj.modelgenerator.HvmBuilder;
import com.tetaj.modelgenerator.Main;
import com.tetaj.modelgenerator.staticmodels.StaticModels;
import com.tetaj.uppaallayouter.UppaalLayouter;

public class JVMModels {

	@Test
	public void testHVMModel() {
		
		String[] args = {
				"../helloWCET-C-AVR/Debug/helloWCET-C-AVR.elf",
				"../helloWCET-C-AVR/src/",
				"/tmp/jvmmodel.xml",
		};
		
		HvmBuilder.main(args);
		
		assertEquals(true, true);
	}
	
	@Test
	public void testJBCModel() {
		
		String[] args = {
				"../helloWCET/bin/",
				"minepump.PeriodicMinepumpTaskHandler#waterLevelDetection()V",
				"../helloWCET/src/",
				"/tmp/jbcmodel.xml",
		};
		
		Main.main(args);
		
		assertEquals(true, true);
	}
	
	@Test
	public void testHardwareModel() throws DuplicateKeyException, XmlSerializationException, IOException, JDOMException, InterruptedException, TransformerException {
		
		NTASystem system = StaticModels.getMetamocAVRModel("assembly");
		
		org.w3c.dom.Document doc = system.toXML();
		
		UppaalLayouter layouter = new UppaalLayouter(doc);
		doc = layouter.getLayoutedXmlW3Document();
		
		TransformerFactory tranFactory = TransformerFactory.newInstance();
		Transformer aTransformer = tranFactory.newTransformer();
		
		Source src = new DOMSource(doc);
		System.out.println("Starting the XML file creation");
		  
		Result dest = new StreamResult("/tmp/hardwaremodel.xml");
		aTransformer.transform(src, dest);
		
		assertEquals(true, true);
	}
	
	@Test
	public void testStaticJVMModel() throws DuplicateKeyException, XmlSerializationException, IOException, JDOMException, InterruptedException, TransformerException {
		
		NTASystem system = StaticModels.getStaticJvmSystem(Constants.HVM_JBC_OPCODE_MNEMONICS, "handle", "jvm", "assembly");
		
		org.w3c.dom.Document doc = system.toXML();
		
		UppaalLayouter layouter = new UppaalLayouter(doc);
		doc = layouter.getLayoutedXmlW3Document();
		
		TransformerFactory tranFactory = TransformerFactory.newInstance();
		Transformer aTransformer = tranFactory.newTransformer();
		
		Source src = new DOMSource(doc);
		System.out.println("Starting the XML file creation");
		  
		Result dest = new StreamResult("/tmp/staticjvmmodel.xml");
		aTransformer.transform(src, dest);
		
		assertEquals(true, true);
	}
	
	@Test
	public void testStaticJVMModel_Tetasarts() throws DuplicateKeyException, XmlSerializationException, IOException, JDOMException, InterruptedException, TransformerException {
		
		NTASystem system = StaticModels.getStaticJvmSystem(Constants.HVM_JBC_OPCODE_MNEMONICS_TETASARTS, "handle", "jvm", "assembly");
		
		org.w3c.dom.Document doc = system.toXML();
		
		UppaalLayouter layouter = new UppaalLayouter(doc);
		doc = layouter.getLayoutedXmlW3Document();
		
		TransformerFactory tranFactory = TransformerFactory.newInstance();
		Transformer aTransformer = tranFactory.newTransformer();
		
		Source src = new DOMSource(doc);
		System.out.println("Starting the XML file creation");
		  
		Result dest = new StreamResult("/tmp/staticjvmmodel.xml");
		aTransformer.transform(src, dest);
		
		assertEquals(true, true);
	}
	
	@Test
	public void testCombineModels() {
		
		String[] args = {
				"/tmp/jvmmodel.xml",
				"/tmp/jbcmodel.xml",
				"/tmp/staticjvmmodel.xml",
				"/tmp/hardwaremodel.xml",				
				"-o",
				"/tmp/combined.xml",
		};
		
		com.tetaj.modelcombiner.Main.main(args);
		
		assertEquals(true, true);
	}
	
}
