package com.tetaj.modelgenerator.uppaalmodelgeneratorircfg;

import java.util.ArrayList;
import java.util.Arrays;

import com.tetaj.cfggenerators.constants.Constants;

import com.jopdesign.wcet.uppaal.model.Location;
import com.jopdesign.wcet.uppaal.model.Template;
import com.jopdesign.wcet.uppaal.model.Transition;

public class DummyUppaalTemplates {
	public static Template getDummyJvmUppaalTemplate(String uniqueIdentifier){
		
		Template template = new Template("DummyVMMetamocStyle", Arrays.asList(""));
		
		Location initLoc = new Location("Idle");
		template.setInitialLocation(initLoc);
		
		template.appendDeclaration("clock time;");
		
		Location executionLoc = new Location("AnalyseAndDecodeJBC");
		executionLoc.setCommited();
		Transition initToExecutionTrans = new Transition(initLoc, executionLoc);
		initToExecutionTrans.getAttrs().setSync(uniqueIdentifier + "_execute?");
		
		template.addLocation(executionLoc);
		template.addTransition(initToExecutionTrans);
		
		
		//Adding all bytecode-JVM channels
		// Exclude the last 51 not implemented bytecodes and 2 implementation specific bytecodes
		// Exclude opcode 187 which is not implemented
		for (int opcode=0; opcode<256-53; opcode++){
			if (opcode != 186) {
				Location bytecodeLoc = new Location("bytecode_" + Constants.JBC_OPCODE_MNEMONICS[opcode]);
				bytecodeLoc.setInvariant("time<=1");
				Transition bytecodeTrans = new Transition(executionLoc, bytecodeLoc);
				bytecodeTrans.getAttrs().appendGuard(uniqueIdentifier + "_instruction==" + uniqueIdentifier.toUpperCase() +"_"+ Constants.JBC_OPCODE_MNEMONICS[opcode].toUpperCase());
				bytecodeTrans.getAttrs().appendUpdate("time=0");
				
				Transition bytecodeReturnTrans = new Transition(bytecodeLoc, initLoc);
				bytecodeReturnTrans.getAttrs().appendGuard("time==1");
				
				template.addLocation(bytecodeLoc);
				template.addTransition(bytecodeTrans);
				template.addTransition(bytecodeReturnTrans);
			}
		}
		
		return template;
		
	}
	public static ArrayList<Template> getDummyHardwareTemplate(String[] instructionNames, String methodId, String hwUniqueIdentifier){
		ArrayList<Template> templates = new ArrayList<Template>();
		
		for(String instruction : instructionNames){
			if(!instruction.equals(Constants.ILLEGAL_OPCODE)){
				Template instructionTemplate = new Template(methodId + instruction, Arrays.asList(""));
				
				Location idle = new Location("Idle");
				instructionTemplate.setInitialLocation(idle);
				Location exec = new Location("Execute");
				Transition exectrans = new Transition(idle,exec);
				exectrans.getAttrs().setSync("invoke_instruction_" + methodId + instruction.toUpperCase() + "?");
				instructionTemplate.addLocation(exec);
				instructionTemplate.addTransition(exectrans);
				
				Location hwInstruction1 = new Location(Constants.AVR_OPCODE_MNEMONICS[35]);
				Transition hwTrans1 = new Transition(exec, hwInstruction1);
				hwTrans1.getAttrs().setSync("fetch!");
				hwTrans1.getAttrs().appendUpdate(hwUniqueIdentifier + "_instruction=" + Constants.AVR_OPCODE_MNEMONICS[35].toUpperCase());
				instructionTemplate.addLocation(hwInstruction1);
				instructionTemplate.addTransition(hwTrans1);
				
				Location hwInstruction2 = new Location(Constants.AVR_OPCODE_MNEMONICS[36]);
				Transition hwTrans2 = new Transition(hwInstruction1, hwInstruction2);
				hwTrans2.getAttrs().setSync("fetch!");
				hwTrans2.getAttrs().appendUpdate(hwUniqueIdentifier + "_instruction=" + Constants.AVR_OPCODE_MNEMONICS[36].toUpperCase());
				instructionTemplate.addLocation(hwInstruction2);
				instructionTemplate.addTransition(hwTrans2);
				
				Location hwInstruction3 = new Location(Constants.AVR_OPCODE_MNEMONICS[37]);
				Transition hwTrans3 = new Transition(hwInstruction2,hwInstruction3);
				hwTrans3.getAttrs().setSync("fetch!");
				hwTrans3.getAttrs().appendUpdate(hwUniqueIdentifier + "_instruction=" + Constants.AVR_OPCODE_MNEMONICS[37].toUpperCase());
				instructionTemplate.addLocation(hwInstruction3);
				instructionTemplate.addTransition(hwTrans3);
				
				Transition returnTrans = new Transition(hwInstruction3, idle);
				returnTrans.getAttrs().setSync("return_instruction_" + methodId + instruction.toUpperCase() + "!");
				instructionTemplate.addTransition(returnTrans);				
				
				templates.add(instructionTemplate);
			}
		}
		
		return templates;
		
	}
}
