package com.tetaj.modelgenerator.uppaalmodelgeneratorircfg;

import static org.junit.Assert.assertEquals;

import java.io.IOException;
import java.util.ArrayList;
import java.util.HashMap;

import javax.xml.transform.Result;
import javax.xml.transform.Source;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerException;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamResult;

import org.junit.Test;
import org.w3c.dom.Document;

import com.tetaj.cfggenerators.JBCToIRCFG;
import com.tetaj.cfggenerators.constants.Constants;
import com.tetaj.ircfg.ControlFlowGraph;
import com.tetaj.ircfg.ControlFlowGraphException;
import com.tetaj.prettyprinter.AppConfig;
import com.tetaj.prettyprinter.IRCFGPrettyPrinter;

import com.jopdesign.libgraph.cfg.GraphException;
import com.jopdesign.libgraph.struct.TypeException;
import com.jopdesign.wcet.uppaal.model.DuplicateKeyException;
import com.jopdesign.wcet.uppaal.model.NTASystem;
import com.jopdesign.wcet.uppaal.model.Template;
import com.jopdesign.wcet.uppaal.model.XmlSerializationException;
import com.tetaj.modelcombiner.UppaalXmlProcessor;
import com.tetaj.modelgenerator.cfganalysis.CfgAnalysis;
import com.tetaj.modelgenerator.cfganalysis.CfgAnalysisException;
import com.tetaj.modelgenerator.staticmodels.StaticModels;
import com.tetaj.modelgenerator.uppaalmodelgeneratorircfg.UppaalModelGenerator;
import com.tetaj.uppaallayouter.UppaalLayouter;
import com.tetaj.uppaalprocessor.UppaalProcessor;

public class UppaalModelGeneratorTest {
	
	@Test 
	public void testClasspath() throws Exception {
		
		String testMethod = "org.testapp.SimpleProgram#TestMethod(I)I";
		JBCToIRCFG cfggen = null;
		boolean cfggenStatus = true;
		try {
			cfggen = new JBCToIRCFG(testMethod, ".");
		} catch (Exception e) {
			cfggenStatus = false;
			e.printStackTrace();
		}
		
		assertEquals(cfggenStatus,true);
		
		UppaalModelGenerator modelGenerator = null;

		ArrayList<String> sourcePaths = AppConfig.getSourcePaths();
		sourcePaths.add("./resources/testappsrc/");
		
		HashMap<String, ControlFlowGraph> cfgMap = cfggen.generateIRCFG();
		
		CfgAnalysis cfgAnalysis = new CfgAnalysis(cfggen.generateIRCFG());
		cfgAnalysis.conductCfgAnalyses();
		
		modelGenerator = new UppaalModelGenerator(testMethod,cfgMap, "jvm", Constants.JBC_OPCODE_MNEMONICS);
		
		NTASystem system = modelGenerator.getUppaalSystem();
		Template initTemplate = modelGenerator.getInitialisationTemplate(testMethod);
		system.addTemplate(initTemplate);
		
		system.appendDeclaration("clock globaltime;");
		system.setSystem(system.getSystem().substring(0, system.getSystem().length()-1) + "," + initTemplate.getId() + ";");
		
		
		
		boolean uppaalToXml = true;
		try {
			system.addTemplate(DummyUppaalTemplates.getDummyJvmUppaalTemplate("jvm"));
			system.setSystem(system.getSystem().substring(0, system.getSystem().length()-1) + "," + DummyUppaalTemplates.getDummyJvmUppaalTemplate("jvm").getId() + ";");
			
			org.w3c.dom.Document doc = system.toXML();
			
			UppaalLayouter layouter = new UppaalLayouter(doc);
			doc = layouter.getLayoutedXmlW3Document();
			
			saveXmlAsFile(doc,"uppaal.xml");
			
		} catch (Exception e) {
			uppaalToXml = false;
			e.printStackTrace();
		}
		assertEquals(uppaalToXml,true);
	}
	@Test
	public void testDynamicDispatch() throws Exception{
		String testMethod = "org.testapp.dynamicdispatch.Main#main()V";
		JBCToIRCFG cfggen = new JBCToIRCFG(testMethod,".");
				
		
		
		ArrayList<String> sourcePaths = AppConfig.getSourcePaths();
		sourcePaths.add("./resources/testappsrc/");
		
		HashMap<String, ControlFlowGraph> cfgMap = cfggen.generateIRCFG();
		
		IRCFGPrettyPrinter ircfgpre = new IRCFGPrettyPrinter("hej.png");
		ircfgpre.createVisualRepresentation(cfgMap.get("org.testapp.dynamicdispatch.Animal#makeSound()Ljava/lang/String;"));
		
		CfgAnalysis cfgAnalysis = new CfgAnalysis(cfggen.generateIRCFG());
		cfgAnalysis.conductCfgAnalyses();
		
		UppaalModelGenerator modelGenerator =new UppaalModelGenerator(testMethod,cfgMap, "jvm", Constants.JBC_OPCODE_MNEMONICS);
		
		NTASystem system = modelGenerator.getUppaalSystem();
		
		boolean uppaalToXml = true;
		try {
			system.addTemplate(DummyUppaalTemplates.getDummyJvmUppaalTemplate("jvm"));
			
			
			org.w3c.dom.Document doc = system.toXML();
			
			UppaalLayouter layouter = new UppaalLayouter(doc);
			doc = layouter.getLayoutedXmlW3Document();
			
			saveXmlAsFile(doc,"uppaal.xml");
			
		} catch (Exception e) {
			uppaalToXml = false;
			e.printStackTrace();
		}
		assertEquals(uppaalToXml,true);
		
		
	}
	
	@Test
	public void testGenerateDummyJVM() throws DuplicateKeyException, XmlSerializationException, TransformerException {
		
		NTASystem system = new NTASystem("dummyJVM");
		
		system.setSystem("system dummyJVM;");
		system.addTemplate(DummyUppaalTemplates.getDummyJvmUppaalTemplate("jvm"));
		
		Document doc = system.toXML();
		
		saveXmlAsFile(doc, "dummyJVM.xml");
	}

	private static void saveXmlAsFile(Document doc, String filepath) throws TransformerException {
		 TransformerFactory tranFactory = TransformerFactory.newInstance();
	     Transformer aTransformer = tranFactory.newTransformer();
	
	     Source src = new DOMSource(doc);
	     System.out.println("Starting the XML file creation");
	  
	     Result dest = new StreamResult(filepath);
	     aTransformer.transform(src, dest);
	}
	
	@Test 
	public void testStaticJvmWithHardware() throws DuplicateKeyException{
		
		String testMethod = "org.testapp.SimpleProgram#TestMethod(I)I";

		UppaalModelGenerator modelGenerator = null;
		try {
			ArrayList<String> sourcePaths = AppConfig.getSourcePaths();
			sourcePaths.add("./resources/testappsrc/");
			
			JBCToIRCFG cfggen = new JBCToIRCFG(testMethod,".");
			HashMap<String, ControlFlowGraph> cfgMap = cfggen.generateIRCFG();
			
			CfgAnalysis cfgAnalysis = new CfgAnalysis(cfggen.generateIRCFG());
			cfgAnalysis.conductCfgAnalyses();
			
			modelGenerator = new UppaalModelGenerator(testMethod,cfgMap, "jvm", Constants.JBC_OPCODE_MNEMONICS);
		} catch (Exception e) {
			e.printStackTrace();
		}
		
		NTASystem system = modelGenerator.getUppaalSystem();

		Template initTemplate = modelGenerator.getInitialisationTemplate(testMethod);
		system.addTemplate(initTemplate);
		
		system.appendDeclaration("clock globaltime;");
		system.setSystem(system.getSystem().substring(0, system.getSystem().length()-1) + "," + initTemplate.getId() + ";");
		
		
		try {
			Template staticJvm = StaticModels.getStaticJvmTemplate(Constants.JBC_OPCODE_MNEMONICS, "handle", "jvm", "assembly");
			system.addTemplate(staticJvm);
			system.setSystem(system.getSystem().substring(0, system.getSystem().length()-1) + "," + staticJvm.getId() + ";");
			
			for(Template template : DummyUppaalTemplates.getDummyHardwareTemplate(Constants.JBC_OPCODE_MNEMONICS, "handle", "avr")){
				system.addTemplate(template);
			}
			for(String instruction : Constants.JBC_OPCODE_MNEMONICS){
				if(!instruction.equals(Constants.ILLEGAL_OPCODE)){
					system.appendDeclaration("urgent chan invoke_instruction_" + "handle" + instruction.toUpperCase() +";");
					system.appendDeclaration("urgent chan return_instruction_" + "handle" + instruction.toUpperCase()+ ";");
				}
			}
			
			org.w3c.dom.Document doc = system.toXML();
			
			UppaalLayouter layouter = new UppaalLayouter(doc);
			doc = layouter.getLayoutedXmlW3Document();
			
			saveXmlAsFile(doc,"uppaal.xml");
			
		} catch (Exception e) {
			e.printStackTrace();
		}
		assertEquals(true,true);
	}
	@Test
	public void testMultipleReturns() throws Exception{
		methodToUppaal("org.testapp.MultipleReturns#multipleReturns(I)I", "uppaal.xml");
		
	}
	
	@Test 
	public void testBreakContinue() throws GraphException, TypeException, ControlFlowGraphException, CfgAnalysisException, IOException, Exception{
		methodToUppaal("org.testapp.BreakAndContinue#breakExample(I)V", "break.xml");
		methodToUppaal("org.testapp.BreakAndContinue#continueExample(I)V", "continue.xml");
		methodToUppaal("org.testapp.BreakAndContinue#mixedExample(I)V", "mixed.xml");
	}
	
	@Test
	public void testInheritance() throws GraphException, TypeException, ControlFlowGraphException, CfgAnalysisException, IOException, Exception{
		methodToUppaal("org.testapp.Inheritance#InheritedMethod(I)V", "inheritance.xml");
	}
	
	@Test
	public void testNativeCall() throws GraphException, TypeException, ControlFlowGraphException, CfgAnalysisException, IOException, Exception{
		methodToUppaal("org.testapp.Native#call()V", "native.xml");
	}
	
	@Test
	public void testStdlibCall() throws GraphException, TypeException, ControlFlowGraphException, CfgAnalysisException, IOException, Exception{
		methodToUppaal("org.testapp.StdLib#call()V", "stdlib.xml");
	}
	
	private void methodToUppaal(String methodName, String uppaalFile) throws GraphException, TypeException,
			ControlFlowGraphException, CfgAnalysisException, Exception,
			IOException {
		String testMethod = methodName;
		JBCToIRCFG cfggen = new JBCToIRCFG(testMethod,".");
				
		
		
		ArrayList<String> sourcePaths = AppConfig.getSourcePaths();
		sourcePaths.add("./resources/testappsrc/");
		
		HashMap<String, ControlFlowGraph> cfgMap = cfggen.generateIRCFG();
		
		CfgAnalysis cfgAnalysis = new CfgAnalysis(cfggen.generateIRCFG());
		cfgAnalysis.conductCfgAnalyses();
		
		UppaalModelGenerator modelGenerator =new UppaalModelGenerator(testMethod,cfgMap, "jvm", Constants.JBC_OPCODE_MNEMONICS);
		
		NTASystem system = modelGenerator.getUppaalSystem();
		
		boolean uppaalToXml = true;
		try {
			system.addTemplate(DummyUppaalTemplates.getDummyJvmUppaalTemplate("jvm"));
			system.addTemplate(modelGenerator.getInitialisationTemplate(testMethod));
			system.appendDeclaration("clock globaltime;");
			system.setSystem(system.getSystem().substring(0,system.getSystem().length()-1) + ", "+ modelGenerator.getInitialisationTemplate(testMethod).getId() + ", " +DummyUppaalTemplates.getDummyJvmUppaalTemplate("jvm").getId()+ ";");
			
			org.w3c.dom.Document doc = system.toXML();
			
			UppaalLayouter layouter = new UppaalLayouter(doc);
			doc = layouter.getLayoutedXmlW3Document();
			
			saveXmlAsFile(doc,uppaalFile);
			
		} catch (Exception e) {
			uppaalToXml = false;
			e.printStackTrace();
		}
		assertEquals(uppaalToXml,true);
		
	}
	
}
