package tetasarts;

import java.io.File;
import java.io.IOException;
import java.io.PrintStream;

import org.apache.bcel.classfile.ClassParser;
import org.apache.bcel.classfile.CodeException;
import org.apache.bcel.classfile.JavaClass;
import org.apache.bcel.classfile.Method;
import org.apache.bcel.generic.BranchInstruction;
import org.apache.bcel.generic.InstructionHandle;
import org.apache.bcel.generic.InstructionList;

import org.apache.bcel.util.ClassPath;

public class BytecodeGenerator {
	private JavaClass mainClass;
	private static BytecodeGenerator bcc;
	private static String outDir = "bOut";
	private static String outFile = "out.b";
 
	public static void main(String[] args) {
		bcc = new BytecodeGenerator();

		for (int i = 0; i < args.length; i++) {
			String classname = args[i];

			if (classname.endsWith(".class")) {
				classname = classname.substring(0, classname.length() - 6)
						.replace('/', '.');
				outFile = classname.substring(0, classname.length() - 6)
						.replace(new String("/"), new String(File.separator))
						+ ".b";
			}

			try {
				File dir = new File(outDir);
				File file = new File(outDir + File.separator + outFile);
				System.out.println(file.getAbsolutePath());
				if (!dir.exists()) {
					if (!dir.mkdir()) {
						System.err.println("Error creating output directory: "
								+ dir.getAbsolutePath());
						System.exit(-1);
					}
				}
				if (!file.exists()) {
					if (!file.createNewFile()) {
						System.err.println("Unable to create outFile: "
								+ file.getAbsolutePath());
						System.exit(-1);
					}
					if (!file.canWrite()) {
						System.err.println("Unable to write to outFile: "
								+ file.getAbsolutePath());
						System.exit(-1);
					}
				}

				bcc.build(classname, new PrintStream(file));
			} catch (Exception e) {
				System.err.println("An error occured!");
				e.printStackTrace();
			}
		}
	}

	public void build(String classname, PrintStream out) {
		
		//String classFile = "bin\\" + classname.replace('.', '\\') + ".class";

		//out.println("Class: " + classname + "\n");
		//System.out.println("Class: " + classname);

		try {
			mainClass = new ClassParser(ClassPath.SYSTEM_CLASS_PATH.getClassFile(classname).getPath()).parse();
		} catch (IOException e) {
			System.err.println("Error loading main class. " + classname);
			// e.printStackTrace();
			System.exit(-1);
		}

		Method[] methods = mainClass.getMethods();
		
		for (Method currentMethod : methods) {
			if (currentMethod.isAbstract()) {
				out.println("Skipping abstract method: "
						+ mainClass.getClassName() + "."
						+ currentMethod.getName()
						+ currentMethod.getSignature());
				out.println("");
				continue;
			}
			out.println("Method: " + currentMethod.getName() + " " + currentMethod.getSignature() + " - Method size: " + currentMethod.getCode().getLength());
			// System.out.println("InstructionList for method " +
			// m.getName()
			// + ":");

			// System.out.println("InstructionList for method " +
			// m.getName()
			// + ":");

			InstructionList l = new InstructionList(currentMethod.getCode()
					.getCode());
			l.setPositions(true);

			InstructionHandle handle = l.getStart();
			StringBuffer res = new StringBuffer();
			do {
				res.append(paddingString(Integer
						.toString(handle.getPosition()), (int)Math.log(currentMethod
						.getCode().getCode().length), ' ', true)).
						append(": ").
						append(paddingString(handle.getInstruction().toString(true), 40, ' ', false)).
						append(" - ");
				if(handle.getInstruction() instanceof BranchInstruction) {
					BranchInstruction bi = (BranchInstruction)handle.getInstruction();
					res.append(bi.getTarget().getPosition()).
					append(" - ");
				}
				res.append("\n");
				
				handle = handle.getNext();
			} while (handle != null);
			
			if(currentMethod.getExceptionTable() != null) {
				if(currentMethod.getExceptionTable().getNumberOfExceptions() > 0) {
					res.append("Exceptions: ").
					append(currentMethod.getExceptionTable().getNumberOfExceptions()).
					append("\n");
					
					for(String name : currentMethod.getExceptionTable().getExceptionNames()) {
						res.append("  ");
						res.append(name).
						append("\n");
					}
					
				}
			} else {
				res.append("No exceptions. \n");
			}
			
			if(currentMethod.getCode().getExceptionTable() != null) {
				if(currentMethod.getCode().getExceptionTable().length > 0) {
					res.append("Code exceptions: ").
					append(currentMethod.getCode().getExceptionTable().length).
					append("\n");
					
					for(CodeException ce : currentMethod.getCode().getExceptionTable()) {
						res.append("  ");
						res.append(ce.toString(currentMethod.getConstantPool())).
						append("\n");
					}
					
				}
			} else {
				res.append("No exceptions. \n");
			}
			
			out.print(res.toString());
			out.println("");
		}

	}

	public synchronized String paddingString(String s, int n, char c,
			boolean paddingLeft) {
		StringBuffer str = new StringBuffer(s);
		int strLength = str.length();
		if (n > 0 && n > strLength) {
			for (int i = 0; i <= n; i++) {
				if (paddingLeft) {
					if (i < n - strLength)
						str.insert(0, c);
				} else {
					if (i > strLength)
						str.append(c);
				}
			}
		}
		return str.toString();
	}
}
