package tetasarts;

import java.io.IOException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Comparator;
import java.util.List;

import org.apache.bcel.classfile.ClassFormatException;
import org.apache.bcel.classfile.ClassParser;
import org.apache.bcel.classfile.JavaClass;
import org.apache.bcel.classfile.Method;
import org.apache.bcel.util.ClassPath;
import org.apache.bcel.util.ClassPath.ClassFile;

public class ClassLookup {
	public static void printClassPath(ClassPath cp){
		for(String s: cp.toString().split(":")){
			System.out.println(s);
		}
	}
	public static void printClassPath(ClassPath cp, String prefix){
		for(String s: cp.toString().split(":")){
			System.out.println(prefix+s);
		}
	}
	
	public static String printableCP(ClassPath cp, String prefix){
		StringBuilder sb = new StringBuilder();
		for(String s: cp.toString().split(":")){
			sb.append(prefix);
			sb.append(s);
			sb.append(System.getProperty("line.separator").toString());
		}
		return sb.toString();
	}
	
	public static JavaClass loadClass(String className){
		ClassFile classFile = getClassFile(className);
		JavaClass jc = null;
		try {
			 jc = new ClassParser(classFile.getInputStream(), classFile.getPath()).parse();
		} catch (ClassFormatException e) {
			e.printStackTrace();
			System.exit(1);
		} catch (IOException e) {
			e.printStackTrace();
			System.exit(1);
		}
		return jc;
	}
	
	public static ClassFile getClassFile(String className){
		ClassFile result = null;
		try {
			result = Options.classpath.getClassFile(className); // naive?
		} catch (IOException e) 
		{
			System.out.println("Cannot find class: "+className);
			List<String> possibleCP = searchSystemCP(className);
			if(possibleCP.size()>0){
				System.out.println("Try adding:");
				for(String s: possibleCP){
					System.out.println("  "+ s);
				}
				System.out.println("to your classpath...");
				// 
			}
			e.printStackTrace();
			System.exit(1);
		}
		if(Options.CHECK_DUPLICATE_CLASSES_IN_CLASSPATH){
			List<JavaClass> jcs = getAllClasses(Options.classpath, className, true);
			assert jcs.size()==1 : "error while looking up class ["+className+ "]"+
								   " in class library. Found "+jcs.size()+" class candidates,"+
								   " using CP: "+System.getProperty("line.separator").toString()+printableCP(Options.classpath, "  ");
		}		
		return result;
	}
	
	public static List<String> searchSystemCP(String className){
		List<String> jclist = new ArrayList<String>();
		for (String s : ClassPath.SYSTEM_CLASS_PATH.toString().split(":")) {
			try {
				ClassFile classFile = new ClassPath(s).getClassFile(className);
				jclist.add(s);
//				JavaClass jc = new ClassParser(classFile.getInputStream(), classFile.getPath()).parse();
			} catch (IOException e) {
//				System.out.println(e.getMessage());
			}
		}
		return jclist;
	}
	
	public static List<JavaClass> getAllClasses(ClassPath classpath,
			String className, boolean acceptSameSignature) {
		List<JavaClass> classes = getAllClasses(classpath, className);
		List<JavaClass> result = new ArrayList<JavaClass>();
		
		if(classes.size()>0)
			result.add(classes.get(0));
		
		if (classes.size() > 1) {
			JavaClass first = classes.get(0);
			for (JavaClass jc : classes) {
				boolean equal = true;
				if (jc == first)
					continue;
				else {
					if(first.isAbstract() != jc.isAbstract() || first.isInterface() != jc.isInterface()){
						equal = false;
					} else {
						Method[] firstMethods = first.getMethods();
						Method[] otherMethods = jc.getMethods();
						if (firstMethods.length == otherMethods.length) {
							Comparator<Method> c = new Comparator<Method>() {
								public int compare(Method o1, Method o2) {
									return (o1.getName()+o1.getSignature()).compareTo(o2.getName()+o2.getSignature());
								}
							};
							Arrays.sort(firstMethods, c);
							Arrays.sort(otherMethods, c);
							for (int i = 0; i < firstMethods.length; i++) {
							//	System.out.println(firstMethods[i].getName() +"  <==>  "+otherMethods[i].getName());
								/**
								 * Return value as defined by given BCELComparator
								 * strategy. By default two method objects are said
								 * to be equal when their names and signatures are
								 * equal.
								 */
								if (!firstMethods[i].equals(otherMethods[i])) {
									equal = false;
									break;
								}
								if(firstMethods[i].isAbstract() != otherMethods[i].isAbstract()){
									equal = false;
									break;
								}
							}
						} else
							equal = false;
					}
				}
				if (!equal) {
					result.add(jc);
				}
			}
		}
		return result;
	}

	public static List<JavaClass> getAllClasses(ClassPath classpath,
			String className) {
		List<JavaClass> jclist = new ArrayList<JavaClass>();
		for (String s : classpath.toString().split(":")) {
			try {
				ClassFile classFile = new ClassPath(s).getClassFile(className);
				JavaClass jc = new ClassParser(classFile.getInputStream(),
						classFile.getPath()).parse();
				jclist.add(jc);

			} catch (IOException e) {
//				System.out.println(e.getMessage());
			}
		}
		return jclist;
	}
}
