package tetasarts;


import java.io.BufferedReader;
import java.io.File;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.PrintStream;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;
import java.util.logging.FileHandler;
import java.util.logging.Level;
import java.util.logging.Logger;
import java.util.logging.SimpleFormatter;

import org.apache.bcel.classfile.ClassParser;
import org.apache.bcel.classfile.JavaClass;
import org.apache.bcel.classfile.Method;
import org.apache.bcel.generic.ObjectType;
import org.apache.bcel.util.ClassPath;
import org.apache.bcel.util.ClassPath.ClassFile;

import soot.coffi.method_info;
import tetasarts.Options.SchedulerType;
import tetasarts.Options.ScjApi;
import tetasarts.analysis.Collapser;
import tetasarts.analysis.SimpleWcetVisitor;
import tetasarts.analysis.SyncMarker;
import tetasarts.generator.Constants;
import tetasarts.generator.JavaTranslator;
import getOptLong.ArgRun;
import getOptLong.ArgumentOption;
import getOptLong.BoolRun;
import getOptLong.BooleanOption;
import getOptLong.Configuration;
import getOptLong.GetOptLong;
import tetasarts.tir.IntermediateData;
import tetasarts.tir.basicblocks.AbstractBasicBlock;
import tetasarts.tir.classinfo.AbstractClassNode;
import tetasarts.tir.classinfo.ClassGraph;
import tetasarts.tir.methods.MethodInformation;
import tetasarts.tir.sets.TemplateSet;
import tetasarts.uppaal.Location;
import tetasarts.uppaal.Transition;
import tetasarts.uppaal.UppaalModel;
import tetasarts.uppaal.UppaalTemplate;
import tetasarts.uppaal.Location.LocationType;
import tetasarts.uppaal.declarations.Channel;
import tetasarts.uppaal.translation.ExplicitModelCreater;
import tetasarts.uppaal.translation.TemplateFinalizerCountVisitor;
import tetasarts.uppaal.translation.TemplateNewCountVisitor;
import uppaal.Automaton;
import uppaal.NTA;

public class Main {
	private static StatisticHandler sh;
	/**
	 * @param args
	 */
	public static void main(String[] argv) throws Exception{
		long start = System.currentTimeMillis();
		
		if (argv.length == 0){
			System.err.println("Invalid arguments, must specify the main class of the system being analyzed");
			return;
		}
		
		String mainClass = null;
		
		Configuration options = buildConfiguration();
		GetOptLong opt = new GetOptLong(argv, options);
		
		if(opt.getNonOptions().size()==1)
			mainClass = opt.getNonOptions().get(0);
		else{
			System.out.println("Error:");
			for(String s : opt.getNonOptions())
				System.out.println("   " + s);
			if(opt.getNonOptions().size() > 1)
				System.out.println(" -- Only one [main] class-file please");
			else
				System.out.println(" -- Please supply [main] class-file");
			System.exit(1);
		}


//		if("".equals(Options.classpath.toString())) {
//			System.out.println("setting classpath to system classpath.");
//			Options.classpath = ClassPath.SYSTEM_CLASS_PATH;
//			//System.exit(0);
//		}
		
			
		if (mainClass.toLowerCase().endsWith(".class")) {
			mainClass = mainClass.substring(0, mainClass.length() - 6).replace('/', '.');
		}
		System.out.println("mainClass: " + mainClass);
		
		setupLogger(Options.outputDir);
		Logger.getLogger("generator").setLevel(Level.SEVERE);
		
		try {
//			System.out.println("using classpath: ");
//			ClassLookup.printClassPath(Options.classpath, "  ");
			new ClassParser(ClassLookup.getClassFile(mainClass).getPath()).parse();
		} catch (Exception e) {
			System.out.println(e);
			System.err.println("Unable to find main class: \"" + mainClass + "\". " + e.toString());
			return;
		}
//		String className = "java.lang.Object";
//		System.out.println("This class: " + className);
//		ClassFile classFile = Options.classpath.getClassFile(className);
//		JavaClass jc = new ClassParser(classFile.getInputStream(), classFile.getPath()).parse();
//		for(Method m: jc.getMethods()){
//			System.out.println(" M: "+m.getName());
//		}
//		
//		System.exit(0);
		//start, create autoflushed printstream and set out and err - for debug, a bit slower
		PrintStream newOut = new PrintStream(System.out, true);
		PrintStream newErr = new PrintStream(System.err, true);
		System.setOut(newOut);
		System.setErr(newErr);
		//end
		
		try {
			
			System.out.println("JT");
			JavaTranslator jt = new JavaTranslator(mainClass);
			System.out.println("done jt - buildIR");
			IntermediateData result = jt.buildIR();
			System.out.println("done IR");
			// start collapsify
			System.out.println("Starting collapsify");
			if(Options.COLLAPSE_BASIC_BLOCKS) {
				if(Options.EXPLICIT_MODEL) {
					System.err.println("Collapsing basic blocks cannot be combined with an explicit model.. yet!");
					System.exit(-1);
				}
				Collapser.collapse(result);
			}
			System.out.println("Collapsify done..");
			
			System.out.println("starting wcetify");
			// start wcetify
			if(!Options.EXPLICIT_MODEL) {
				for(AbstractClassNode acn : result.getClassGraph().getNodes()){
					for(MethodInformation mi : acn.getMethods().values()){
					//	if(mi.isGenerated()){
							SyncMarker.markGraph(mi.getInstructionGraph());
							SimpleWcetVisitor sw = new SimpleWcetVisitor(mi);
							for(AbstractBasicBlock abb : mi.getInstructionGraph().getNodes()){
								abb.accept(sw);
							}
					//	}
					}
				}
			}
			System.out.println("wcetify done");
			
			// Start UPPAALify
			
			UppaalModel uppaalModel = tetasarts.uppaal.translation.UppaalTranslation.generateUppaalModel(result);

			
/** FINALIZATION **/
			// finalizers enabled?
			if(Options.CONSIDER_FINALIZERS){
				// Fix Finalizers if any 
				Set<Entry<MethodInformation, Integer>> finalizersToConsider = TemplateFinalizerCountVisitor.finalizerIndexMap.entrySet();
				if(finalizersToConsider.size() > 0){
					TemplateSet runTemplates = TemplateSet.globalSet.getMapFromMethodSignature("\\(\\)Z").getMapFromMethodName("run");
					Iterator<UppaalTemplate> templateIt = runTemplates.iterator();
					while(templateIt.hasNext()){
						UppaalTemplate runTemplate = templateIt.next();
						System.out.println(runTemplate.getMethodInformation().getName());
						Location initLoc = runTemplate.getInit();
						List<Location> endLocs = runTemplate.getPreds(initLoc);
						assert endLocs.size() == 1 : " only one end location";
						Location endLoc = endLocs.get(0);
						
						//TODO: Get the correct execution time for this
						int finalizerMethodInvokeExecutionTime = 10; // set execution time to 10
						Location finCalling = new Location("FinalizeCalling");
						Location runFinalizer = new Location("running_finalizer");
						Location retFinalizer = new Location("returnFrom_finalizer");
						Location ensureDone = new Location("considerAllFinalizers");
						ensureDone.setType(LocationType.COMMITED);
						finCalling.setInvariant("executionTime <= "+finalizerMethodInvokeExecutionTime+" && executionTime'==running[tID]"); 
						retFinalizer.setInvariant("executionTime<=methodSwitchCost && executionTime'==running[tID]");
						
						
						runTemplate.addLocation(finCalling);
						runTemplate.addLocation(runFinalizer);
						runTemplate.addLocation(retFinalizer);
						runTemplate.addLocation(ensureDone);
						
	
						// generate invoke transition(s)
						for(Entry <MethodInformation, Integer> finalizerEntry : finalizersToConsider){
							Transition cllF2runF = new Transition(finCalling.getId(), runFinalizer.getId());
							cllF2runF.addGuard("executionTime=="+finalizerMethodInvokeExecutionTime);
							cllF2runF.addGuard("finalizeCount[tID]["+finalizerEntry.getValue()+"]>0");
							String channelName = UppaalModel.getChannelIdentifier(finalizerEntry.getKey());
							Channel c = new Channel(channelName, finalizerEntry.getKey().getCanonicalName(), "tID");
							cllF2runF.setSync(c, true);
							cllF2runF.setUpdate("executionTime=0");
							runTemplate.addTransition(cllF2runF);
						}
						// generate return transition(s)
						for(Entry <MethodInformation, Integer> finalizerEntry : finalizersToConsider){
							Transition runF2retF = new Transition(runFinalizer.getId(), retFinalizer.getId());
	//						t.addGuard("executionTime=="+finalizerMethodInvokeExecutionTime);
	
							String channelName = UppaalModel.getChannelIdentifier(finalizerEntry.getKey());
							Channel c = new Channel(channelName, finalizerEntry.getKey().getCanonicalName(), "tID");
							runF2retF.setSync(c, false);
							
							runF2retF.setUpdate("executionTime=0,\nmethodSwitchCost="+finalizerMethodInvokeExecutionTime); // get methodSwitchTime
							runF2retF.addUpdate("finalizeCount[tID]["+finalizerEntry.getValue()+"]--");
							runTemplate.addTransition(runF2retF);
						}
	
	
						// relink transitions pred(endDummy) --> callFinalizer
						
						List<Transition> transitions = new LinkedList<Transition>();
	
						List<Location> endPreds = runTemplate.getPreds(endLoc);
						for(Location l: endPreds){
							for(Transition t:runTemplate.getTransitions(l.getId(), endLoc.getId())){
								t.setTarget(ensureDone.getId());
							}
							transitions.addAll(runTemplate.getTransitions(l.getId(), endLoc.getId()));
						}
						Transition doneLink = new Transition(retFinalizer.getId(), ensureDone.getId());
						doneLink.addGuard("executionTime==methodSwitchCost");
						doneLink.addUpdate("executionTime=0");
						runTemplate.addTransition(doneLink);
						Transition doneTrueLink = new Transition(ensureDone.getId(), endLoc.getId());
						doneTrueLink.addGuard("finalized(tID)");
						runTemplate.addTransition(doneTrueLink);
	//					
						Transition doneFalseLink = new Transition(ensureDone.getId(), finCalling.getId());
						doneFalseLink.addGuard("!finalized(tID)");
						runTemplate.addTransition(doneFalseLink);
						
	//					List<Transition> lastTransitions = runTemplate.getTransitions(endLoc.getId(),initLoc.getId());
	//					assert lastTransitions.size() == 1 : " only one last transition";
	//					lastTransitions.get(0).setSource(retFinalizer.getId());
	//					runTemplate.addTransition();
					}
				}

				// TODO Sort classes according to index
				uppaalModel.getDeclarations().addString("// Class mapping:");
				Set<Entry<ObjectType, Integer>> classEntrySet = TemplateNewCountVisitor.classIndexMap.entrySet();
				for(Entry<ObjectType, Integer> classIndex : classEntrySet){
					uppaalModel.getDeclarations().addString("//   ["+classIndex.getValue()+"]: " + classIndex.getKey());
				}
				uppaalModel.getDeclarations().addString("// end.");
				uppaalModel.getDeclarations().addString("typedef int[0,"+(TemplateNewCountVisitor.getTotalClassCount())+"] ClassID;");
				uppaalModel.getDeclarations().addString("int instanceCount[ThreadID][ClassID];");
				
				// TODO Sort finalizers according to index
				uppaalModel.getDeclarations().addString("// Finalizer mapping:");
				for(Entry<MethodInformation, Integer> methodIndex : TemplateFinalizerCountVisitor.finalizerIndexMap.entrySet()){
					uppaalModel.getDeclarations().addString("//   ["+methodIndex.getValue()+"]: " + methodIndex.getKey());
				}
				uppaalModel.getDeclarations().addString("// end.");
				uppaalModel.getDeclarations().addString("typedef int[0,"+(TemplateFinalizerCountVisitor.getTotalFinalizerCount())+"] ClassID;");
				uppaalModel.getDeclarations().addString("int finalizeCount[ThreadID][ClassID];");
				
				uppaalModel.getDeclarations().addString("bool finalized(ThreadID tid){\n"+
				    "	bool result = true;\n"+
				    "	for(i:ClassID){\n"+
				    "		result &= finalizeCount[tid][i]==0;\n"+
				    "	}\n"+
				    "	return result;\n"+
				"}");

//				uppaalModel.getDeclarations().add

			}
/** END FINALIZATION	**/		
			
			/** 
			 * New counting 
			 */
			
			if(Options.COUNT_NEWS){
				uppaalModel.getDeclarations().addString("int liveObjects = 0;");
			}
			
			/**
			 * End newcounting
			 */
			
			
			/** WCET/UTILIZATION/IDLE **/
			
			if(Options.UTILIZATION_MONITOR){
				uppaalModel.getDeclarations().addString("clock util, idle;");
			}
			/** END WCET/UTILIZATION/IDLE **/
			
			String outfile;
						
			outfile = Options.outputDir+ Options.outputFile;

			if(outfile.toLowerCase().endsWith(".xml")) {
				outfile = outfile.substring(0, outfile.lastIndexOf(".xml"));
			}
			String outputFile = outfile+".xml";
			uppaalModel.writeModelToFile(outfile + ".xml");
			/*
			 * This is by far not the optimal place to do model combining!
			 * However, it is the only place where all information about
			 * the program uppaal model is available
			 */
			if(Options.EXPLICIT_MODEL) {
				//this seems stupid but it is the easiest way to use the
				// combiner which assumes jupaal is used
				NTA programModel = new NTA(outputFile);
				
				ExplicitModelCreater explicitCreater = new ExplicitModelCreater(programModel, Options.PLATFORM_TYPE);
				NTA explicitModel = explicitCreater.generateExplicitModel();
				
				//remove following comment when everything works!
				
				/*File programModelFile = new File(outputFile);
				if(!programModelFile.delete()) {
					System.err.println("could not delete old file containing the program model!");
				}*/
				explicitModel.writeModelToFile(outputFile);
			}
			
//			uppaalModel.writeModelToFile("/tmp/hmm" + ".xml");

			if(Options.LAY_TEMPLATES_NICELY) {
				uppaal.NTA nta = new uppaal.NTA(outfile+".xml");
				{ // dont taint scope
					StringBuilder sb = new StringBuilder();
					for(Automaton a : nta.getAutomata()){
						sb.append("-t " + a.getName().toString()+" ");
					}
					
					String line="", errline="";
					
					ProcessBuilder pb = new ProcessBuilder("/bin/bash", "-c", Options.LAYOUT_UPPAAL+"/bin/layout_uppaal"+" "+sb.toString()+" "+outfile+".xml /tmp/output.xml");
					for(String c : pb.command()){
						System.out.print(c+ " ");
					}
					System.out.println();
					Map<String, String> env = pb.environment();
					env.put("PYTHONPATH", ".");
					pb.directory(new File(Options.LAYOUT_UPPAAL));
					Process p = pb.start();
			        BufferedReader stderr = new BufferedReader(new InputStreamReader(p.getErrorStream()));
			        BufferedReader stdin  = new BufferedReader(new InputStreamReader(p.getInputStream()));
			        while ((line = stdin.readLine()) != null || (errline = stderr.readLine())!=null) {
			          if(line != null && ! line.equals("")) System.out.println(line);
			          if(errline   != null && ! errline.equals(""))  System.out.println(errline);
			          line = errline ="";
			        }
			        stdin.close();
			        stderr.close();
				}
				
	//			for(uppaal.Automaton a : nta.getAutomata()){
	//				a.removeStumpAngles(angleThreshold, lengthThreshold)
	//			}
				

				nta.writeModelToFile(outputFile);
				

				
				
			}
			uppaalModel.writeNotDeadlockToFile(outfile + ".q");
			System.out.println("Done");	
//			System.out.println(((MethodInformation)MethodSet.globalSet.getMapFromName("^finalize$").toArray()[0]).getCanonicalClassName());
			
			
//			sarts.uppaal.UppaalModel.correctXML(outfile + ".xml");
		} catch (Exception e) {
			// TODO Auto-generated catch block
			System.out.flush();
			System.err.flush();
			Thread.sleep(100);
			e.printStackTrace();
			throw e;
		}
		
		System.out.flush();
		System.err.flush();
		System.out.println("Warnings and Errors: (see generator.log for details)");
		System.out.println(sh.getStats());
		long end = System.currentTimeMillis();
		long time = end-start;
		long ms = time %1000;
		long rest = (time-ms);
		long secs = (rest/1000) % 60;
		rest = rest - (secs*1000);
		long mins = rest/60000;
		System.out.println("Generation time: "+mins+"m "+ secs + "s "+ms+ "ms");
		
	}
	
	private static Configuration buildConfiguration() {
		Configuration options = new Configuration();
		
		options.add(new BooleanOption(
				"--no-basicblock-collapse", 
				"-b", 
				"Turn off basicblock collapsing.", 
				new BoolRun() {
					public void run() {
						Options.COLLAPSE_BASIC_BLOCKS = false;
					}
				})
		);
		
		options.add(new BooleanOption(
				"--report-missing-wcet", 
				"-r", 
				"Report discovered instructions with absent WCET information.", 
				new BoolRun() {
					public void run() {
						Options.REPORT_MISSING_WCET = true;
					}
				})
		);
		
		options.add(new BooleanOption(
				"--include-java-class-lib", 
				"-ijc", 
				"If the dependency extent enters the Java Class Library, also generate TAs for those.", 
				new BoolRun() {
					public void run() {
						Options.ANALYSE_JAVA_CLASS_LIB = true;
					}
				})
		);
		
		options.add(new BooleanOption(
				"--full-class-loading", 
				"-F", 
				"Try to load -all- classes.", 
				new BoolRun() {
					public void run() {
						Options.FULL_CLASS_LOADING = true;
					}
				})
		);
		
		options.add(new BooleanOption(
				"--lazy-method-generation", 
				"-m", 
				"Lazily generate intermediate representation of methods (TESTING). Default: "+Options.LAZY_METHODS, 
				new BoolRun() {
			public void run() {
				Options.LAZY_METHODS = true;
			}
		})
		);
		
		options.add(new BooleanOption(
				"--layout-templates", 
				"-l", 
				"Lay out the resulting templates nicely. This requires \"layout_uppaal\" to be in your path. layout_uppaal is part of pyuppaal.", 
				new BoolRun() {
					
					@Override
					public void run() {
						Options.LAY_TEMPLATES_NICELY = true;
					}
				}));
		
		options.add(new ArgumentOption(
				   
				"--layout-uppaal-path", 
				"-L", 
				"PATH",
				"Set the path to the layout_uppaal program. Default: ["+Options.LAYOUT_UPPAAL+"]", 
				new ArgRun() {
					public boolean run(String argument) {
						Options.LAYOUT_UPPAAL = argument;
						return true;
					}
				})			
		);


		options.add(new BooleanOption(
				"--use-jop-opcode-methods", 
				"-u", 
				"Generate and use methods for special JOP instructions such as new. This makes the model more acurate and \"true\", but also more complex and will increase verification time.", 
				new BoolRun() {
					public void run() {
						Options.USE_JOP_OPCODE_METHODS = true;
					}
				})
		);

		
		options.add(new ArgumentOption(
				   
				"--classpath", 
				"-C", 
				"CLASSPATH",
				"Sets the classpath used for the analysis. Default is system path.", 
				new ArgRun() {
					public boolean run(String argument) {
						Options.classpath = new ClassPath(
								(Options.classpath.toString()==""?"":Options.classpath.toString()+":")+argument);
						return true;
					}
				})			
		);

		options.add(new BooleanOption(
				"--count-new-objects", 
				"-c", 
				"Increment a counter for every new-instruction encountered.", 
				new BoolRun() {
					public void run() {
						Options.COUNT_NEWS = true;
					}
				})
		);
		
		
		options.add(new BooleanOption(
				"--inherit-classpath", 
				"-i", 
				"In the generation and analysis, inherit classpath of SARTS.", 
				new BoolRun() {
					public void run() {
						System.out.println("Inheriting path");
						System.out.println(Options.classpath);
						Options.classpath = new ClassPath((Options.classpath.toString()==""?"":Options.classpath.toString()+":")+ClassPath.SYSTEM_CLASS_PATH);//System.getProperty("java.class.path"));
						System.out.println(Options.classpath);
					}
				})
		);

		options.add(new BooleanOption(
				"--check-duplicate-classes", 
				"-cd", 
				"If enabled, we assert that there is only one candidate pr classin the entire classpath.", 
				new BoolRun() {
					public void run() {
						Options.CHECK_DUPLICATE_CLASSES_IN_CLASSPATH = true;
					}
				})
		);

		options.add(new BooleanOption(
				"--include-finalizers", 
				"-f", 
				"In the generation and analysis, include analysis of finalizers. With private scoped memory semantics (finalizers are executed at the end of a release)", 
				new BoolRun() {
					public void run() {
						Options.CONSIDER_FINALIZERS = true;
						System.out.println("Finalizers enabled...");
					}
				})
		);
		
		options.add(new BooleanOption(
				"--include-utilization", 
				"-util", 
				"Include determination of utilization.", 
				new BoolRun() {
					public void run() {
						Options.UTILIZATION_MONITOR = true;
					}
				})
		);
		
		options.add(new BooleanOption(
				"--use-statistical-model-checking", 
				"-smc", 
				"Use model for statistical model checking.", 
				new BoolRun() {
					public void run() {
						Options.STATISTICAL_MODEL_CHECKING = true;
					}
				})
		);
		
		options.add(new BooleanOption(
				"--use-scj", 
				"-scj", 
				"Use \"official\" SCJ instead of SCJ2.", 
				new BoolRun() {
					public void run() {
						
						Options.SCJ_API = ScjApi.SCJ;
						
						Constants.PERIODIC_THREAD = "javax.safetycritical.PeriodicEventHandler";
						Constants.PERIODIC_PARAMETERS = "javax.safetycritical.PeriodicParameters";
						Constants.SPORADIC_PARAMETERS = "javax.safetycritical.SporadicParameters";
						Constants.SPORADIC_THREAD = "javax.safetycritical.SporadicThread";
						
						Constants.RELATIVETIME = "javax.safetycritical.RelativeTime";
						Constants.RELATIVETIMEMICRO = "javax.safetycritical.RelativeTimeMicro";
						Constants.RELATIVETIMEMILLI = "javax.safetycritical.RelativeTimeMilli";
						Constants.RELATIVETIMENANO = "javax.safetycritical.RelativeTimeNano";
						Constants.RELATIVETIMESECONDS = "javax.safetycritical.RelativeTimeSeconds";
						
						Constants.RT_THREAD_ADAPTER = "javax.safetycritical.RtThreadAdapter";
						Constants.RT_SW_THREAD_ADAPTER = "javax.safetycritical.RtSwEventAdapter";
						Constants.REALTIME_SYSTEM = "javax.safetycritical.RealtimeSystem";
					}
				})
		);
		
		
		options.add(new ArgumentOption(
				   
				"--output-path", 
				"-o", 
				"OUTPUT_DIR",
				"Sets the output directory used for the analysis.", 
				new ArgRun() {
					public boolean run(String argument) {
						Options.outputDir = argument+(argument.endsWith(File.separator)?"":File.separator);
						System.out.println("Output dir: "+Options.outputDir);
						File f = new File(Options.outputDir);
						if(!f.exists()){
							System.err.println("Outputdir does not exist, creating...");
							if(!f.mkdirs()){
								System.err.println("Failed to create output directory");
								return false;
							}
						}
						if(!f.isDirectory()){
							System.err.println("Outputdir is not a directory");
							return false;
						} 
						return true;
					}
				})			
		);
		
		options.add(new ArgumentOption(
				"--main-entry-point", 
				"-mep", "<method-name of main class>", 
				"Sets the main entry point for the analysis. \"main\" is default", 
				new ArgRun() {
			
					@Override
					public boolean run(String argument) {
						Options.MAIN_ENTRY_POINT = argument; 
						return true;
					}
		}));
		
		options.add(new ArgumentOption(
				"--devirtualization-analysis", 
				"-da", 
				"[naive|cha|rta|vta|rtavta]", 
				"Sets the analysis used for approximating receivers of virtual method calls. CHA increases the analysis time considerably. Naive is default", 
				new ArgRun() {
					
					private static final String NAIVE = "naive";
					private static final String CHA = "cha";
					private static final String RTA = "rta";
					private static final String VTA = "vta";
					private static final String RTAVTA = "rtavta";
					
					@Override
					public boolean run(String argument) {
						if(NAIVE.equals(argument)) {
							Options.DEVIRTUALIZATION_METHOD = Options.DevirtualizationType.NAIVE;
						} else if(CHA.equals(argument)) {
							Options.DEVIRTUALIZATION_METHOD = Options.DevirtualizationType.CHA;
						} else if(RTA.equals(argument)) {
							Options.DEVIRTUALIZATION_METHOD = Options.DevirtualizationType.RTA;
						} else if(VTA.equals(argument)) {
							Options.DEVIRTUALIZATION_METHOD = Options.DevirtualizationType.VTA;
						} else if(RTAVTA.equals(argument)) {
							Options.DEVIRTUALIZATION_METHOD = Options.DevirtualizationType.RTAVTA;
						} else {
							System.err.println("The provided argument [" + argument + "] to --devirtualization is not valid. Use: [naive] or [cha].");
							System.exit(-1);
							return false;
						}
						return true;
					}
				}));
		
		options.add(new ArgumentOption(
				"--scheduler",
				"-s",
				"[fps|edf|fifo]",
				"Sets the scheduler used in the analysed system. FPS is used by default",
				new ArgRun() {
					
					private static final String FPS = "fps";
					private static final String EDF = "edf";
					private static final String FIFO = "fifo";
										
					@Override
					public boolean run(String argument) {
						if(FPS.equals(argument)) {
							Options.SCHEDULER_TYPE = SchedulerType.FPS;
						} else if(EDF.equals(argument)) {
							Options.SCHEDULER_TYPE = SchedulerType.EDF;
						} else if(FIFO.equals(argument)) {
							Options.SCHEDULER_TYPE = SchedulerType.FIFO;
						} else {
							return false;
						}
						return true;
					}
				}));

		options.add(new ArgumentOption(
				"--platform", 
				"-p", 
				"[jop|hvmavr]", 
				"Sets the platform on which the system is hosted. jop is default.", 
				new ArgRun() {
					
					private static final String JOP = "jop";
					private static final String HVMAVR = "hvmavr";
					
					@Override
					public boolean run(String argument) {
						if(JOP.equals(argument)) {
							Options.PLATFORM_TYPE = Options.PlatformType.JOP;
						} else if(HVMAVR.equals(argument)) {
							Options.PLATFORM_TYPE = Options.PlatformType.HVMAVR;
						} else {
							System.err.println("The provided argument [" + argument + "] to --platform is not valid. Use: [jop] or [hvmavr].");
							System.exit(-1);
							return false;
						}
						return true;
					}
				}));
		
		options.add(new BooleanOption(
				"--explicit-model", 
				"-e", 
				"Make an explicit resulting model i.e. do not include execution times on transitions, but have an explicit underlying platform model.\n Default is inlining as in the tradition SARTS approach", 
				new BoolRun() {
					public void run() {
						Options.EXPLICIT_MODEL = true;
					}
				})
		);
		
		options.add(new ArgumentOption(
				"--hw-clock-freq", 
				"-cf",
				"[0-9]*",
				"Sets the clock frequency of the hardware to be simulated", 
				new ArgRun() {
					
			
					@Override
					public boolean run(String argument) {
						try{
							Options.MHZ = Integer.parseInt(argument);
						}
						catch(NumberFormatException e) {
							System.err.println("Could not parse string provided to the -cf/--hw-clock-freq option");
							return false;
						}
						return true;
					}
				}));
		
		options.add(new BooleanOption(
				"--wcrtanalysis", 
				"-wcrt", 
				"Generate queries for determining the WCRTs of all tasks", 
				new BoolRun() {
					public void run() {
						Options.GENERATE_WCRT = true;
					}
				})
		);
		
		options.add(new BooleanOption(
				"--blockingtimeanalysis", 
				"-bta", 
				"Generates queries for determining the blocking time of all tasks.", 
				new BoolRun() {
					public void run() {
						Options.GENERATE_BLOCKING = true;
					}
				})
		);
		
		options.add(new BooleanOption(
				"--wcetanalysis", 
				"-wcet", 
				"Generates queries for determining the WCET of all tasks.", 
				new BoolRun() {
					public void run() {
						Options.GENERATE_WCET = true;
					}
				})
		);
		
		options.add(new BooleanOption(
				"--precise-loop", 
				"-lo", 
				"Use the loop identification algorithm from Dragon book", 
				new BoolRun() {
					public void run() {
						Options.PRECISE_LOOP = true;
					}
				})
		);
		
		options.add(new BooleanOption(
				"--print-system-classpath", 
				"-pscp", 
				"Print the full system classpath.", 
				new BoolRun() {
					public void run() {
						ClassLookup.printClassPath(ClassPath.SYSTEM_CLASS_PATH, "  ");
					}
				})
		);
		options.add(new BooleanOption(
				"--print-working-directory", 
				"-pwd", 
				"Print the working directory.", 
				new BoolRun() {
					public void run() {
						System.out.println("current dir: " + System.getProperty("user.dir"));
					}
				})
		);
		
		options.add(new BooleanOption(
				"--warn-athrow", 
				"-wa", 
				"Warn when visiting athrow instructions...", 
				new BoolRun() {
					public void run() {
						Options.WARN_ATHROW = true;
					}
				})
		);
		
		return options;
	}
	
	private static void setupLogger(String outDir) {
		Logger logger = Logger.getLogger("generator");
		FileHandler fh = null;
		try {
			fh = new FileHandler(outDir + "generator.log");
		} catch (IOException e) {
			System.err.println("Error opening log file! " + outDir + "generator.log");
			System.exit(-1);
		}
		fh.setFormatter(new SimpleFormatter());
		logger.addHandler(fh);
		sh = new StatisticHandler();
		logger.addHandler(sh);
	}
	
	public static void printClass(AbstractClassNode classNode, int indent, ClassGraph cg) {
		StringBuffer sb = new StringBuffer();
		sb.append('|');
		for(int i=0; i<indent-1; i++) {
			sb.append('-');
		}
		sb.append(classNode.getCanonicalName()).append(" ").append(cg.getSuccs(classNode).size());
		System.out.println(sb);
		for(AbstractClassNode c : cg.getSuccs(classNode)) {
			printClass(c, indent+2, cg);
		}
	}
}