package tetasarts;


import java.io.IOException;
import java.io.PrintStream;
import java.util.logging.FileHandler;
import java.util.logging.Logger;
import java.util.logging.SimpleFormatter;

import org.apache.bcel.util.ClassPath;

import tetasarts.analysis.Collapser;
import tetasarts.analysis.SimpleWcetVisitor;
import tetasarts.analysis.SyncMarker;
import tetasarts.generator.Constants;
import tetasarts.generator.JavaTranslator;
import tetasarts.tir.IntermediateData;
import tetasarts.tir.basicblocks.AbstractBasicBlock;
import tetasarts.tir.classinfo.AbstractClassNode;
import tetasarts.tir.methods.MethodInformation;
import tetasarts.uppaal.UppaalModel;
import tetasarts.uppaal.translation.analysis.IDevirtualizationAnalysis;
import tetasarts.uppaal.translation.analysis.NaiveDevirtualizationAnalysis;
import tetasarts.uppaal.translation.tasks.ConcreteMethodDecorator;

public class MethodWCET {
	private static StatisticHandler sh;

	/**
	 * @param args
	 */
	public static void main(String[] args) throws Exception{
		Constants.USE_WCET = true;
		Constants.DISABLE_TASKS = true;
		setupLogger();
		if("".equals(Options.classpath.toString())) {
			System.out.println("setting classpath to system classpath.");
			Options.classpath = ClassPath.SYSTEM_CLASS_PATH;
		}
		if (args.length == 4){
			String className = args[0];
			String methodName = args[1];
			int minWcet = Integer.parseInt(args[2]);
			int maxWcet = Integer.parseInt(args[3]);
			
			if (className.endsWith(".class")) {
				className = className.substring(0, className.length() - 6).replace('/', '.');
			} else {
				className = className.replace('/', '.');
			}
				
			//start, create autoflushed printstream and set out and err - for debug, a bit slower
			PrintStream newOut = new PrintStream(System.out, true);
			PrintStream newErr = new PrintStream(System.err, true);
			System.setOut(newOut);
			System.setErr(newErr);
			//end
			
			try {
				JavaTranslator jt = new JavaTranslator(className);
				IntermediateData result = jt.buildIR();
				
				// start collapsify
				Collapser.collapse(result);
				
				// start wcetify
				for(AbstractClassNode acn : result.getClassGraph().getNodes()){
					for(MethodInformation mi : acn.getMethods().values()){
//						System.out.println("Marking: "+ acn.getCanonicalName()  + mi.getName());
						if(mi.isGenerated()){
							SyncMarker.markGraph(mi.getInstructionGraph());
	//						System.out.println("Done Marking: "+ acn.getCanonicalName() + mi.getName());
	//						Mapper.Map(mi.getInstructionGraph());
							SimpleWcetVisitor sw = new SimpleWcetVisitor(mi);
							for(AbstractBasicBlock abb : mi.getInstructionGraph().getNodes()){
								abb.accept(sw);
							}
						}
					}
				}
				
				MethodInformation method = null;
				// start wcetify
				for(AbstractClassNode acn : result.getClassGraph().getNodes()){
					if (acn.getCanonicalName().equals(className)){
						for(MethodInformation mi : acn.getMethods().values()){
							if (mi.getName().equals(methodName) || (mi.getName() + mi.getSignature()).equals(methodName)){
								if (method != null){
									System.err.println("The selected method is not unique");
									System.exit(-1);
								}
								method = mi;
							}
						}
					}
				}
				if(method == null) {
					System.err.println("method not found!");
					System.exit(-1);
				}
				NaiveDevirtualizationAnalysis naiveDevirtualization = new NaiveDevirtualizationAnalysis(result.getClassGraph());
				ConcreteMethodDecorator cmd = new ConcreteMethodDecorator(naiveDevirtualization);
				cmd.decorateMethods(method, result.getClassGraph());
				
				// Start UPPAALify
				UppaalModel model = tetasarts.uppaal.translation.UppaalTranslation.generateUppaalModel(result);
				
				model.writeWcetModelToFile("/tmp/wcet.xml", method);
				model.writeWcetQueriesToFile("/tmp/wcet.q", method, minWcet, maxWcet);
			
//				sarts.uppaal.UppaalModel.correctXML("output.xml");
			} catch (Exception e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}
			
			System.out.flush();
			System.err.flush();
			System.out.println("Warnings and Errors: (see generator.log for details)");
			System.out.println(sh.getStats());
		} else {
			System.err.println("Invalid arguments, use <class> <method> <minWcet> <maxWcet>");
		}
	}

	private static void setupLogger() {
		Logger logger = Logger.getLogger("generator");
		FileHandler fh = null;
		try {
			fh = new FileHandler("generator.log");
		} catch (IOException e) {
			System.err.println("Error opening log file!");
			System.exit(-1);
		}
		fh.setFormatter(new SimpleFormatter());
		logger.addHandler(fh);
		sh = new StatisticHandler();
		logger.addHandler(sh);
	}
}





