package tetasarts.analysis;

import java.util.Collection;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

import tetasarts.exceptions.IncompatibleMergeException;
import tetasarts.exceptions.NodeNotFoundException;
import tetasarts.tir.IntermediateData;
import tetasarts.tir.basicblocks.*;
import tetasarts.tir.classinfo.AbstractClassNode;
import tetasarts.tir.methods.LazyMethodInformation;
import tetasarts.tir.methods.MethodInformation;

public class Collapser{

	public static void collapse(IntermediateData intermediate){
		new Collapser(intermediate);
	}

	private class CollapseVisitor extends AbstractBasicBlockVisitor{
		private int merges = 0;
		public int getMerges(){
			return merges;
		}
		public void resetMerges(){
			merges =0;
		}
		BasicBlockGraph basicBlockGraph;
		public CollapseVisitor(MethodInformation mi) {
			//if(mi.isGenerated()){
				BasicBlockGraph graph = basicBlockGraph  = mi.getInstructionGraph();
				nodesToVisit = new LinkedList<AbstractBasicBlock>();
				for(AbstractBasicBlock head : graph.getHeads()){
					nodesToVisit.add(head);
				}
				while(nodesToVisit.size()!=0){
					AbstractBasicBlock next = nodesToVisit.remove(0);
					next.accept(this);
				}
			//}
		}
		List<AbstractBasicBlock> nodesToVisit;
		
		Map<AbstractBasicBlock, Boolean> knownNodes = new HashMap<AbstractBasicBlock, Boolean>();
		
		@Override
		public void defaultVisit(AbstractBasicBlock s) {
			knownNodes.put(s, Boolean.TRUE);
			assert basicBlockGraph.getSuccs(s)!=null : "Successorlist is null!";
			for(AbstractBasicBlock succ : basicBlockGraph.getSuccs(s)){
				
				if(knownNodes.get(succ) == null)
					nodesToVisit.add(succ);
			}
			if(basicBlockGraph.getSuccs(s).size()==1){
				AbstractBasicBlock succ = basicBlockGraph.getSuccs(s).get(0);
				if(s.canMerge(succ)){
					try {
						AbstractBasicBlock newNode = basicBlockGraph.merge(s, succ);
						while(nodesToVisit.remove(succ));
						while(nodesToVisit.remove(s));
						nodesToVisit.add(0, newNode);
						merges++;
						return;
					} catch (NodeNotFoundException e) {
						e.printStackTrace();
					} catch (IncompatibleMergeException e) {
						e.printStackTrace();
					}
				} else {
				}
			}
		}
	}
	
	CollapseVisitor visitor;
	
	private Collapser(IntermediateData intermediate){
		Collection<AbstractClassNode> classNodes = intermediate.getClassGraph().getNodes();
		for(AbstractClassNode aclass : classNodes){
			for(MethodInformation methodInfo : aclass.getMethods().values()){
				//"Collapsing: "+ methodInfo.getName());
				int i = 0;
				do{
					i++;
					//"Pass  #" + i);
					visitor = new CollapseVisitor(methodInfo);
//					for(AbstractBasicBlock basicBlock : methodInfo.getInstructionGraph().getNodes()){
//						basicBlock.accept(visitor);
//					}
				}while(visitor.getMerges()!=0);
				//("Collapsing done: "+ methodInfo.getName());
			}
		}
	}
}
