package tetasarts.analysis;

import java.util.HashSet;
import java.util.Set;

import tetasarts.tir.basicblocks.AbstractBasicBlock;
import tetasarts.tir.basicblocks.BasicBlockGraph;
import tetasarts.tir.basicblocks.basicBlockTypes.MonitorEnterBasicBlock;
import tetasarts.tir.basicblocks.basicBlockTypes.MonitorExitBasicBlock;


public class SyncMarker{
	public static void markGraph(BasicBlockGraph graph){
		if(graph.getHeads().size()!=1){
		//	System.err.println("Graph does not have exactly one head: " + graph + " " + graph.getHeads().size());
			return;
		}
		new SyncMarker(graph);
	}
	
	private SyncMarker(BasicBlockGraph graph) {
		this.graph = graph;
		AbstractBasicBlock head = graph.getHeads().get(0);
		mark(head, 0);
	}
	private BasicBlockGraph graph;
	Set<AbstractBasicBlock> knownBlocks = new HashSet<AbstractBasicBlock>();
	
	private void mark(AbstractBasicBlock b, int level){
		if(startsSynchronized(b)){
			level++;
		} else if(endsSynchronized(b)){
			if (level > 0)
				level--;
		} else {
			if(level>0){
				b.addSuffix("SYNC"+level);
				b.setSynchronized(true);
			} else {
				b.setSynchronized(false);
			}
		}
		knownBlocks.add(b);
		for(AbstractBasicBlock succ : graph.getSuccs(b)){
			if(!knownBlocks.contains(succ))mark(succ, level);
		}
	}
	
	
	private boolean endsSynchronized(AbstractBasicBlock block){
		try {
			MonitorExitBasicBlock b = (MonitorExitBasicBlock)block;
		} catch (ClassCastException e) {
			return false;
		}
		return true;
	}
	private boolean startsSynchronized(AbstractBasicBlock block){
		try {
			MonitorEnterBasicBlock b = (MonitorEnterBasicBlock)block;
		} catch (ClassCastException e) {
			return false;
		}
		return true;
	}
}
