package tetasarts.generator;

import java.io.IOException;
import java.io.InputStream;
import java.io.InputStream;
import java.util.Iterator;
import java.util.Stack;
import java.util.logging.Logger;

import org.apache.bcel.Repository;
import org.apache.bcel.classfile.ClassFormatException;
import org.apache.bcel.classfile.ClassParser;
import org.apache.bcel.classfile.ConstantPool;
import org.apache.bcel.classfile.JavaClass;
import org.apache.bcel.classfile.Method;
import org.apache.bcel.generic.*;
import org.apache.bcel.util.ClassPath.ClassFile;

import tetasarts.ClassLookup;
import tetasarts.Options;
import tetasarts.tir.classinfo.*;


public class ClassGraphBuilder implements Visitor {
	private ClassGraph currentClassGraph;
	private Logger logger = Logger.getLogger("generator");
	private AbstractClassNode objectNode;
	private Method currentMethod;
	private Stack<Method> methodStack;
	
	public ClassGraphBuilder() {
		methodStack = new Stack<Method>();
	}
	
	public void buildClassGraph(String className, ClassGraph cg) {
		//System.out.println("buildClassGraph "+ className);
		currentClassGraph = cg;
		if(currentClassGraph.getClassFromCanonicalName(Constants.JAVA_LANG_OBJECT) == null) {
			try {
//				JavaClass c = new ClassParser("Object.class").parse();

//				System.exit(0);
				JavaClass c = ClassLookup.loadClass(Constants.JAVA_LANG_OBJECT);
//				try {
//					c = Repository.lookupClass(Constants.JAVA_LANG_OBJECT);
//				} catch (ClassNotFoundException e) {
//					// TODO Auto-generated catch block
//					e.printStackTrace();
//				}
//				System.out.println("loaded");
				buildClass(Constants.JAVA_LANG_OBJECT, c);
//				objectNode = currentClassGraph.getClassFromCanonicalName(Constants.JAVA_LANG_OBJECT);
			} catch (ClassFormatException e) {
				e.printStackTrace();
				System.exit(0);
			}
		} else {
			objectNode = currentClassGraph.getClassFromCanonicalName(Constants.JAVA_LANG_OBJECT);
		}
		buildClass(className);
	}
	
	private void buildClass(String className) {
		if(currentClassGraph.getClassFromCanonicalName(className) != null) {
			return;
		}
		JavaClass jc = ClassLookup.loadClass(className);
		if((jc.getPackageName().contains("java.") || jc.getPackageName().contains("javax.")) && !Options.ANALYSE_JAVA_CLASS_LIB) {
			AbstractClassNode targetNode = new NormalClass(className, currentClassGraph);
			addClass(targetNode, objectNode);
		} else {
			buildClass(className, jc);
		}
		
//		JavaClass jc = null;
//		try {
//			if(Options.FULL_CLASS_LOADING){
//				jc = Repository.lookupClass(className);
//			} else {
//				System.out.println("This class: " + className);
//				ClassFile classFile = Options.classpath.getClassFile(className);
//				jc = new ClassParser(classFile.getInputStream(), classFile.getPath()).parse();
//				
//				//jc = new ClassParser(classFile.getPath()).parse();
//			}
//			
////			System.out.println("looking up: "+className);
////			System.out.println("cn : " +jc.getClassName());
////			System.out.println("scn: " + jc.getSuperclassName());
//		} catch (ClassNotFoundException e) {
//			logger.severe("Error loading class: \"" + className + "\" - Creating dummy class. " + e.toString());
//			System.err.println("ERROR");
//			System.out.println("ERROR");
//			AbstractClassNode targetNode = new NormalClass(className, currentClassGraph);
//			addClass(targetNode, objectNode);
//			System.exit(0);
//			return;
//		} catch (IOException e) {
//			// TODO Auto-generated catch block
////			e.printStackTrace();
//			return;
//		}

//		System.out.println("buildclass: " + jc.getClassName());
		
	}
	
	private void addClass(AbstractClassNode targetNode, AbstractClassNode superClass) {
		if(!currentClassGraph.contains(targetNode)) {
			currentClassGraph.addNode(targetNode);
		}
		currentClassGraph.addEdge(superClass, targetNode);
	}
	
	private void buildClass(String className, JavaClass javaClass) {
		//System.out.println("building class: " + className);
		if(currentClassGraph.getClassFromCanonicalName(className) != null) {
			return;
		}
		AbstractClassNode targetNode = null;
		AbstractClassNode superNode = null;

		String superClassName = javaClass.getSuperclassName();
		if(superClassName.equals(Constants.JAVA_LANG_OBJECT)) {
			targetNode = new NormalClass(className, currentClassGraph);
			targetNode.setJavaClass(javaClass);
			superNode = objectNode;
		} else {
			JavaClass superClass = null;
			try{
//				ClassFile classFile = Options.classpath.getClassFile(superClassName);
//				JavaClass superClass = new ClassParser(classFile.getPath()).parse();
//				System.out.println("looking up superclass: "+superClassName);
				superClass = Repository.lookupClass(superClassName);
			} catch (Exception e) {
				e.printStackTrace();
				logger.severe("Error loading superclass of \"" + className + "\"! Assuming java.lang.Object");
				System.out.println(superClassName);
				System.out.println(className);
				System.out.println(currentMethod);
				System.out.println("classpath:");
				for(String s: Repository.getRepository().getClassPath().toString().split(":")){
					System.out.println("  "+s);
				}
				targetNode = new NormalClass(className, currentClassGraph);
				targetNode.setJavaClass(javaClass);
				superNode = objectNode;
				Thread.dumpStack();
				System.exit(0);
			}
			String currentSuperClassName = superClass.getClassName();
			buildClass(currentSuperClassName, superClass);
	
			superNode = currentClassGraph.getClassFromCanonicalName(currentSuperClassName);
	
			if(superNode instanceof SporadicThread || className.equals(Constants.SPORADIC_THREAD)) {
				// sporadic
				targetNode = new SporadicThread(className, currentClassGraph);
					
			} else if (superNode instanceof PeriodicThread || className.equals(Constants.PERIODIC_THREAD)) {
				// periodic
				targetNode = new PeriodicThread(className, currentClassGraph);
			} else {
				// normal
				targetNode = new NormalClass(className, currentClassGraph);
			}
			targetNode.setJavaClass(javaClass);
		}
		
		addClass(targetNode, superNode);

		visit(javaClass);

	}
	
	int visitcount = 0;
	
	int count =0 ;
	
	private void visit(JavaClass jc) {

		if(currentMethod != null)
			methodStack.push(currentMethod);
		Method[] methods = jc.getMethods();
		
		// PRINT callstack?
//		for(Method m:methodStack){
//			System.out.print(m.getName()+".");
//		}
//		System.out.println();
		for (Method current : methods) {
			currentMethod = current;
			if (currentMethod.isAbstract()) {
//				System.out.println("Abstract method");
				continue;
			}
			if (currentMethod.isNative()) {
//				System.out.println("NATIVE");
				continue;
			}
			

			InstructionList l = new InstructionList(currentMethod.getCode().getCode());
			l.setPositions(true);

			InstructionHandle handle = l.getStart();
			do {
//				System.out.println("handle: "+handle.getInstruction());
//				System.out.println("BH currentMethod: " + jc.getClassName()+"::"+ currentMethod.getName());
				handle.accept(this);
//				System.out.println("AH currentMethod: " + jc.getClassName()+"::"+ currentMethod.getName());
				handle = handle.getNext();
			} while (handle != null);
		}
		if(!methodStack.isEmpty())
			currentMethod = methodStack.pop();
		else {
			currentMethod = null;
		}
	}
	
	@Override
	public void visitLoadClass(LoadClass obj) {
//		System.out.println("visitLoadClass+ "+obj.getClass().getName());
		ConstantPool cp = currentMethod.getConstantPool();
//		System.out.println("current method: "+currentMethod.getName());
//		if(currentMethod.getName().equals("cloneArray")){
//			System.out.println("hest");
//		}
		CPInstruction inst = (CPInstruction)obj;
//		System.out.println(inst.getName());
		
		ConstantPoolGen cpg = new ConstantPoolGen(cp);
//		System.out.println(obj);
//		System.out.println(obj.getType(cpg));
		ObjectType objectType = obj.getLoadClassType(cpg);
		if(objectType == null){ // We dont care about primitive types at this point
			// NEW -- we are still asked to load primitive types..
//			System.out.println("This is a primitive type...");
//			System.out.println(obj.getType(cpg));
		} else {
			String classname = obj.getLoadClassType(cpg).getClassName();
			buildClass(classname);
		}
	}

	/******************************************************
	 * 
	 * The rest of the visit methods are unused.
	 * 
	 ******************************************************/

	@Override
	public void visitAALOAD(AALOAD obj) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitAASTORE(AASTORE obj) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitACONST_NULL(ACONST_NULL obj) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitALOAD(ALOAD obj) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitANEWARRAY(ANEWARRAY obj) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitARETURN(ARETURN obj) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitARRAYLENGTH(ARRAYLENGTH obj) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitASTORE(ASTORE obj) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitATHROW(ATHROW obj) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitAllocationInstruction(AllocationInstruction obj) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitArithmeticInstruction(ArithmeticInstruction obj) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitArrayInstruction(ArrayInstruction obj) {

	}

	@Override
	public void visitBALOAD(BALOAD obj) {

	}

	@Override
	public void visitBASTORE(BASTORE obj) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitBIPUSH(BIPUSH obj) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitBREAKPOINT(BREAKPOINT obj) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitBranchInstruction(BranchInstruction obj) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitCALOAD(CALOAD obj) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitCASTORE(CASTORE obj) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitCHECKCAST(CHECKCAST obj) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitCPInstruction(CPInstruction obj) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitConstantPushInstruction(ConstantPushInstruction obj) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitConversionInstruction(ConversionInstruction obj) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitD2F(D2F obj) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitD2I(D2I obj) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitD2L(D2L obj) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitDADD(DADD obj) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitDALOAD(DALOAD obj) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitDASTORE(DASTORE obj) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitDCMPG(DCMPG obj) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitDCMPL(DCMPL obj) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitDCONST(DCONST obj) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitDDIV(DDIV obj) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitDLOAD(DLOAD obj) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitDMUL(DMUL obj) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitDNEG(DNEG obj) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitDREM(DREM obj) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitDRETURN(DRETURN obj) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitDSTORE(DSTORE obj) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitDSUB(DSUB obj) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitDUP(DUP obj) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitDUP2(DUP2 obj) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitDUP2_X1(DUP2_X1 obj) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitDUP2_X2(DUP2_X2 obj) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitDUP_X1(DUP_X1 obj) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitDUP_X2(DUP_X2 obj) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitExceptionThrower(ExceptionThrower obj) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitF2D(F2D obj) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitF2I(F2I obj) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitF2L(F2L obj) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitFADD(FADD obj) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitFALOAD(FALOAD obj) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitFASTORE(FASTORE obj) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitFCMPG(FCMPG obj) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitFCMPL(FCMPL obj) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitFCONST(FCONST obj) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitFDIV(FDIV obj) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitFLOAD(FLOAD obj) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitFMUL(FMUL obj) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitFNEG(FNEG obj) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitFREM(FREM obj) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitFRETURN(FRETURN obj) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitFSTORE(FSTORE obj) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitFSUB(FSUB obj) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitFieldInstruction(FieldInstruction obj) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitFieldOrMethod(FieldOrMethod obj) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitGETFIELD(GETFIELD obj) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitGETSTATIC(GETSTATIC obj) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitGOTO(GOTO obj) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitGOTO_W(GOTO_W obj) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitGotoInstruction(GotoInstruction obj) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitI2B(I2B obj) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitI2C(I2C obj) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitI2D(I2D obj) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitI2F(I2F obj) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitI2L(I2L obj) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitI2S(I2S obj) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitIADD(IADD obj) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitIALOAD(IALOAD obj) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitIAND(IAND obj) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitIASTORE(IASTORE obj) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitICONST(ICONST obj) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitIDIV(IDIV obj) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitIFEQ(IFEQ obj) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitIFGE(IFGE obj) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitIFGT(IFGT obj) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitIFLE(IFLE obj) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitIFLT(IFLT obj) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitIFNE(IFNE obj) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitIFNONNULL(IFNONNULL obj) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitIFNULL(IFNULL obj) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitIF_ACMPEQ(IF_ACMPEQ obj) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitIF_ACMPNE(IF_ACMPNE obj) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitIF_ICMPEQ(IF_ICMPEQ obj) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitIF_ICMPGE(IF_ICMPGE obj) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitIF_ICMPGT(IF_ICMPGT obj) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitIF_ICMPLE(IF_ICMPLE obj) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitIF_ICMPLT(IF_ICMPLT obj) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitIF_ICMPNE(IF_ICMPNE obj) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitIINC(IINC obj) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitILOAD(ILOAD obj) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitIMPDEP1(IMPDEP1 obj) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitIMPDEP2(IMPDEP2 obj) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitIMUL(IMUL obj) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitINEG(INEG obj) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitINSTANCEOF(INSTANCEOF obj) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitINVOKEINTERFACE(INVOKEINTERFACE obj) {
		// TODO Auto-generated method stub
//		System.out.println("Visit: INVOKEINTERFACE"+obj.getClass().getName());
	}

	@Override
	public void visitINVOKESPECIAL(INVOKESPECIAL obj) {
		// TODO Auto-generated method stub
//		System.out.println("Visit: INVOKESPECIAL"+obj.getClass().getName());
	}

	@Override
	public void visitINVOKESTATIC(INVOKESTATIC obj) {
		// TODO Auto-generated method stub
//		System.out.println("Visit: INVOKESTATIC"+obj.getClass().getName());
	}

	@Override
	public void visitINVOKEVIRTUAL(INVOKEVIRTUAL obj) {
//		System.out.println("Visit: INVOKEVIRTUAL"+obj.getClass().getName());
	}

	@Override
	public void visitIOR(IOR obj) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitIREM(IREM obj) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitIRETURN(IRETURN obj) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitISHL(ISHL obj) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitISHR(ISHR obj) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitISTORE(ISTORE obj) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitISUB(ISUB obj) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitIUSHR(IUSHR obj) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitIXOR(IXOR obj) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitIfInstruction(IfInstruction obj) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitInvokeInstruction(InvokeInstruction obj) {
//		System.out.println("Visit: InvokeInstruction"+obj.getClass().getName());
	}

	@Override
	public void visitJSR(JSR obj) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitJSR_W(JSR_W obj) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitJsrInstruction(JsrInstruction obj) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitL2D(L2D obj) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitL2F(L2F obj) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitL2I(L2I obj) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitLADD(LADD obj) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitLALOAD(LALOAD obj) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitLAND(LAND obj) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitLASTORE(LASTORE obj) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitLCMP(LCMP obj) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitLCONST(LCONST obj) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitLDC(LDC obj) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitLDC2_W(LDC2_W obj) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitLDIV(LDIV obj) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitLLOAD(LLOAD obj) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitLMUL(LMUL obj) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitLNEG(LNEG obj) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitLOOKUPSWITCH(LOOKUPSWITCH obj) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitLOR(LOR obj) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitLREM(LREM obj) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitLRETURN(LRETURN obj) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitLSHL(LSHL obj) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitLSHR(LSHR obj) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitLSTORE(LSTORE obj) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitLSUB(LSUB obj) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitLUSHR(LUSHR obj) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitLXOR(LXOR obj) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitLoadInstruction(LoadInstruction obj) {
//		System.out.println("Visit: LoadInstruction"+obj.getClass().getName());
	}

	@Override
	public void visitLocalVariableInstruction(LocalVariableInstruction obj) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitMONITORENTER(MONITORENTER obj) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitMONITOREXIT(MONITOREXIT obj) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitMULTIANEWARRAY(MULTIANEWARRAY obj) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitNEW(NEW obj) {
//		System.out.println("Visit: NEW"+obj.getClass().getName());
	}

	@Override
	public void visitNEWARRAY(NEWARRAY obj) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitNOP(NOP obj) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitPOP(POP obj) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitPOP2(POP2 obj) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitPUTFIELD(PUTFIELD obj) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitPUTSTATIC(PUTSTATIC obj) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitPopInstruction(PopInstruction obj) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitPushInstruction(PushInstruction obj) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitRET(RET obj) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitRETURN(RETURN obj) {
//		System.out.println("Visit: RETURN"+obj.getClass().getName());
	}

	@Override
	public void visitReturnInstruction(ReturnInstruction obj) {
//		System.out.println("Visit: ReturnInstruction"+obj.getClass().getName());
	}

	@Override
	public void visitSALOAD(SALOAD obj) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitSASTORE(SASTORE obj) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitSIPUSH(SIPUSH obj) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitSWAP(SWAP obj) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitSelect(Select obj) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitStackConsumer(StackConsumer obj) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitStackInstruction(StackInstruction obj) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitStackProducer(StackProducer obj) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitStoreInstruction(StoreInstruction obj) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitTABLESWITCH(TABLESWITCH obj) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitTypedInstruction(TypedInstruction obj) {
//		System.out.println("Visit: TypedInstruction"+obj.getClass().getName());
	}

	@Override
	public void visitUnconditionalBranch(UnconditionalBranch obj) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitVariableLengthInstruction(VariableLengthInstruction obj) {
		// TODO Auto-generated method stub

	}

}
