package tetasarts.generator;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;
import java.util.logging.Logger;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.apache.bcel.Repository;
import org.apache.bcel.classfile.Attribute;
import org.apache.bcel.classfile.ClassFormatException;
import org.apache.bcel.classfile.ClassParser;
import org.apache.bcel.classfile.CodeException;
import org.apache.bcel.classfile.JavaClass;
import org.apache.bcel.classfile.LineNumberTable;
import org.apache.bcel.classfile.Method;
import org.apache.bcel.generic.*;
import org.apache.bcel.util.ClassPath.ClassFile;

import tetasarts.ClassLookup;
import tetasarts.Options;
import tetasarts.tir.IntermediateData;
import tetasarts.tir.basicblocks.AbstractBasicBlock;
import tetasarts.tir.basicblocks.BasicBlockGraph;
import tetasarts.tir.basicblocks.LoopBound;
import tetasarts.tir.basicblocks.basicBlockTypes.IfBasicBlock;
import tetasarts.tir.basicblocks.basicBlockTypes.LoopBasicBlock;
import tetasarts.tir.basicblocks.basicBlockTypes.MethodCallingBasicBlock;
import tetasarts.tir.basicblocks.basicBlockTypes.MonitorEnterBasicBlock;
import tetasarts.tir.basicblocks.basicBlockTypes.MonitorExitBasicBlock;
import tetasarts.tir.basicblocks.basicBlockTypes.NewBasicBlock;
import tetasarts.tir.basicblocks.basicBlockTypes.SimpleBasicBlock;
import tetasarts.tir.basicblocks.basicBlockTypes.SpecialInvokeBasicBlock;
import tetasarts.tir.basicblocks.basicBlockTypes.SporadicInvokeBasicBlock;
import tetasarts.tir.basicblocks.basicBlockTypes.SwitchBasicBlock;
import tetasarts.tir.classinfo.AbstractClassNode;
import tetasarts.tir.methods.MethodInformation;
import tetasarts.uppaal.translation.analysis.CHASootDevirtualizationAnalysis;
import tetasarts.uppaal.translation.analysis.CallSite;
import tetasarts.uppaal.translation.analysis.Callee;
import tetasarts.uppaal.translation.analysis.IDevirtualizationAnalysis;
import tetasarts.uppaal.translation.analysis.NaiveDevirtualizationAnalysis;

public class LocationBuilder implements Visitor {

	protected Map<Integer, String> currentLineMap = new HashMap<Integer, String>();
	protected Method currentMethod;
	protected JavaClass currentClass;
	protected MethodInformation currentMethodTarget;
	protected BasicBlockGraph currentInstructionGraph;
	protected HashSet<String> knownClasses;
	protected IntermediateData intermediateData;
	protected Map<Integer, AbstractBasicBlock> instructionNodeMap;
	protected JavaTranslator javaTranslator;
	protected InstructionHandle currentHandle;
	protected Map<AbstractBasicBlock, LoopBound> loopBoundMap;
	protected Logger logger;
	protected Map<Integer, CodeException> exceptionMap;
	protected boolean inExceptionHandler;
	
	protected IDevirtualizationAnalysis devirtualizer;

	public void setCurrentClass(JavaClass jc) {
		currentClass = jc;
		loadLineMap(jc);
	}
	public JavaClass getCurrentClass() {
		return currentClass;
	}
	
	public void setCurrentMethodTarget(MethodInformation mi) {
		currentMethodTarget = mi;
	}
	public MethodInformation getCurrentMethodTarget() {
		return currentMethodTarget;
	}

	protected int getCurrentSourceLineNumber() {
		if(currentMethod.getLineNumberTable()== null) // some methods have no line-number table...
			return 0;
		assert currentMethod != null : "currentmethod is null!";
		assert currentHandle != null : "No current handle -- null";
//		assert currentMethod.getLineNumberTable() != null : "No line-number table!";
		LineNumberTable tab = currentMethod.getLineNumberTable();
		return tab.getSourceLine(currentHandle.getPosition());
	}

	protected String getCurrentSourceLine() {
		return currentLineMap.get(new Integer(getCurrentSourceLineNumber()) - 1);
	}
	protected String getSourceLine(int offset) {
		return currentLineMap.get(new Integer(getCurrentSourceLineNumber()) - 1 - offset);
	}


	public LocationBuilder(JavaTranslator jt) {
		javaTranslator = jt;
		knownClasses = jt.getKnownClasses();
		intermediateData = jt.getIntermediateData();
		

		
		logger = Logger.getLogger("generator");
	}

	public void visitMethod(Method method, MethodInformation methodTarget) {
		currentMethod = method;
		exceptionMap = new HashMap<Integer, CodeException>();
		if(currentMethod.getCode().getExceptionTable() != null) {
			for(CodeException ce : currentMethod.getCode().getExceptionTable()) {
				exceptionMap.put(ce.getHandlerPC(), ce);
			}
		}
		inExceptionHandler = false;
		
		instructionNodeMap = javaTranslator.instructionNodeMap;
		loopBoundMap = new HashMap<AbstractBasicBlock, LoopBound>();

		currentMethodTarget = methodTarget;
//		currentInstructionGraph = currentMethodTarget.getInstructionGraph();
		currentInstructionGraph = currentMethodTarget.instructionGraph;

		
		InstructionList l = new InstructionList(currentMethod.getCode().getCode());
		l.setPositions(true);

		InstructionHandle handle = l.getStart();
		do {
//			if(handle.getPosition()==824){
//				System.out.println("hmm "+ handle.getClass());
//				System.exit(0);
//			}

			visit(handle);

			handle = handle.getNext();
		} while (handle != null);
	}
	
	public void loadLineMap(JavaClass jc) {
		String classname = jc.getClassName();
		String packageName = "".equals(jc.getPackageName()) ? "" : jc.getPackageName() + File.separator;
		String sourceFile;
			//TODO MUST BE FIXED!!
//			System.out.println("classnamelookup: "+Repository.lookupClass(classname).getFileName());
			//sourceFile = Repository.lookupClass(classname).getSourceFileName();
//		try {
			ClassFile classFile = ClassLookup.getClassFile(classname);
			sourceFile = classFile.getPath().replace(File.separator + "bin" + File.separator, File.separator + "src" + File.separator).replace(".class", ".java");
//			
//		} catch (IOException e) {
//			sourceFile = "src" + File.separator + packageName.replace('.', File.separatorChar) + jc.getSourceFileName();
//			e.printStackTrace();
//		}
			
			
//			sourceFile = Options.classpath.getClassFile(classname).getPath().replace(File.separator + "bin" + File.separator, File.separator + "src" + File.separator).replace(".class", ".java");
//		} catch (IOException e1) {
//			sourceFile = "src" + File.separator + packageName.replace('.', File.separatorChar) + jc.getSourceFileName();
//			System.out.println("Setting Source file to: "+ sourceFile);
//			e1.printStackTrace();
		//} catch (ClassNotFoundException e) {
			// TODO Auto-generated catch block
			//sourceFile = "src" + File.separator + packageName.replace('.', File.separatorChar) + jc.getSourceFileName();
			//e.printStackTrace();
		//}
		
//		String classname = jc.getClassName();
//		String packageName = "".equals(jc.getPackageName()) ? "" : jc.getPackageName() + File.separator;
//		String sourceFile = "src" + File.separator + packageName.replace('.', File.separatorChar) + jc.getSourceFileName();
		Map<Integer, String> lineMap = new HashMap<Integer, String>();
		try {
			FileReader fileReader = new FileReader(sourceFile);
			BufferedReader r = new BufferedReader(fileReader);
			int i = 0;
			String line;
			while ((line = r.readLine()) != null) {
				lineMap.put(i++, line);
			}
		} catch (FileNotFoundException e) {
			logger.info("Unable to find source code for class " + classname + " - filename: " + sourceFile);
		} catch (IOException e) {
			logger.info("Error while reading source code for class " + classname + " - filename: " + sourceFile);
		}
		this.currentLineMap = lineMap;
	}

	public void visit(InstructionHandle iHand) {
		currentHandle = iHand;
		//TODO: It seems the below returns messes some things up... 
		if(exceptionMap.containsKey(currentHandle.getPosition()) && !currentHandle.hasTargeters()) {
//			System.out.println("Has no tageters?" + currentHandle.getPosition());
			inExceptionHandler = true;
//			System.out.println("exceptionMap: "+iHand);
			return;
		}
		
		if(inExceptionHandler) {
			if(!currentHandle.hasTargeters() && 
					   !(currentHandle.getInstruction() instanceof ATHROW) &&
			   !(currentHandle.getInstruction() instanceof ReturnInstruction)) {
//				System.out.println("No Target: "+iHand);
				return;
			}
			inExceptionHandler = false;
//			System.out.println("Exceptionhandler exit: "+ iHand);
		}
		
		iHand.accept(this);
	}
	
	protected void createBasicBlock(Instruction obj) {
		
		AbstractBasicBlock bb = null;

		if(Options.USE_JOP_OPCODE_METHODS && (Constants.JAVA_OPCODE_METHOD.containsKey(obj.getOpcode()))) {
			currentHandle.setInstruction(new INVOKESTATIC(1));
			CallSite callSite = new CallSite(currentClass.getClassName(), currentMethod, currentHandle);
			Callee callee = new  Callee(Constants.JOP_JVM_CLASS,
					Constants.JAVA_OPCODE_METHOD.get(obj.getOpcode()), 
					Constants.JAVA_OPCODE_SIGNATURE.get(obj.getOpcode()));
			
			bb = new MethodCallingBasicBlock(callSite, callee);
			
			/*bb = new MethodCallingBasicBlock(
					currentClass.getClassName(),
					currentMethod.getName(),
					currentMethod.getSignature(),
					Constants.JOP_JVM_CLASS,
					Constants.JAVA_OPCODE_METHOD.get(obj.getOpcode()), 
					Constants.JAVA_OPCODE_SIGNATURE.get(obj.getOpcode()));
			currentHandle.setInstruction(new INVOKESTATIC(1));
			*/
		} else {
			bb = new SimpleBasicBlock();
		}

		bb.addBytecode(currentHandle);
		currentInstructionGraph.addNode(bb);
		instructionNodeMap.put(new Integer(currentHandle.getPosition()), bb);
	}

	
	public void visitATHROW(ATHROW obj) {

		// OMG
		if(Options.WARN_ATHROW)
			logger.severe("Instruction ATHROW not supported! method: " + currentMethodTarget.getCanonicalName() + " inst: " + currentHandle);
//		createBasicBlock(obj);
		//System.exit(-1);
	}

	
	public void visitBREAKPOINT(BREAKPOINT obj) {

		logger.severe("Breakpoint instruction not implemented on underlying hardware! " + obj);
		createBasicBlock(obj);
	}

	
	public void visitBranchInstruction(BranchInstruction obj) {
		AbstractBasicBlock bb = null;
		// OMG (GotoInstruction, IfInstruction, JsrInstruction, Select)
//		System.out.println("branch");
		
		// OMGOMG JsrInstruction
		
		if(obj instanceof Select) {
			
			Select selIns = (Select)obj;
			
			//bb = new IfBasicBlock();
			createBasicBlock(obj);
//			System.err.println("LocationBuilder: Select instructions not supported! " + obj);
//			System.exit(-1);
			return;
		}
		
		if(!Options.PRECISE_LOOP && obj.getIndex() < 0) { // backward jump = Loop

			AbstractBasicBlock target = instructionNodeMap.get(obj.getTarget().getPosition());
			//System.out.println("backward jump");
			assert target != null : "Target of branchinstruction is null! Method: " + currentMethodTarget.getCanonicalName() + " instr: " + currentHandle;
			
			LoopBound lb = loopBoundMap.get(target);
			if(lb == null) {
				int loopBound = getLoopBound(getSourceLine(1));
				lb = new LoopBound(target, loopBound);
				target.setTarget(true);
				loopBoundMap.put(target, lb);
			}
			
			bb = new LoopBasicBlock(lb);
			lb.addLoopNode((LoopBasicBlock)bb);
		}		
		else {
//			System.out.println("if!");
//			System.out.println(currentHandle.toString(true));
			bb = new IfBasicBlock();
		}
		
		bb.addBytecode(currentHandle);
		currentInstructionGraph.addNode(bb);
		instructionNodeMap.put(currentHandle.getPosition(), bb);
//		if(obj instanceof Select) {
//			System.out.println("Select instructions not supported! " + obj);
//			throw new RuntimeException("Select instructions not supported! " + obj);
//		}

	}
	
	private Pattern loopBoundPattern = Pattern.compile("([0-9]+)");
	protected int getLoopBound(String sourceLine) {
        int loopBound = -1;
        if(sourceLine == null) {
                logger.info("No sourceline for " + currentMethodTarget.getCanonicalName() + " on line " + getCurrentSourceLineNumber());
                return -1;
        }
        try {
                int i = sourceLine.indexOf("@loopbound");
                if (i == -1) {
                        throw new Exception();
                }

                Matcher pMatcher = loopBoundPattern.matcher(sourceLine);
				while(pMatcher.find()) {
					loopBound = Integer.parseInt(pMatcher.group());
					return loopBound;
				}
                /*
                sourceLine = sourceLine.substring(i + 4).trim();

                if (!sourceLine.startsWith("loop")) {
                        throw new Exception();
                }

                sourceLine = sourceLine.substring(4).trim();

                if (!sourceLine.startsWith("=")) {
                        throw new Exception();
                }

                sourceLine = sourceLine.substring(1).trim();

                loopBound = Integer.parseInt(sourceLine);
                */
        } catch (Exception e) {
                logger.info("Syntax error in loop bound in file " + currentClass.getFileName() + " on line "
                                + getCurrentSourceLineNumber() + 
                                ". Supported syntax is: // @loopbound = <integer>");
        }
        return loopBound;
	}
	
	
	public void visitCPInstruction(CPInstruction obj) {

//	    ANEWARRAY, CHECKCAST, FieldOrMethod, INSTANCEOF, LDC, LDC2_W, MULTIANEWARRAY, NEW
		// OMG exception thrower. see http://jakarta.apache.org/bcel/apidocs/org/apache/bcel/generic/package-tree.html
		if(!(obj instanceof InvokeInstruction) && !(obj instanceof NEW))
			createBasicBlock(obj);
	}

	
	public void visitInvokeInstruction(InvokeInstruction obj) {
		// goto more specific instructions, special, static and invokevirtual
	}

	
	public void visitMONITORENTER(MONITORENTER obj) {
		// OMG exception thrower. see http://jakarta.apache.org/bcel/apidocs/org/apache/bcel/generic/package-tree.html
		MonitorEnterBasicBlock mebb = new MonitorEnterBasicBlock();
		mebb.addBytecode(currentHandle);
		
		currentInstructionGraph.addNode(mebb);
		instructionNodeMap.put(currentHandle.getPosition(), mebb);
	}

	
	public void visitMONITOREXIT(MONITOREXIT obj) {
		// OMG exception thrower. see http://jakarta.apache.org/bcel/apidocs/org/apache/bcel/generic/package-tree.html
		MonitorExitBasicBlock mebb = new MonitorExitBasicBlock();
		mebb.addBytecode(currentHandle);
		
		currentInstructionGraph.addNode(mebb);
		instructionNodeMap.put(currentHandle.getPosition(), mebb);
	}

	
	public void visitNEWARRAY(NEWARRAY obj) {
		// OMG exception thrower. see http://jakarta.apache.org/bcel/apidocs/org/apache/bcel/generic/package-tree.html

		createBasicBlock(obj);
	}

	
	public void visitSelect(Select obj) {
//		System.out.println("Select");
//		System.out.println("Position: "+currentHandle.getPosition());
//	    System.out.println("Post instruction: " + obj.getTarget());
//	    System.out.println("nested instructions: ");
//		for(InstructionHandle ih: obj.getTargets()){
//			System.out.println("  " + ih.getPosition());
//		}
		// TODO fix! implemented in Java
		AbstractBasicBlock bb = null;
//		System.out.println(currentHandle.toString(true));
		bb = new SwitchBasicBlock();
	
		bb.addBytecode(currentHandle);
		currentInstructionGraph.addNode(bb);
		instructionNodeMap.put(currentHandle.getPosition(), bb);
//		throw new RuntimeException("quit");
//		System.exit(0);
//		AbstractBasicBlock bb = new MethodCallingBasicBlock(
//				currentClass.getClassName(),
//				currentMethod.getName(),
//				currentMethod.getSignature(),
//				Constants.JOP_JVM_CLASS,
//				Constants.JAVA_OPCODE_METHOD.get(obj.getOpcode()), 
//				Constants.JAVA_OPCODE_SIGNATURE.get(obj.getOpcode()));
//
////		currentHandle.setInstruction(new INVOKESTATIC(1));
////		InstructionHandle g = currentHandle.getInstructionHandle(new INVOKESTATIC(1));
//		
//		bb.addBytecode(currentHandle);
//		currentInstructionGraph.addNode(bb);
//		instructionNodeMap.put(new Integer(currentHandle.getPosition()), bb);
	}

	
	public void visitIMPDEP1(IMPDEP1 obj) {

		// OMG JOP specific instruction implemented in Java. See JOP decumentation
		createBasicBlock(obj);
	}

	
	public void visitIMPDEP2(IMPDEP2 obj) {

		// OMG
		logger.severe("IMPDEP2 not supported in underlying hardware!");
		createBasicBlock(obj);
	}

	
	public void visitINVOKEINTERFACE(INVOKEINTERFACE obj) {

		// TODO Implement this
//		logger.severe("INVOKEINTERFACE not supported in translator!");
//		System.exit(-1);
		ConstantPoolGen cpg = new ConstantPoolGen(currentMethod.getConstantPool());
		String calledClassName = obj.getReferenceType(cpg).toString();
		AbstractClassNode calledClass = javaTranslator.classGraph.getClassFromCanonicalName(calledClassName);
		String calledMethodName = obj.getMethodName(cpg);
		String calledMethodSignature = obj.getSignature(cpg);

		if(Constants.INVOKEINTERFACEDEBUG){
			System.out.println("calledClassName      : " + calledClassName);
			System.out.println("calledMethodName     : " + calledMethodName);
			System.out.println("calledMethodSignature: " + calledMethodSignature);
			
			System.out.println("currentMethod: "+ currentMethod.getName());
			System.out.println("currentClass : " + currentClass.getClassName());
			//System.exit(0);		
		}
		
		AbstractBasicBlock bb = null;
		if(calledClassName.equals(Constants.JOP_NATIVE)) {
			bb = new SimpleBasicBlock(); // the correct native instruction is found in TransitionBuilder
		} else {
			if (calledClassName.equals(Constants.REALTIME_SYSTEM) && 
					calledMethodName.equals("fire") &&
					calledMethodSignature.equals("(I)V")){
				String currentLine = getCurrentSourceLine();
				try {
					currentLine = currentLine.substring(currentLine.indexOf(Constants.REALTIME_SYSTEM_FIRE + "(") + 
							(Constants.REALTIME_SYSTEM_FIRE + "(").length());
					currentLine = currentLine.substring(0, currentLine.indexOf(")"));
					
					CallSite callSite = new CallSite(currentClass.getClassName(), currentMethod, currentHandle);

					bb = new SporadicInvokeBasicBlock(
							callSite, 
							Integer.parseInt(currentLine));
					
					/*bb = new SporadicInvokeBasicBlock(
							currentClass.getClassName(),
							currentMethod.getName(),
							currentMethod.getSignature(),
							calledClass.getCanonicalName(),
							calledMethodName, 
							calledMethodSignature,
							Integer.parseInt(currentLine));
							*/
				} catch (Exception e) {
					logger.severe("Error retrieving correct sporadic task. In file: " + 
							currentClass.getSourceFileName() + " on line: "
							+ getCurrentSourceLineNumber() + ". " + 
							" Supported syntax:" +
							" RealtimeSystem.fire(<int_literal>);");
					
					e.printStackTrace();
	
					System.exit(-1);
				}
			} else {
				CallSite callSite = new CallSite(currentClass.getClassName(), currentMethod, currentHandle);
				
				bb = new MethodCallingBasicBlock(callSite);
			}
		}
		
		bb.addBytecode(currentHandle);
		currentInstructionGraph.addNode(bb);
		instructionNodeMap.put(new Integer(currentHandle.getPosition()), bb);
	}

	
	public void visitRET(RET obj) {

		createBasicBlock(obj);
		// OMG returnBasicBlock
	}

	
	public void visitReturnInstruction(ReturnInstruction obj) {
		// OMG exception thrower. see http://jakarta.apache.org/bcel/apidocs/org/apache/bcel/generic/package-tree.html

		createBasicBlock(obj);
		// OMG returnBasicBlock
	}
	
	
	public void visitARRAYLENGTH(ARRAYLENGTH obj) {

		// OMG exception thrower. see http://jakarta.apache.org/bcel/apidocs/org/apache/bcel/generic/package-tree.html
		createBasicBlock(obj);
	}

	
	public void visitArithmeticInstruction(ArithmeticInstruction obj) {

		// OMG exception thrower. see http://jakarta.apache.org/bcel/apidocs/org/apache/bcel/generic/package-tree.html
		createBasicBlock(obj);
	}

	
	public void visitArrayInstruction(ArrayInstruction obj) {

		// OMG exception thrower. see http://jakarta.apache.org/bcel/apidocs/org/apache/bcel/generic/package-tree.html
		createBasicBlock(obj);
	}

	/************************
	 * 
	 * Simple visit methods
	 * 
	 ************************/

	
	public void visitStackInstruction(StackInstruction obj) {

		createBasicBlock(obj);
	}
	
	
	public void visitNOP(NOP obj) {

		createBasicBlock(obj);
	}

	
	public void visitSIPUSH(SIPUSH obj) {

		createBasicBlock(obj);
	}

	
	public void visitLCMP(LCMP obj) {

		createBasicBlock(obj);
	}

	
	public void visitLCONST(LCONST obj) {

		createBasicBlock(obj);
	}

	
	public void visitLocalVariableInstruction(LocalVariableInstruction obj) {

		createBasicBlock(obj);
	}

	
	public void visitConversionInstruction(ConversionInstruction obj) {

		createBasicBlock(obj);
	}

	
	public void visitDCMPG(DCMPG obj) {

		createBasicBlock(obj);
	}

	
	public void visitDCMPL(DCMPL obj) {

		createBasicBlock(obj);
	}

	
	public void visitDCONST(DCONST obj) {

		createBasicBlock(obj);
	}

	
	public void visitFCMPG(FCMPG obj) {

		createBasicBlock(obj);
	}

	
	public void visitFCMPL(FCMPL obj) {

		createBasicBlock(obj);
	}

	
	public void visitFCONST(FCONST obj) {

		createBasicBlock(obj);
	}

	
	public void visitICONST(ICONST obj) {

		createBasicBlock(obj);
	}

	
	public void visitBIPUSH(BIPUSH obj) {

		createBasicBlock(obj);
	}

	
	public void visitACONST_NULL(ACONST_NULL obj) {

		createBasicBlock(obj);
	}

	/******************************************************
	 * 
	 * The rest of the visit methods are unused.
	 * 
	 ******************************************************/

	
	public void visitStackProducer(StackProducer obj) {


	}

	
	public void visitStoreInstruction(StoreInstruction obj) {


	}

	
	public void visitTABLESWITCH(TABLESWITCH obj) {
//		System.err.println("TABLESWITCH");
	}

	
	public void visitTypedInstruction(TypedInstruction obj) {


	}

	
	public void visitUnconditionalBranch(UnconditionalBranch obj) {


	}

	
	public void visitVariableLengthInstruction(VariableLengthInstruction obj) {


	}

	
	public void visitAALOAD(AALOAD obj) {

		
	}

	
	public void visitAASTORE(AASTORE obj) {


	}

	
	public void visitALOAD(ALOAD obj) {


	}

	
	public void visitANEWARRAY(ANEWARRAY obj) {


	}

	
	public void visitARETURN(ARETURN obj) {


	}

	
	public void visitASTORE(ASTORE obj) {


	}

	
	public void visitAllocationInstruction(AllocationInstruction obj) {


	}

	
	public void visitBALOAD(BALOAD obj) {


	}

	
	public void visitBASTORE(BASTORE obj) {


	}

	
	public void visitCALOAD(CALOAD obj) {


	}

	
	public void visitCASTORE(CASTORE obj) {


	}

	
	public void visitCHECKCAST(CHECKCAST obj) {


	}

	
	public void visitConstantPushInstruction(ConstantPushInstruction obj) {


	}

	
	public void visitD2F(D2F obj) {


	}

	
	public void visitD2I(D2I obj) {


	}

	
	public void visitD2L(D2L obj) {


	}

	
	public void visitDADD(DADD obj) {


	}

	
	public void visitDALOAD(DALOAD obj) {


	}

	
	public void visitDASTORE(DASTORE obj) {


	}

	
	public void visitDDIV(DDIV obj) {


	}

	
	public void visitDLOAD(DLOAD obj) {


	}

	
	public void visitDMUL(DMUL obj) {


	}

	
	public void visitDNEG(DNEG obj) {


	}

	
	public void visitDREM(DREM obj) {


	}

	
	public void visitDRETURN(DRETURN obj) {


	}

	
	public void visitDSTORE(DSTORE obj) {


	}

	
	public void visitDSUB(DSUB obj) {


	}

	
	public void visitDUP(DUP obj) {


	}

	
	public void visitDUP2(DUP2 obj) {


	}

	
	public void visitDUP2_X1(DUP2_X1 obj) {


	}

	
	public void visitDUP2_X2(DUP2_X2 obj) {


	}

	
	public void visitDUP_X1(DUP_X1 obj) {


	}

	
	public void visitDUP_X2(DUP_X2 obj) {


	}

	
	public void visitExceptionThrower(ExceptionThrower obj) {


	}

	
	public void visitF2D(F2D obj) {


	}

	
	public void visitF2I(F2I obj) {


	}

	
	public void visitF2L(F2L obj) {


	}

	
	public void visitFADD(FADD obj) {


	}

	
	public void visitFALOAD(FALOAD obj) {


	}

	
	public void visitFASTORE(FASTORE obj) {


	}

	
	public void visitFDIV(FDIV obj) {


	}

	
	public void visitFLOAD(FLOAD obj) {


	}

	
	public void visitFMUL(FMUL obj) {


	}

	
	public void visitFNEG(FNEG obj) {


	}

	
	public void visitFREM(FREM obj) {


	}

	
	public void visitFRETURN(FRETURN obj) {


	}

	
	public void visitFSTORE(FSTORE obj) {


	}

	
	public void visitFSUB(FSUB obj) {


	}

	
	public void visitFieldInstruction(FieldInstruction obj) {


	}

	
	public void visitFieldOrMethod(FieldOrMethod obj) {


	}

	
	public void visitGETFIELD(GETFIELD obj) {


	}

	
	public void visitGETSTATIC(GETSTATIC obj) {


	}

	
	public void visitGOTO(GOTO obj) {


	}

	
	public void visitGOTO_W(GOTO_W obj) {


	}

	
	public void visitGotoInstruction(GotoInstruction obj) {


	}

	
	public void visitI2B(I2B obj) {


	}

	
	public void visitI2C(I2C obj) {


	}

	
	public void visitI2D(I2D obj) {


	}

	
	public void visitI2F(I2F obj) {


	}

	
	public void visitI2L(I2L obj) {


	}

	
	public void visitI2S(I2S obj) {


	}

	
	public void visitIADD(IADD obj) {


	}

	
	public void visitIALOAD(IALOAD obj) {


	}

	
	public void visitIAND(IAND obj) {


	}

	
	public void visitIASTORE(IASTORE obj) {


	}

	
	public void visitIDIV(IDIV obj) {


	}

	
	public void visitIFEQ(IFEQ obj) {


	}

	
	public void visitIFGE(IFGE obj) {


	}

	
	public void visitIFGT(IFGT obj) {


	}

	
	public void visitIFLE(IFLE obj) {


	}

	
	public void visitIFLT(IFLT obj) {


	}

	
	public void visitIFNE(IFNE obj) {


	}

	
	public void visitIFNONNULL(IFNONNULL obj) {


	}

	
	public void visitIFNULL(IFNULL obj) {


	}

	
	public void visitIF_ACMPEQ(IF_ACMPEQ obj) {


	}

	
	public void visitIF_ACMPNE(IF_ACMPNE obj) {


	}

	
	public void visitIF_ICMPEQ(IF_ICMPEQ obj) {


	}

	
	public void visitIF_ICMPGE(IF_ICMPGE obj) {


	}

	
	public void visitIF_ICMPGT(IF_ICMPGT obj) {


	}

	
	public void visitIF_ICMPLE(IF_ICMPLE obj) {


	}

	
	public void visitIF_ICMPLT(IF_ICMPLT obj) {


	}

	
	public void visitIF_ICMPNE(IF_ICMPNE obj) {


	}

	
	public void visitIINC(IINC obj) {


	}

	
	public void visitILOAD(ILOAD obj) {


	}

	
	public void visitIMUL(IMUL obj) {


	}

	
	public void visitINEG(INEG obj) {


	}

	
	public void visitINSTANCEOF(INSTANCEOF obj) {


	}

	
	public void visitINVOKESPECIAL(INVOKESPECIAL obj) {

		// TODO This is a marker for visitINVOKESPECIAL - fix this
		
		ConstantPoolGen cpg = new ConstantPoolGen(currentMethod.getConstantPool());
		String calledClassName = obj.getReferenceType(cpg).toString();
//		System.out.println("looking for calledClassName: "+calledClassName);
		int classindex= cpg.lookupClass(calledClassName);
//		System.out.println(cpg.getConstant(classindex).getClass().getCanonicalName());
//		System.out.println(cpg.getConstant(classindex));
//		try {
//			JavaClass jc = Repository.lookupClass(calledClassName);
////			System.out.println("attributes");
////			for(Attribute a : jc.getAttributes()){
////				System.out.println("  "+a);
////			}
////			System.out.println("interfaces: ");
////			for(String s : jc.getInterfaceNames()){
////				System.out.println("  "+s);
////			}
//		} catch (ClassNotFoundException e1) {
//			// TODO Auto-generated catch block
//			e1.printStackTrace();
//			System.exit(0);
//		}
		
		String calledMethodName = obj.getMethodName(cpg);
//		System.out.println("calledMethod               : "+calledMethodName);
		String calledMethodSignature = obj.getSignature(cpg);
//		System.out.println("calledMethodSignature      : "+calledMethodSignature);
		AbstractClassNode calledClass = javaTranslator.classGraph.getClassFromCanonicalName(calledClassName);
		
		AbstractBasicBlock bb = null;
		if(calledClassName.equals(Constants.JOP_NATIVE)) {
			bb = new SimpleBasicBlock(); // the correct native instruction is found in TransitionBuilder
		} else {
			
			CallSite callSite = new CallSite(currentClass.getClassName(), currentMethod, currentHandle);
			
			if (calledClassName.equals(Constants.REALTIME_SYSTEM) && 
					calledMethodName.equals("fire") &&
					calledMethodSignature.equals("(I)V")){
				String currentLine = getCurrentSourceLine();
				try {
					currentLine = currentLine.substring(currentLine.indexOf(Constants.REALTIME_SYSTEM_FIRE + "(") + 
							(Constants.REALTIME_SYSTEM_FIRE + "(").length());
					currentLine = currentLine.substring(0, currentLine.indexOf(")"));
					
					
					bb = new SporadicInvokeBasicBlock(
							callSite,
							Integer.parseInt(currentLine));
				} catch (Exception e) {
					logger.severe("Error retrieving correct sporadic task. In file: " + 
							currentClass.getSourceFileName() + " on line: "
							+ getCurrentSourceLineNumber() + ". " + 
							" Supported syntax:" +
							" RealtimeSystem.fire(<int_literal>);");
					
					e.printStackTrace();
	
					System.exit(-1);
				}
			} else {
				bb = new SpecialInvokeBasicBlock(callSite);
			}
		}
		
		bb.addBytecode(currentHandle);
		currentInstructionGraph.addNode(bb);
		instructionNodeMap.put(new Integer(currentHandle.getPosition()), bb);		
		
//		obj.getName(cpg)

	}

	
	public void visitINVOKESTATIC(INVOKESTATIC obj) {
		// TODO Invoke static should be fixed - though it works atm, it _is_ a special case of invoke
		
		ConstantPoolGen cpg = new ConstantPoolGen(currentMethod.getConstantPool());
		String calledClassName = obj.getReferenceType(cpg).toString();
		AbstractClassNode calledClass = javaTranslator.classGraph.getClassFromCanonicalName(calledClassName);
		String calledMethodName = obj.getMethodName(cpg);
		String calledMethodSignature = obj.getSignature(cpg);
		
		AbstractBasicBlock bb = null;
		///TODO: fixfix
/*  UPDATE THIS  <<<<<<< .mine
//		if(calledClassName.equals(Constants.JOP_NATIVE)) {
//			bb = new SimpleBasicBlock(); // the correct native instruction is found in TransitionBuilder
//		} else {
//			if (calledClassName.equals(Constants.REALTIME_SYSTEM) && 
//					calledMethodName.equals("fire") &&
//					calledMethodSignature.equals("(I)V")){
//				String currentLine = getCurrentSourceLine();
//				try {
//					currentLine = currentLine.substring(currentLine.indexOf(Constants.REALTIME_SYSTEM_FIRE + "(") + 
//							(Constants.REALTIME_SYSTEM_FIRE + "(").length());
//					currentLine = currentLine.substring(0, currentLine.indexOf(")"));
//					bb = new SporadicInvokeBasicBlock(
//							calledMethodName, 
//							calledMethodSignature, 
//							calledClass,
//							Integer.parseInt(currentLine));
//				} catch (Exception e) {
//					logger.severe("Error retrieving correct sporadic task. In file: " + 
//							currentClass.getSourceFileName() + " on line: "
//							+ getCurrentSourceLineNumber() + ". " + 
//							" Supported syntax:" +
//							" RealtimeSystem.fire(<int_literal>);");
//					
//					e.printStackTrace();
//	
//					System.exit(-1);
//				}
//			} else {
//				bb = new InvokeInterfaceBasicBlock(calledMethodName, calledMethodSignature, calledClass);
//			}
//		}
======= */
		if(calledClassName.equals(Constants.JOP_NATIVE)) {
			bb = new SimpleBasicBlock(); // the correct native instruction is found in TransitionBuilder
		} else {
			if (calledClassName.equals(Constants.REALTIME_SYSTEM) && 
					calledMethodName.equals("fire") &&
					calledMethodSignature.equals("(I)V")){
				String currentLine = getCurrentSourceLine();
				try {
					currentLine = currentLine.substring(currentLine.indexOf(Constants.REALTIME_SYSTEM_FIRE + "(") + 
							(Constants.REALTIME_SYSTEM_FIRE + "(").length());
					currentLine = currentLine.substring(0, currentLine.indexOf(")"));
					
					CallSite callSite = new CallSite(currentClass.getClassName(), currentMethod, currentHandle);

					bb = new SporadicInvokeBasicBlock(
							callSite,
							Integer.parseInt(currentLine));
				} catch (Exception e) {
					logger.severe("Error retrieving correct sporadic task. In file: " + 
							currentClass.getSourceFileName() + " on line: "
							+ getCurrentSourceLineNumber() + ". " + 
							" Supported syntax:" +
							" RealtimeSystem.fire(<int_literal>);");
					
					e.printStackTrace();
	
					System.exit(-1);
				}
			} else {
				CallSite callSite = new CallSite(currentClass.getClassName(), currentMethod, currentHandle);
				
				bb = new MethodCallingBasicBlock(callSite);
			}
		}
		
		bb.addBytecode(currentHandle);
		currentInstructionGraph.addNode(bb);
		instructionNodeMap.put(new Integer(currentHandle.getPosition()), bb);

	}

	
	public void visitINVOKEVIRTUAL(INVOKEVIRTUAL obj) {
		// TODO Current implementation is a standard invoke... - check up on this
		
		ConstantPoolGen cpg = new ConstantPoolGen(currentMethod.getConstantPool());
		String calledClassName = obj.getReferenceType(cpg).toString();
		AbstractClassNode calledClass = javaTranslator.classGraph.getClassFromCanonicalName(calledClassName);
		String calledMethodName = obj.getMethodName(cpg);
		String calledMethodSignature = obj.getSignature(cpg);
		
		AbstractBasicBlock bb = null;
		if(calledClassName.equals(Constants.JOP_NATIVE)) {
			bb = new SimpleBasicBlock(); // the correct native instruction is found in TransitionBuilder
		} else {
			if (calledClassName.equals(Constants.REALTIME_SYSTEM) && 
					calledMethodName.equals("fire") &&
					calledMethodSignature.equals("(I)V")){
				String currentLine = getCurrentSourceLine();
				try {
					currentLine = currentLine.substring(currentLine.indexOf(Constants.REALTIME_SYSTEM_FIRE + "(") + 
							(Constants.REALTIME_SYSTEM_FIRE + "(").length());
					currentLine = currentLine.substring(0, currentLine.indexOf(")"));
					
					CallSite callSite = new CallSite(currentClass.getClassName(), currentMethod, currentHandle);

					bb = new SporadicInvokeBasicBlock(
							callSite,
							Integer.parseInt(currentLine));
				} catch (Exception e) {
					logger.severe("Error retrieving correct sporadic task. In file: " + 
							currentClass.getSourceFileName() + " on line: "
							+ getCurrentSourceLineNumber() + ". " + 
							" Supported syntax:" +
							" RealtimeSystem.fire(<int_literal>);");
					
					e.printStackTrace();
	
					System.exit(-1);
				}
			} else {
				CallSite callSite = new CallSite(currentClass.getClassName(), currentMethod, currentHandle);
				
				bb = new MethodCallingBasicBlock(callSite);
			}
		}
		
		bb.addBytecode(currentHandle);
		currentInstructionGraph.addNode(bb);
		instructionNodeMap.put(new Integer(currentHandle.getPosition()), bb);

	}

	
	public void visitIOR(IOR obj) {


	}

	
	public void visitIREM(IREM obj) {


	}

	
	public void visitIRETURN(IRETURN obj) {


	}

	
	public void visitISHL(ISHL obj) {


	}

	
	public void visitISHR(ISHR obj) {


	}

	
	public void visitISTORE(ISTORE obj) {


	}

	
	public void visitISUB(ISUB obj) {


	}

	
	public void visitIUSHR(IUSHR obj) {


	}

	
	public void visitIXOR(IXOR obj) {


	}

	
	public void visitIfInstruction(IfInstruction obj) {


	}

	
	public void visitJSR(JSR obj) {


	}

	
	public void visitJSR_W(JSR_W obj) {


	}

	
	public void visitJsrInstruction(JsrInstruction obj) {


		// OMG 
	}

	
	public void visitL2D(L2D obj) {


	}

	
	public void visitL2F(L2F obj) {


	}

	
	public void visitL2I(L2I obj) {


	}

	
	public void visitLADD(LADD obj) {


	}

	
	public void visitLALOAD(LALOAD obj) {


	}

	
	public void visitLAND(LAND obj) {


	}

	
	public void visitLASTORE(LASTORE obj) {


	}

	
	public void visitLDC(LDC obj) {


	}

	
	public void visitLDC2_W(LDC2_W obj) {


	}

	
	public void visitLDIV(LDIV obj) {


	}

	
	public void visitLLOAD(LLOAD obj) {


	}

	
	public void visitLMUL(LMUL obj) {


	}

	
	public void visitLNEG(LNEG obj) {


	}

	
	public void visitLOOKUPSWITCH(LOOKUPSWITCH obj) {
//		System.err.println("LOOKUPSWITCH");
	}

	
	public void visitLOR(LOR obj) {


	}

	
	public void visitLREM(LREM obj) {


	}

	
	public void visitLRETURN(LRETURN obj) {


	}

	
	public void visitLSHL(LSHL obj) {


	}

	
	public void visitLSHR(LSHR obj) {


	}

	
	public void visitLSTORE(LSTORE obj) {


	}

	
	public void visitLSUB(LSUB obj) {


	}

	
	public void visitLUSHR(LUSHR obj) {


	}

	
	public void visitLXOR(LXOR obj) {


	}

	
	public void visitLoadClass(LoadClass obj) {


	}

	
	public void visitLoadInstruction(LoadInstruction obj) {


	}

	
	public void visitMULTIANEWARRAY(MULTIANEWARRAY obj) {


	}

	
	public void visitNEW(NEW obj) {

		//TODO Proper implementation of new -- "NewBasicBlock" - decide information needed...		
		if(!(Options.CONSIDER_FINALIZERS || Options.COUNT_NEWS)){
			createBasicBlock(obj);
//			System.exit(0);
		}
		else {
			ConstantPoolGen cpg = new ConstantPoolGen(currentMethod.getConstantPool());
			try{
				System.out.println(obj.getLoadClassType(cpg));
			}catch(Exception e){
				e.printStackTrace();
				System.exit(0);
			}
			assert obj.getLoadClassType(cpg).toString().equals(obj.getType(cpg).toString()) : "I assume LoadClassType and Type is the same"; // WRONG! - but inform me when it isnt
			
			NewBasicBlock newBasicBlock = new NewBasicBlock(obj.getLoadClassType(cpg));
			newBasicBlock.addBytecode(currentHandle);
			currentInstructionGraph.addNode(newBasicBlock);
			instructionNodeMap.put(new Integer(currentHandle.getPosition()), newBasicBlock);
		}
	}

	
	public void visitPOP(POP obj) {


	}

	
	public void visitPOP2(POP2 obj) {


	}

	
	public void visitPUTFIELD(PUTFIELD obj) {


	}

	
	public void visitPUTSTATIC(PUTSTATIC obj) {


	}

	
	public void visitPopInstruction(PopInstruction obj) {


	}

	
	public void visitPushInstruction(PushInstruction obj) {


	}

	
	public void visitRETURN(RETURN obj) {


	}

	
	public void visitSALOAD(SALOAD obj) {


	}

	
	public void visitSASTORE(SASTORE obj) {


	}

	
	public void visitSWAP(SWAP obj) {


	}

	
	public void visitStackConsumer(StackConsumer obj) {


	}

}
