package tetasarts.generator;


import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.logging.Logger;

import org.apache.bcel.classfile.CodeException;
import org.apache.bcel.classfile.JavaClass;
import org.apache.bcel.classfile.Method;
import org.apache.bcel.generic.*;

import sun.awt.EmbeddedFrame;
import tetasarts.Options;
import tetasarts.tir.basicblocks.AbstractBasicBlock;
import tetasarts.tir.basicblocks.BasicBlockGraph;
import tetasarts.tir.basicblocks.LoopBound;
import tetasarts.tir.basicblocks.basicBlockTypes.EmptyBasicBlock;
import tetasarts.tir.basicblocks.basicBlockTypes.IfBasicBlock;
import tetasarts.tir.basicblocks.basicBlockTypes.LoopBasicBlock;
import tetasarts.tir.basicblocks.basicBlockTypes.SwitchBasicBlock;
import tetasarts.tir.methods.MethodInformation;

public class TransitionBuilder implements Visitor {

	protected Method currentMethod;
	protected InstructionHandle currentHandle;
	protected JavaClass currentClass;
	protected BasicBlockGraph currentInstructionGraph;
	protected MethodInformation currentMethodTarget;
	protected Map<Integer, AbstractBasicBlock> instructionNodeMap;
	protected JavaTranslator javaTranslator;
	protected AbstractBasicBlock previousBasicBlock = null;
	protected AbstractBasicBlock exitNode;
	protected Logger logger;
	protected Map<Integer, CodeException> exceptionMap;
	protected boolean inExceptionHandler;

	public void setCurrentClass(JavaClass jc) {
		currentClass = jc;
	}
	public JavaClass getCurrentClass() {
		return currentClass;
	}

	public MethodInformation getCurrentMethodTarget() {
		return currentMethodTarget;
	}
	public void setCurrentMethodTarget(MethodInformation currentMethodTarget) {
		this.currentMethodTarget = currentMethodTarget;
	}
	
	
	public TransitionBuilder(JavaTranslator jt) {
		javaTranslator = jt;
		logger = Logger.getLogger("generator");
	}

	public void visitMethod(Method method, MethodInformation methodTarget) {
		currentMethod = method;
		exceptionMap = new HashMap<Integer, CodeException>();
		if(currentMethod.getCode().getExceptionTable() != null) {
			for(CodeException ce : currentMethod.getCode().getExceptionTable()) {
				exceptionMap.put(ce.getHandlerPC(), ce);
			}
		}
		inExceptionHandler = false;
		
		currentMethodTarget = methodTarget;
		currentInstructionGraph = currentMethodTarget.instructionGraph;
		instructionNodeMap = javaTranslator.instructionNodeMap;
		
		InstructionList l = new InstructionList(method.getCode().getCode());
		l.setPositions(true);

		InstructionHandle handle = l.getStart();

		AbstractBasicBlock entranceNode = new EmptyBasicBlock();
		currentInstructionGraph.addNode(entranceNode);
		previousBasicBlock = entranceNode;
		
		exitNode = new EmptyBasicBlock();
		currentInstructionGraph.addNode(exitNode);
		
		do {
			visit(handle);

			handle = handle.getNext();
		} while (handle != null);
		
		handle = l.getEnd();
		AbstractBasicBlock lastBlock = instructionNodeMap.get(handle.getPosition());
		if (lastBlock != null) {
//			if(currentInstructionGraph.getSuccs(lastBlock).size() == 0) {
			if(! currentInstructionGraph.hasSuccessors(lastBlock)){
				currentInstructionGraph.addEdge(lastBlock, exitNode);
			}
		}
		
		if(currentInstructionGraph.getPreds(exitNode).size() == 0) {
			handle = l.getEnd();
			AbstractBasicBlock abb = instructionNodeMap.get(handle.getPosition());
			if (abb == null)
				abb = entranceNode;
			
			currentInstructionGraph.addEdge(abb, exitNode);
		}
		
//		currentInstructionGraph.SESEify();
//		assert false : "STOP HERE";
	}

	public void visit(InstructionHandle iHand) {
		currentHandle = iHand;
		
		if(exceptionMap.containsKey(currentHandle.getPosition()) && !currentHandle.hasTargeters()) {
			inExceptionHandler = true;
			return;
		}
		
		if(inExceptionHandler) {
			if(!currentHandle.hasTargeters() && 
					   /*!(currentHandle.getInstruction() instanceof BranchInstruction) ||*/
					   !(currentHandle.getInstruction() instanceof ATHROW) &&
					   !(currentHandle.getInstruction() instanceof ReturnInstruction)) {
				return;
			}
			inExceptionHandler = false;
			return;
		}

		iHand.accept(this);
	}
	
	private void findExitNodes(LoopBound loopBound) {
		LoopBasicBlock lastLoopBlock;
		Iterator<LoopBasicBlock> i = loopBound.getLoopNodeIterator();
		assert i.hasNext() : "No loopnodes in loopbound object. ";

		do {
			lastLoopBlock = i.next();
		} while(i.hasNext());

		int lastNodePosition = lastLoopBlock.getBytecodes().get(0).getPosition();
		
		InstructionHandle targetInstruction = loopBound.getTargetNode().getBytecodes().get(0);
		InstructionHandle currentInstruction = targetInstruction.getNext();
		do { // for all instructions between the target and the last loopnode
			
			// if we meet a jump instruction, check if the target is outside the loop
			if(currentInstruction.getInstruction() instanceof BranchInstruction) {
				BranchInstruction bi = (BranchInstruction)currentInstruction.getInstruction();
				if(bi.getTarget().getPosition() > lastNodePosition) { // target outside loop?
					AbstractBasicBlock abb = instructionNodeMap.get(currentInstruction.getPosition());
//					assert abb != null : "instructionNodeMap returned null! "+currentInstruction.getClass() + " -- "+ currentInstruction + " -- " + currentInstructionGraph.toString();
					if(abb!= null)
						loopBound.addExitNode(abb);
//					assert instructionNodeMap.get(bi.getTarget().getPosition()) != null: "exit is null! inst: " + bi + " - method: " + currentMethodTarget.getCanonicalName();
				}
			}
			
			if(currentInstruction.getNext() == null) { // if we don't have a next, it must be a return or goto
				if(currentInstruction.getInstruction() instanceof BranchInstruction) { 
					// if it is a goto, update the next instruction to be the target of the goto
					currentInstruction = ((BranchInstruction)currentInstruction.getInstruction()).getTarget();
				} else {
					break;
				}

			} else {
				currentInstruction = currentInstruction.getNext();
			}
		} while (currentInstruction.getPosition() != lastNodePosition);

		// if the last loopnode is NOT a goto, then it can also exit the loop. 
		if(!(lastLoopBlock.getBytecodes().get(0).getInstruction() instanceof GotoInstruction)) {
			loopBound.addExitNode(lastLoopBlock);
		}
	}
	
	public AbstractBasicBlock linkToPrevious(Instruction i) {
		AbstractBasicBlock abb = instructionNodeMap.get(currentHandle.getPosition());
		assert abb != null : "Current AbstractBasicBlock not found in instructionNodeMap! " + currentHandle.getPosition();
		if(previousBasicBlock != null) {
			currentInstructionGraph.addEdge(previousBasicBlock, abb);
		}
//		if(b){
//			System.out.println("linking "+ abb.getName() + " -> " + previousBasicBlock);
//			if(previousBasicBlock != null && previousBasicBlock.getBytecodes().size()>0)
//				System.err.println("linking "+ currentHandle.getPosition() + " -> " + previousBasicBlock.getBytecodes().get(0).getPosition());
//		}
//
		AbstractBasicBlock prev = previousBasicBlock;
		previousBasicBlock = abb;
		return prev;
	}

	boolean b = true;
	@Override
	public void visitSelect(Select arg0) {
		b = true;
//		AbstractBasicBlock prev = linkToPrevious(arg0);
		
//		if(prev != null)
//			System.out.println("visitSelect prev: " + prev.getBytecodes());
//		System.err.println("Linked...");
//		
//		System.out.println("ARG          : " + arg0.getClass());
//		System.out.println("CurrentHandle: " + currentHandle);
//		System.err.println("TARGET:      : " + arg0.getTarget().getPosition());
		{
//			System.out.println("Position: " +arg0.getTarget().getPosition());
			AbstractBasicBlock abb = instructionNodeMap.get(arg0.getTarget().getPosition());
			assert abb != null : "instructionNodeMap returned null!";
			abb.setTarget(true);
		}

		
		AbstractBasicBlock abb = instructionNodeMap.get(currentHandle.getPosition());
		assert abb != null : "instructionNodeMap returned null!";

		List<Integer> l = new ArrayList<Integer>();
//		System.out.println("TARGET: " + ((Select)arg0).getTarget());
		for(InstructionHandle ih : ((Select)arg0).getTargets()) {
//			System.out.println("  ih:"+ih);
			if(! l.contains(ih.getPosition())){
				l.add(ih.getPosition());
				AbstractBasicBlock currentBB = instructionNodeMap.get(ih.getPosition());
				assert currentBB != null : "instructionNodeMap returned null!";
				currentBB.setTarget(true);
				currentInstructionGraph.addEdge(abb, currentBB);
//				System.err.println("sL "+ abb.getName() + " -> " + currentBB.getName());
//				System.err.println("sL "+ currentHandle.getPosition() + " -> " + ih.getPosition());
			} else {
//				System.err.println(ih.getPosition() + " already linked");
			}
		}
		
		/** Also link DEFAULT case: **/
		
		InstructionHandle ih = ((Select)arg0).getTarget();
		if(! l.contains(ih.getPosition())){
			l.add(ih.getPosition());
			AbstractBasicBlock currentBB = instructionNodeMap.get(ih.getPosition());
			assert currentBB != null : "instructionNodeMap returned null!";
			currentBB.setTarget(true);
			currentInstructionGraph.addEdge(abb, currentBB);
		} else {
//			System.err.println(ih.getPosition() + " already linked");
//			System.exit(1);
		}
		
		
		b = false;
//		AbstractBasicBlock target = instructionNodeMap.get(arg0.getTarget().getPosition());
//		assert target != null : "instructionNodeMap returned null!";
//		currentInstructionGraph.addEdge(abb, target);
		
//		System.exit(0);

/**		
		if((abb instanceof LoopBasicBlock)) {
			LoopBasicBlock loopBasicBlock = (LoopBasicBlock)abb;
			LoopBound loopBound = loopBasicBlock.getLoopBound();

			// link current to target
			currentInstructionGraph.addEdge(loopBasicBlock, loopBound.getTargetNode());

			// find exit nodes
			if(!loopBound.getExitNodeIterator().hasNext()) { // exit nodes not found yet
				findExitNodes(loopBound);
			}
			
		} else if((abb instanceof IfBasicBlock)) {

			
		} 
/*		else if((arg0 instanceof Select)) {
			// TODO fix! implemented in Java
			
			//AbstractBasicBlock bb = new MethodCallingBasicBlock(Constants.JAVA_OPCODE_METHOD.get(arg0.getOpcode()), Constants.JAVA_OPCODE_SIGNATURE.get(arg0.getOpcode()), javaTranslator.classGraph.getClassFromCanonicalName(Constants.JOP_JVM_CLASS));

//			currentHandle.swapInstruction(new INVOKESTATIC(1));
//			abstractBasicBlock.addBytecode(currentHandle);
			abb.getBytecodes().get(0).swapInstruction(new INVOKESTATIC(1));
			
			for(InstructionHandle ih : ((Select)arg0).getTargets()) {
				AbstractBasicBlock currentBB = instructionNodeMap.get(ih.getPosition());
				assert currentBB != null : "instructionNodeMap returned null!";
				currentBB.setTarget(true);
				currentInstructionGraph.addEdge(abb, currentBB);
			}
		}
		else {
			logger.severe("Unknown branching type. " + arg0);
		}
		*/
		// don't link next to current if current is a goto instruction
//		if(arg0 instanceof GotoInstruction) {
//			previousBasicBlock = null;
//		}
		
		/**
		System.out.println("Select");
		System.out.println("Position: "+currentHandle.getPosition());
	    System.out.println("Post instruction: " + obj.getTarget());
	    System.out.println("nested instructions: ");
		for(InstructionHandle ih: obj.getTargets()){
			System.out.println("  " + ih.getPosition());
		}
		// TODO fix! implemented in Java
		AbstractBasicBlock bb = null;
		System.out.println(currentHandle.toString(true));
		bb = new SwitchBasicBlock();
	
		bb.addBytecode(currentHandle);
		currentInstructionGraph.addNode(bb);
		instructionNodeMap.put(currentHandle.getPosition(), bb);
		
		**/

	}
	
	@Override
	public void visitBranchInstruction(BranchInstruction arg0) {
		if(arg0 instanceof Select) {
//			System.err.println("TransitionBuilder: Select instructions not supported! " + arg0);
			return;
		}
		
		linkToPrevious(arg0);
//		System.out.println("ARG          : " + arg0.getClass());
//		System.out.println("CurrentHandle: " + currentHandle);
		{
//			System.out.println("Position: " +arg0.getTarget().getPosition());
			AbstractBasicBlock abb = instructionNodeMap.get(arg0.getTarget().getPosition());
//			System.out.println(currentClass);
//			System.out.println(currentMethod);
			Set<Integer> ks = instructionNodeMap.keySet();
			ArrayList<Integer> al = new ArrayList<Integer>(ks);
			Collections.sort(al);
//			for(int i: al){
//				System.out.println("known: "+i);
//			}
//			System.err.println(arg0.getTarget().getPosition());
			assert abb != null : "instructionNodeMap returned null! -- "+ arg0.getTarget().getPosition();
			abb.setTarget(true);
		}

		AbstractBasicBlock abb = instructionNodeMap.get(currentHandle.getPosition());
		assert abb != null : "instructionNodeMap returned null!";

		if((abb instanceof LoopBasicBlock)) {
			LoopBasicBlock loopBasicBlock = (LoopBasicBlock)abb;
			LoopBound loopBound = loopBasicBlock.getLoopBound();

			// link current to target
			currentInstructionGraph.addEdge(loopBasicBlock, loopBound.getTargetNode());

			// find exit nodes
			if(!Options.PRECISE_LOOP) {
				if(!loopBound.getExitNodeIterator().hasNext()) { // exit nodes not found yet
					findExitNodes(loopBound);
				}
			}
				
		} else if((abb instanceof IfBasicBlock)) {
			AbstractBasicBlock target = instructionNodeMap.get(arg0.getTarget().getPosition());
			assert target != null : "instructionNodeMap returned null!";
			currentInstructionGraph.addEdge(abb, target);
			
		} 
/*		else if((arg0 instanceof Select)) {
			// TODO fix! implemented in Java
			
			//AbstractBasicBlock bb = new MethodCallingBasicBlock(Constants.JAVA_OPCODE_METHOD.get(arg0.getOpcode()), Constants.JAVA_OPCODE_SIGNATURE.get(arg0.getOpcode()), javaTranslator.classGraph.getClassFromCanonicalName(Constants.JOP_JVM_CLASS));

//			currentHandle.swapInstruction(new INVOKESTATIC(1));
//			abstractBasicBlock.addBytecode(currentHandle);
			abb.getBytecodes().get(0).swapInstruction(new INVOKESTATIC(1));
			
			for(InstructionHandle ih : ((Select)arg0).getTargets()) {
				AbstractBasicBlock currentBB = instructionNodeMap.get(ih.getPosition());
				assert currentBB != null : "instructionNodeMap returned null!";
				currentBB.setTarget(true);
				currentInstructionGraph.addEdge(abb, currentBB);
			}
		}*/ else {
			logger.severe("Unknown branching type. " + arg0);
		}
		
		// don't link next to current if current is a goto instruction
		if(arg0 instanceof GotoInstruction) {
			previousBasicBlock = null;
		}
	}

	@Override
	public void visitCPInstruction(CPInstruction arg0) {

		if((arg0 instanceof InvokeInstruction)) {
			InvokeInstruction obj = (InvokeInstruction)arg0;
			ConstantPoolGen cpg = new ConstantPoolGen(currentMethod.getConstantPool());
			String calledClassName = obj.getReferenceType(cpg).toString();
			String calledMethodName = obj.getMethodName(cpg);
			String calledMethodSignature = obj.getSignature(cpg);
			
			if(calledClassName.equals(Constants.JOP_NATIVE)) {
				AbstractBasicBlock abb = instructionNodeMap.get(currentHandle.getPosition());
				assert abb != null : "instructionNodeMap returned null!";
				abb.getBytecodes().get(0).setInstruction(getNativeBytecode(calledClassName, calledMethodName, calledMethodSignature));
			}
		}
		linkToPrevious(arg0);
	}

	private Instruction getNativeBytecode(String className, String methodName, String methodSignature) {

		Short opcode = Constants.NATIVE_METHOD_OPCODE.get(methodName);
		
		if(opcode != null) {
			return new NativeInstruction(opcode.shortValue());
		} else {
			logger.severe("unknown native method: " + className
					+ "." + methodName + methodSignature + 
					" adding nop instead. Method: " + currentMethodTarget.getCanonicalName());
			return new NativeInstruction(Constants.JOPSYS_NOP);
		}
	}

	@Override
	public void visitReturnInstruction(ReturnInstruction arg0) {

		linkToPrevious(arg0);
		
		AbstractBasicBlock abb = instructionNodeMap.get(currentHandle.getPosition());
		assert abb != null : "instructionNodeMap returned null!";

		currentInstructionGraph.addEdge(abb, exitNode);
		
		previousBasicBlock = null;
	}

	@Override
	public void visitATHROW(ATHROW arg0) {

		// OMG
		//logger.severe("Instruction ATHROW not supported! " + arg0);
//		linkToPrevious(arg0);
	}

	@Override
	public void visitBREAKPOINT(BREAKPOINT arg0) {

		//logger.severe("Breakpoint instruction not implemented on underlying hardware! " + arg0);
		linkToPrevious(arg0);
	}

	@Override
	public void visitIMPDEP1(IMPDEP1 arg0) {
		// TODO Auto-generated method stub
		// Java implemented

		// OMG JOP specific instruction implemented in Java. See JOP decumentation
	}

	@Override
	public void visitIMPDEP2(IMPDEP2 arg0) {
		// TODO Auto-generated method stub

		// OMG
		//logger.severe("IMPDEP2 not supported in underlying hardware!");
		linkToPrevious(arg0);
	}

	@Override
	public void visitRET(RET arg0) {

		linkToPrevious(arg0);
		
		// OMG return
	}

	/***********************
	 * 
	 * Simple visit methods
	 * 
	 ***********************/
	
	@Override
	public void visitACONST_NULL(ACONST_NULL arg0) {

		linkToPrevious(arg0);
	}

	@Override
	public void visitARRAYLENGTH(ARRAYLENGTH arg0) {

		linkToPrevious(arg0);
	}

	@Override
	public void visitArithmeticInstruction(ArithmeticInstruction arg0) {
		// OMG Exception thrower

		linkToPrevious(arg0);
	}

	@Override
	public void visitArrayInstruction(ArrayInstruction arg0) {

		linkToPrevious(arg0);
	}

	@Override
	public void visitBIPUSH(BIPUSH arg0) {

		linkToPrevious(arg0);
	}

	@Override
	public void visitConversionInstruction(ConversionInstruction arg0) {

		linkToPrevious(arg0);
	}

	@Override
	public void visitDCMPG(DCMPG arg0) {

		linkToPrevious(arg0);
	}

	@Override
	public void visitDCMPL(DCMPL arg0) {

		linkToPrevious(arg0);
	}

	@Override
	public void visitDCONST(DCONST arg0) {

		linkToPrevious(arg0);
	}

	@Override
	public void visitFCMPG(FCMPG arg0) {

		linkToPrevious(arg0);
	}

	@Override
	public void visitFCMPL(FCMPL arg0) {

		linkToPrevious(arg0);
	}

	@Override
	public void visitFCONST(FCONST arg0) {

		linkToPrevious(arg0);
	}

	@Override
	public void visitICONST(ICONST arg0) {

		linkToPrevious(arg0);
	}

	@Override
	public void visitLCMP(LCMP arg0) {

		linkToPrevious(arg0);
	}

	@Override
	public void visitLCONST(LCONST arg0) {

		linkToPrevious(arg0);
	}

	@Override
	public void visitLocalVariableInstruction(LocalVariableInstruction arg0) {

		linkToPrevious(arg0);
	}

	@Override
	public void visitMONITORENTER(MONITORENTER arg0) {

		linkToPrevious(arg0);
	}

	@Override
	public void visitMONITOREXIT(MONITOREXIT arg0) {

		linkToPrevious(arg0);
	}

	@Override
	public void visitNEWARRAY(NEWARRAY arg0) {

		linkToPrevious(arg0);
	}

	@Override
	public void visitNOP(NOP arg0) {

		linkToPrevious(arg0);
	}

	@Override
	public void visitSIPUSH(SIPUSH arg0) {

		linkToPrevious(arg0);
	}

	@Override
	public void visitStackInstruction(StackInstruction arg0) {

		linkToPrevious(arg0);
	}

	/******************************************************
	 * 
	 * The rest of the visit methods are unused.
	 * 
	 ******************************************************/

	@Override
	public void visitStackProducer(StackProducer arg0) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitStoreInstruction(StoreInstruction arg0) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitTABLESWITCH(TABLESWITCH arg0) {
		// TODO Auto-generated method stub
	}

	@Override
	public void visitTypedInstruction(TypedInstruction arg0) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitUnconditionalBranch(UnconditionalBranch arg0) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitVariableLengthInstruction(VariableLengthInstruction arg0) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitAALOAD(AALOAD arg0) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitAASTORE(AASTORE arg0) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitALOAD(ALOAD arg0) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitANEWARRAY(ANEWARRAY arg0) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitARETURN(ARETURN arg0) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitASTORE(ASTORE arg0) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitAllocationInstruction(AllocationInstruction arg0) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitBALOAD(BALOAD arg0) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitBASTORE(BASTORE arg0) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitCALOAD(CALOAD arg0) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitCASTORE(CASTORE arg0) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitCHECKCAST(CHECKCAST arg0) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitConstantPushInstruction(ConstantPushInstruction arg0) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitD2F(D2F arg0) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitD2I(D2I arg0) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitD2L(D2L arg0) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitDADD(DADD arg0) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitDALOAD(DALOAD arg0) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitDASTORE(DASTORE arg0) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitDDIV(DDIV arg0) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitDLOAD(DLOAD arg0) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitDMUL(DMUL arg0) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitDNEG(DNEG arg0) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitDREM(DREM arg0) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitDRETURN(DRETURN arg0) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitDSTORE(DSTORE arg0) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitDSUB(DSUB arg0) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitDUP(DUP arg0) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitDUP2(DUP2 arg0) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitDUP2_X1(DUP2_X1 arg0) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitDUP2_X2(DUP2_X2 arg0) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitDUP_X1(DUP_X1 arg0) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitDUP_X2(DUP_X2 arg0) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitExceptionThrower(ExceptionThrower arg0) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitF2D(F2D arg0) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitF2I(F2I arg0) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitF2L(F2L arg0) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitFADD(FADD arg0) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitFALOAD(FALOAD arg0) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitFASTORE(FASTORE arg0) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitFDIV(FDIV arg0) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitFLOAD(FLOAD arg0) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitFMUL(FMUL arg0) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitFNEG(FNEG arg0) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitFREM(FREM arg0) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitFRETURN(FRETURN arg0) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitFSTORE(FSTORE arg0) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitFSUB(FSUB arg0) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitFieldInstruction(FieldInstruction arg0) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitFieldOrMethod(FieldOrMethod arg0) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitGETFIELD(GETFIELD arg0) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitGETSTATIC(GETSTATIC arg0) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitGOTO(GOTO arg0) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitGOTO_W(GOTO_W arg0) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitGotoInstruction(GotoInstruction arg0) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitI2B(I2B arg0) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitI2C(I2C arg0) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitI2D(I2D arg0) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitI2F(I2F arg0) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitI2L(I2L arg0) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitI2S(I2S arg0) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitIADD(IADD arg0) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitIALOAD(IALOAD arg0) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitIAND(IAND arg0) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitIASTORE(IASTORE arg0) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitIDIV(IDIV arg0) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitIFEQ(IFEQ arg0) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitIFGE(IFGE arg0) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitIFGT(IFGT arg0) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitIFLE(IFLE arg0) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitIFLT(IFLT arg0) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitIFNE(IFNE arg0) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitIFNONNULL(IFNONNULL arg0) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitIFNULL(IFNULL arg0) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitIF_ACMPEQ(IF_ACMPEQ arg0) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitIF_ACMPNE(IF_ACMPNE arg0) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitIF_ICMPEQ(IF_ICMPEQ arg0) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitIF_ICMPGE(IF_ICMPGE arg0) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitIF_ICMPGT(IF_ICMPGT arg0) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitIF_ICMPLE(IF_ICMPLE arg0) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitIF_ICMPLT(IF_ICMPLT arg0) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitIF_ICMPNE(IF_ICMPNE arg0) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitIINC(IINC arg0) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitILOAD(ILOAD arg0) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitIMUL(IMUL arg0) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitINEG(INEG arg0) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitINSTANCEOF(INSTANCEOF arg0) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitINVOKEINTERFACE(INVOKEINTERFACE arg0) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitINVOKESPECIAL(INVOKESPECIAL arg0) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitINVOKESTATIC(INVOKESTATIC arg0) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitINVOKEVIRTUAL(INVOKEVIRTUAL arg0) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitIOR(IOR arg0) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitIREM(IREM arg0) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitIRETURN(IRETURN arg0) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitISHL(ISHL arg0) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitISHR(ISHR arg0) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitISTORE(ISTORE arg0) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitISUB(ISUB arg0) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitIUSHR(IUSHR arg0) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitIXOR(IXOR arg0) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitIfInstruction(IfInstruction arg0) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitInvokeInstruction(InvokeInstruction arg0) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitJSR(JSR arg0) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitJSR_W(JSR_W arg0) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitJsrInstruction(JsrInstruction arg0) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitL2D(L2D arg0) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitL2F(L2F arg0) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitL2I(L2I arg0) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitLADD(LADD arg0) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitLALOAD(LALOAD arg0) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitLAND(LAND arg0) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitLASTORE(LASTORE arg0) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitLDC(LDC arg0) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitLDC2_W(LDC2_W arg0) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitLDIV(LDIV arg0) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitLLOAD(LLOAD arg0) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitLMUL(LMUL arg0) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitLNEG(LNEG arg0) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitLOOKUPSWITCH(LOOKUPSWITCH arg0) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitLOR(LOR arg0) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitLREM(LREM arg0) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitLRETURN(LRETURN arg0) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitLSHL(LSHL arg0) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitLSHR(LSHR arg0) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitLSTORE(LSTORE arg0) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitLSUB(LSUB arg0) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitLUSHR(LUSHR arg0) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitLXOR(LXOR arg0) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitLoadClass(LoadClass arg0) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitLoadInstruction(LoadInstruction arg0) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitMULTIANEWARRAY(MULTIANEWARRAY arg0) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitNEW(NEW arg0) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitPOP(POP arg0) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitPOP2(POP2 arg0) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitPUTFIELD(PUTFIELD arg0) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitPUTSTATIC(PUTSTATIC arg0) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitPopInstruction(PopInstruction arg0) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitPushInstruction(PushInstruction arg0) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitRETURN(RETURN arg0) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitSALOAD(SALOAD arg0) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitSASTORE(SASTORE arg0) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitSWAP(SWAP arg0) {
		// TODO Auto-generated method stub

	}

	@Override
	public void visitStackConsumer(StackConsumer arg0) {
		// TODO Auto-generated method stub

	}

}
