package tetasarts.platforms;

import java.io.File;
import java.io.IOException;
import java.util.HashMap;
import java.util.logging.Logger;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.xml.sax.SAXException;

public class HVMAVR {
	
	private static final String HVMAVR_PATH = "/hvmavr.xml";
	private String hvmavrXMLPath = getClass().getResource(HVMAVR_PATH).getPath();
	HashMap<String, InstructionTimingInfo> timingInfoMap;
	
	private static HVMAVR instance = null;
	protected HVMAVR() throws SAXException, IOException, ParserConfigurationException {
		this.timingInfoMap = new HashMap<String, InstructionTimingInfo>();
		

		File fXmlFile = new File(hvmavrXMLPath);
		DocumentBuilderFactory dbFactory = DocumentBuilderFactory.newInstance();
		DocumentBuilder dBuilder = dbFactory.newDocumentBuilder();
		Document timingmodelDoc = dBuilder.parse(fXmlFile);
		
		NodeList nList = timingmodelDoc.getElementsByTagName("JavaByteCode");
		for(int i = 0; i < nList.getLength(); i++) {
			Node nNode = nList.item(i);	 
			if (nNode.getNodeType() == Node.ELEMENT_NODE) {
				Element jbcEle = (Element) nNode;
				int opcode = Integer.parseInt(jbcEle.getElementsByTagName("opcode").item(0).getTextContent());
				String mnemonic = jbcEle.getElementsByTagName("mnemonic").item(0).getTextContent();
				int bcet = Integer.parseInt(jbcEle.getElementsByTagName("BCET").item(0).getTextContent());
				int wcet = Integer.parseInt(jbcEle.getElementsByTagName("WCET").item(0).getTextContent());
				
				timingInfoMap.put(mnemonic, new InstructionTimingInfo(mnemonic, opcode, bcet, wcet));
			}
		}
		   
	}
	public static HVMAVR getInstance() {
		if(instance == null) {
			try {
				instance = new HVMAVR();
			} catch (SAXException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			} catch (IOException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			} catch (ParserConfigurationException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}
		}
		return instance;
	}
	
	public InstructionTimingInfo getInstructionTiming(String mnemonic) {
		InstructionTimingInfo inf = timingInfoMap.get(mnemonic);
		if(inf == null) {
			Logger log = Logger.getLogger("platformlogger");
			log.severe("Could not find timing info for bytecode with mnemonic: " + mnemonic + ". Making dummy!");
			inf = new InstructionTimingInfo(mnemonic, -1, 100, 100);
		}
		return inf;
	}

}
