package tetasarts.tir.basicblocks;

import java.util.*;

import tetasarts.tir.Graph;
import tetasarts.tir.basicblocks.basicBlockTypes.EmptyBasicBlock;

public class BasicBlockGraph extends Graph<AbstractBasicBlock,IBasicBlockVistitor>{
	String source = null;
	
	public BasicBlockGraph(String source) {
		this.source = source;
	}
	
	@Override
	public String toString() {
		return source;
	}
	
	private class HeadsFinderVisitor extends AbstractBasicBlockVisitor{
		private List<AbstractBasicBlock> heads;
		public void defaultVisit(AbstractBasicBlock s) {
			if(getPreds(s).size() == 0){
				heads.add(s);
			}
		}
		public List<AbstractBasicBlock> getHeads(){
			heads = new ArrayList<AbstractBasicBlock>();
			for(AbstractBasicBlock block : getNodes()){
				block.accept(this);
			}
			return Collections.unmodifiableList(heads);
		}
	}
	
	private class TailsFinderVisitor extends AbstractBasicBlockVisitor{
		private List<AbstractBasicBlock> tails;
		public void defaultVisit(AbstractBasicBlock s) {
			if(getSuccs(s).size() == 0){
				tails.add(s);
			}
		}
		public List<AbstractBasicBlock> getTails(){
			tails = new ArrayList<AbstractBasicBlock>();
			for(AbstractBasicBlock block : getNodes()){
				block.accept(this);
			}
			return Collections.unmodifiableList(tails);
		}
	}
	
	public void SESEify(){
		System.out.println("SESEify");
		BasicBlockGraph grph = this;
		if(grph.getHeads().size() != 1){
			System.out.println("adding to head");
			EmptyBasicBlock SEHead = new EmptyBasicBlock();
			List<AbstractBasicBlock> heads = grph.getHeads();
			grph.addNode(SEHead);
			for(AbstractBasicBlock subhead : heads){
				boolean added = grph.addEdge(SEHead, subhead);
				assert added : "Superhead not added?";
			}
			for(AbstractBasicBlock p : grph.getPreds(SEHead)) {
				System.out.println("pred: "+p);
			}
		}
		
		if(grph.getTails().size() != 1){
			System.out.println("adding to tail");
			EmptyBasicBlock SETail = new EmptyBasicBlock();
			List<AbstractBasicBlock> tails = grph.getTails();
			grph.addNode(SETail);
			for(AbstractBasicBlock subtail : tails){
				boolean added = grph.addEdge(subtail, SETail);
				assert added : "Superhead not added?";
			}
			for(AbstractBasicBlock p : grph.getSuccs(SETail)) {
				System.out.println("succ: "+p);
			}
		}
		
		assert getHeads().size() == 1 : "Nr of heads on graph was not 1, count : " + getHeads().size();
		assert getTails().size() == 1 : "Nr of tails on graph was not 1, count : " + getTails().size();
	}

	public List<AbstractBasicBlock> getHeads(){
		return new HeadsFinderVisitor().getHeads();
	}
	
	public List<AbstractBasicBlock> getTails(){
		return new TailsFinderVisitor().getTails();
	}
}
