package tetasarts.tir.classinfo;

import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.logging.Logger;

import org.apache.bcel.classfile.JavaClass;

import tetasarts.tir.Node;
import tetasarts.tir.methods.MethodInformation;

public abstract class AbstractClassNode extends Node<IClassVisitor>{
//	static {
//		classGraph = new ClassGraph();
//	}
//	static ClassGraph classGraph;		
	private Logger logger = Logger.getLogger("generator");

	private String canonicalName;
	
	JavaClass javaClass;
	
	public JavaClass getJavaClass() {
		return javaClass;
	}

	public void setJavaClass(JavaClass javaClass) {
		this.javaClass = javaClass;
	}
	
	
	private ClassGraph classGraph;
	
	public AbstractClassNode(String canonicalName, ClassGraph classGraph){
		this.canonicalName = canonicalName;
		this.classGraph = classGraph;
	}
	
	public MethodInformation getConcreteMethodImplementation(String methodName, String methodSignature) {
		AbstractClassNode current = this;
		MethodInformation methodInformation = current.getMethod(methodName, methodSignature);
		
		// find first concrete implementation
		while(methodInformation == null) {
			List<AbstractClassNode> list = classGraph.getPreds(current);
			if(list.size() == 0) {
				logger.info("No parent implementation for method: " + getCanonicalName() + "." + methodInformation.getCanonicalName() + methodInformation.getSignature());
				break;
			}
			assert list.size() == 1 : "Class contains more than one predecessor! (multiple inheritance?)";
			current = list.get(0);
			methodInformation = current.getMethod(methodName, methodSignature);
		}
		
		return methodInformation;
	}

	
	public String getCanonicalName(){
		return canonicalName;
	}
	
	private Map<String, MethodInformation> methods = new HashMap<String, MethodInformation>();
	public Map<String, MethodInformation> getMethods(){
		return methods;
	}
	
	public MethodInformation getMethod(String methodName, String methodSignature) {
		return methods.get(methodName + methodSignature);
	}
	
	public MethodInformation getMethod(String methodNameAndSignature) {
		return methods.get(methodNameAndSignature);
	}
	
	public void addMethod(MethodInformation method){
		methods.put(method.getName() + method.getSignature(), method);
	}
	
	@Override
	public abstract void accept(IClassVisitor v); 
	
	public boolean isAbstract(){
		return false;
	}
	
	public boolean inheritsFrom(String className, ClassGraph classGraph) {
		if(canonicalName.equals(className)) {
			return false;
		}
		boolean result = false;
		AbstractClassNode parent = this;
		
		while (!result) {
			if(parent.getCanonicalName().equals(className)) {
				result = true;
			}
			
			if(classGraph.getPreds(parent).size() == 0) {
				return false;
			}
			for(AbstractClassNode current : classGraph.getPreds(parent)) {
				parent = current;
			}
		}
		
		return result;
	}

//	public abstract AbstractClassNode clone();
	
//	@Override
//	public AbstractClassNode clone() {
//		AbstractClassNode result = null;
//		try {
//			result = (AbstractClassNode)super.clone();
//			
//		} catch (CloneNotSupportedException e) {
//			// TODO Auto-generated catch block
//			e.printStackTrace();
//		}
//		return result;
//	}
}
