package tetasarts.tir.classinfo;

import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.logging.Logger;

import tetasarts.tir.Graph;
import tetasarts.tir.methods.MethodInformation;


public class ClassGraph extends Graph<AbstractClassNode, IClassVisitor>{
	private Map<String, AbstractClassNode> nameToClassMap = new HashMap<String, AbstractClassNode>();

	public ClassGraph() {
		super();
	}
	
	public AbstractClassNode getClassFromCanonicalName(String name){
		return nameToClassMap.get(name);
	}
	
	public Set<AbstractClassNode> getClassesFromSimpleName(String name) {
		Set<AbstractClassNode> result = new HashSet<AbstractClassNode>();
		
		for(AbstractClassNode current : nameToClassMap.values()) {
			int index = current.getCanonicalName().lastIndexOf(".");
			if(index != -1) {
				String simpleName = current.getCanonicalName().substring(index + 1);
				if(simpleName.equals(name)) {
					result.add(current);
				}
			}
		}
		
		return result;
	}

	public boolean contains(String canonicalName) {
		return nameToClassMap.containsKey(canonicalName);
	}
	
	@Override
	public boolean addNode(AbstractClassNode node) {
		nameToClassMap.put(node.getCanonicalName(),node);
		return super.addNode(node);
	}
}
