package tetasarts.tir.methods;

import java.util.LinkedList;
import java.util.List;

import org.apache.bcel.classfile.JavaClass;
import org.apache.bcel.classfile.Method;

import tetasarts.Options;
import tetasarts.generator.LocationBuilder;
import tetasarts.generator.LoopBoundAnalysis;
import tetasarts.generator.TransitionBuilder;
import tetasarts.tir.basicblocks.AbstractBasicBlock;
import tetasarts.tir.basicblocks.BasicBlockGraph;

public class LazyMethodInformation extends MethodInformation {

	JavaClass jc;
	Method currentMethod;
	LocationBuilder locationBuilder;
	TransitionBuilder transitionBuilder;

	public LazyMethodInformation(JavaClass jc, Method currentMethod,
			LocationBuilder locationBuilder, TransitionBuilder transitionBuilder) {
		// TODO Auto-generated constructor stub
		super(jc.getClassName(), currentMethod.getName(), currentMethod.getSignature(), currentMethod.getCode().getCode().length);
		this.jc = jc;
		this.currentMethod = currentMethod;
		this.locationBuilder = locationBuilder;
		this.transitionBuilder = transitionBuilder;
		setGenerated(false);
	}
	
	public BasicBlockGraph getInstructionGraph(){
		// lazily generate instructiongraph...

		if(!isGenerated()){
//			System.out.println("Generating "+this.getCanonicalName());
//			assert false : "STOP";
			// save old state -- probably not needed
			JavaClass lbClass       = locationBuilder.getCurrentClass();
			MethodInformation lbMIT = locationBuilder.getCurrentMethodTarget(); 
			JavaClass tbClass       = transitionBuilder.getCurrentClass();
			MethodInformation tbMIT = transitionBuilder.getCurrentMethodTarget();
			
			locationBuilder.setCurrentClass(jc);
			transitionBuilder.setCurrentClass(jc);
			
			locationBuilder.visitMethod(currentMethod, this);
			
			transitionBuilder.visitMethod(currentMethod, this);
			if(Options.PRECISE_LOOP) {
				LoopBoundAnalysis loopboundanalysis = new LoopBoundAnalysis(this.instructionGraph, currentMethod, jc);
				loopboundanalysis.conductAnalysis();
			}
			locationBuilder.setCurrentClass(lbClass);
			locationBuilder.setCurrentMethodTarget(lbMIT);
			transitionBuilder.setCurrentClass(tbClass);
			transitionBuilder.setCurrentMethodTarget(tbMIT);
			setGenerated(true);
		}
		
		return instructionGraph;
	}
}
