package tetasarts.uppaal;

import org.jdom.*;

class Label extends Name {
	private LabelKind kind;

	public String toString(){
		return getName();
	}
	/**
	 * The different possible kinds/types of labels.
	 */
	enum LabelKind {   
		SELECT, GUARD, SYNCHRONISATION, ASSIGNMENT, INVARIANT, COMMENTS, EXPONENTIAL_RATE
	};
	
	/**
	 * Creates a label of a specific type but without any name or position
	 * @param kind The type of label
	 */
	Label(LabelKind kind){
		this("", kind, 0, 0);
	}

	/**
	 * Creates a new label element with the specified name and kind/type.
	 * The position is set to 0,0
	 * @param name The name of the label element
	 * @param kind The kind/type of label element
	 */
	Label(String name, LabelKind kind) {
		this(name, kind, 0, 0);
	}

	/**
	 * Creates a new label element with the specified name and kind/type and the position of the element
	 * @param name The name of the label element
	 * @param kind The kind/type of label element
	 * @param x The x position of the label location
	 * @param y The y position of the label location
	 */
	Label(String name, LabelKind kind, int x, int y) {
		super(name, x, y);
		this.kind = kind;
	}

	/**
	 * Creates an XML Element object corresponding to the Label object
	 * @return XML Element
	 */
	Element generateXMLElement() {
		Element result = super.generateXMLElement();
		switch (this.kind) {
		case SELECT:
			result.setAttribute("kind", "select");
			break;
		case GUARD:
			result.setAttribute("kind", "guard");
			break;
		case SYNCHRONISATION:
			result.setAttribute("kind", "synchronisation");
			break;
		case ASSIGNMENT:
			result.setAttribute("kind", "assignment");
			break;
		case INVARIANT:
			result.setAttribute("kind", "invariant");
			break;
		case COMMENTS:
			result.setAttribute("kind", "comments");
			break;
		case EXPONENTIAL_RATE:
			result.setAttribute("kind", "exponentialrate");
			break;
		default:
		}

		return result;
	}
}
