package tetasarts.uppaal;

import java.util.LinkedList;
import java.util.List;

import org.jdom.*;

import tetasarts.uppaal.Label.LabelKind;

public class Location {
	//location (name?, label*, urgent?, committed?)
	//attributes
	//ID must be unique
	private int id;
	private int posX;
	private int posY;
	//Color is possible in the DTD but we don't want to use it
	//private int color;
	//name
	private Name name;
	
	//label*
	private Label comment;
	private Label invariant;
	private Label exponentialRate;
	
	//urgent?
	//committed?
	/**
	 * The different types of locations
	 */
	public enum LocationType { NORMAL, URGENT, COMMITED };
	private LocationType type;
	
	/**
	 * Creates a new location with the given name and id
	 * The name and id must be unique in the template, which is not checked here
	 * @param name The name of the location
	 * @param id The id of the location
	 */
	public Location(String name){
		this(name, LocationType.NORMAL, 0, 0);
	}
	
	List<String> codeLines = new LinkedList<String>();
	public void addCodeLine(String line){
		if(!codeLines.contains(line)) 
			codeLines.add(line);
	}
		
	/**
	 * Creates a new location with all the properties set.
	 * The name and id must be unique in the template, which is not checked here
	 * @param name The name of the location
	 * @param type The type of the location
	 * @param x The x position of the Location
	 * @param y The y position of the Location
	 */
	public Location(String name, LocationType type, int x, int y){
		this.id = generateId();
		this.name = new Name(name + this.id, x, y);
		this.type = type;
		this.posX = x;
		this.posY = y;
		this.comment = new Label("", LabelKind.COMMENTS, x, y);
		this.invariant = new Label("", LabelKind.INVARIANT, x, y);
		this.exponentialRate = new Label("", LabelKind.EXPONENTIAL_RATE, x, y);
	}

	public int getId() {
		return id;
	}

	public int getPosX() {
		return posX;
	}

	public void setPosX(int posX) {
		int diff = posX - this.posX;
		this.name.setPosX(name.getPosX()+diff);
		this.invariant.setPosX(invariant.getPosX()+diff);
		this.posX = posX;
	}

	public int getPosY() {
		return posY;
	}

	public void setPosY(int posY) {
		int diff = posY - this.posY;
		this.name.setPosY(name.getPosY()+diff);
		this.invariant.setPosY(invariant.getPosY()+diff);
		this.posY = posY;
	}

	public Name getName() {
		return name;
	}

	public void setName(Name name) {
		this.name = name;
	}

	public LocationType getType() {
		return type;
	}

	public void setType(LocationType type) {
		this.type = type;
	}
	
	public Label getComment() {
		StringBuilder sb = new StringBuilder(comment.getName());
		for(String codeString : codeLines){
			//newLabelName = newLabelName + codeString;
			sb.append("\n"+codeString);
		}
		String newLabelName = sb.toString();
//		System.out.println("HMM:" + sb.toString());
//		System.exit(0);
		Label newComment = new Label(newLabelName, LabelKind.COMMENTS, comment.getPosX(), comment.getPosY()); 
		return newComment;
	}

	public void setComment(String comment) {
		this.comment.setName(comment);
	}

	public Label getInvariant() {
		return invariant;
	}

	public void setInvariant(String invariant) {
		this.invariant.setName(invariant);
	}
	
	public void setInvariant(String invariant, int x, int y){
		this.invariant.setName(invariant);
		this.invariant.setPosX(x);
		this.invariant.setPosY(y);
	}
	
	public void setInvarinatPosition(int x, int y){
		this.invariant.setPosX(x);
		this.invariant.setPosY(y);
	}
	
	public Label getExponentialRate() {
		return exponentialRate;
	}

	public void setExponentialRate(String expRate) {
		this.exponentialRate.setName(expRate);
	}
	
	public void setExponentialRate(String expRate, int x, int y){
		this.exponentialRate.setName(expRate);
		this.exponentialRate.setPosX(x);
		this.exponentialRate.setPosY(y);
	}
	
	public void setExponentialRatePosition(int x, int y){
		this.exponentialRate.setPosX(x);
		this.exponentialRate.setPosY(y);
	}
	
	/**
	 * Creates an XML Element object corresponding to the Location object
	 * @return XML Element
	 */
	Element generateXMLElement(){
		Element result = new Element("location");
		
		result.setAttribute("id", "id"+this.id);
		result.setAttribute("x", Integer.toString(this.posX));
		result.setAttribute("y", Integer.toString(this.posY));
		
		result.addContent(getName().generateXMLElement());
		
		result.addContent(getComment().generateXMLElement());
		result.addContent(getInvariant().generateXMLElement());
		result.addContent(getExponentialRate().generateXMLElement());
		
		if(this.type == LocationType.COMMITED){
			result.addContent(new Element("committed"));
		} else if (this.type == LocationType.URGENT){
			result.addContent(new Element("urgent"));
		}		
		
		return result;
	}
	// overrides
	public String toString(){
		return name.toString();
	}
	
	// analysis specific
	private int wcet;
	public void setWCET(int wcet){
		this.wcet = wcet;
	}
	public int getWCET(){
		return wcet;
	}
	
	private static int idCount = 0;
	
	private static int generateId() {
		return idCount++;
	}
}
