package tetasarts.uppaal;

import java.util.*;

import org.jdom.*;

import tetasarts.uppaal.Label.LabelKind;
import tetasarts.uppaal.declarations.Channel;

public class Transition {
	//transition (source, target, label*, nail*)
	//attributes
	//The following attributes are allowed in the DTD but we don't want to use it
	//private int id;
	//private int posX;
	//private int posY;
	//private int color;
	
	//source
	//TODO source must be in the template
	private int source;
	
	//target
	//TODO target must be in the template
	private int target;
	
	//label*
	private Label select;
	private Label guard;
	private Label sync;
	private Label update;
	
	//nail*
	private ArrayList<Nail> nails;
	
	/**
	 * Creates a new transition between the source and target.
	 * These must exist in the template
	 * @param sourceID The source location
	 * @param targetID The target location
	 */
	public Transition(int sourceID, int targetID){
		this(sourceID, targetID, new ArrayList<Nail>());
	}
	/**
	 * Creates a new transition between the source and target.
	 * These must exist in the template
	 * @param sourceID The source location
	 * @param targetID The target location
	 * @param labels The labels for the transition
	 * @param nails A list of nails/positions, to change the apperance of the transition
	 */
	public Transition(int sourceID, int targetID, ArrayList<Nail> nails){
		this.source = sourceID;
		this.target = targetID;
		this.nails = nails;
		this.select = new Label(LabelKind.SELECT);
		this.guard = new Label(LabelKind.GUARD);
		this.sync = new Label(LabelKind.SYNCHRONISATION);
		this.update = new Label(LabelKind.ASSIGNMENT);
	}

	public int getSource() {
		return source;
	}

	public void setSource(int source) {
		this.source = source;
	}

	public int getTarget() {
		return target;
	}

	public void setTarget(int target) {
		this.target = target;
	}

	public ArrayList<Nail> getNails() {
		return nails;
	}

	public void setNails(ArrayList<Nail> nails) {
		this.nails = nails;
	}
	
	public void addNail(Nail nail){
		this.nails.add(nail);
	}
	
	public Label getSelect() {
		return select;
	}
	public void setSelect(String select, int x, int y){
		this.select.setName(select);
		this.select.setPosX(x);
		this.select.setPosY(y);
	}
	public Label getGuard() {
		return guard;
	}
	public void setGuard(String guard, int x, int y){
		this.guard.setName(guard);
		this.guard.setPosX(x);
		this.guard.setPosY(y);
	}

	public void setGuard(String guard){
		setGuard(guard, 0,0);
	}
	
	public void addGuard(String guard){
		if (this.guard.getName()=="")
			this.guard.setName(guard);
		else 
			this.guard.setName(this.guard.getName() + " && \n" + guard);
	}
	
	public Label getSync() {
		return sync;
	}
	public void setSync(Channel channel, Boolean signal, int x, int y){
		if (signal)
			sync.setName(channel.getIdentifier() + "!");
		else 
			sync.setName(channel.getIdentifier() + "?");
		
		sync.setPosX(x);
		sync.setPosY(y);
	}
	
	public void setSync(Channel channel, Boolean signal){
		if (signal)
			sync.setName(channel.getIdentifier() + "!");
		else 
			sync.setName(channel.getIdentifier() + "?");
	}
	
	public void setSyncRunOrFire(String name){
		sync.setName(name);
	}
	
	public Label getUpdate() {
		return update;
	}
	public void setUpdate(String update, int x, int y){
		this.update.setName(update);
		this.update.setPosX(x);
		this.update.setPosY(y);
	}	
	public void setUpdate(String update){
		setUpdate(update, 0,0);
	}
	
	public void addUpdate(String update){
		if (this.update.getName()=="")
			this.update.setName(update);
		else 
			this.update.setName(this.update.getName() + ",\n" + update);
	}
	
	/**
	 * Creates an XML Element object corresponding to the Transition object
	 * @return XML Element
	 */
	Element generateXMLElement(){
		Element result = new Element("transition");
		
		Element sourceElement = new Element("source");
		sourceElement.setAttribute("ref", "id"+this.source);
		result.addContent(sourceElement);
		
		Element targetElement = new Element("target");
		targetElement.setAttribute("ref", "id"+this.target);
		result.addContent(targetElement);
		
		result.addContent(this.select.generateXMLElement());
		result.addContent(this.guard.generateXMLElement());
		result.addContent(this.sync.generateXMLElement());
		result.addContent(this.update.generateXMLElement());
		
		for (Nail nail : nails) {
			result.addContent(nail.generateXMLElement());
		}
		
		return result;		
	}
}
