package tetasarts.uppaal;


import java.io.IOException;
import java.io.PrintStream;
import java.util.*;

import org.jdom.*;
import org.jdom.input.JDOMParseException;
import org.jdom.input.SAXBuilder;
import org.jdom.output.Format;
import org.jdom.output.XMLOutputter;

import com.sun.tools.javac.comp.Todo;

import tetasarts.Options;
import tetasarts.Options.SchedulerType;
import tetasarts.generator.Constants;
import tetasarts.tir.classinfo.PeriodicTask;
import tetasarts.tir.classinfo.SporadicTask;
import tetasarts.tir.methods.MethodInformation;
import tetasarts.uppaal.Location.LocationType;
import tetasarts.uppaal.declarations.Bool;
import tetasarts.uppaal.declarations.Channel;
import tetasarts.uppaal.declarations.Clock;
import tetasarts.uppaal.declarations.Declarations;
import tetasarts.uppaal.declarations.Int;
import tetasarts.uppaal.schedulers.AbstractSchedulerUppaalModel;
import tetasarts.uppaal.schedulers.EdfScheduler;
import tetasarts.uppaal.schedulers.FifoScheduler;
import tetasarts.uppaal.schedulers.FpsScheduler;
import tetasarts.uppaal.schedulers.SchedulerTemplate;
import tetasarts.uppaal.translation.ExplicitModelCreater;
import tetasarts.uppaal.translation.PeriodicThreadDeadline;
import tetasarts.uppaal.translation.SporadicThreadDeadline;
import tetasarts.uppaal.translation.ThreadDeadline;

public class UppaalModel {
	// nta (imports?, declaration?, template+, instantiation?, system)
	// imports? Does not work with UPPAAL....
	// private String imports;
	// declarations?
	private Declarations declarations = new Declarations();
	// template+
	private ArrayList<UppaalTemplate> templates;
	// instantiation? Is not used, because it is redundant when we have system
	// private String instantiation;
	// system
	private String system;

	public int nrPeriodicThreads = 0;
	public int nrSporadicThreads = 0;
	public long LCM = 0;//Integer.MAX_VALUE/4;
	
	private Set<PeriodicTask> periodicTasks;
	private Set<SporadicTask> sporadicTasks;

	private AbstractSchedulerUppaalModel schedulerModel = null;
	
	/**
	 * Creates an empty model
	 */
	public UppaalModel() {
		this(new Declarations(), new ArrayList<UppaalTemplate>(), "");
	}

	/**
	 * Creates a model with all properties set
	 * 
	 * @param declarations
	 *            Global declarations
	 * @param templates
	 *            The different templates in the model
	 * @param system
	 *            Instantiation of the different templates
	 */
	public UppaalModel(Declarations declarations,
			ArrayList<UppaalTemplate> templates, String system) {
		this.declarations = declarations;
		this.declarations.addBoolean(new Bool("notImplementedMethods"));
		this.declarations.addBoolean(new Bool("loopAnnotationNotSet"));
		this.templates = templates;
		this.system = system;
	}

	public Declarations getDeclarations() {
		return declarations;
	}

	public void setDeclarations(Declarations declarations) {
		this.declarations = declarations;
	}

	public ArrayList<UppaalTemplate> getTemplates() {
		return templates;
	}

	public void setTemplates(ArrayList<UppaalTemplate> templates) {
		this.templates = templates;
	}

	public String getSystem() {
		return system;
	}

	public void setSystem(String system) {
		this.system = system;
	}
	
	public Set<PeriodicTask> getPeriodicTasks() {
		return periodicTasks;
	}

	public void setPeriodicTasks(Set<PeriodicTask> periodicTasks) {
		this.periodicTasks = periodicTasks;
	}

	public Set<SporadicTask> getSporadicTasks() {
		return sporadicTasks;
	}

	public void setSporadicTasks(Set<SporadicTask> sporadicTasks) {
		this.sporadicTasks = sporadicTasks;
	}

	public void addTemplate(UppaalTemplate template) {
		templates.add(template);
	}

	private void updateLabelPositions() {
		int spaceBetweenLabels = 15; //The size between lines in UPPAAL
		int locationDistance = 30;

		for (UppaalTemplate template : templates) {
			for (Location location : template.getLocations()) {
				int x = location.getPosX();
				int y = location.getPosY();
				int offset = locationDistance;
				if (location.getName().getName() != "") {
					location.getName().setPosX(x - spaceBetweenLabels);
					location.getName().setPosY(y - offset);
					offset += spaceBetweenLabels;
				}
				if (location.getInvariant().getName() != "") {
					location.getInvariant().setPosX(x - spaceBetweenLabels);
					location.getInvariant().setPosY(y - offset);
				}
			}
			for (Transition transition : template.getTransitions()) {
				int sourceID = transition.getSource();
				int targetID = transition.getTarget();
				int sourceX = 0;
				int sourceY = 0;
				int targetX = 0;
				int targetY = 0;

				ArrayList<Nail> nails = transition.getNails();
				int nrNails = nails.size();
				if (nrNails == 0) {
					for (Location location : template.getLocations()) {
						if (location.getId() == sourceID) {
							sourceX = location.getPosX();
							sourceY = location.getPosY();
						}
						if (location.getId() == targetID) {
							targetX = location.getPosX();
							targetY = location.getPosY();
						}
					}
				} else if (nrNails == 1) {
					for (Location location : template.getLocations()) {
						if (location.getId() == sourceID) {
							sourceX = location.getPosX();
							sourceY = location.getPosY();
							break;
						}
					}
					targetX = nails.get(0).getPosX();
					targetY = nails.get(0).getPosY();
				} else {
					sourceX = nails.get((nrNails / 2) - 1).getPosX();
					sourceY = nails.get((nrNails / 2) - 1).getPosY();
					targetX = nails.get(nrNails / 2).getPosX();
					targetY = nails.get(nrNails / 2).getPosY();
				}

				int x = sourceX + (targetX - sourceX) / 2;
				int y = sourceY + (targetY - sourceY) / 2;

				// TODO test if it works
				int offset = -locationDistance;
				if (transition.getSelect().getName() != "") {
					transition.getSelect().setPosX(x - locationDistance);
					transition.getSelect().setPosY(y + offset);
					//offset must be added with spaceBetweenLabels as many times there are newlines
					offset += spaceBetweenLabels * transition.getSelect().getName().split("\n").length;
				}
				if (transition.getGuard().getName() != "") {
					transition.getGuard().setPosX(x - locationDistance);
					transition.getGuard().setPosY(y + offset);
					//offset must be added with spaceBetweenLabels as many times there are newlines
					offset += spaceBetweenLabels * transition.getGuard().getName().split("\n").length;
				}
				if (transition.getSync().getName() != "") {
					transition.getSync().setPosX(x - locationDistance);
					transition.getSync().setPosY(y + offset);
					//offset must be added with spaceBetweenLabels as many times there are newlines
					offset += spaceBetweenLabels * transition.getSync().getName().split("\n").length;
				}
				if (transition.getUpdate().getName() != "") {
					transition.getUpdate().setPosX(x - locationDistance);
					transition.getUpdate().setPosY(y + offset);
					//offset must be added with spaceBetweenLabels as many times there are newlines
					offset += spaceBetweenLabels * transition.getUpdate().getName().split("\n").length;
				}
			}
		}
	}

	final int xInterval = 200;
	final int yInterval = 125;
	final int nailX1 = 25;
	final int nailY1 = 25;
	final int nailX2 = 25;
	final int nailY2 = -25;

	public void userFriendlyOutput() {
		for (UppaalTemplate template : templates) {
			if (template.getInit() == null) {
				int initId = Integer.MAX_VALUE;
				Location initLoc = null;
				for (Location location : template.getLocations()) {
					// if no init is set, the location with the lowest ID is
					// chosen
					if (location.getId() < initId) {
						initId = location.getId();
						initLoc = location;
					}
					// reset all positions
					location.setPosX(-1);
					location.setPosY(-1);
				}
				initLoc.setPosX(0);
				initLoc.setPosY(0);
				template.setInit(initLoc);
			} else {
				for (Location location : template.getLocations()) {
					// if no init is set, the location with the lowest ID is
					// chosen
					location.setPosX(-1);
					location.setPosY(-1);
				}
				template.getInit().setPosX(0);
				template.getInit().setPosY(0);
			}

			updateLocations(template, template.getInit());
			
			//Add nails to loops
			for (Transition transition : template.getTransitions()) {
				if (transition.getSource() == transition.getTarget()) {
					for (Location location : template.getLocations()) {
						if (location.getId() == transition.getSource()){
							transition.addNail(new Nail(location.getPosX() + nailX1,
									location.getPosY() + nailY1));
							transition.addNail(new Nail(location.getPosX() + nailX2,
									location.getPosY() + nailY2));
						}
					}
				}
			}
		}
		
		
		// Updates the position of the labels after the locations and
		// transitions has been moved
		updateLabelPositions();
	}

	private void updateLocations(UppaalTemplate template,
			Location currentLocation) {
		ArrayList<Location> succ = new ArrayList<Location>();
		int locationOffset = 0;
		for (Transition transition : template.getTransitions()) {
			// Only change location if it is not a loop
			if (transition.getSource() != transition.getTarget()) {
				if (transition.getSource() == currentLocation.getId()) {
					// Find the opposite location
					for (Location location : template.getLocations()) {
						if (location.getId() == transition.getTarget()) {
							// only update the position if it has not already
							// been updated
							if (location.getPosX() == -1) {
								location.setPosX(currentLocation.getPosX()
										+ xInterval * locationOffset);
								location.setPosY(currentLocation.getPosY()
										+ yInterval);
								succ.add(location);
								locationOffset++;
								// Check if it is a A -> B -> A set of
								// transitions
								for (Transition trans2 : template
										.getTransitions()) {
									if (trans2.getSource() == transition
											.getTarget()
											&& trans2.getTarget() == transition
													.getSource()) {
										if (trans2.getNails().size() == 0
												&& transition.getNails().size() == 0) {
											// if it is a A -> B -> A then we
											// want B to be to the right of A,
											// but it must not interfere with
											// other locations so subtract 50
											// from the X position
											location.setPosX(currentLocation
													.getPosX()
													+ xInterval
													* (locationOffset) - 50);
											location.setPosY(currentLocation
													.getPosY());
											trans2.addNail(new Nail(
													currentLocation.getPosX()
															+ nailX1,
													currentLocation.getPosY()
															+ nailY1));
											transition.addNail(new Nail(
													currentLocation.getPosX()
															+ nailX2,
													currentLocation.getPosY()
															+ nailY2));
											break;
										}
									}
								}
								break;
							} else {
								// Check if the location is above the current
								// location and at the same X coordinate
								if (currentLocation.getPosX() == location
										.getPosX()
										&& currentLocation.getPosY() > location
												.getPosY()) {
									currentLocation.setPosX(currentLocation
											.getPosX()
											+ xInterval - 50);
								}
							}
						}
					}
				}
			}
		}
		for (Location location : succ) {
			updateLocations(template, location);
		}
	}

	/**
	 * Creates a valid UPPAAL XML document, containing the entire model
	 * 
	 * @return The UPPAAL model as an XML document
	 */
	private Document generateXMLDocument() {
		userFriendlyOutput();
//		updateLabelPositions();

		Element nta = new Element("nta");
		
		Element system = new Element("system");

		
		switch(Options.SCHEDULER_TYPE) {
			case FIFO:
				schedulerModel = new FifoScheduler();
				break;
			case EDF:
				schedulerModel = new EdfScheduler();
				break;
			case FPS:
				schedulerModel = new FpsScheduler();
				break;
			default:
				String msg = "Not capable of setting the scheduler.";
				System.err.println(msg);
				System.exit(-1);
		}
		
		/*
		 * First, build the declarations
		 */
		Element declarations = new Element("declaration");
		
		// Adds the amount of threads
		Declarations decl = new Declarations();
		StringBuilder totalThreadsStr = new StringBuilder();
		totalThreadsStr.append("const int totalThreads = ");
		if(this.nrPeriodicThreads > 0) {
			String periodicThreadStr = "periodicThreads";
			decl.addInteger(new Int(periodicThreadStr, this.nrPeriodicThreads,
					true));
			totalThreadsStr.append(periodicThreadStr);
			if(this.nrSporadicThreads > 0)
				totalThreadsStr.append(" + ");
			else
				totalThreadsStr.append(";");
		}
		if(this.nrSporadicThreads > 0) {
			decl.addInteger(new Int("sporadicThreads", this.nrSporadicThreads,
					true));
			totalThreadsStr.append("sporadicThreads;");
		}
		
		// add total threads
		decl.addString(totalThreadsStr.toString());
		
		// Add utilization
		if(Options.UTILIZATION_MONITOR) {
			decl.addString("//Utilization monitor uses this");
			decl.addBoolean(new Bool("idling", true, false));
		}
		// Add schedulerID
		decl.addInteger(new Int("schedulerID", 0, true));
		

		// Add IDs for periodic and sporadic tasks
		if(this.nrPeriodicThreads > 0)
			decl.addString("typedef int[1,periodicThreads] PeriodicID;");
		if(this.nrSporadicThreads > 0) {
			if(this.nrPeriodicThreads > 0) {
				decl.addString("typedef int[periodicThreads + 1, totalThreads] SporadicID;");
			} else {
				decl.addString("typedef int[1, sporadicThreads] SporadicID;");
			}					
		}
		decl.addString("typedef int[1,totalThreads] ThreadID;");
		
		// Add scheduler-specific arrays
		decl.addString("int running[totalThreads+1];");
		decl.addString("//Scheduler");
		decl.addString("bool schedulable[ThreadID];");
		decl.addString("bool blocked[ThreadID];");
		if(this.nrSporadicThreads > 0)
			decl.addString("bool fireable[SporadicID];");
		
		// add channel arrays for periodic and sporadic tasks
		decl.addString(((Options.STATISTICAL_MODEL_CHECKING) ? "broadcast " : "") + "chan run[ThreadID];");
		if(this.nrSporadicThreads > 0)
			decl.addString(((Options.STATISTICAL_MODEL_CHECKING) ? "broadcast " : "") + "chan fire[SporadicID];");
		
		// add broadcast channel for initialising the system
		decl.addChannel(new Channel("GO", "", true));
		
		// add synchronisation booleans
		decl.addString("// bools for synchronization");
		decl.addBoolean(new Bool("synchronized", false, false));
		//decl.addBoolean(new Bool("interruptWaiting", false, false));
		decl.addString("\n");
		
		// add globalclock used for determining when the hyper period is exceeded
		decl.addClock(new Clock("globalClock"));
		
		//If using SMC, we use the following constant to denote the exponential rate
		if(Options.STATISTICAL_MODEL_CHECKING)
			decl.addString("const int " + Constants.EXPONENTIAL_RATE_STR + " = 1000000;");
		
		// LCM calculation
		long periods[] = new long[periodicTasks.size()+sporadicTasks.size()];
		int count = 0;
		for(Iterator<PeriodicTask> iter = periodicTasks.iterator();iter.hasNext();){
			periods[count] = iter.next().getPeriod();
			count++;
		}
		for(Iterator<SporadicTask> iter = sporadicTasks.iterator();iter.hasNext();){
			periods[count] = iter.next().getMinInterarrival();
			count++;
		}
		this.LCM = LCMCalc.computeLCM(periods);
		System.out.println("LCM: "+ this.LCM);
		decl.addInteger(new Int("exerciseBound", (int)(this.LCM*0.06), true));
		// End LCM Calculation
		
		// Add functions
		// TODO: these may not be general? FIFO cannot use synchronization for anything?
		StringBuilder funcBuilder = new StringBuilder();
		funcBuilder
			.append("void runScheduler(){\n")
			.append("	int i;\n")
			.append("	for (i = 0; i <= totalThreads; i++){\n")
			.append("		running[i] = 0;\n")
			.append("	}\n")
			.append("	running[schedulerID] = 1;\n")
			.append("}\n\n")
			.append("int monitorDepth = 0;\n")
			.append("void monitorEnter(){\n")
			.append("	synchronized = true;\n")
			.append("	monitorDepth++;\n")
			.append("}\n\n")
			.append("void monitorExit(){\n")
			.append("	monitorDepth--;\n")
			.append("	if (monitorDepth == 0){\n")
			.append("		synchronized = false;\n")
			.append("		for(i:ThreadID) {\n")
			.append("			blocked[i] = false;\n")
			.append("		}\n")
			.append("		runScheduler();\n")
			.append("	}\n")
			.append("}\n");
		decl.addString(funcBuilder.toString());
		
		// get declarations that apply specifically for the scheduler
		decl.addString(schedulerModel.getDeclarations().toString());
		
		
		// This is kind of hackish, but we do not want to include priorities unless we are using
		// FPS scheduling
		if(Options.SCHEDULER_TYPE == SchedulerType.FPS) {
			decl.addString(this.declarations.toString());
		}
		else {
			String[] additionalDeclarations = this.declarations.toString().split("\n");
			for(String str : additionalDeclarations) {
				if(!str.toLowerCase().contains("threadpriority"))
					decl.addString(str);
			}
		}
		if(Options.EXPLICIT_MODEL) {
			decl.addString(((Options.STATISTICAL_MODEL_CHECKING) ? "broadcast" : "urgent") + " chan jvm_execute;");
			decl.addString("int[0,255] jvm_instruction;");
		}
	
		Collections.sort(templates);
		for (UppaalTemplate template : templates) {
			if(Options.EXPLICIT_MODEL) {
				Declarations de = template.getDeclaration();
				// We need to use an iterator because it is the only
				// way of removing items safely while doing iteration
				Iterator<Int> iter = de.getIntegers().iterator();
				Int instructionInt = null;
				while(iter.hasNext()) {
					instructionInt = iter.next();
					if(instructionInt.getId().contains("JVM_")) {
						boolean intAlreadyInDecl = false;
						for(Int integer : decl.getIntegers()) {
							if(integer.getId().equals(instructionInt.getId())) {
								intAlreadyInDecl = true;
								break;
							}
						}
						if(!intAlreadyInDecl)
							decl.addInteger(new Int(instructionInt.getId(), instructionInt.getValue(), instructionInt.isConstant()));
						iter.remove();					
					}
				}
			}
		}
		
		// Set declaration!
		declarations.addContent(decl.toString());
		nta.addContent(declarations);
		
		/*
		 * Second, build the templates and, meanwhile, build the system declarations!
		 */
		StringBuilder systemDeclBuilder = new StringBuilder("//Initialization\n");
		StringBuilder systemBuilder = new StringBuilder("\nsystem\n");
		
		// Add the scheduler, periodic, and sporadic controller templates
		SchedulerTemplate schedTemplate = schedulerModel.getScheduler();
		nta.addContent(schedTemplate.generateXMLElement());
		systemBuilder.append("       ").append(schedTemplate.getName()).append(",\n");
		if(this.nrPeriodicThreads > 0)
			nta.addContent(schedulerModel.getPeriodicController().generateXMLElement());
		if(this.nrSporadicThreads > 0)
			nta.addContent(schedulerModel.getSporadicController().generateXMLElement());
		
		
		// build all templates of public methods and add the system declarations			
		Collections.sort(templates);
		for (UppaalTemplate template : templates) {
			
			if (template instanceof PublicMethod) {
				systemDeclBuilder.
					append(template.getName()).
					append("_threads(const ThreadID tID) = ").
					append(template.getName()).
					append("(tID);\n");
				systemBuilder.append("       ").append(template.getName()).append("_threads, \n");

			}
			nta.addContent(template.generateXMLElement());
		}

		
		
		if(this.nrSporadicThreads > 0) {
			//systemBuilder.append("//Sporadic Threads\n");
			systemDeclBuilder.append("//Sporadic Threads\n");
		}
		for (SporadicTask sporadicTask : sporadicTasks) {
			String templateName;
			/*
			 * TODO: aperiodiceventhandler is not accounted for yet...
			 */
			if(Options.SCJ_API == Options.ScjApi.SCJ) {
				templateName = convertString("Template_", sporadicTask.getInstanceOf().getMethod("run", "()Z").getCanonicalName());
			} else
				templateName = convertString("Template_", sporadicTask.getInstanceOf().getMethod("run", "()Z").getCanonicalName());				
			
			systemDeclBuilder.
				append(templateName).
				append("_").
				append(sporadicTask.getTID()).
				append(" = ").
				append(templateName).
				append("(").
				append(sporadicTask.getTID()).
				append(");\n");
		
			//systemBuilder.append("       ").append(templateName).append("_").append(sporadicTask.getTID()).append(", ");
			
			systemDeclBuilder.
				append(schedulerModel.getSporadicController().getName()).
				append(sporadicTask.getTID()).
				append(" = ").
				append(schedulerModel.getSporadicController().getName()).
				append("(").
				append(sporadicTask.getTID()).
				append(", ").
				append(convertToClockCycles(sporadicTask.getMinInterarrival())).append(", ").
				append(convertToClockCycles(sporadicTask.getDeadline())).append(");\n");
			
			/*systemBuilder.
				append(schedulerModel.getSporadicController().getName()).
				append(sporadicTask.getTID()).
				append(", \n");*/
		}
		
		if(this.nrPeriodicThreads > 0) {
			//systemBuilder.append("//Periodic Threads\n");
			systemDeclBuilder.append("//Periodic Threads\n");
		}
		for (PeriodicTask periodicTask : periodicTasks) {
			String templateName;
			if(Options.SCJ_API == Options.ScjApi.SCJ) {
				templateName = convertString("Template_", periodicTask.getInstanceOf().getMethod("handleAsyncEvent", "()V").getCanonicalName());
			} else {
				templateName = convertString("Template_", periodicTask.getInstanceOf().getMethod("run", "()Z").getCanonicalName());				
			}
			systemDeclBuilder.
				append(templateName).
				append("_").
				append(periodicTask.getTID()).
				append(" = ").
				append(templateName).
				append("(").
				append(periodicTask.getTID()).
				append(");\n");
			
			//systemBuilder.append("       ").append(templateName).append("_").append(periodicTask.getTID()).append(", ");
			
			systemDeclBuilder.
				append(schedulerModel.getPeriodicController().getName()).
				append(periodicTask.getTID()).
				append(" = ").
				append(schedulerModel.getPeriodicController().getName()).
				append("(").
				append(periodicTask.getTID()).append(", ").
				append(convertToClockCycles(periodicTask.getPeriod())).append(", ").
				append(convertToClockCycles(periodicTask.getDeadline())).append(", ").
				append(convertToClockCycles(periodicTask.getOffset())).append(");\n");
			
			/*systemBuilder.
				append(schedulerModel.getPeriodicController().getName()).
				append(periodicTask.getTID()).
				append(", \n");
				*/
		}
		
		// Add periodic and sporadic tasks to system definition and make priorities
		ArrayList<ThreadDeadline> threadDeadlines = new ArrayList<ThreadDeadline>();
		for(PeriodicTask pTask : periodicTasks) {
			threadDeadlines.add(new PeriodicThreadDeadline(pTask));
		}
		for(SporadicTask spoTask : sporadicTasks) {
			threadDeadlines.add(new SporadicThreadDeadline(spoTask));
		}
		Collections.sort(threadDeadlines);
		Collections.reverse(threadDeadlines);
		for (ThreadDeadline threadDeadline : threadDeadlines) {
			String templateName = "";
			if(threadDeadline instanceof PeriodicThreadDeadline) {
				
				PeriodicThreadDeadline pDeadline = (PeriodicThreadDeadline)threadDeadline;
				
				if(Options.SCJ_API == Options.ScjApi.SCJ) {
					templateName = convertString("Template_", pDeadline.getpTask().getInstanceOf().getMethod("handleAsyncEvent", "()V").getCanonicalName());
				} else {
					templateName = convertString("Template_", pDeadline.getpTask().getInstanceOf().getMethod("run", "()Z").getCanonicalName());				
				}
				
				systemBuilder.append("       ").append(schedulerModel.getPeriodicController().getName()).
												append(pDeadline.getpTask().getTID()).
												append((Options.STATISTICAL_MODEL_CHECKING) ? ", " : " < ");
				
				systemBuilder.append(templateName)
							 .append("_")
							 .append(pDeadline.getpTask().getTID())
							 .append((Options.STATISTICAL_MODEL_CHECKING) ? ",\n" : " <\n");

				
			} else if(threadDeadline instanceof SporadicThreadDeadline) {
				
				SporadicThreadDeadline spoDeadline = (SporadicThreadDeadline) threadDeadline;
				
				if(Options.SCJ_API == Options.ScjApi.SCJ) {
					templateName = convertString("Template_", spoDeadline.getSporadicTask().getInstanceOf().getMethod("run", "()Z").getCanonicalName());
				} else {
					templateName = convertString("Template_", spoDeadline.getSporadicTask().getInstanceOf().getMethod("run", "()Z").getCanonicalName());				
				}
				
				systemBuilder.append("       ").append(templateName).append("_").append(spoDeadline.getSporadicTask().getTID()).append((Options.STATISTICAL_MODEL_CHECKING) ? ", " : " < ");
				
				systemBuilder.
					append(schedulerModel.getSporadicController().getName()).
					append(spoDeadline.getSporadicTask().getTID()).
					append((Options.STATISTICAL_MODEL_CHECKING) ? ",\n" : " <\n");
			}		
		}
		// Finally, remove the last priority.
		int index = systemBuilder.lastIndexOf((Options.STATISTICAL_MODEL_CHECKING) ? ",\n" : " <\n");
		if(index != -1) {
			systemBuilder.replace(index, systemBuilder.length(), ",\n");
		}
		
		// add UTIL/IDLE
		if(Options.UTILIZATION_MONITOR){
			UppaalTemplate t = new UppaalTemplate("Utilization");
			
			Location l = new Location("UtilizationMonitor");
			l.setInvariant("util' == !idling && idle' == idling");
			t.addLocation(l);
			t.setInit(l);
			
			nta.addContent(t.generateXMLElement());
		}
		
		if(Options.UTILIZATION_MONITOR){
			systemDeclBuilder.append("//Utilization\n");
			systemDeclBuilder.append("  UtilizationMonitor = Utilization();\n");
			systemDeclBuilder.append("//End utilization\n");
		}

		//end add util/idle
		
		if(Options.UTILIZATION_MONITOR){
			systemBuilder.append("       ").append("UtilizationMonitor");
			
		}

		systemBuilder.append(";");
		String systemStr = systemBuilder.toString();
		if(systemStr.substring(systemStr.length() - 3).equals(",\n;")) {
			systemStr = systemStr.substring(0, systemStr.length() - 3) + ";";
		}
		
		systemDeclBuilder.append(systemStr);

		this.setSystem(systemDeclBuilder.toString());
		
		//initializes all templates
		system.addContent(this.getSystem());
		nta.addContent(system);

		DocType docType = new DocType("nta",
				"-//Uppaal Team//DTD Flat System 1.1//EN",
				"http://www.it.uu.se/research/group/darts/uppaal/flat-1_1.dtd");

		Document xmlDocument = new Document(nta);
		xmlDocument.setDocType(docType);

		return xmlDocument;

	}
	
	private Document generateWCETDocument(MethodInformation methodInformation) {
		
		Element nta = new Element("nta");
		
		Element declarations = new Element("declaration");
		declarations.addContent("typedef int[0,0] ThreadID;\n" +
				"int running[ThreadID];\n"+
				"int depth = 0;\n" +
				"void monitorEnter(){ depth++; }\n"+
				"void monitorExit(){ depth--; }\n");
		
		declarations.addContent(this.declarations.toString());
		nta.addContent(declarations);
		
		Element system = new Element("system");

		StringBuilder systemDeclBuilder = new StringBuilder("\n//\n//Initialization\n");
		StringBuilder systemBuilder = new StringBuilder("\n//\nsystem\n");
		
		MethodInformation method = new MethodInformation("startingTemplate", "startingTemplate", "startingTemplate", 1);
		//Initial template to start the correct method
		UppaalTemplate startingTemplate = new PublicMethod(method);
		startingTemplate.setParameter("const ThreadID tID");
		Location initial = new Location("initial", LocationType.COMMITED, -200, 0);
		//To ensure no ID is added after the name
		initial.setName(new Name("initial"));
		startingTemplate.addLocation(initial);
		startingTemplate.setInit(initial);
		Location temp = new Location("temp");
		//To ensure no ID is added after the name
		temp.setName(new Name("temp"));
		startingTemplate.addLocation(temp);
		Location end = new Location("end", LocationType.COMMITED, 200, 0);
		//To ensure no ID is added after the name
		end.setName(new Name("end"));
		startingTemplate.addLocation(end);
		Transition t1 = new Transition(initial.getId(), temp.getId());
		t1.setSync(new Channel(UppaalModel.getChannelIdentifier(methodInformation), methodInformation.getCanonicalName(), "tID"), true);
		t1.addUpdate("running[tID] = 1");
		Transition t2 = new Transition(temp.getId(), end.getId());
		t2.setSync(new Channel(UppaalModel.getChannelIdentifier(methodInformation), methodInformation.getCanonicalName(), "tID"), false);
		t2.addUpdate("running[tID] = 0");
		Transition loop = new Transition(end.getId(), end.getId());
		startingTemplate.addTransition(t1);
		startingTemplate.addTransition(t2);
		startingTemplate.addTransition(loop);
		//End initial template
		addTemplate(startingTemplate);
		
		//updateLabelPositions();
		userFriendlyOutput();
		
		Collections.sort(templates);
		for (UppaalTemplate template : templates) {
			systemDeclBuilder.
				append(template.getName()).
				append("_threads(const ThreadID tID) = ").
				append(template.getName()).
				append("(tID);\n");
			systemBuilder.append("       ").append(template.getName()).append("_threads, \n");
			nta.addContent(template.generateXMLElement());
		}
		systemBuilder.replace(systemBuilder.lastIndexOf(","), systemBuilder.length(), ";");
		//systemBuilder.append(UppaalModel.convertString("Template_", "startingTemplate") + ";");
		
		systemDeclBuilder.append(systemBuilder.toString());
		this.setSystem(systemDeclBuilder.toString());

		//initializes all templates
		system.addContent(this.getSystem());
		nta.addContent(system);

		DocType docType = new DocType("nta",
				"-//Uppaal Team//DTD Flat System 1.1//EN",
				"http://www.it.uu.se/research/group/darts/uppaal/flat-1_1.dtd");

		Document xmlDocument = new Document(nta);
		xmlDocument.setDocType(docType);

		return xmlDocument;
	}

	/**
	 * Writes the entire XML document of the model to a file
	 * 
	 * @param outputFile
	 *            The target file
	 */
	public void writeModelToFile(String outputFile) {
		Format format = Format.getPrettyFormat();

		XMLOutputter output = new XMLOutputter(format);
		try {
			java.io.PrintStream printStream = new PrintStream(outputFile);
			output.output(generateXMLDocument(), printStream);
			printStream.close();
		} catch (IOException e) {
			e.printStackTrace();
		}
	}
	
	public void writeNotDeadlockToFile(String outputFile) {
		try {
			java.io.PrintStream printStream = new PrintStream(outputFile);
			
			printStream.append("A[] not (");
			String perName = schedulerModel.getPeriodicController().getName().toString();
		    Iterator<PeriodicTask> perTaskIterator = periodicTasks.iterator();
			while (perTaskIterator.hasNext()) {
		        PeriodicTask task = perTaskIterator.next();
		        printStream.append(perName + task.getTID() + ".DeadlineOverrun" + ((perTaskIterator.hasNext() || sporadicTasks.size() > 0) ? " || " : ""));
		    }
			
			String spoName = schedulerModel.getSporadicController().getName().toString();
			Iterator<SporadicTask> spoTaskIterator = sporadicTasks.iterator();
			while (spoTaskIterator.hasNext()) {
		        SporadicTask task = spoTaskIterator.next();
		        printStream.append(spoName + task.getTID() + ".DeadlineOverrun" + ((spoTaskIterator.hasNext()) ? " || " : ""));
		    }
			printStream.append(")\n");
			
			
			//printStream.append("A[] !deadlock\n");

			if(Options.GENERATE_WCRT) {
				String pName = schedulerModel.getPeriodicController().getName().toString();
			    Iterator<PeriodicTask> pTaskIterator = periodicTasks.iterator();
				while (pTaskIterator.hasNext()) {
			        PeriodicTask task = pTaskIterator.next();
			        printStream.append("sup {" + pName + task.getTID() + ".ExecutingThread} : " + pName + task.getTID() + ".wcrt\n");
			    }
				
				String sName = schedulerModel.getSporadicController().getName().toString();
				Iterator<SporadicTask> sTaskIterator = sporadicTasks.iterator();
				while (sTaskIterator.hasNext()) {
			        SporadicTask task = sTaskIterator.next();
			        printStream.append("sup {" + sName + task.getTID() + ".ExecutingThread} : " + sName + task.getTID() + ".wcrt\n");
			    }				
			}
			
			if(Options.GENERATE_BLOCKING) {
				String pName = schedulerModel.getPeriodicController().getName().toString();
			    Iterator<PeriodicTask> pTaskIterator = periodicTasks.iterator();
				while (pTaskIterator.hasNext()) {
			        PeriodicTask task = pTaskIterator.next();
			        printStream.append("sup {" + pName + task.getTID() + ".ExecutingThread} : " + pName + task.getTID() + ".blockingTime\n");
			    }
				
				String sName = schedulerModel.getSporadicController().getName().toString();
				Iterator<SporadicTask> sTaskIterator = sporadicTasks.iterator();
				while (sTaskIterator.hasNext()) {
			        SporadicTask task = sTaskIterator.next();
			        printStream.append("sup {" + sName + task.getTID() + ".ExecutingThread} : " + sName + task.getTID() + ".blockingTime\n");
			    }				
			}
			
			if(Options.GENERATE_WCET) {
				String pName = schedulerModel.getPeriodicController().getName().toString();
			    Iterator<PeriodicTask> pTaskIterator = periodicTasks.iterator();
				while (pTaskIterator.hasNext()) {
			        PeriodicTask task = pTaskIterator.next();
			        printStream.append("sup {" + pName + task.getTID() + ".ExecutingThread} : " + pName + task.getTID() + ".wcet\n");
			    }
				
				String sName = schedulerModel.getSporadicController().getName().toString();
				Iterator<SporadicTask> sTaskIterator = sporadicTasks.iterator();
				while (sTaskIterator.hasNext()) {
			        SporadicTask task = sTaskIterator.next();
			        printStream.append("sup {" + sName + task.getTID() + ".ExecutingThread} : " + sName + task.getTID() + ".wcet\n");
			    }				
			}
			
			if(Options.UTILIZATION_MONITOR)
				printStream.append("sup: util,idle");
			
			
			// Add WCRT query -- This needs to be done with a proper wcet clock, not released time which will always be max 
			/*if(Options.GENERATE_WCRT) {
				printStream.append("sup: ");
			    Iterator<PeriodicTask> pTaskIterator = periodicTasks.iterator();
				while (pTaskIterator.hasNext()) {
			        PeriodicTask task = pTaskIterator.next();
			        printStream.append("PeriodicThread" + task.getTID()+".releasedTime"+(pTaskIterator.hasNext()?",":""));
			    }
			    Iterator<SporadicTask> sTaskIterator = sporadicTasks.iterator();
			    printStream.append(sTaskIterator.hasNext()?",":"");
				while (sTaskIterator.hasNext()) {
			        SporadicTask task = sTaskIterator.next();
			        printStream.append("PeriodicThread" + task.getTID()+".releasedTime"+(sTaskIterator.hasNext()?",":""));
			    }
			}*/
			
			
			printStream.close();
		} catch (IOException e) {
			e.printStackTrace();
		}
	}
	
	@Deprecated
	public void writeWcetModelToFile(String outputFile, MethodInformation methodInformation) {
		Format format = Format.getPrettyFormat();

		XMLOutputter output = new XMLOutputter(format);
		try {
			java.io.PrintStream printStream = new PrintStream(outputFile);
			output.output(generateWCETDocument(methodInformation), printStream);
			printStream.close();
		} catch (IOException e) {
			e.printStackTrace();
		}
	}
	
	@Deprecated
	public void writeWcetQueriesToFile(String outputFile, MethodInformation methodInformation,
			int minWcet, int maxWcet) {
		try {
			java.io.PrintStream printStream = new PrintStream(outputFile);
			printStream.append("A[] !deadlock\n");
			printStream.append("E<> " + UppaalModel.getTemplateIdentifier(methodInformation) + "_threads(0).wcet >= " + minWcet + " && " +
					UppaalModel.convertString("Template_", "startingTemplate") + "_threads(0).end\n");
			printStream.append(UppaalModel.convertString("Template_", "startingTemplate") + "_threads(0).end --> " + UppaalModel.convertString("Template_", methodInformation.getCanonicalName()) + "_threads(0).wcet <= " + maxWcet);
			printStream.close();
		} catch (IOException e) {
			e.printStackTrace();
		}
	}

	public static Boolean correctXML(String inputFile) {
		// Check whether the output is valid UPPAAL XML
		System.out.println("Validating XML file...");
		try {
			SAXBuilder b = new SAXBuilder();
			b.setValidation(true);
			String msg = "Correct XML!";
			try {
				b.build(inputFile);
			} catch (JDOMParseException e) {
				msg = e.getMessage();
			}
			System.out.println(msg);
		} catch (Exception e) {
			e.printStackTrace();
			return false;
		}
		System.out.println("XML validation complete.");
		return true;
	}
	
	
	//TODO Reconsider this! -- replace by getChannelIdentifier
	static Hashtable<String, String> nameMap = new Hashtable<String, String>();
	static int ID = 0;
	private static String convertString(String prefix, String input){
		input = input.trim();
		if (nameMap.containsKey(input)){
			return prefix + nameMap.get(input);
		} else {
			ID++;
			nameMap.put(input, String.format("%04d",ID));
			return prefix + String.format("%04d",ID);
		}
	}
	
	private static Map<MethodInformation, String> methodChannelMap = new HashMap<MethodInformation, String>();
	public static String getChannelIdentifier(MethodInformation methodInformation){
		if(!methodChannelMap.containsKey(methodInformation))
			methodChannelMap.put(methodInformation, UppaalModel.convertString("chan_", methodInformation.getCanonicalName()));
		return methodChannelMap.get(methodInformation);
	}

	private static Map<MethodInformation, String> methodTemplateMap = new HashMap<MethodInformation, String>();
	public static String getTemplateIdentifier(MethodInformation methodInformation){
		if(!methodTemplateMap.containsKey(methodInformation))
			methodTemplateMap.put(methodInformation, UppaalModel.convertString("Template_", methodInformation.getCanonicalName()));
		return methodTemplateMap.get(methodInformation);
	}
	
	
	
	private final int maxSize = (int)(Math.pow(2, 30)-1);
	private long convertToClockCycles(long nanos){
		
		
		long clockCycles = (nanos*Options.MHZ)/1000;
		assert (clockCycles/maxSize)==0 : "Period, deadlines, or offset too large"; 
		return clockCycles;
	}
	
	public static String getOriginalName(String newName) {
		for(String a : nameMap.keySet()) {
			if(nameMap.get(a).equals(newName))
				return a;
		}
		return null;
	}


}
