package tetasarts.uppaal.modelcombiner;

import java.util.Arrays;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import tetasarts.uppaal.modelcombiner.jvmgeneration.JvmGenerator;
import uppaal.Automaton;
import uppaal.Location;
import uppaal.NTA;
import uppaal.Name;
import uppaal.Transition;
import uppaal.labels.Update;

public class Main {

private static String DEFAULT_OUTPUT_PATH = "./tetajUppaalModel.xml";
	
	public static void main(String[] argv) {
		String uppaalModelOutputPath = DEFAULT_OUTPUT_PATH;
		
		if(argv.length < 2) {
			System.out.println("Usage: java ModelCombiner <UPPAAL Model>* [-o Output UPPAAL model location]");
			return;
		}
		
		int outputPathSwitch;
		String[] xmlFilePaths;
		
		if ((outputPathSwitch = Arrays.asList(argv).indexOf("-o")) != -1) {
			uppaalModelOutputPath = argv[outputPathSwitch + 1];
			
			int xmlFilesNum = argv.length - 2;
			xmlFilePaths = new String[xmlFilesNum];
			
			for(int i = 0; i < argv.length - 2; i++)
				xmlFilePaths[i] = argv[i];
		} else {
			xmlFilePaths = argv;
		}
		
		LinkedList<NTA> ntaSystems = new LinkedList<NTA>();
				
		for(String filepath : xmlFilePaths)
			ntaSystems.add(new NTA(filepath));
		
		NTA jbcAutomaton = null, jbcProgramAutomaton = null;
		for(NTA nta : ntaSystems) {
			for(Automaton aut : nta.getAutomata()) {
				if(aut.getName().getName().contains("handle")
				   && aut.getName().getName().contains("OPCODE_T")) {
					jbcAutomaton = nta;
				}
				else if(aut.getName().getName().contains("Template")) {
					jbcProgramAutomaton = nta;
				}
			}
		}
		
		assert jbcAutomaton != null;
		assert jbcProgramAutomaton != null;
		
		jbcAutomaton = JvmGenerator.generate(jbcAutomaton, findUsedJBCOpcodes(jbcProgramAutomaton));
		Combiner ntaCombiner = new Combiner(ntaSystems);
		System.out.println("Combining models...");
		NTA combNTA = ntaCombiner.combineNtaSystems();
		System.out.println("Combining models finished");
		
		System.out.println("Writing model to file: " + uppaalModelOutputPath);
		combNTA.writePrettyLayoutModelToFile(uppaalModelOutputPath);
		System.out.println("Finished");
		
		
	}
	
	private static HashSet<String> findUsedJBCOpcodes(NTA ntaSystem) {
		final Pattern JBCOpcodePattern = Pattern.compile("JVM_([a-zA-Z0-9_]+)");
		List<Automaton> automata = ntaSystem.getAutomata();
		HashSet<String> usedOpcodes = new HashSet<String>();

		for(Automaton aut : automata) {
			for(Transition trans : aut.getTransitions()) {
				Update updateLbl = trans.getUpdate();
				if(updateLbl != null) {
					Matcher regExMatcher = JBCOpcodePattern.matcher(updateLbl.toString());
					if(regExMatcher.find()) {
						usedOpcodes.add(regExMatcher.group(1));
					}
				}
			}
		/*	
			for(Location loc : aut.getLocations()) {
				Name locationName = loc.getName();
				if(locationName != null) {
					Matcher regExMatcher = JBCOpcodePattern.matcher(locationName.getName());
					if(regExMatcher.find()) {
						usedOpcodes.add(regExMatcher.group(1));
					}
				}
			}*/
		}
		return usedOpcodes;
	}
	
	
	
}
