package tetasarts.uppaal.modelcombiner.jvmgeneration;

import java.util.Arrays;
import java.util.HashSet;
import java.util.LinkedList;

import uppaal.Automaton;
import uppaal.Location;
import uppaal.Location.LocationType;
import uppaal.Transition;
import uppaal.labels.Synchronization;
import uppaal.labels.Synchronization.SyncType;
import uppaal.labels.Update;

import tetasarts.uppaal.modelcombiner.optimization.ModelCleanerOptimization;
import tetasarts.uppaal.modelcombiner.optimization.NTATetaSARTSOptimizer;
import tetasarts.uppaal.modelcombiner.optimization.TetaSARTSOptimization;
import tetasarts.uppaal.modelcombiner.optimization.sync.AutomatonDependencyGraph;
import tetasarts.uppaal.modelcombiner.optimization.sync.SynchronizationResolver;
import uppaal.Automaton;
import uppaal.NTA;

public class JvmGenerator {

	public static NTA generate(NTA jbcImplementationNTA, HashSet<String> usedOpcodes) {
		StaticJVM staticJvm = new StaticJVM("staticjvm");
		Automaton jvmAutomaton = staticJvm.getAutomaton(usedOpcodes);

		
		NTA n = new NTA();
		n.addAutomaton(jvmAutomaton);
		//n.writePrettyLayoutModelToFile("/home/luckow/static.xml");
		//jbcImplementationNTA.writePrettyLayoutModelToFile("/home/luckow/test.xml");
		
		jbcImplementationNTA.addAutomaton(jvmAutomaton);
		jbcImplementationNTA.addAutomaton(getMetamocInterface());
		
		jbcImplementationNTA.getSystemDeclaration().getSystemInstances().add("staticjvm <");
		jbcImplementationNTA.getSystemDeclaration().getSystemInstances().add("hwinterface <");
		
		/*LinkedList<String> decls = (LinkedList<String>)*/
		/*if(decls.size() > 0) {
			String lastEle = (String) decls.getLast();
			
			lastEle = lastEle.replace(';', ',');		
			
			decls.removeLast();
			decls.addLast(lastEle);
			decls.addLast("\tstaticjvm;");
		} else
			decls.addLast("system \tstaticjvm;");
			*/
		AutomatonDependencyGraph depGraph = new AutomatonDependencyGraph(jbcImplementationNTA);
		
		LinkedList<TetaSARTSOptimization> optimizations = new LinkedList<TetaSARTSOptimization>();
		//optimizations.add(new ModelCleanerOptimization(usedOpcodes));
		optimizations.add(new SynchronizationResolver(depGraph));
		optimizations.add(new ModelCleanerOptimization(depGraph, usedOpcodes));
		
		NTATetaSARTSOptimizer optimizer = new NTATetaSARTSOptimizer(jbcImplementationNTA, optimizations);
		
		//optimizer.optimize();

		return jbcImplementationNTA;
	}
	
	private static Automaton getMetamocInterface() {
		Automaton template = new Automaton("hwinterface");
		
		Location idle = new Location(template, "Idle");
		template.setInit(idle);
		Location exec = new Location(template, "ExecuteMetamoc");
		
		Transition trans1 = new Transition(template, idle, exec);
		trans1.setSync(new Synchronization("assembly_execute", SyncType.RECEIVER));
		trans1.setUpdate(new Update("instrtype[PIPELINE_FETCH_STAGE] = assembly_instruction"));
		
		Transition trans2 = new Transition(template,exec, idle);
		trans2.setSync(new Synchronization("fetch", SyncType.INITIATOR));
	
		return template;		
	}
	
	
}
