package tetasarts.uppaal.modelcombiner.jvmgeneration;

import java.util.HashSet;

import uppaal.Automaton;
import uppaal.Location;
import uppaal.Location.LocationType;
import uppaal.Transition;
import uppaal.labels.Synchronization;
import uppaal.labels.Synchronization.SyncType;

class StaticJVM {
	
	private Location chooseJbcLocation, jbcFinishLocation;
	private Automaton staticJVMAutomaton;
	
	public StaticJVM(String name) {
		
		staticJVMAutomaton = new Automaton(name);
		
		Location idleLoc = new Location(staticJVMAutomaton, "Idle");
		staticJVMAutomaton.setInit(idleLoc);
		
		Location executionLoc = new Location(staticJVMAutomaton, "ExecuteJBC");
		executionLoc.setType(LocationType.COMMITTED);
		
		Transition execTransition = new Transition(staticJVMAutomaton, idleLoc, executionLoc);
		execTransition.setSync(new Synchronization("jvm_execute", SyncType.RECEIVER));
		
		Location initInterpretationLoc = new Location(staticJVMAutomaton, "InitialiseInterpretationLoop");
		Transition execInitInterpretationTrans = new Transition(staticJVMAutomaton, executionLoc,initInterpretationLoc);
		execInitInterpretationTrans.setSync(new Synchronization("invoke_assembly_process_jbc_pre", SyncType.INITIATOR));
		
		this.chooseJbcLocation = new Location(staticJVMAutomaton, "ChooseJBCImplementation");
		this.chooseJbcLocation.setType(LocationType.COMMITTED);
		
		Transition waitInterpretationTrans = new Transition(staticJVMAutomaton, initInterpretationLoc, chooseJbcLocation);
		waitInterpretationTrans.setSync(new Synchronization("return_assembly_process_jbc_pre", SyncType.RECEIVER));
			
		this.jbcFinishLocation = new Location(staticJVMAutomaton,"jbcFinish");
		this.jbcFinishLocation.setType(LocationType.COMMITTED);
		
		Location cleanupInterpretationLoc = new Location(staticJVMAutomaton,"CleanupInterpretationLoop");
		
		Transition cleanupInterpretationTrans = new Transition(staticJVMAutomaton, this.jbcFinishLocation, cleanupInterpretationLoc);
		cleanupInterpretationTrans.setSync(new Synchronization("invoke_assembly_process_jbc_post", SyncType.INITIATOR));
		
		Transition cleanupInterpretationWaitTrans = new Transition(staticJVMAutomaton, cleanupInterpretationLoc, idleLoc);
		cleanupInterpretationWaitTrans.setSync(new Synchronization("return_assembly_process_jbc_post", SyncType.RECEIVER));
	}
	
	public Automaton getAutomaton(HashSet<String> usedOpcodes) {
		for(String instruction : usedOpcodes){
			/*if(!instruction.equals(Constants.ILLEGAL_OPCODE) && !instruction.contains("JOP") 
			&& !instruction.contains("NONE") && !instruction.contains("none") 
			&& !instruction.contains("jop")) //&& !instruction.contains("getstatic") && !instruction.contains("putstatic"))
			{*/
				//if(instruction.contains("invokespecial")) {
					Location newJbcLoc = new Location(staticJVMAutomaton, "JVM_Execute_"+instruction.toUpperCase());
					
					Transition newJbcTrans = new Transition(staticJVMAutomaton, this.chooseJbcLocation, newJbcLoc);
					newJbcTrans.setGuard("jvm_instruction == JVM_" + instruction.toUpperCase());
					newJbcTrans.setSync(new Synchronization("invoke_assembly_handle" + instruction.toUpperCase() + "_OPCODE", SyncType.INITIATOR));
					
					Transition jbcReturnTrans = new Transition(staticJVMAutomaton,newJbcLoc,this.jbcFinishLocation);
					jbcReturnTrans.setSync(new Synchronization("return_assembly_handle" + instruction.toUpperCase() + "_OPCODE", SyncType.RECEIVER));
				//}
			//}
		}		
		
		return this.staticJVMAutomaton;
	}	
}
