package tetasarts.uppaal.modelcombiner.optimization.sync;

import java.util.Arrays;
import java.util.Collection;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;

import uppaal.Automaton;
import uppaal.NTA;
import uppaal.Transition;
import uppaal.labels.Synchronization.SyncType;

public class AutomatonDependencyGraph {

	
	HashMap<Automaton, AutomatonNode> graphAutomata;
	List<Automaton> automata;
	AutomatonNode rootNode;
	
	
	private class TransAutomatonRelation {
		private final Transition trans;
		private final Automaton aut;
		public TransAutomatonRelation(Transition trans, Automaton originAutomaton) {
			this.trans = trans;
			this.aut = originAutomaton;
		}
		public Transition getTrans() {
			return trans;
		}
		public Automaton getAutomaton() {
			return aut;
		}
	}
	
	public AutomatonDependencyGraph(NTA ntaSystem) {
		
		this.graphAutomata = new HashMap<Automaton, AutomatonNode>();
		this.automata = ntaSystem.getAutomata();
		
		createGraph();		
	}
	public Collection<AutomatonNode> getAutomata() {
		return this.graphAutomata.values();
	}
	public AutomatonNode getRootNode() {
		return this.rootNode;
	}
	
	private void createGraph() {
		HashMap<String, LinkedList<TransAutomatonRelation>> syncReceivers = new HashMap<String, LinkedList<TransAutomatonRelation>>();
		HashMap<String, LinkedList<TransAutomatonRelation>> syncInitiators = new HashMap<String, LinkedList<TransAutomatonRelation>>();
		for(Automaton aut : automata) {
			
			AutomatonNode autGraphNode = new AutomatonNode(aut);
			this.graphAutomata.put(aut, autGraphNode);
			
			if(aut.getName().getName().equals("staticjvm")) {
				this.rootNode = autGraphNode;
			}			
			
			for(Transition trans : aut.getTransitions()) {
				if(trans.getSync() != null) {
					String channelName = trans.getSync().getChannelName();
					if(!channelName.equals("assembly_execute")) {
						SyncType sType = trans.getSync().getSyncType();
						if(sType == SyncType.INITIATOR) {
							LinkedList<TransAutomatonRelation> syncTransitions = syncInitiators.get(channelName);
							if(syncTransitions != null)
								syncTransitions.add(new TransAutomatonRelation(trans, aut));
							else
								syncInitiators.put(channelName, new LinkedList<TransAutomatonRelation>(Arrays.asList(new TransAutomatonRelation(trans, aut))));
						}
						else if(sType == SyncType.RECEIVER) {
							LinkedList<TransAutomatonRelation> syncTransitions = syncReceivers.get(channelName);
							if(syncTransitions != null)
								syncTransitions.add(new TransAutomatonRelation(trans, aut));
							else
								syncReceivers.put(channelName, new LinkedList<TransAutomatonRelation>(Arrays.asList(new TransAutomatonRelation(trans, aut))));				
						}
					}
				}
			}
		}
		retrievePairedSynchronisations(syncInitiators, syncReceivers);
	}
	
	private void retrievePairedSynchronisations(HashMap<String, LinkedList<TransAutomatonRelation>> syncInitiators, 
			 HashMap<String, LinkedList<TransAutomatonRelation>> syncReceivers) {
		for(String channelName : syncInitiators.keySet()) {
			LinkedList<TransAutomatonRelation> syncTransitions = syncInitiators.get(channelName);			
			if(syncTransitions.size() >= 1) {
				for(TransAutomatonRelation syncInitTransRel : syncTransitions) {
					Automaton initAut = syncInitTransRel.getAutomaton();
					AutomatonNode initAutGraphNode = this.graphAutomata.get(initAut);
					
					LinkedList<TransAutomatonRelation> receiverTransitions = syncReceivers.get(channelName);
					if(receiverTransitions != null) {
						for(TransAutomatonRelation syncRecTransRel : receiverTransitions) {
							Automaton recAut = syncRecTransRel.getAutomaton();						
							AutomatonNode recAutGraphNode = this.graphAutomata.get(recAut);			

							initAutGraphNode.addOutgoingEdge(new DependencyEdge(initAutGraphNode, recAutGraphNode, syncInitTransRel.getTrans(), syncRecTransRel.getTrans()));
						}
					}
				}
			}
		}	
	}
}
