package tetasarts.uppaal.modelcombiner.optimization.sync;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.util.HashMap;

import att.grappa.Attribute;
import att.grappa.Edge;
import att.grappa.Graph;
import att.grappa.Node;

public class PrettyPrinter {

	private Graph overallGraph;
	private HashMap<String, Node> dependencyNodeToGrappaNodeMap = new HashMap<String, Node>();
	private String outputLocation;
	
	public PrettyPrinter(String graphName, AutomatonDependencyGraph autGraph, String outputLoc) {
		outputLocation = outputLoc;
		
		overallGraph = new Graph(graphName);
		Attribute graphAttr = new Attribute(Attribute.SUBGRAPH, Attribute.LABEL_ATTR, graphName);
		overallGraph.setAttribute(graphAttr);
		
		try {
			processGraph(autGraph, outputLocation);
		} catch (FileNotFoundException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
		
		
	}
	private void processGraph(AutomatonDependencyGraph autGraph, String outputLocation) throws FileNotFoundException {

		AutomatonNode rootNode = autGraph.getRootNode();
		traverseDepthFirstAndCreateGraph(rootNode);
		
		FileOutputStream fo = new FileOutputStream(new File(outputLocation));
		overallGraph.printGraph(fo);
		
		try {
			fo.close();
		} catch (IOException e) {
			e.printStackTrace();
		}
		
		convertDotFileToPng();
		
	}
	
	private void convertDotFileToPng() {
		int filenameBeginIndex = this.outputLocation.lastIndexOf('/') + 1;
		String filename = this.outputLocation.substring(filenameBeginIndex);
		String path = this.outputLocation.substring(0, filenameBeginIndex);
		String strippedFileName = filename.replace(".dot", "");
		
		String dotCmd = "dot " + this.outputLocation + " -Tpng -o " + path + strippedFileName + ".png";
		
		try {
			Runtime.getRuntime().exec(dotCmd);
			
		} catch (IOException e) {
			e.printStackTrace();
		}
	}
	
	
	private Node prevNode = null;
	private void traverseDepthFirstAndCreateGraph(AutomatonNode node) {
		Node sourceNode = insertNewNodeToGraph(node);
		if(prevNode != null)
			addEdgeToNode(prevNode, sourceNode);
		
		for(DependencyEdge edge : node.getoutgoingEdges()) {
			if(edge.getInitSyncTransition().getSync().getChannelName().contains("invoke")) {
				AutomatonNode targetNode = edge.getDestination();
				if(targetNode == null) return;
				else {
					prevNode = sourceNode;
					traverseDepthFirstAndCreateGraph(targetNode);
				}
			}
		}
		
	}
	
	private Node insertNewNodeToGraph(AutomatonNode node) {
		String autName = node.getAutomaton().getName().getName();
		if(dependencyNodeToGrappaNodeMap.containsKey(autName)) {
			return dependencyNodeToGrappaNodeMap.get(autName);
		}
		else {
			Node targetNode = new Node(this.overallGraph, getNodeName(node));
			Attribute attr = new Attribute(Attribute.NODE, Attribute.LABEL_ATTR, getLabelString(node));
			targetNode.setAttribute(attr);
			dependencyNodeToGrappaNodeMap.put(autName, targetNode);
			return targetNode;
		}
	}
	
	private void addEdgeToNode(Node sourceNode, Node targetNode) {
		Edge basicBlockEdge = new Edge(overallGraph, sourceNode, targetNode);
		overallGraph.addEdge(basicBlockEdge);
	}
	
	private String getNodeName(AutomatonNode node) {
		return node.getAutomaton().getName().getName();
	}
	
	private String getLabelString(AutomatonNode node) {
		StringBuilder sb = new StringBuilder();

		sb.append(node.getAutomaton().getName().getName());
		/*if(basicBlock.hasProperty(DFST_BASIC_BLOCK_ID_PROPERTY)) {
			sb.append("Depth First ordering ID (" + basicBlock.getProperty(DFST_BASIC_BLOCK_ID_PROPERTY) + ")\r");
		}
		int currentLine = 0;
		for(Instruction instruction : basicBlock.getInstructions()){
			
			int instructionLineNum = instruction.getLinenumber();
			
			if(currentLine != instructionLineNum){
				sb.append("Line Number: " + instructionLineNum + ")\r");
				currentLine = instructionLineNum;
			}
			
			sb.append("\t" + instruction.getMnemonic() + " " + instruction.getOperands() + "\r");	
		}*/
		return sb.toString();
	}
	
	
}
