package tetasarts.uppaal.schedulers;

import tetasarts.Options;
import tetasarts.generator.Constants;
import tetasarts.uppaal.Location;
import tetasarts.uppaal.Nail;
import tetasarts.uppaal.Name;
import tetasarts.uppaal.Transition;
import tetasarts.uppaal.UppaalTemplate;
import tetasarts.uppaal.Location.LocationType;
import tetasarts.uppaal.declarations.Channel;
import tetasarts.uppaal.declarations.Clock;
import tetasarts.uppaal.declarations.Declarations;

public abstract class AbstractSchedulerUppaalModel {

	public abstract SchedulerTemplate getScheduler();
	public abstract PeriodicControllerTemplate getPeriodicController();
	public abstract SporadicControllerTemplate getSporadicController();
	public abstract Declarations getDeclarations();
	
	protected PeriodicControllerTemplate generatePeriodicControllerTemplate(String insertFuncSignature, String doneFuncSignature) {
		PeriodicControllerTemplate periodicThreadTemplate = new PeriodicControllerTemplate("PeriodicThread");
		periodicThreadTemplate.setParameter("const PeriodicID pID, const int period, const int deadline, const int offset");		
		
		// Create declarations
		Declarations decl = new Declarations();
		decl.addClock(new Clock("releasedTime"));
		if(Options.GENERATE_WCRT) {
			decl.addClock(new Clock("wcrt"));
		}
		if(Options.GENERATE_BLOCKING) {
			decl.addClock(new Clock("blockingTime"));
		}
		if(Options.GENERATE_WCET) {
			decl.addClock(new Clock("wcet"));
		}
		periodicThreadTemplate.setDeclaration(decl);
		
		// Create all locations
		Location initLoc = createLoc("", LocationType.NORMAL, -160, -176);
		periodicThreadTemplate.addLocation(initLoc);
		periodicThreadTemplate.setInit(initLoc);
		
		Location checkForOffsetLoc = createLoc("CheckForOffset", LocationType.NORMAL, -160, -80);
		checkForOffsetLoc.setInvariant("releasedTime <= offset", checkForOffsetLoc.getPosX(), checkForOffsetLoc.getPosY() - 20);
		periodicThreadTemplate.addLocation(checkForOffsetLoc);
		
		Location toBeScheduledLoc = createLoc("", LocationType.COMMITED, -160, 0);
		periodicThreadTemplate.addLocation(toBeScheduledLoc);
		
		Location readyToBeScheduledLoc = createLoc("ReadyToBeScheduled", LocationType.COMMITED, 360, 0);
		periodicThreadTemplate.addLocation(readyToBeScheduledLoc);
		
		Location executingThreadLoc = createLoc("ExecutingThread", LocationType.NORMAL, 360, 136);
		String blockingInvariantStr = ((Options.GENERATE_BLOCKING) ? "blockingTime' == blocked[pID]" : "");
		String wcetInvariantStr = ((Options.GENERATE_WCET) ? "wcet' == running[pID]" : "");
		String invariantStr = blockingInvariantStr;
		if(Options.GENERATE_WCET && !invariantStr.equals("")) {
			invariantStr += " &&\n";
		}
		invariantStr += wcetInvariantStr;
		
		executingThreadLoc.setInvariant(invariantStr, executingThreadLoc.getPosX(), executingThreadLoc.getPosY() - 20);
		if(Options.STATISTICAL_MODEL_CHECKING) {
			executingThreadLoc.setExponentialRate(Constants.EXPONENTIAL_RATE_STR);
		}
		periodicThreadTemplate.addLocation(executingThreadLoc);
		
		Location doneLoc = createLoc("DONE", LocationType.NORMAL, 360, 272);
		doneLoc.setInvariant("releasedTime <= period", doneLoc.getPosX(), doneLoc.getPosY() - 20);
		periodicThreadTemplate.addLocation(doneLoc);
		
		Location deadlockLoc = createLoc("DeadlineOverrun", LocationType.COMMITED, 56, 136);
		periodicThreadTemplate.addLocation(deadlockLoc);
		
		// Create all transitions
		Transition initCheckTrans = new Transition(initLoc.getId(), checkForOffsetLoc.getId());
		int x = (initLoc.getPosX() + checkForOffsetLoc.getPosX()) / 2 - 20;
		int y = (initLoc.getPosY() + checkForOffsetLoc.getPosY()) / 2;
		//initCheckTrans.setGuard("offset != 0", x, y - 30);
		initCheckTrans.setSync(new Channel("GO", "x", true), false, x, y - 15);
		periodicThreadTemplate.addTransition(initCheckTrans);
		
		/*Transition initToBeScheduledTrans = new Transition(initLoc.getId(), toBeScheduledLoc.getId());
		x = ((initLoc.getPosX() + toBeScheduledLoc.getPosX()) / 2) - 30;
		y = ((initLoc.getPosY() + toBeScheduledLoc.getPosY()) / 2) - 30;
		initToBeScheduledTrans.setSync(new Channel("GO", "x", true), false, x, y + 30);
		initToBeScheduledTrans.setGuard("offset == 0", x, y + 15);
		periodicThreadTemplate.addTransition(initToBeScheduledTrans);
		*/
		
		Transition checktoBeScheduledTrans = new Transition(checkForOffsetLoc.getId(), toBeScheduledLoc.getId());
		x = ((checkForOffsetLoc.getPosX() + toBeScheduledLoc.getPosX()) / 2);
		y = (checkForOffsetLoc.getPosY() + toBeScheduledLoc.getPosY()) / 2;
		checktoBeScheduledTrans.setGuard("releasedTime == offset", x, y);
		periodicThreadTemplate.addTransition(checktoBeScheduledTrans);
		
		Transition toBeScheduledToReadyTrans = new Transition(toBeScheduledLoc.getId(), readyToBeScheduledLoc.getId());
		x = ((toBeScheduledLoc.getPosX() + readyToBeScheduledLoc.getPosX()) / 2) - 50;
		y = ((toBeScheduledLoc.getPosY() + readyToBeScheduledLoc.getPosY()) / 2) + 10;
		toBeScheduledToReadyTrans.setUpdate("schedulable[pID] = true,\nreleasedTime = 0,\n" + insertFuncSignature + ((Options.GENERATE_WCRT) ? ",\nwcrt = 0" : "") + ((Options.GENERATE_BLOCKING) ? ",\nblockingTime = 0" : "") + ((Options.GENERATE_WCET) ? ",\nwcet = 0" : ""), x, y);
		periodicThreadTemplate.addTransition(toBeScheduledToReadyTrans);
		
		Transition readyToExecutingTrans = new Transition(readyToBeScheduledLoc.getId(), executingThreadLoc.getId());
		x = ((readyToBeScheduledLoc.getPosX() + executingThreadLoc.getPosX()) / 2) + 10;
		y = ((readyToBeScheduledLoc.getPosY() + executingThreadLoc.getPosY()) / 2);
		readyToExecutingTrans.setSync(new Channel("run[pID]", "", false), true, x, y);
		periodicThreadTemplate.addTransition(readyToExecutingTrans);
		
		Transition executingToDeadlockTrans = new Transition(executingThreadLoc.getId(), deadlockLoc.getId());
		x = ((executingThreadLoc.getPosX() + deadlockLoc.getPosX()) / 2) - 60;
		y = ((executingThreadLoc.getPosY() + deadlockLoc.getPosY()) / 2) - 20;
		executingToDeadlockTrans.setGuard("releasedTime > deadline", x, y);
		periodicThreadTemplate.addTransition(executingToDeadlockTrans);
	/*	
		Transition deadlockLoopTrans = new Transition(deadlockLoc.getId(), deadlockLoc.getId());
		deadlockLoopTrans.addNail(new Nail(24, 168));
		deadlockLoopTrans.addNail(new Nail(80, 168));
		periodicThreadTemplate.addTransition(deadlockLoopTrans);
		*/
		Transition executingToDoneTrans = new Transition(executingThreadLoc.getId(), doneLoc.getId());
		x = ((executingThreadLoc.getPosX() + doneLoc.getPosX()) / 2) + 10;
		y = ((executingThreadLoc.getPosY() + doneLoc.getPosY()) / 2);
		executingToDoneTrans.setGuard("releasedTime <= deadline", x, y - 30);
		executingToDoneTrans.setSync(new Channel("run[pID]", "", false), false, x, y - 15);
		executingToDoneTrans.setUpdate("schedulable[pID] = false,\n" + doneFuncSignature + ((Options.GENERATE_WCRT) ? ",\nwcrt = 0" : "") + ((Options.GENERATE_BLOCKING) ? ",\nblockingTime = 0" : "") + ((Options.GENERATE_WCET) ? ",\nwcet = 0" : ""), x, y);
		periodicThreadTemplate.addTransition(executingToDoneTrans);
		
		Transition doneToBeScheduledTrans = new Transition(doneLoc.getId(), toBeScheduledLoc.getId());
		Nail doneToBeScheduledNail = new Nail(-160, 272);
		doneToBeScheduledTrans.addNail(doneToBeScheduledNail);
		x = ((doneLoc.getPosX() + doneToBeScheduledNail.getPosX()) / 2) - 40;
		y = ((doneLoc.getPosY() + doneToBeScheduledNail.getPosY()) / 2) - 20;
		doneToBeScheduledTrans.setGuard("releasedTime == period", x, y);
		periodicThreadTemplate.addTransition(doneToBeScheduledTrans);

		return periodicThreadTemplate;
	}
	
	protected SporadicControllerTemplate generateSporadicControllerTemplate(String insertFuncSignature, String doneFuncSignature) {
		SporadicControllerTemplate sporadicThreadTemplate = new SporadicControllerTemplate("SporadicThread");
		sporadicThreadTemplate.setParameter("const SporadicID sID, const int minIA, const int deadline");		
		// Create declarations
		Declarations decl = new Declarations();
		decl.addClock(new Clock("releasedTime"));
		if(Options.GENERATE_WCRT) {
			decl.addClock(new Clock("wcrt"));
		}
		if(Options.GENERATE_BLOCKING) {
			decl.addClock(new Clock("blockingTime"));
		}
		if(Options.GENERATE_WCET) {
			decl.addClock(new Clock("wcet"));
		}
		sporadicThreadTemplate.setDeclaration(decl);
		
		// Create all locations
		Location initLoc = createLoc("", LocationType.NORMAL, -264, -144);
		sporadicThreadTemplate.addLocation(initLoc);
		sporadicThreadTemplate.setInit(initLoc);
		
		Location readyToBeFiredLoc = createLoc("ReadyToBeFired", LocationType.NORMAL, -96, -144);
		sporadicThreadTemplate.addLocation(readyToBeFiredLoc);
		
		Location readyToBeScheduledLoc = createLoc("ReadyToBeScheduled", LocationType.COMMITED, 448, -144);
		sporadicThreadTemplate.addLocation(readyToBeScheduledLoc);
		
		Location executingThreadLoc = createLoc("ExecutingThread", LocationType.NORMAL, 448, -8);
		String blockingInvariantStr = ((Options.GENERATE_BLOCKING) ? "blockingTime' == blocked[sID]" : "");
		String wcetInvariantStr = ((Options.GENERATE_WCET) ? "wcet' == running[sID]" : "");
		String invariantStr = blockingInvariantStr;
		if(Options.GENERATE_WCET && !invariantStr.equals("")) {
			invariantStr += " &&\n";
		}
		invariantStr += wcetInvariantStr;
		
		executingThreadLoc.setInvariant(invariantStr, executingThreadLoc.getPosX(), executingThreadLoc.getPosY() - 20);
		if(Options.STATISTICAL_MODEL_CHECKING) {
			executingThreadLoc.setExponentialRate(Constants.EXPONENTIAL_RATE_STR);
		}
		sporadicThreadTemplate.addLocation(executingThreadLoc);
		
		Location deadlineOverrunLoc = createLoc("DeadlineOverrun", LocationType.COMMITED, 200, -8);
		sporadicThreadTemplate.addLocation(deadlineOverrunLoc);
		
		Location doneLoc = createLoc("DONE", LocationType.NORMAL, 448, 120);
		doneLoc.setInvariant("releasedTime <= minIA", doneLoc.getPosX(), doneLoc.getPosY() - 20);
		sporadicThreadTemplate.addLocation(doneLoc);
		
		// Create all transitions
		Transition initReadyFiredTrans = new Transition(initLoc.getId(), readyToBeFiredLoc.getId());
		int x = (initLoc.getPosX() + readyToBeFiredLoc.getPosX()) / 2 - 70;
		int y = (initLoc.getPosY() + readyToBeFiredLoc.getPosY()) / 2;
		initReadyFiredTrans.setSync(new Channel("GO", "x", true), false, x, y - 30);
		initReadyFiredTrans.setUpdate("fireable[sID] = true", x, y - 15);
		sporadicThreadTemplate.addTransition(initReadyFiredTrans);
		
		Transition readyFiredReadyScheduledTrans = new Transition(readyToBeFiredLoc.getId(), readyToBeScheduledLoc.getId());
		x = (readyToBeFiredLoc.getPosX() + readyToBeScheduledLoc.getPosX()) / 2 - 70;
		y = (readyToBeFiredLoc.getPosY() + readyToBeScheduledLoc.getPosY()) / 2;
		readyFiredReadyScheduledTrans.setSync(new Channel("fire[sID]", "x", false), false, x, y);
		readyFiredReadyScheduledTrans.setUpdate("schedulable[sID] = true,\nfireable[sID] = false,\nreleasedTime = 0,\n" + insertFuncSignature + ((Options.GENERATE_WCRT) ? ",\nwcrt = 0" : "")  + ((Options.GENERATE_BLOCKING) ? ",\nblockingTime = 0" : "")  + ((Options.GENERATE_WCET) ? ",\nwcet = 0" : ""), x, y + 15);
		sporadicThreadTemplate.addTransition(readyFiredReadyScheduledTrans);
		
		Transition readyScheduledExecutingTrans = new Transition(readyToBeScheduledLoc.getId(), executingThreadLoc.getId());
		x = ((readyToBeScheduledLoc.getPosX() + executingThreadLoc.getPosX()) / 2) + 10;
		y = (readyToBeScheduledLoc.getPosY() + executingThreadLoc.getPosY()) / 2;
		readyScheduledExecutingTrans.setSync(new Channel("run[sID]", "x", false), true, x, y);
		sporadicThreadTemplate.addTransition(readyScheduledExecutingTrans);
		
		Transition executingDeadlineTrans = new Transition(executingThreadLoc.getId(), deadlineOverrunLoc.getId());
		x = (executingThreadLoc.getPosX() + deadlineOverrunLoc.getPosX()) / 2 - 80;
		y = (executingThreadLoc.getPosY() + deadlineOverrunLoc.getPosY()) / 2 - 20;
		executingDeadlineTrans.setGuard("releasedTime > deadline", x, y);
		sporadicThreadTemplate.addTransition(executingDeadlineTrans);
		
	/*	Transition deadlockTrans = new Transition(deadlineOverrunLoc.getId(), deadlineOverrunLoc.getId());
		deadlockTrans.addNail(new Nail(152, 40));
		deadlockTrans.addNail(new Nail(240, 40));
		sporadicThreadTemplate.addTransition(deadlockTrans);
		*/
		Transition executingDoneTrans = new Transition(executingThreadLoc.getId(), doneLoc.getId());
		x = ((executingThreadLoc.getPosX() + doneLoc.getPosX()) / 2) + 10;
		y = ((executingThreadLoc.getPosY() + doneLoc.getPosY()) / 2) - 20;
		executingDoneTrans.setGuard("releasedTime <= deadline", x, y - 15);
		executingDoneTrans.setSync(new Channel("run[sID]", "x", false), false, x, y);
		executingDoneTrans.setUpdate("schedulable[sID] = false,\n" + doneFuncSignature +((Options.GENERATE_WCRT) ? ",\nwcrt = 0" : "")  + ((Options.GENERATE_BLOCKING) ? ",\nblockingTime = 0" : "")  + ((Options.GENERATE_WCET) ? ",\nwcet = 0" : ""), x, y + 15);
		sporadicThreadTemplate.addTransition(executingDoneTrans);
		
		
		Transition doneReadyFireTrans = new Transition(doneLoc.getId(), readyToBeFiredLoc.getId());
		Nail n = new Nail(-96, 120);
		doneReadyFireTrans.addNail(n);
		x = ((n.getPosX() + doneLoc.getPosX()) / 2) - 50;
		y = ((n.getPosY() + doneLoc.getPosY()) / 2) - 50;
		doneReadyFireTrans.setGuard("releasedTime == minIA", x, y);
		doneReadyFireTrans.setUpdate("releasedTime = 0,\nfireable[sID] = true", x, y + 15);
		sporadicThreadTemplate.addTransition(doneReadyFireTrans);

		return sporadicThreadTemplate;
	}
	
	protected Location createLoc(String name, LocationType locType, int x, int y) {
		/*
		 * This is kind'a stupid - however, it is an easy solution to avoid problems with the location names
		 * where the unique location id is concatenated to the name.... and for the scheduler 
		 * (and controller templates) there will not be name clashes on the locations :)
		 */
		Location loc = new Location(name, locType, x, y);
		loc.setName(new Name(name, x, y));
		return loc;
	}
	
}
