package tetasarts.uppaal.translation;

import java.util.HashSet;
import java.util.List;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import tetasarts.Options.PlatformType;
import tetasarts.uppaal.UppaalModel;
import tetasarts.uppaal.modelcombiner.Combiner;
import tetasarts.uppaal.modelcombiner.jvmgeneration.JvmGenerator;
import uppaal.Automaton;
import uppaal.Location;
import uppaal.NTA;
import uppaal.Name;
import uppaal.Transition;
import uppaal.labels.Update;

public class ExplicitModelCreater {

	//private static final String HVM_MODEL_PATH = "/hvmmodelOptimized.xml";
	private static final String HVM_MODEL_PATH = "/hvmmodel_opt_v2.xml";
	private static final String AVR_MODEL_PATH = "/metamocAvr.xml";
	private static final String JOP_MODEL_PATH = "/jopmodel.xml";
	
	/*
	private NTA programModel;
	private NTA jvmModel;
	private NTA hwModel;
	*/
	private Combiner modelCombiner;
	
	public ExplicitModelCreater(NTA programModel, PlatformType platformType) {
		this.modelCombiner = new Combiner();
		modelCombiner.addNTA(programModel);
		switch(platformType) {
			case HVMAVR:
				NTA hvmModel = new NTA(getClass().getResource(HVM_MODEL_PATH).getPath());
				modelCombiner.addNTA(JvmGenerator.generate(hvmModel, findUsedJBCOpcodes(programModel)));
				modelCombiner.addNTA(new NTA(getClass().getResource(AVR_MODEL_PATH).getPath()));
				break;
			case JOP:
				NTA jopModel = new NTA(getClass().getResource(JOP_MODEL_PATH).getPath());
				//modelCombiner.addNTA(jopModel);
				modelCombiner.addNTA(JvmGenerator.generate(jopModel, findUsedJBCOpcodes(programModel)));
				break;
		}
	}
	
	public NTA generateExplicitModel() {
		NTA explicitModel = modelCombiner.combineNtaSystems();
		//explicitModel.getDeclarations().add("urgent chan jvm_execute;");
		
		return explicitModel;
	}
	
	private HashSet<String> findUsedJBCOpcodes(NTA ntaSystem) {
		final Pattern JBCOpcodePattern = Pattern.compile("JVM_([a-zA-Z0-9_]+)");
		List<Automaton> automata = ntaSystem.getAutomata();
		HashSet<String> usedOpcodes = new HashSet<String>();

		for(Automaton aut : automata) {
			for(Transition trans : aut.getTransitions()) {
				Update updateLbl = trans.getUpdate();
				if(updateLbl != null) {
					Matcher regExMatcher = JBCOpcodePattern.matcher(updateLbl.toString());
					if(regExMatcher.find()) {
						usedOpcodes.add(regExMatcher.group(1));
					}
				}
			}
		/*	
			for(Location loc : aut.getLocations()) {
				Name locationName = loc.getName();
				if(locationName != null) {
					Matcher regExMatcher = JBCOpcodePattern.matcher(locationName.getName());
					if(regExMatcher.find()) {
						usedOpcodes.add(regExMatcher.group(1));
					}
				}
			}*/
		}
		return usedOpcodes;
	}

}
