package tetasarts.uppaal.translation;

import java.util.Hashtable;

import org.apache.bcel.generic.InstructionHandle;

import tetasarts.Options;
import tetasarts.generator.Constants;
import tetasarts.tir.basicblocks.AbstractBasicBlock;
import tetasarts.tir.basicblocks.AbstractBasicBlockVisitor;
import tetasarts.tir.basicblocks.basicBlockTypes.MethodCallingBasicBlock;
import tetasarts.uppaal.Location;
import tetasarts.uppaal.Transition;
import tetasarts.uppaal.UppaalTemplate;
import tetasarts.uppaal.Location.LocationType;

public class LocationCreaterVisitor extends AbstractBasicBlockVisitor{
	private UppaalTemplate template;
	private Hashtable<AbstractBasicBlock, Location> locationMap;
	
	public LocationCreaterVisitor(UppaalTemplate template, Hashtable<AbstractBasicBlock, Location> locationMap){
		this.template = template;
		this.locationMap = locationMap;
	}
	
	@Override
	public void visitMethodCallingBasicBlock(MethodCallingBasicBlock s) {
		try{
			String callInstr = s.getName();
			String calleeName = s.getPossibleCallees().iterator().next().getName();
			
			calleeName = calleeName.replace("<", "").replace(">", "");
			
			Location loc = new Location(callInstr + "_" + calleeName);

			setComment(loc,s);
			template.addLocation(loc);
			locationMap.put(s, loc);
			
			Location loc1 = new Location("running_" + calleeName);
			Location loc2 = new Location("returningFrom_" + calleeName);
			if(Options.EXPLICIT_MODEL) {
				loc2.setType(LocationType.COMMITED);
			}
			template.addLocation(loc1);
			template.addLocation(loc2);
			
			Transition transition1 = new Transition(loc.getId(), loc1.getId());
			template.addTransition(transition1);
			
			Transition transition2 = new Transition(loc1.getId(), loc2.getId());
			template.addTransition(transition2);
		} catch (RuntimeException e){
			System.out.println(s.getCallSite().getCallSiteClassName());
			System.out.println(s.getCallSite().getCallSiteMethod());
			System.out.println(s.getName());
			e.printStackTrace();
			throw e;
		}
	}
	
	@Override
	public void defaultVisit(AbstractBasicBlock s) {
		String locationName = (!s.getBytecodes().isEmpty()) ? s.getBytecodes().get(0).getInstruction().getName() : s.getName();
		
		Location loc = new Location(locationName);

		setComment(loc,s);
		template.addLocation(loc);
		locationMap.put(s, loc);
		
		super.defaultVisit(s);
	}
	
	private void setComment(Location loc, AbstractBasicBlock s) {
		StringBuilder sb = new StringBuilder();
		for (InstructionHandle instructionHandle : s.getBytecodes()) {
			sb.append(instructionHandle.toString()).append("\n");
		}
		loc.setComment(sb.toString());
	}
}
