package tetasarts.uppaal.translation;

import java.util.HashMap;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.logging.Logger;

import org.apache.bcel.generic.Instruction;

import com.sun.tools.javac.main.JavacOption.Option;

import tetasarts.Options;
import tetasarts.Options.PlatformType;
import tetasarts.analysis.SimpleWcetVisitor;
import tetasarts.generator.Constants;
import tetasarts.platforms.HVMAVR;
import tetasarts.platforms.InstructionTimingInfo;
import tetasarts.tir.basicblocks.AbstractBasicBlock;
import tetasarts.tir.basicblocks.AbstractBasicBlockVisitor;
import tetasarts.tir.basicblocks.BasicBlockGraph;
import tetasarts.tir.basicblocks.LoopBound;
import tetasarts.tir.basicblocks.basicBlockTypes.LoopBasicBlock;
import tetasarts.tir.basicblocks.basicBlockTypes.MethodCallingBasicBlock;
import tetasarts.tir.basicblocks.basicBlockTypes.MonitorEnterBasicBlock;
import tetasarts.tir.basicblocks.basicBlockTypes.MonitorExitBasicBlock;
import tetasarts.tir.basicblocks.basicBlockTypes.NewBasicBlock;
import tetasarts.tir.basicblocks.basicBlockTypes.SporadicInvokeBasicBlock;
import tetasarts.tir.classinfo.ClassGraph;
import tetasarts.tir.classinfo.SporadicTask;
import tetasarts.tir.methods.MethodInformation;
import tetasarts.uppaal.Location;
import tetasarts.uppaal.Transition;
import tetasarts.uppaal.UppaalModel;
import tetasarts.uppaal.UppaalTemplate;
import tetasarts.uppaal.Location.LocationType;
import tetasarts.uppaal.declarations.Channel;
import tetasarts.uppaal.declarations.Clock;
import tetasarts.uppaal.declarations.Declarations;
import tetasarts.uppaal.declarations.Int;

public class TemplateDecoratorVisitor extends AbstractBasicBlockVisitor{
	private UppaalTemplate template;
	private Hashtable<AbstractBasicBlock, Location> locationMap;
	private BasicBlockGraph basicBlockGraph;
	private ClassGraph classGraph;
	private Set<SporadicTask> sporadicTasks;
	private Map<Integer, LoopBound> loopBounds = new HashMap<Integer, LoopBound>();
	private int loopCount = 0;
	private Logger logger = Logger.getLogger("generator");
	
	public TemplateDecoratorVisitor(
			UppaalTemplate template, 
			Hashtable<AbstractBasicBlock, 
			Location> locationMap, 
			BasicBlockGraph basicBlockGraph, 
			ClassGraph classGraph,
			Set<SporadicTask> sporadicTasks){
		this.template = template;
		this.locationMap = locationMap;
		this.basicBlockGraph = basicBlockGraph;
		this.classGraph = classGraph;
		this.sporadicTasks = sporadicTasks;
	}
	
	@Override
	public void visitSporadicInvokeBasicBlock(
			SporadicInvokeBasicBlock sporadicInvokeBasicBlock) {
		defaultVisit(sporadicInvokeBasicBlock);
		if (!Constants.DISABLE_TASKS){
			int fID = 0;
			for (SporadicTask sporadicTask : sporadicTasks) {
				if(sporadicTask.getEvent() == sporadicInvokeBasicBlock.getEvent()){
					fID = sporadicTask.getTID();
					break;
				}
			}
			assert fID != 0 : "Invoke of non existing sporadic thread";
			
			Location currentLocation = locationMap.get(sporadicInvokeBasicBlock);

			for (Location location : template.getSuccs(currentLocation)){
				for (Transition transition : template.getTransitions(currentLocation.getId(), location.getId())){
					transition.addGuard("fireable[" + fID + "] == true");
					transition.setSyncRunOrFire("fire[" + fID + "]!");
				}
				//Only added to illustrate old sporadic approach
//				Transition oldTrans = new Transition(currentLocation.getId(), location.getId());
//				oldTrans.addGuard("fireable[" + fID + "] == false");
//				String wcetInst = "inst" + currentLocation.getId();
//				oldTrans.addGuard("executionTime == " + wcetInst);
//				oldTrans.addUpdate("executionTime = 0");
//				template.addTransition(oldTrans);
			}
			
			template.addBacktrack(currentLocation, "fireable[" + fID + "] == false");
		}
	}

	@Override
	public void visitMethodCallingBasicBlock(MethodCallingBasicBlock s) {
		defaultVisit(s);
		
		Location currentLocation = locationMap.get(s);
		//Set<MethodInformation> concreteMethods = s.findConcreteMethods(classGraph);
		Set<MethodInformation> concreteMethods = s.getPossibleCallees();

		//Happens if no implementation of the method has been found
		if(concreteMethods.size() == 0) {
			logger.severe("No method implementation found at callsite: " + s.getCallSite().getCallSiteClassName() + "." + s.getCallSite().getCallSiteMethod() + "[linenumber: " + s.getCallSite().getInvokeInstructionHandle().getPosition() + "]");
			
			Location intermediateLocation = template.getSuccs(currentLocation).get(0);
			for (Transition t : template.getIncommingTransitions(intermediateLocation)) {
				t.addUpdate("notImplementedMethods = true");
			}
			
			if(!Options.EXPLICIT_MODEL) {
				intermediateLocation.setInvariant("executionTime <= 0");
				
				Location intermediateLocation2 = template.getSuccs(intermediateLocation).get(0);
				intermediateLocation2.setInvariant("executionTime <= 0");
			}
			return;
		}
		
		Location intermediateLocation1 = template.getSuccs(currentLocation).get(0);
		
		List<Transition> incommingTransitions = template.getIncommingTransitions(intermediateLocation1);
		List<Transition> outgoingTransitions = template.getOutgoingTransitions(intermediateLocation1);

		assert concreteMethods.size() == incommingTransitions.size() : "Incomming transitions not equal to number of methods";
		assert concreteMethods.size() == outgoingTransitions.size() : "Outgoing transitions not equal to number of methods";
		
		//TODO update with exact method, i.e. which class the method is in
		int i = 0;
		for(MethodInformation currentMethod : concreteMethods) {
			String channelName = UppaalModel.getChannelIdentifier(currentMethod);
			Transition methodCallTrans = incommingTransitions.get(i);
			
			if(Options.EXPLICIT_MODEL) {
				int originalTargetLocID = methodCallTrans.getTarget();			
				
				Location methodCallLoc = new Location("readyToCall_" + channelName);
				methodCallLoc.setType(LocationType.COMMITED);
				template.addLocation(methodCallLoc);
				
				methodCallTrans.setTarget(methodCallLoc.getId());
				
				Transition newMethodCallTrans = new Transition(methodCallLoc.getId(), originalTargetLocID);
				newMethodCallTrans.setSync(new Channel(channelName, currentMethod.getCanonicalName(), "tID"), true);
				template.addTransition(newMethodCallTrans);
			} else {
				methodCallTrans.setSync(new Channel(channelName, currentMethod.getCanonicalName(), "tID"), true);
			}
			if (currentMethod.isSynchronized())
				incommingTransitions.get(i).addUpdate("monitorEnter()");
			
			outgoingTransitions.get(i).setSync(new Channel(channelName, currentMethod.getCanonicalName(),"tID"), false);
			
			if (currentMethod.isSynchronized())
				outgoingTransitions.get(i).addUpdate("monitorExit()");
			if(!Options.EXPLICIT_MODEL) {
				outgoingTransitions.get(i).addUpdate("executionTime = 0");
				if(Options.PLATFORM_TYPE == PlatformType.JOP) {
					//TODO use method cache to determine miss or hit
					//The cost for a miss
					int l = -1;
					String returnType = currentMethod.getSignature().substring(currentMethod.getSignature().lastIndexOf(')')+1);
					//IRETURN
					if (returnType.equals("B") || returnType.equals("C") || returnType.equals("I") || returnType.equals("S") || returnType.equals("Z")){
						l = SimpleWcetVisitor.calculateI(false, template.getMethodSizeInWords());
						if (l <= 10)
							l = 0;
						else
							l -= 10;
					//LRETURN
					} else if (returnType.equals("J")){
						l = SimpleWcetVisitor.calculateI(false, template.getMethodSizeInWords());
						if (l <= 11)
							l = 0;
						else
							l -= 11;
					//FRETURN
					} else if (returnType.equals("F")){
						l = SimpleWcetVisitor.calculateI(false, template.getMethodSizeInWords());
						if (l <= 10)
							l = 0;
						else
							l -= 10;
					//DRETURN
					} else if (returnType.equals("D")){
						l = SimpleWcetVisitor.calculateI(false, template.getMethodSizeInWords());
						if (l <= 11)
							l = 0;
						else
							l -= 11;
					//ARETURN
					} else if (returnType.startsWith("L") || returnType.startsWith("[")){
						l = SimpleWcetVisitor.calculateI(false, template.getMethodSizeInWords());
						if (l <= 10)
							l = 0;
						else
							l -= 10;
					//RETURN
					} else {
						l = SimpleWcetVisitor.calculateI(false, template.getMethodSizeInWords());
						if (l <= 9)
							l = 0;
						else
							l -= 9;
					}
					assert l != -1 : "return instruction not found";
					outgoingTransitions.get(i).addUpdate("methodSwitchCost = " + l);
				}
			}
			//outgoingTransitions.get(i).addUpdate("methodSwitchCost = 0");
		    i++;
		}
		
		//Add the cost of the return to the method.
		if(!Options.EXPLICIT_MODEL) { 
			String methodSwitchString = "";
			if(Options.PLATFORM_TYPE == PlatformType.JOP) {
				methodSwitchString = "methodSwitchCost";
			} else if(Options.PLATFORM_TYPE == PlatformType.HVMAVR) {
				methodSwitchString = "0";
			}
			
			Location intermediateLocation2 = template.getSuccs(intermediateLocation1).get(0);
			intermediateLocation2.setInvariant("executionTime <= " + methodSwitchString + " &&\nexecutionTime' == running[tID]");
			
			for (Transition transition : template.getOutgoingTransitions(intermediateLocation2)) {
				transition.addGuard("executionTime == " + methodSwitchString);
				transition.addUpdate("executionTime = 0");
			}
		}
	}

	@Override
	public void visitMonitorEnterBasicBlock(
			MonitorEnterBasicBlock monitorEnterBasicBlock) {
		defaultVisit(monitorEnterBasicBlock);
		
		Location currentLocation = locationMap.get(monitorEnterBasicBlock);
		
		for (Transition transition : template.getIncommingTransitions(currentLocation)) {
			transition.addUpdate("monitorEnter()");
		}
	}

	@Override
	public void visitMonitorExitBasicBlock(
			MonitorExitBasicBlock monitorExitBasicBlock) {
		defaultVisit(monitorExitBasicBlock);
		
		Location currentLocation = locationMap.get(monitorExitBasicBlock);

		for (Transition transition : template.getOutgoingTransitions(currentLocation)){
			transition.addUpdate("monitorExit()");
		}
	}

	@Override
	public void visitLoopBasicBlock(LoopBasicBlock loopBasicBlock) {
		defaultVisit(loopBasicBlock);
		
		LoopBound loopBound = loopBasicBlock.getLoopBound();
		//If the loop bound is not already added, add it
		if (!loopBounds.containsValue(loopBound)){
			if (loopBound.getLoopBound() < 0)
				logger.severe("Loop bound not annotated in method: " + template.getMethodInformation().getCanonicalName());
			loopBounds.put(loopCount, loopBound);
			template.getDeclaration().addInteger(new Int("loopBound" + loopCount));
			
			AbstractBasicBlock loopTargetNode = loopBound.getTargetNode();
			//System.err.println("looptargetnode: " + loopTargetNode.getBytecodes().get(0));
			Location loopTargetLocation = locationMap.get(loopTargetNode);
			loopTargetLocation.addCodeLine("\nloopTargetLocation");
			assert loopTargetLocation != null : "loopTargetLocation is null!";
			
			Iterator<LoopBasicBlock> loopNodes = loopBound.getLoopNodeIterator();
			//Update the loop counter each time the loop is run
			while(loopNodes.hasNext()){
				LoopBasicBlock loopNode = loopNodes.next();
				Location loopLocation = locationMap.get(loopNode);
				loopLocation.addCodeLine("\nloopLocation");
				//Continue the loop if loopbound is not met
				for (Transition transition : template.getTransitions(loopLocation.getId(), loopTargetLocation.getId())) {
					transition.addGuard("loopBound" + loopCount + " < " + loopBound.getLoopBound());
					transition.addUpdate("loopBound" + loopCount + "++");
				}
				
//				//Add the guard to leave the loop
//				for (Location location : template.getSuccs(loopLocation)){
//					if (!location.equals(loopTargetLocation)){
//						for (Transition transition : template.getTransitions(loopLocation.getId(), location.getId())) {
//							if (loopBound.getLoopBound() < 0){
//								transition.addGuard("loopBound" + loopCount + " == 0");
//								//Resets the counter when exit the loop
//								transition.addUpdate("loopBound" + loopCount + " = 0");
//								transition.addUpdate("loopAnnotationNotSet = true");
//							} else {
//								transition.addGuard("loopBound" + loopCount + " == " + loopBound.getLoopBound());
//								//Resets the counter when exit the loop
//								transition.addUpdate("loopBound" + loopCount + " = 0");
//							}
//						}
//					}
//				}
			}
			
			Iterator<AbstractBasicBlock> exitNodes = loopBound.getExitNodeIterator();
			//Add exit conditions for all exit nodes
			while(exitNodes.hasNext()){
				AbstractBasicBlock exitNode = exitNodes.next();
				assert exitNode != null : "No exit node found";
				
				Location exitLocation = locationMap.get(exitNode);
				exitLocation.addCodeLine("\nexitLocation");
				AbstractBasicBlock succNode = null;
				int position = 0;
				//locates the successor nodes which exits the loop
				for (AbstractBasicBlock currentNode : basicBlockGraph.getSuccs(exitNode)) {
					if (currentNode.getBytecodes().get(0).getPosition() >= position){
						succNode = currentNode;
						position = currentNode.getBytecodes().get(0).getPosition();
					}
				}
				
				assert succNode != null : "No successor node found " + exitNode.getBytecodes().get(0) + " " + ((Object)exitNode) + " " + template.getName();
				
				Location exitReturnLocation = locationMap.get(succNode);
				exitReturnLocation.addCodeLine("\nexitReturnLocation");
					
				//Add the guard to leave the loop
				for (Transition transition : template.getTransitions(exitLocation.getId(), exitReturnLocation.getId())) {
					if (loopBound.getLoopBound() == -1){
						transition.addGuard("loopBound" + loopCount + " == 0");
						//Resets the counter when exit the loop
						transition.addUpdate("loopBound" + loopCount + " = 0");
						transition.addUpdate("loopAnnotationNotSet = true");
					} else {
						transition.addGuard("loopBound" + loopCount + " == " + loopBound.getLoopBound());
						//Resets the counter when exit the loop
						transition.addUpdate("loopBound" + loopCount + " = 0");
					}
				}
			}
			
			//To make a unique counter for each loop in a template
			loopCount++;
		}
	}
	
	@Override
	public void defaultVisit(AbstractBasicBlock s) {
		Location currentLocation = locationMap.get(s);
		String wcetInst = "inst" + currentLocation.getId();
		if(!Options.EXPLICIT_MODEL) {
			if(Options.PLATFORM_TYPE == PlatformType.JOP) {
				if (Constants.USE_WCET){
					currentLocation.setInvariant("executionTime <= " + (Constants.INLINE_WCET?s.getWCET():wcetInst) + 
						" &&\nexecutionTime' == running[tID] &&\nwcet' == running[tID]");
				}else /*if(!Options.EXPLICIT_MODEL)*/{
					currentLocation.setInvariant("executionTime <= " + (Constants.INLINE_WCET?s.getWCET():wcetInst) + 
						" &&\nexecutionTime' == running[tID]");
				}
				if(!Constants.INLINE_WCET) {
					template.getDeclaration().addInteger(new Int(wcetInst, s.getWCET(), true));
				}
			} else if(Options.PLATFORM_TYPE == PlatformType.HVMAVR) {
				if(s.getBytecodes().isEmpty()) {
					currentLocation.setInvariant("executionTime <= " + s.getWCET() + 
							" &&\nexecutionTime' == running[tID]");
				} else {
					Instruction inst = s.getBytecodes().get(0).getInstruction();
					HVMAVR hvmavrPlatform = HVMAVR.getInstance();
					
					InstructionTimingInfo timingInfo = hvmavrPlatform.getInstructionTiming(inst.getName());

					if (Constants.USE_WCET){
						currentLocation.setInvariant("executionTime <= " + timingInfo.getWcet() + 
								" &&\nexecutionTime' == running[tID] &&\nwcet' == running[tID]");
					}
					else {
						currentLocation.setInvariant("executionTime <= " + timingInfo.getWcet() + 
								" &&\nexecutionTime' == running[tID]");
					}
					
				}
			}
		}
		if(Options.STATISTICAL_MODEL_CHECKING) {
			currentLocation.setExponentialRate(Constants.EXPONENTIAL_RATE_STR);
		}
		for (Transition transition : template.getOutgoingTransitions(currentLocation)) {
			if(!Options.EXPLICIT_MODEL) {
				if(Options.PLATFORM_TYPE == PlatformType.JOP) {
					transition.addGuard("executionTime == " + (Constants.INLINE_WCET?s.getWCET():wcetInst));
				} else if(Options.PLATFORM_TYPE == PlatformType.HVMAVR) {
					if(s.getBytecodes().isEmpty()) {
						transition.addGuard("executionTime == " + s.getWCET());
					} else {
						Instruction inst = s.getBytecodes().get(0).getInstruction();
						HVMAVR hvmavrPlatform = HVMAVR.getInstance();
						
						InstructionTimingInfo timingInfo = hvmavrPlatform.getInstructionTiming(inst.getName());
						if(timingInfo.getBcet() != timingInfo.getWcet()) {
							transition.addGuard("executionTime >= " + timingInfo.getBcet() + " &&\n" +
												"executionTime <= " + timingInfo.getWcet());
						}
						else {
							transition.addGuard("executionTime <= " + timingInfo.getWcet());
						}
					}
				}
				transition.addUpdate("executionTime = 0");
			}
			else if(Options.EXPLICIT_MODEL) {
				if(s.getBytecodes().size() > 0) {
					Instruction ins = s.getBytecodes().get(0).getInstruction();
					transition.addGuard("running[tID] == true");
					String instructionStr = "JVM_" + ins.getName().toUpperCase();
					transition.addUpdate("jvm_instruction = " + instructionStr);
					transition.setSync(new Channel("jvm_execute", "bla", false), true);
					Declarations tempDecl = template.getDeclaration();
					boolean intAlreadyInDecl = false;
					for(Int integer : tempDecl.getIntegers()) {
						if(integer.getId().equals(instructionStr)) {
							intAlreadyInDecl = true;
						}
					}
					if(!intAlreadyInDecl)
						template.getDeclaration().addInteger(new Int(instructionStr, ins.getOpcode(), true));
				}
			}
		}
		
		super.defaultVisit(s);
	}

	
	
	public void generateLoopBacktrack() {
		for(Integer loopCount : loopBounds.keySet()) {
			LoopBound loopBound = loopBounds.get(loopCount);
			
			Iterator<LoopBasicBlock> loopNodes = loopBound.getLoopNodeIterator();
			
			while(loopNodes.hasNext()){
				LoopBasicBlock loopNode = loopNodes.next();
				Location loopLocation = locationMap.get(loopNode);

				if(template.getSuccs(loopLocation).size() == 1) {
					// add backtrack
					String guard;
					
					if (loopBound.getLoopBound() == -1){
						guard = "loopBound" + loopCount + " == 0";
					} else {
						guard = "loopBound" + loopCount + " == " + loopBound.getLoopBound();
					}

					template.addBacktrack(loopLocation, guard);
				}
			}

		
			Iterator<AbstractBasicBlock> exitNodes = loopBound.getExitNodeIterator();
			
			while(exitNodes.hasNext()){
				AbstractBasicBlock exitNode = exitNodes.next();
				Location exitLocation = locationMap.get(exitNode);

				if(template.getSuccs(exitLocation).size() == 1) {
					// add backtrack
					String guard = "loopBound" + loopCount + " < " + loopBound.getLoopBound();

					template.addBacktrack(exitLocation, guard);
				}
			}
		}
	}
}
