package tetasarts.uppaal.translation;

import java.util.HashMap;
import java.util.Hashtable;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.logging.Logger;

import org.apache.bcel.generic.ObjectType;

import tetasarts.generator.Constants;
import tetasarts.tir.basicblocks.AbstractBasicBlock;
import tetasarts.tir.basicblocks.AbstractBasicBlockVisitor;
import tetasarts.tir.basicblocks.BasicBlockGraph;
import tetasarts.tir.basicblocks.LoopBound;
import tetasarts.tir.basicblocks.basicBlockTypes.NewBasicBlock;
import tetasarts.tir.classinfo.AbstractClassNode;
import tetasarts.tir.classinfo.ClassGraph;
import tetasarts.tir.classinfo.NormalClass;
import tetasarts.tir.classinfo.SporadicTask;
import tetasarts.tir.methods.MethodInformation;
import tetasarts.uppaal.Location;
import tetasarts.uppaal.Transition;
import tetasarts.uppaal.UppaalModel;
import tetasarts.uppaal.UppaalTemplate;
import tetasarts.uppaal.declarations.Int;

public class TemplateFinalizerCountVisitor extends AbstractBasicBlockVisitor{
	private UppaalTemplate template;
	private Hashtable<AbstractBasicBlock, Location> locationMap;
	private BasicBlockGraph basicBlockGraph;
	private ClassGraph classGraph;
	private UppaalModel uppaalModel;
	
	public static Map <ObjectType, Integer> classIndexMap = new HashMap<ObjectType, Integer>();
	public static Map <MethodInformation, Integer> finalizerIndexMap = new HashMap<MethodInformation, Integer>();
	private static int finalizerCount = 0;
	public TemplateFinalizerCountVisitor(UppaalTemplate template,
			Hashtable<AbstractBasicBlock, Location> locationMap,
			BasicBlockGraph basicBlockGraph, ClassGraph classGraph,
			UppaalModel uppaalModel) {
		this.uppaalModel = uppaalModel;
		this.template = template;
		this.locationMap = locationMap;
		this.basicBlockGraph = basicBlockGraph;
		this.classGraph = classGraph;
	}

	private int getFinalizerIndex(MethodInformation finalizerMethod){
		if(!finalizerIndexMap.containsKey(finalizerMethod)){
			finalizerIndexMap.put(finalizerMethod, finalizerCount++);
		}
		return finalizerIndexMap.get(finalizerMethod);
	}
	
	public static int getTotalFinalizerCount(){
		return finalizerCount;
	}
	
	@Override
	public void visitNewBasicBlock(NewBasicBlock newBasicBlock) {
		
		Location currentLocation = locationMap.get(newBasicBlock);
		List<Transition> transitions=  template.getOutgoingTransitions(currentLocation);
		AbstractClassNode abstractClassInfo = classGraph.getClassFromCanonicalName(newBasicBlock.getObjectType().getClassName());
		MethodInformation finalizer = abstractClassInfo.getConcreteMethodImplementation("finalize", "()V");
		
		for(Transition t : transitions){
			t.addUpdate("finalizeCount[tID]["+getFinalizerIndex(finalizer)+"] ++");
		}
		
		super.visitNewBasicBlock(newBasicBlock);
	}	
}
