package tetasarts.uppaal.translation;

import java.util.HashMap;
import java.util.Hashtable;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.logging.Logger;

import org.apache.bcel.generic.ObjectType;

import tetasarts.Options;
import tetasarts.generator.Constants;
import tetasarts.tir.basicblocks.AbstractBasicBlock;
import tetasarts.tir.basicblocks.AbstractBasicBlockVisitor;
import tetasarts.tir.basicblocks.BasicBlockGraph;
import tetasarts.tir.basicblocks.LoopBound;
import tetasarts.tir.basicblocks.basicBlockTypes.NewBasicBlock;
import tetasarts.tir.classinfo.ClassGraph;
import tetasarts.tir.classinfo.SporadicTask;
import tetasarts.uppaal.Location;
import tetasarts.uppaal.Transition;
import tetasarts.uppaal.UppaalModel;
import tetasarts.uppaal.UppaalTemplate;
import tetasarts.uppaal.declarations.Int;

public class TemplateNewCountVisitor extends AbstractBasicBlockVisitor{
	private UppaalTemplate template;
	private Hashtable<AbstractBasicBlock, Location> locationMap;
	private BasicBlockGraph basicBlockGraph;
	private ClassGraph classGraph;
	private UppaalModel uppaalModel;
	
	public static Map <ObjectType, Integer> classIndexMap = new HashMap<ObjectType, Integer>();
	public static Map <String, Integer> finalizerIndexMap = new HashMap<String, Integer>();
	private static int classCount = 0;
	public TemplateNewCountVisitor(UppaalTemplate template,
			Hashtable<AbstractBasicBlock, Location> locationMap,
			BasicBlockGraph basicBlockGraph, ClassGraph classGraph,
			UppaalModel uppaalModel) {
		this.uppaalModel = uppaalModel;
		this.template = template;
		this.locationMap = locationMap;
		this.basicBlockGraph = basicBlockGraph;
		this.classGraph = classGraph;
	}

	private int getIndex(NewBasicBlock newBasicBlock){
		ObjectType objectType = newBasicBlock.getObjectType();
		if(!classIndexMap.containsKey(objectType)){
			classIndexMap.put(objectType, classCount++);
		}
		return classIndexMap.get(objectType);
	}
	
	public static int getTotalClassCount(){
		return classCount;
	}
	
	@Override
	public void visitNewBasicBlock(NewBasicBlock newBasicBlock) {
		Location currentLocation = locationMap.get(newBasicBlock);
		List<Transition> transitions=  template.getOutgoingTransitions(currentLocation);
		if(Options.CONSIDER_FINALIZERS){
			for(Transition t : transitions){
				t.addUpdate("instanceCount[tID]["+getIndex(newBasicBlock)+"] ++");
			}
		}
		if(Options.COUNT_NEWS){
			for(Transition t : transitions){
				t.addUpdate(Options.NEW_COUNTER_ID + "++");
			}
		} else {
			System.err.println("ERROR");
			System.exit(0);
		}
		super.visitNewBasicBlock(newBasicBlock);
	}
	
}
