package tetasarts.uppaal.translation;

import java.util.Hashtable;
import java.util.Set;

import tetasarts.tir.basicblocks.AbstractBasicBlock;
import tetasarts.tir.basicblocks.AbstractBasicBlockVisitor;
import tetasarts.tir.basicblocks.BasicBlockGraph;
import tetasarts.tir.basicblocks.basicBlockTypes.MethodCallingBasicBlock;
import tetasarts.tir.basicblocks.basicBlockTypes.NewBasicBlock;
import tetasarts.tir.classinfo.ClassGraph;
import tetasarts.tir.methods.MethodInformation;
import tetasarts.uppaal.Location;
import tetasarts.uppaal.Transition;
import tetasarts.uppaal.UppaalTemplate;

public class TransitionCreaterVisitor extends AbstractBasicBlockVisitor{
	private UppaalTemplate template;
	private Hashtable<AbstractBasicBlock, Location> locationMap;
	private BasicBlockGraph basicBlockGraph;
	private ClassGraph classGraph;
	
	public TransitionCreaterVisitor(UppaalTemplate template, Hashtable<AbstractBasicBlock, Location> locationMap, BasicBlockGraph graph, ClassGraph classGraph){
		this.template = template;
		this.locationMap = locationMap;
		this.basicBlockGraph = graph;
		this.classGraph = classGraph;
	}

	@Override
	public void visitMethodCallingBasicBlock(MethodCallingBasicBlock s) {
		//Set<MethodInformation> concreteMethods = s.findConcreteMethods(classGraph);
		
		Set<MethodInformation> concreteMethods = null;
		try {
			concreteMethods = s.getPossibleCallees();
		} catch (Exception e) {
			System.err.println(e.getMessage());
			e.printStackTrace();
			System.exit(1);
		}
		
		Location currentLocation = locationMap.get(s);
		Location intermediateLocation1 = template.getSuccs(currentLocation).get(0);
		Location intermediateLocation2 = template.getSuccs(intermediateLocation1).get(0);
		template.removeTransition(currentLocation, intermediateLocation1);
		template.removeTransition(intermediateLocation1, intermediateLocation2);
		
		//add transitions for each possible method
		for(int i = 0; i < concreteMethods.size(); i++) {
			Transition intermediateTransition = new Transition(currentLocation.getId(), intermediateLocation1.getId());
			template.addTransition(intermediateTransition);
			Transition transition = new Transition(intermediateLocation1.getId(), intermediateLocation2.getId());
			template.addTransition(transition);
		}
		//Add transitions to each successor node
		for (AbstractBasicBlock succNode : basicBlockGraph.getSuccs(s)) {
			Location succLocation = locationMap.get(succNode);
			Transition transition = new Transition(intermediateLocation2.getId(), succLocation.getId());
			template.addTransition(transition);
		}
		
		if(concreteMethods.size() == 0) {
			Transition intermediateTransition = new Transition(currentLocation.getId(), intermediateLocation1.getId());
			template.addTransition(intermediateTransition);

			Transition transition = new Transition(intermediateLocation1.getId(), intermediateLocation2.getId());
			template.addTransition(transition);
			
			for (AbstractBasicBlock succNode : basicBlockGraph.getSuccs(s)) {
				Location succLocation = locationMap.get(succNode);
				transition = new Transition(intermediateLocation2.getId(), succLocation.getId());
				template.addTransition(transition);
			}
		}
	}
	
//	@Override
//	public void visitLoopBasicBlock(LoopBasicBlock s) {
//		defaultVisit(s);
//		
//		LoopBound loopBound = s.getLoopBound();
//		//If the loop bound is not already added, add it
//		if (!loopBounds.contains(loopBound)){
//			loopBounds.add(loopBound);
//			
////			Location loopTargetLocation = locationMap.get(loopBound.getTargetNode());
//			
//			AbstractBasicBlock returnNode = null;
//			int position = 0;
//			Iterator<AbstractBasicBlock> exitNodes = loopBound.getExitNodeIterator();
//			while(exitNodes.hasNext()){
//				AbstractBasicBlock exitNode = exitNodes.next();
//				
//				//locates the successor nodes which exits the loop
//				for (AbstractBasicBlock currentNode : basicBlockGraph.getSuccs(exitNode)) {
//					if (currentNode.getBytecodes().get(0).getPosition() >= position){
//						returnNode = currentNode;
//						position = currentNode.getBytecodes().get(0).getPosition();
//					}
//				}
//				
//				if(basicBlockGraph.getSuccs(exitNode).size() == 1) {
//					
//				}
//				
//				//Add a transition to allow the exit node to continue the loop if loopbound is not met
////				Location exitLocation = locationMap.get(exitNode);
////				Transition continueLoopTransition = new Transition(exitLocation.getId(), loopTargetLocation.getId());
////				template.addTransition(continueLoopTransition);
//			}
//			
//			if (returnNode == null){
//				logger.severe("No return node found for the loop, probably an infinite loop which is not used");
//			} else {
//				//allow the loop to exit if loopbound is met
//				Location returnLocation = locationMap.get(returnNode);
//				Iterator<LoopBasicBlock> loopNodes = loopBound.getLoopNodeIterator();
//				while (loopNodes.hasNext()){
//					LoopBasicBlock loopNode = loopNodes.next();
//					
//					Location loopLocation = locationMap.get(loopNode);
//					Transition exitLoopTransition = new Transition(loopLocation.getId(), returnLocation.getId());
//					template.addTransition(exitLoopTransition);
//				}
//			}
//		}
//	}
	
	@Override
	public void defaultVisit(AbstractBasicBlock s) {
		addOutgoingTransitions(s, 1);
		super.defaultVisit(s);
	}
	
//	@Override
//	public void visitNewBasicBlock(NewBasicBlock newBasicBlock) {
//		addOutgoingTransitions(newBasicBlock, 1);
//	}
	
	private void addOutgoingTransitions(AbstractBasicBlock node, int count) {
		assert count > 0 : "count less than or equal to zero";
		Location currentLocation = locationMap.get(node);
		for (AbstractBasicBlock succNode : basicBlockGraph.getSuccs(node)) {
			Location succLocation = locationMap.get(succNode);
			for(int i = 0; i < count; i++) {
				Transition transition = new Transition(currentLocation.getId(), succLocation.getId());
				template.addTransition(transition);
			}
		}
	}
}
