package tetasarts.uppaal.translation.analysis;

import java.util.HashSet;
import java.util.List;
import java.util.Set;
import java.util.logging.Logger;

import org.apache.bcel.generic.ArrayType;
import org.apache.bcel.generic.ConstantPoolGen;
import org.apache.bcel.generic.InstructionHandle;
import org.apache.bcel.generic.InvokeInstruction;
import org.apache.bcel.generic.ObjectType;
import org.apache.bcel.generic.Type;

import tetasarts.tir.basicblocks.basicBlockTypes.MethodCallingBasicBlock;
import tetasarts.tir.classinfo.AbstractClassNode;
import tetasarts.tir.classinfo.ClassGraph;
import tetasarts.tir.methods.MethodInformation;

public class NaiveDevirtualizationAnalysis implements IDevirtualizationAnalysis {

	protected ClassGraph classGraph;
	private Logger logger = Logger.getLogger("generator");
	
	
	public NaiveDevirtualizationAnalysis(ClassGraph classGraph) {
		this.classGraph = classGraph;		
	}
	
	@Override
	public Set<MethodInformation> devirtualizeMethodCall(CallSite callSite) {
		Set<MethodInformation> result = new HashSet<MethodInformation>();
		
		InvokeInstruction callsiteInstruction = (InvokeInstruction)callSite.getInvokeInstructionHandle().getInstruction();
		ConstantPoolGen conPoolGen = new ConstantPoolGen(callSite.getCallSiteMethod().getConstantPool());
		String calleeMethodName = "";
		try {
			calleeMethodName = callsiteInstruction.getMethodName(conPoolGen);
		} catch (Exception e) {
			e.printStackTrace();
			System.exit(1);
			// TODO: handle exception
		}
		String calleeMethodSignature = callsiteInstruction.getSignature(conPoolGen);
//		System.out.println(callsiteInstruction.getReferenceType(conPoolGen));

		if(callsiteInstruction.getReferenceType(conPoolGen) instanceof ArrayType)
			return result;
		assert callsiteInstruction.getReferenceType(conPoolGen) instanceof ObjectType : "referencetype is not of objecttype";

		String calleeClassName = ((ObjectType) callsiteInstruction.getReferenceType(conPoolGen)).getClassName();
		AbstractClassNode current = classGraph.getClassFromCanonicalName(calleeClassName);
		//AbstractClassNode current = classGraph.getClassFromCanonicalName(callSite.getCallSiteClassName());
		
		
		MethodInformation methodInformation = current.getMethod(calleeMethodName, calleeMethodSignature);
		
		// find first concrete implementation
		while(methodInformation == null) {
			List<AbstractClassNode> list = classGraph.getPreds(current);
			if(list.size() == 0) {
				logger.info("No parent implementation for method: " + calleeClassName + "." + calleeMethodName + calleeMethodSignature);
				break;
			}
			assert list.size() == 1 : "Class contains more than one predecessor! (multiple inheritance?)";
			current = list.get(0);
			methodInformation = current.getMethod(calleeMethodName, calleeMethodSignature);
		}
		if(methodInformation != null) {
			result.add(methodInformation);
		}
		
		/* Originally this was set to test for specialinvoke instead of
		 * invokestatic. However, why? The following below is WRONG. it is treated as specialinvoke!
		 */
		if (!callSite.isStaticInvoke() && !callsiteInstruction.getName().toLowerCase().contains("special")) 
			findOverriddenMethods(classGraph, current, calleeMethodName, calleeMethodSignature, result);
		
		if(result.size() == 0) {
			logger.info("Could not find implementation for method: " + calleeClassName + "." + calleeMethodName + calleeMethodSignature);
		}
		for(MethodInformation i : result) {
			if(i.getName().contains("minimal")) {
				System.out.println("why?");
			}
		}
		return result;		
		
	}


	private void findOverriddenMethods(ClassGraph classGraph, AbstractClassNode classNode, String calleeMethodName, String calleeMethodSignature, Set<MethodInformation> result) {
	List<AbstractClassNode> succs = classGraph.getSuccs(classNode);
	for(AbstractClassNode currentClass : succs) {
		MethodInformation currentMethodImplementation = currentClass.getMethod(calleeMethodName, calleeMethodSignature);
		if(currentMethodImplementation != null) {
			result.add(currentMethodImplementation);
		}
		findOverriddenMethods(classGraph, currentClass, calleeMethodName, calleeMethodSignature, result);
	}
}
	
	
	
	

}
