package tetasarts.uppaal.translation.analysis.bceltosoot;

import java.util.HashMap;

import org.apache.bcel.generic.ArrayType;
import org.apache.bcel.generic.BasicType;
import org.apache.bcel.generic.ObjectType;
import org.apache.bcel.generic.ReferenceType;

public class BcelToSootHelper {

	private static final HashMap<org.apache.bcel.generic.Type, soot.Type> bcelBasicTypeToSootBasicTypeMap = 
			new HashMap<org.apache.bcel.generic.Type, soot.Type>();
	
	static {
		bcelBasicTypeToSootBasicTypeMap.put(BasicType.INT, soot.IntType.v());
		bcelBasicTypeToSootBasicTypeMap.put(BasicType.BOOLEAN, soot.BooleanType.v());
		bcelBasicTypeToSootBasicTypeMap.put(BasicType.BYTE, soot.ByteType.v());
		bcelBasicTypeToSootBasicTypeMap.put(BasicType.CHAR, soot.CharType.v());
		bcelBasicTypeToSootBasicTypeMap.put(BasicType.DOUBLE, soot.DoubleType.v());
		bcelBasicTypeToSootBasicTypeMap.put(BasicType.FLOAT, soot.FloatType.v());
		bcelBasicTypeToSootBasicTypeMap.put(BasicType.LONG, soot.LongType.v());
		bcelBasicTypeToSootBasicTypeMap.put(BasicType.VOID, soot.VoidType.v());
		bcelBasicTypeToSootBasicTypeMap.put(BasicType.SHORT, soot.ShortType.v());
		bcelBasicTypeToSootBasicTypeMap.put(BasicType.STRING, soot.RefType.v("java.lang.String"));
		bcelBasicTypeToSootBasicTypeMap.put(BasicType.STRINGBUFFER, soot.RefType.v("java.lang.StringBuffer"));
	}
	
	
	public static soot.Type bcelTypeToSootType(org.apache.bcel.generic.Type bcelType) {
		if(bcelType instanceof BasicType) {
			return bcelBasicTypeToSootBasicTypeMap.get(bcelType);
		}
		else if(bcelType instanceof ReferenceType) 
		{
			if(bcelType instanceof ArrayType) {
				ArrayType bcelArrayType = (ArrayType) bcelType;
				
				return soot.ArrayType.v(BcelToSootHelper.bcelTypeToSootType(bcelArrayType.getBasicType()), bcelArrayType.getDimensions());
			}
			else if(bcelType instanceof ObjectType) 
			{
				ObjectType bcelObjectType = (ObjectType) bcelType;
				
				return soot.RefType.v(bcelObjectType.getClassName());				
			}
		}
		
		return null;
	}
	
}
