package tetasarts.uppaal.translation.tasks;

import java.util.LinkedList;
import java.util.List;
import java.util.Set;

import tetasarts.generator.Constants;
import tetasarts.tir.basicblocks.AbstractBasicBlockVisitor;
import tetasarts.tir.basicblocks.basicBlockTypes.MethodCallingBasicBlock;
import tetasarts.tir.classinfo.ClassGraph;
import tetasarts.tir.methods.MethodInformation;
import tetasarts.uppaal.translation.analysis.Callee;
import tetasarts.uppaal.translation.analysis.IDevirtualizationAnalysis;


public class ConcreteMethodDecorator extends AbstractBasicBlockVisitor {
	private ClassGraph classGraph;
	private IDevirtualizationAnalysis devirtualizationAnalysis;
	private List<MethodInformation> methodQueue;
	private List<MethodInformation> doneMethods;

	public ConcreteMethodDecorator(IDevirtualizationAnalysis devirtualizationAnalysis) {
		this.devirtualizationAnalysis = devirtualizationAnalysis;
		methodQueue = new LinkedList<MethodInformation>();
	}
	
	/**TODO document this */
	public void decorateMethods(MethodInformation runMethod, ClassGraph classGraph) {
		assert classGraph != null : "classGraph is null";
		this.classGraph = classGraph;
		doneMethods = new LinkedList<MethodInformation>();
		assert runMethod != null : "runMetod is null";
//		assert runMethod.isGenerated() : "Run-method ("+runMethod.getCanonicalName()+") has not been generated!";
		methodQueue.add(runMethod);
		
		while(!methodQueue.isEmpty()) {
			MethodInformation mi = methodQueue.remove(0);
			doneMethods.add(mi);
			assert mi != null : "Method information is null!";
			if(mi == null) {
				continue;
			}
			if (!mi.getCanonicalName().equals(Constants.REALTIME_SYSTEM + ".fire(I)V")){
				mi.setUsed(true);
				mi.getInstructionGraph().accept(this);
			}
		}
	}
	
	@Override
	public void visitMethodCallingBasicBlock(MethodCallingBasicBlock s) {
		//Set<MethodInformation> concreteMethods = s.findConcreteMethods(classGraph);
		
		/*
		 * This may be the case for JOP native methods
		 */
		if(s.hasUnresolvedCallee()) {
			Callee unresolvedCallee = s.getUnresolvedCallee();
			MethodInformation jopMethod = this.classGraph
					.getClassFromCanonicalName(unresolvedCallee.getCalleeClassName())
					.getMethod(unresolvedCallee.getCalleeMethodname(), 
							unresolvedCallee.getCalleeMethodSignature());
			
			s.setPossibleCallees(jopMethod);			
		}
		if(!s.isPossibleCalleesSet() && !s.hasUnresolvedCallee())
			s.setPossibleCallees(devirtualizationAnalysis.devirtualizeMethodCall(s.getCallSite()));

		for(MethodInformation mi : s.getPossibleCallees()) {
			if(!doneMethods.contains(mi) && !methodQueue.contains(mi)) {
				assert mi != null : "Found method is null!";
				methodQueue.add(mi);
			}
		}
	}
}
