package javax.scj;


public class PeriodicParameters {
	/**
	 * The period
	 */
	private int period;
	/**
	 * The deadline
	 */
	private int deadline;
	/**
	 * The offset time, indicating how long to wait before starting after the initialization phase
	 */
	private int offset;

	/**
	 * Creates a PeriodicParameters object.
	 * @param period Defines the period of the periodic thread.
	 * @param deadline Defines the deadline of the periodic thread.
	 * @param offset Defines the start time of the periodic thread. This is used as an offset from the initialization phase.
	 */
	public PeriodicParameters(int period, int deadline,
			int offset) {
		this.period = period;
		this.deadline = deadline;
		this.offset = offset;
	}
	
	/**
	 * Creates a PeriodicParameters object.
	 * Offset is initialized to 0.
	 * @param period Defines the period of the periodic thread.
	 * @param deadline Defines the deadline of the periodic thread.
	 */
	public PeriodicParameters(int period, int deadline){
		this.period = period;
		this.deadline = deadline;
		this.offset = 0;
	}
	
	/**
	 * Creates a PeriodicParameters object.
	 * The deadline is the same as the period, and offset is initialized to 0.
	 * @param period Defines the period of the periodic thread.
	 */
	public PeriodicParameters(int period){
		this.period = period;
		this.deadline = period;
		this.offset = 0;
	}

	/**
	 * Get the period.
	 * @return The period.
	 */
	public final int getPeriod() {
		return period;
	}

	/**
	 * Get the deadline.
	 * @return The deadline.
	 */
	public final int getDeadline() {
		return deadline;
	}
	
	/**
	 * Get the offset time.
	 * @return The offset time.
	 */
	public final int getOffset() {
		return offset;
	}
}
