package javax.scj;

import java.util.Vector;

import com.jopdesign.sys.Const;
import com.jopdesign.sys.Native;
import com.jopdesign.sys.RtThreadImpl;

import joprt.RtThread;
import joprt.SwEvent;

public final class RealtimeSystem {

	/**
	 * This class is uninstantiable.
	 */
	private RealtimeSystem() {
	}

	static RtThreadAdapter[] periodicTasks;
	static RtSwEventAdapter[] sporadicTasks;

	static int globalTime;
	static int lastUpdateTime;
	static Object mutex = new Object();

	/**
	 * Starts the real-time system (the mission). All periodic and sporadic
	 * RT-events are scheduled.
	 */
	public static void start() {
		RtThreadImpl.init();

		Vector periodicList = new Vector();
		Vector sporadicList = new Vector();
		
		int cnt = RealtimeThread.threadList.size();
		for (int i = 0; i < cnt; ++i) {
			RealtimeThread realtimeThread = (RealtimeThread)RealtimeThread.threadList.elementAt(i);
			
			if (realtimeThread instanceof PeriodicThread) {
				PeriodicThread periodicThread = (PeriodicThread) realtimeThread;
				
				periodicList.addElement(new RtThreadAdapter(periodicThread, cnt - i, periodicThread.period, periodicThread.offset));
			} else if (realtimeThread instanceof SporadicThread) {
				SporadicThread sporadicThread = (SporadicThread) realtimeThread;
				
				sporadicList.addElement(new RtSwEventAdapter(sporadicThread, cnt - i,
						sporadicThread.minInterarrival));
			}
		}
		periodicTasks = new RtThreadAdapter[periodicList.size()];
		periodicList.copyInto(periodicTasks);
		
		sporadicTasks = new RtSwEventAdapter[sporadicList.size()];
		sporadicList.copyInto(sporadicTasks);

		RtThread.startMission();
	}

	/**
	 * Stop the real-time system (mission). When the event run() method returns
	 * true the cleanup() methods are invoked until returning true for a clean
	 * shutdown.
	 * 
	 */
	public static void stop() {
		for (int i = 0; i < periodicTasks.length; ++i) {
			periodicTasks[i].shutdown();
		}
		for (int i = 0; i < sporadicTasks.length; ++i) {
			sporadicTasks[i].shutdown();
			sporadicTasks[i].handle();
		}

	}

	/**
	 * Schedule a software event
	 * 
	 * @param event
	 *            The event representing the
	 */
	public static void fire(int event) {
		for (int i = 0; i < sporadicTasks.length; ++i) {
			RtSwEventAdapter rsa = sporadicTasks[i];
			if (rsa.re.event == event) {
				rsa.fire();
				break;
			}
		}
	}
	
	/**
	 * Return the elapsed time from system startup in micro seconds.
	 * Wraps around all 4295 seconds.
	 * 
	 * @return
	 */
	public static int currentTimeMicros() {
		return Native.rdMem(Const.IO_US_CNT);
	}
}

class RtThreadAdapter extends RtThread {

	boolean runReturn;
	boolean cleanupReturn;
	boolean shutdown;
	PeriodicThread re;

	public RtThreadAdapter(PeriodicThread re, int prio, int us, int off) {
		super(prio, us, off);
		this.re = re;
	}

	void shutdown() {
		shutdown = true;
	}

	public void run() {
		for (;;) {
			if (runReturn && shutdown) {
				if (!cleanupReturn) {
					cleanupReturn = re.cleanup();
				} else {
					break;
				}
			} else {
				runReturn = re.run();
			}
			waitForNextPeriod();
		}
	}
}

class RtSwEventAdapter extends SwEvent {

	boolean runReturn;
	boolean cleanupReturn;
	boolean shutdown;
	SporadicThread re;

	public RtSwEventAdapter(SporadicThread re, int prio, int us) {
		super(prio, us);
		this.re = re;
		// we assume true for a handler that was never called
		runReturn = true;
	};

	void shutdown() {
		shutdown = true;
	}

	public void handle() {

		// who invokes it for cleanup?
		if (runReturn && shutdown) {
			if (!cleanupReturn) {
				cleanupReturn = re.cleanup();
			}
		} else {
			runReturn = re.run();
		}
	}
}
