// *************************************************************************
// File name  : SporadicParameters.java
// Start date : 27/11/2006.
// Programmer : Hans So.
// Java       : Java 1.5.0_08 
// Description: 
//
// Revision history:
//  date     init comment
//
// *************************************************************************
package javax.scj;


public class SporadicParameters {
	/**
	 * The event which must be fired to invoke the sporadic thread.
	 */
	private int event;
	/**
	 * The minimum interarrival time.
	 */
	private int minInterarrival;
	/**
	 * The deadline.
	 */
	private int deadline;

	/**
	 * Creates a new SporadicParameters object.
	 * @param event Defines the event, which must must fired to invoke the sporadic thread.
	 * @param minInterarrival Defines the minimum interarrival time of the sporadic thread.
	 * @param deadline Defines the deadline of the sporadic thread.
	 */
	public SporadicParameters(int event, int minInterarrival,
			int deadline) {
		this.event = event;
		this.minInterarrival = minInterarrival;
		this.deadline = deadline;
	}
	
	/**
	 * Creates a new SporadicParameters object.
	 * The deadline is the same as the minimum interarrival time.
	 * @param event Defines the event, which must must fired to invoke the sporadic thread.
	 * @param minInterarrival Defines the minimum interarrival time of the sporadic thread.
	 */
	public SporadicParameters(int event, int minInterarrival){
		this.event = event;
		this.minInterarrival = minInterarrival;
		this.deadline = minInterarrival;
	}

	/**
	 * Get the event.
	 * @return The event.
	 */
	public final int getEvent() {
		return event;
	}

	/**
	 * Get the minimum interarrival time.
	 * @return The minimum interarrival time.
	 */
	public final int getMinInterarrival() {
		return minInterarrival;
	}

	/**
	 * Get the deadline.
	 * @return The deadline.
	 */
	public final int getDeadline() {
		return deadline;
	}
}
