package com.tetaj.cfggenerators;

import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Set;


import com.jopdesign.libgraph.cfg.GraphException;
import com.jopdesign.libgraph.cfg.block.BasicBlock.Edge;
import com.jopdesign.libgraph.cfg.statements.common.InvokeStmt;
import com.jopdesign.libgraph.struct.MethodInfo;
import com.jopdesign.libgraph.struct.TypeException;
import com.tetaj.cfggenerators.constants.Constants;
import com.tetaj.cfggenerators.jophelper.CFGGenerator;
import com.tetaj.cfggenerators.jophelper.Utility;
import com.tetaj.ircfg.BasicBlock;
import com.tetaj.ircfg.ControlFlowGraph;
import com.tetaj.ircfg.ControlFlowGraphException;
import com.tetaj.ircfg.Instruction;

public class JBCToIRCFG implements IIRCFGGenerators{

	private HashMap<String, ControlFlowGraph> ircfg;
	
	public JBCToIRCFG(String fqMethodName, String classPath) throws GraphException, TypeException, ControlFlowGraphException{
		CFGGenerator cfggen = new CFGGenerator(fqMethodName, classPath);
		
		this.ircfg = translateCFG(cfggen.getCfgMap());
		
	}

	private HashMap<String, ControlFlowGraph> translateCFG(HashMap<String, com.jopdesign.libgraph.cfg.ControlFlowGraph> cfgMap) throws ControlFlowGraphException {
		
		HashMap<String, ControlFlowGraph> cfgs = new HashMap<String, ControlFlowGraph>();
		
		for(String fqMethodName : cfgMap.keySet()){
			com.jopdesign.libgraph.cfg.ControlFlowGraph jopCFG = cfgMap.get(fqMethodName);
			
			String sourceFilePath = getSourceFilePathFromFqName(fqMethodName);
			
			ControlFlowGraph irCFG = cfgs.get(jopCFG.getSignature());
			
			if(irCFG==null)
				irCFG = new ControlFlowGraph(sourceFilePath, fqMethodName);
			
			for(com.jopdesign.libgraph.cfg.block.BasicBlock jopBB : (List<com.jopdesign.libgraph.cfg.block.BasicBlock>)jopCFG.getBlocks()){

				BasicBlock irBB = new BasicBlock(jopBB.getBlockIndex());
				
				handleInstructions(cfgs, jopBB, irBB);
				
				irCFG.addBasicBlock(irBB);
				
			}
			
			List<BasicBlock> irBbs = irCFG.getSortedBasicBlockList();
			
			for(com.jopdesign.libgraph.cfg.block.BasicBlock jopBB : (List<com.jopdesign.libgraph.cfg.block.BasicBlock>)jopCFG.getBlocks()){
				
				BasicBlock irBb = irBbs.get(jopBB.getBlockIndex());
				
				if(jopBB.hasNextBlock()){
					irBb.setFallthrough(irBbs.get(jopBB.getNextBlockEdge().getTargetBlock().getBlockIndex()));
				}
				for(Edge target : (List<Edge>)jopBB.getTargetEdges()){
					irBb.addBranch(irBbs.get(target.getTargetBlock().getBlockIndex()));
				}
			}
			if(irCFG.getBasicBlocks().size()==0){
				irCFG.addBasicBlock(new BasicBlock(0));
			}
			cfgs.put(fqMethodName, irCFG);
		}
		
		return cfgs;
	}

	private String getSourceFilePathFromFqName(String fqMethodName) {
		String packageName = Utility.getPackageName(fqMethodName);
		String className = Utility.getClassName(fqMethodName);
		String sourceFilePath = getFilePath(packageName, className);
		return sourceFilePath;
	}
	private void handleInstructions(HashMap<String, ControlFlowGraph> cfgs,	com.jopdesign.libgraph.cfg.block.BasicBlock jopBB, BasicBlock irBB) {
		int instrId=0;
		for(com.jopdesign.libgraph.cfg.statements.stack.StackStatement statement : (List<com.jopdesign.libgraph.cfg.statements.stack.StackStatement>)jopBB.getStackCode().getStatements()){
			Instruction irInstr = new Instruction(instrId++);
			irInstr.setOpcode(statement.getOpcode());
			if(statement.getOpcode()>=0)
				irInstr.setMnemonic(Constants.JBC_OPCODE_MNEMONICS[statement.getOpcode()]);
			irInstr.setLinenumber(statement.getLineNumber());
			
			irBB.addInstruction(irInstr);

			if(statement instanceof InvokeStmt){
				if(((InvokeStmt) statement).getMethodInfo()!=null){
					MethodInfo info = ((InvokeStmt) statement).getMethodInfo();
					irInstr.setMethodinvoke(true);
					
					
					
					//Check if instruction is Native Invoke
					if(!info.isNative()){
						ControlFlowGraph methodCfg = cfgs.get(((InvokeStmt) statement).getMethodInfo().getFQMethodName());
						if(methodCfg==null){
							String sourceFilePath = getSourceFilePathFromFqName(((InvokeStmt) statement).getMethodInfo().getFQMethodName());
							methodCfg = new ControlFlowGraph(sourceFilePath, ((InvokeStmt) statement).getMethodInfo().getFQMethodName());
							
							cfgs.put(((InvokeStmt) statement).getMethodInfo().getFQMethodName(), methodCfg);
						}
						irBB.addMethodInvokeCfg(methodCfg,irInstr.getId());
					}else if(info.isNative()){
						irBB.setNativeCallName(info.getFQMethodName());
						irInstr.setNativeCall(true);
					}	
				}else{
					System.out.println("MethodInfo is NULL");
				}
			}
		}
	}

	@Override
	public HashMap<String, ControlFlowGraph> generateIRCFG() {
		return this.ircfg;
	}

	private String getFilePath(String packageName, String className) {
		StringBuilder path = new StringBuilder();
		
		path.append(packageName.replace('.', '/'));
		path.append("/");
		path.append(className);
		path.append(".java");
		
		return path.toString();
	}
}
