package com.tetaj.cfggenerators.jophelper.jopircfg;

import java.util.ArrayList;
import java.util.Collection;
import java.util.List;

import org.apache.log4j.Logger;



import com.jopdesign.libgraph.cfg.ControlFlowGraph;
import com.jopdesign.libgraph.cfg.GraphException;
import com.jopdesign.libgraph.cfg.block.BasicBlock;
import com.jopdesign.libgraph.cfg.block.BasicBlock.Edge;
import com.jopdesign.libgraph.cfg.statements.Statement;
import com.jopdesign.libgraph.cfg.statements.common.InvokeStmt;
import com.jopdesign.libgraph.demo.SimpleConfig;
import com.jopdesign.libgraph.struct.AppStruct;
import com.jopdesign.libgraph.struct.ClassInfo;
import com.jopdesign.libgraph.struct.MethodCode;
import com.jopdesign.libgraph.struct.MethodInfo;
import com.jopdesign.libgraph.struct.MethodInvocation;
import com.jopdesign.libgraph.struct.TypeException;
import com.jopdesign.libgraph.struct.bcel.BcelClassLoader;
import com.jopdesign.libgraph.struct.bcel.BcelMethodCode;
import com.jopdesign.libgraph.struct.type.MethodSignature;
import com.tetaj.cfggenerators.constants.Constants;
import com.tetaj.cfggenerators.jophelper.Utility;

public class IRCFG {
	
	private String classPath;
	
	private String methodName;
	private String methodSignature;
	private String className;
	
	private ArrayList<MethodInfo> invokedMethodInfos;
	private ControlFlowGraph cfg;

	public IRCFG(String fqMethodName, String classPath) throws GraphException, TypeException{
		this.methodName = Utility.getMethodName(fqMethodName);
		this.methodSignature = Utility.getMethodSignature(fqMethodName);
		this.className = Utility.getFullClassName(fqMethodName);
		
		this.classPath = classPath;
		this.invokedMethodInfos = null;
		this.cfg = new ControlFlowGraph(null);
		
		this.buildGraph();
		this.buildInvokedMehthodList();
	}
	
	public List<MethodInfo> getInvokedMethodInfos() {
		return this.invokedMethodInfos;
	}
	
	public ControlFlowGraph getCFG(){
		return this.cfg;
	}
	
	private void buildGraph() throws GraphException, TypeException{
		
		Logger logger = Logger.getLogger(BcelMethodCode.class);
		logger.debug("Log4j has been initialized");
		
		AppStruct appStruct = new AppStruct(new BcelClassLoader(), new SimpleConfig() );
		
		appStruct.setClassPath(this.classPath);
		
		appStruct.loadClassInfo(this.className);
		
		appStruct.initClassInfos(true);
		
		ClassInfo classInfo = appStruct.getClassInfo(this.className,false);
				
		MethodInfo methodInfo = classInfo.getMethodInfo(MethodSignature.createFullName(this.methodName, this.methodSignature));
		
		//TODO Only for debug purposes
		if(methodInfo==null){
			System.out.println("Method not found. Searched for:");
			System.out.println(MethodSignature.createFullName(this.methodName, this.methodSignature));
			System.out.println("Class contains following methods:"); 
			for(MethodInfo containedMethod : (Collection<MethodInfo>)classInfo.getMethodInfos() )
				System.out.println(containedMethod.getFQMethodName());
		}
		
		MethodCode mc = methodInfo.getMethodCode();
		
		// TODO NATIVE CALLS: decide how we do this
		if (mc != null) {
			mc.compileGraph();
			this.cfg = mc.getGraph();
		}
		
		//printCFGtoConsole(cfg);			
	}
	
	//To be removed
	private void printCFGtoConsole(ControlFlowGraph cfg) {
		for(BasicBlock basicBlock : (List<BasicBlock>)cfg.getBlocks()){
			
			
			System.out.println("Block Index: " +basicBlock.getBlockIndex());
			
			int curLine = 0;
			int opcode = 0;
			for(int i= 0; i<basicBlock.getStackCode().size();i++){
				
				if(curLine!=basicBlock.getStackCode().getStackStatement(i).getLineNumber()){
					System.out.println("\t Line: " + basicBlock.getStackCode().getStackStatement(i).getLineNumber());
					curLine = basicBlock.getStackCode().getStackStatement(i).getLineNumber();
				}
				
				opcode = basicBlock.getStackCode().getStackStatement(i).getOpcode();
				if(opcode>=0){
					System.out.println("\t\t Bytecode: " + Constants.JBC_OPCODE_MNEMONICS[opcode]);
				}
				System.out.println("\t\t Type: " + basicBlock.getStackCode().getStackStatement(i).getClass());
			}
			
			System.out.println("\t\t Har du nabo: " + basicBlock.hasNextBlock());
			
			if(basicBlock.getNextBlockEdge()!=null)
				System.out.println("\t\t Next edge: " + basicBlock.getNextBlockEdge().getTargetBlock().getBlockIndex());
			
			System.out.println("\t\t Target num: " + basicBlock.getTargetCount());
			for(Edge target : (List<Edge>)basicBlock.getTargetEdges()){
					System.out.println("\t\t Target ID: " + target.getTargetBlock().getBlockIndex());
					
			}
			
		}
	}
	
	private void buildInvokedMehthodList(){
		this.invokedMethodInfos = new ArrayList<MethodInfo>();
		
		for(BasicBlock basicBlock : (List<BasicBlock>)this.cfg.getBlocks()){
			for(Statement statement : (List<Statement>)basicBlock.getCodeBlock().getStatements()){
				if(statement instanceof InvokeStmt){
					if(((InvokeStmt) statement).getMethodInfo()!=null){
						this.invokedMethodInfos.add(((InvokeStmt) statement).getMethodInfo());
					}	
				}
			}
		}
	}
}
