package com.tetaj.ircfg;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;

public class ControlFlowGraph {

	private String sourceFilePath;
	private String methodName;
	private BasicBlock entryBlock;
	private List<BasicBlock> basicBlocks;
	
	public int highestBlockIndex = 0;
	
	public ControlFlowGraph(String sourceFilePath, String methodName){
		this.sourceFilePath = sourceFilePath;
		this.methodName = methodName;
		this.entryBlock = null;	
		this.basicBlocks = new ArrayList<BasicBlock>();
	}
	
	public void setSourceFilePath(String sourceFilePath) {
		this.sourceFilePath = sourceFilePath;
	}

	public String getSourceFilePath() {
		return sourceFilePath;
	}

	public void setMethodName(String methodName) {
		this.methodName = methodName;
	}

	public String getMethodName() {
		return methodName;
	}

	public BasicBlock getEntryBlock() {
		return getSortedBasicBlockList().get(0);
	}
	
	public List<BasicBlock> getMethodInvokeList(){
		List<BasicBlock> allBasicBlocks = getSortedBasicBlockList();
		
		List<BasicBlock> methodInvokeBasicBlocks = new ArrayList<BasicBlock>();
		
		for(BasicBlock bb: allBasicBlocks){
			if(bb.isMethodInvoke()){
				methodInvokeBasicBlocks.add(bb);
			}
		}
		
		return methodInvokeBasicBlocks;
	}
	
	public BasicBlock getBasicBlock(int id){
		List<BasicBlock> allBasicBlocks = getSortedBasicBlockList();
		
		return allBasicBlocks.get(id);	
	}
	
	public List<BasicBlock> getSortedBasicBlockList(){
		
		Collections.sort(this.basicBlocks);	
		
		return this.basicBlocks;
	}
	

	public void addBasicBlock(BasicBlock basicBlock) {
		if(basicBlock.getId()==0)
			this.entryBlock=basicBlock;
		
		this.basicBlocks.add(basicBlock);
	}
	public void addBasicBlock(List<BasicBlock> basicBlockList) {
		this.basicBlocks = basicBlockList;
	}
	
	public List<BasicBlock> getBasicBlocks() {
		return basicBlocks;
	}

	public boolean removeBasicBlock(BasicBlock basicBlock) {
		return this.basicBlocks.remove(basicBlock);
		
	}

	public void resetIds() {
		int counter = 0;

		for (BasicBlock basicBlock : this.getSortedBasicBlockList()) {
			basicBlock.setId(counter);
			counter++;
		}
	}
}
