package com.tetaj.modelgenerator;

import uppaal.Automaton;
import uppaal.Declaration;
import uppaal.Location;
import uppaal.Location.LocationType;
import uppaal.NTA;
import uppaal.Transition;
import uppaal.labels.Invariant;
import uppaal.labels.Synchronization;
import uppaal.labels.Synchronization.SyncType;
import uppaal.labels.Update;

import com.tetaj.cfggenerators.constants.Constants;

import com.tetaj.modelgenerator.jop.JopWcetCalculator;
import com.tetaj.modelgenerator.jop.NotImplementedException;


public class JopBuilder {

	public static void main(String args[]) 
	{
		if (args.length < 1) {
			System.out.println("Usage: <targetName>");
			System.exit(1);
		}
		
		String targetFileName = args[0];
		

		NTA ntaSystem = new NTA();
		short jbcInsOpcode = 0;
		
		for(String jbcOpcodeMnemonic : Constants.JBC_OPCODE_MNEMONICS) {
			if(jbcOpcodeMnemonic.equals(Constants.ILLEGAL_OPCODE))
				continue;
			try {
				int jbcWcet = JopWcetCalculator.getWCET(jbcInsOpcode);
			
				String autName = "handle" + jbcOpcodeMnemonic.toUpperCase() + "_OPCODE_T";
				Automaton jbcAutomaton = new Automaton(autName);
				
				String executionTimeClock = "jbcExecutionTime";
				String wcetJbcVar = jbcOpcodeMnemonic + "_wcet";
				jbcAutomaton.setDeclaration(new Declaration("clock " + executionTimeClock + "; \nconst int " + wcetJbcVar + " = " + jbcWcet + ";"));
				
				Location idleLocation = new Location(jbcAutomaton, "Idle");
				jbcAutomaton.setInit(idleLocation);
				
				Location execLocation = new Location(jbcAutomaton, "Executing_" + jbcOpcodeMnemonic.toUpperCase()); 
				
				Transition idleExecTrans = new Transition(jbcAutomaton, idleLocation, execLocation);
				
				String invokeChannelName = "invoke_assembly_handle" + jbcOpcodeMnemonic.toUpperCase() + "_OPCODE";
				
				Synchronization invokeSynchronisation = new Synchronization(invokeChannelName, SyncType.RECEIVER);
				idleExecTrans.setSync(invokeSynchronisation);
				Update resetExecTimeUpdate = new Update(executionTimeClock + " = 0");
				idleExecTrans.setUpdate(resetExecTimeUpdate);

				execLocation.setInvariant(new Invariant(executionTimeClock + " <= " + wcetJbcVar));

				Location returnLocation = new Location(jbcAutomaton, "Return");
				returnLocation.setType(LocationType.COMMITTED);
				
				Transition execReturnTrans = new Transition(jbcAutomaton, execLocation, returnLocation);
				execReturnTrans.setGuard(executionTimeClock + " == " + wcetJbcVar);
				execReturnTrans.setUpdate(resetExecTimeUpdate);
				
				Transition returnIdletrans = new Transition(jbcAutomaton, returnLocation, idleLocation);
				
				String returnChannelName = "return_assembly_handle" + jbcOpcodeMnemonic.toUpperCase() + "_OPCODE";
				
				Synchronization returnSynchronisation = new Synchronization(returnChannelName, SyncType.INITIATOR);
				returnIdletrans.setSync(returnSynchronisation);
				
				ntaSystem.addAutomaton(jbcAutomaton);
				
				
				ntaSystem.getDeclarations().add(new Declaration("urgent chan " + invokeChannelName + ";"));
				ntaSystem.getDeclarations().add(new Declaration("urgent chan " + returnChannelName + ";"));
				
				ntaSystem.getSystemDeclaration().addSystemInstance(autName);
				
			} catch (NotImplementedException e) {
				
			}
			jbcInsOpcode++;
		}
		//This is just to compensate for the structure of the staticjvm model that assumes
		//templates for pre- and postprocessing of the java bytecode
		Automaton preprocessAut = new Automaton("process_jbc_pre_T");
		Location initLocPre = new Location(preprocessAut, "idle");
		preprocessAut.setInit(initLocPre);
		Location execLocPre = new Location(preprocessAut, "exec");
		execLocPre.setType(LocationType.COMMITTED);
		
		Transition idleExecTransPre = new Transition(preprocessAut, initLocPre, execLocPre);
		idleExecTransPre.setSync(new Synchronization("invoke_assembly_process_jbc_pre", SyncType.RECEIVER));
		
		Transition execIdleTransPre = new Transition(preprocessAut, execLocPre, initLocPre);
		execIdleTransPre.setSync(new Synchronization("return_assembly_process_jbc_pre", SyncType.INITIATOR));
		
		ntaSystem.addAutomaton(preprocessAut);
		
		
		Automaton postprocessAut = new Automaton("process_jbc_post_T");
		Location initLoc = new Location(postprocessAut, "idle");
		postprocessAut.setInit(initLoc);
		Location execLoc = new Location(postprocessAut, "exec");
		execLoc.setType(LocationType.COMMITTED);
		
		Transition idleExecTrans = new Transition(postprocessAut, initLoc, execLoc);
		idleExecTrans.setSync(new Synchronization("invoke_assembly_process_jbc_post", SyncType.RECEIVER));
		
		Transition execIdleTrans = new Transition(postprocessAut, execLoc, initLoc);
		execIdleTrans.setSync(new Synchronization("return_assembly_process_jbc_post", SyncType.INITIATOR));
		
		ntaSystem.addAutomaton(postprocessAut);
		

		
		System.out.println("writing to: " + targetFileName);
		
		ntaSystem.writePrettyLayoutModelToFile(targetFileName);
		System.out.println("finished");
		/*
		String systemString = "system "; 
		
		//Create templates for all methods
		for(String fqMethodName : cfgMap.keySet()){
			ControlFlowGraph cfg = cfgMap.get(fqMethodName);
					
			MethodTemplate methodTemplate = new MethodTemplate(fqMethodName, cfg, this.uniqueIdentifier);
			system.addTemplate(methodTemplate.getTemplate());
			channels.add("urgent chan " + "invoke_" + this.uniqueIdentifier + "_" + FQNameHelper.getUppaalFriendlyName(fqMethodName) + ";");
			channels.add("urgent chan " + "return_" + this.uniqueIdentifier + "_" + FQNameHelper.getUppaalFriendlyName(fqMethodName) + ";");
			systemString += "\n\t" + methodTemplate.getUppaalName() + ",";
			
		}
		
		//Adding all method channels
		for (String channel : channels){
			system.appendDeclaration(channel);
		}
		
		for (int i = 0; i<this.instructionMnemonics.length; i++){
			if(this.instructionMnemonics[i] != Constants.ILLEGAL_OPCODE){
				system.appendDeclaration("const int " + this.uniqueIdentifier.toUpperCase() + "_" + this.instructionMnemonics[i].toUpperCase() + " = " + i +";");		
			}
		}
				
		system.appendDeclaration("urgent chan " + this.uniqueIdentifier + "_execute;");
		system.appendDeclaration("int " + this.uniqueIdentifier + "_instruction;");
		
		systemString = systemString.substring(0,systemString.length()-1) + ";";
		system.setSystem(systemString);
				
		return system;
		*/
		
		
	}
	
	
}
