package com.tetaj.modelgenerator.cfganalysis;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;

import com.tetaj.ircfg.BasicBlock;
import com.tetaj.ircfg.ControlFlowGraph;
import com.tetaj.ircfg.Edge;
import com.tetaj.ircfg.LoopInfo;

public class ProgressMeasureAnalysis implements IAnalysis {

	private HashMap<String, ControlFlowGraph> CFGtable;
	
	public ProgressMeasureAnalysis(HashMap<String, com.tetaj.ircfg.ControlFlowGraph> CFGTable) {
		this.CFGtable = CFGTable;
	}
	
	//Requires Loop Bound Analysis to be run first
	@Override
	public void conductAnalysis() throws CfgAnalysisException {
		
		ArrayList<Edge> edges = new ArrayList<Edge>();
		
		//Add all edges to list
		for(ControlFlowGraph cfg : this.CFGtable.values()){
			for(BasicBlock bb : cfg.getBasicBlocks()){
				
				for(Edge branchEdge : bb.getBranchEdges()){
					if(!edges.contains(branchEdge)){
						edges.add(branchEdge);
					}	
				}
				if(bb.getFallthroughEdge() != null && !edges.contains(bb.getFallthroughEdge())){
					edges.add(bb.getFallthroughEdge());
				}
			}
		}
		
		//Add PM update to all edges
		for(Edge edge : edges){
			if(edge != null && edge.getLoopExitInfo().size()==0){
				edge.setProperty("pm", "1");
			}
			else if(edge != null && edge.getLoopExitInfo().size()==1){
				LoopInfo info = edge.getLoopExitInfo().get(0);
				
				ArrayList<BasicBlock> loopbbs = info.getBasicBlocks();
				
				//Find and add method invoke Bbs
				ArrayList<BasicBlock> tempBbs = new ArrayList<BasicBlock>();
				for(BasicBlock bb : loopbbs){
					if(bb.isMethodInvoke()){
						addMethodInvokeBbs(bb.getMethodInvokeCfg(),tempBbs);
					}
				}
				//Add all method invoke bbs to list
				loopbbs.addAll(tempBbs);
				  
				
				int longestPath = findLongestPath(edge.getSource(), loopbbs);
				longestPath++;
				
				edge.setProperty("pm", "" + (info.getLoopBound()*longestPath+1) + "-" + longestPath + "*");
			}			
		}

	}
	
	private int findLongestPath(BasicBlock source,	ArrayList<BasicBlock> loopbbs) {
		instantiateLongestPath(loopbbs.size(), loopbbs);
		
		
		return longestPath(source, loopbbs);

	}
	private int longestPath(BasicBlock bb, ArrayList<BasicBlock> loopbbs) {
		int dist, max = 0;
		  bb.setProperty("VISITED", true);
		  
		  ArrayList<BasicBlock> neighbours = new ArrayList<BasicBlock>();
		  findNeighbours(bb, neighbours, loopbbs);
		  
		  for (BasicBlock neighbour : neighbours)
		    if (!(Boolean)neighbour.getProperty("VISITED")) {
		      dist = 1 + longestPath(neighbour, loopbbs);
		      if (dist > max)
		        max = dist; 
		    }
		  bb.setProperty("VISITED", false);
		  
		  return(max);
	}
			
			
	private void findNeighbours(BasicBlock bb, ArrayList<BasicBlock> neighbours,ArrayList<BasicBlock> loopbbs) {
		
		for(Edge edge : bb.getBranchEdges()){
			if(loopbbs.contains(edge.getTarget()) && !neighbours.contains(edge.getTarget())){
				neighbours.add(edge.getTarget());
			}
		}
		if(bb.getFallthroughEdge()!=null){
			if(loopbbs.contains(bb.getFallthroughEdge().getTarget()) && !neighbours.contains(bb.getFallthroughEdge().getTarget())){
				neighbours.add(bb.getFallthroughEdge().getTarget());
			}
		}
		if(bb.isMethodInvoke()){
			for(ControlFlowGraph cfg : bb.getMethodInvokeCfg().values()){
				neighbours.add(cfg.getEntryBlock());
			}
		}
	
	}


	private void instantiateLongestPath(int max, ArrayList<BasicBlock> loopbbs) {	
		for(BasicBlock bb : loopbbs){
			bb.setProperty("VISITED", false);
		}
	}
		
	



	private void addMethodInvokeBbs(HashMap<Integer, ControlFlowGraph> hashMap, ArrayList<BasicBlock> tempBbs) {
		for(ControlFlowGraph cfg : hashMap.values()){
			tempBbs.addAll(cfg.getBasicBlocks());
			
			for(BasicBlock bb : cfg.getBasicBlocks()){
				if(bb.isMethodInvoke()){
					addMethodInvokeBbs(bb.getMethodInvokeCfg(),tempBbs);
				}
			}
		}
	}

}
