package com.tetaj.modelgenerator.staticmodels;

import java.util.ArrayList;
import java.util.Arrays;

import com.tetaj.cfggenerators.constants.Constants;

import com.jopdesign.wcet.uppaal.model.DuplicateKeyException;
import com.jopdesign.wcet.uppaal.model.Location;
import com.jopdesign.wcet.uppaal.model.NTASystem;
import com.jopdesign.wcet.uppaal.model.Template;
import com.jopdesign.wcet.uppaal.model.Transition;

public class StaticModels {
	public static Template getStaticJvmTemplate(String[] instructionNames, String methodId, String jbcUniqueIdentifier, String hwUniqueIdentifier){
		Template template = new Template("StaticJvm", Arrays.asList(""));
		
		Location idleLoc = new Location("Idle");
		template.setInitialLocation(idleLoc);
		
		Location executionLoc = new Location("ExecuteJBC");
		executionLoc.setCommited();
		template.addLocation(executionLoc);
		
		Transition execTransition = new Transition(idleLoc, executionLoc);
		execTransition.getAttrs().setSync(jbcUniqueIdentifier + "_execute?");
		template.addTransition(execTransition);
		
		Location initInterpretationLoc = new Location("InitialiseInterpretationLoop");
		template.addLocation(initInterpretationLoc);
		Transition initInterpretationTrans = new Transition(executionLoc,initInterpretationLoc);
		initInterpretationTrans.getAttrs().setSync("invoke_assembly_process_jbc_pre!");
		template.addTransition(initInterpretationTrans);
		
		Location choseJbcImplementationLoc = new Location("ChoseJBCImplementation");
		choseJbcImplementationLoc.setCommited();
		template.addLocation(choseJbcImplementationLoc);
		
		Transition waitInterpretationTrans = new Transition(initInterpretationLoc,choseJbcImplementationLoc);
		waitInterpretationTrans.getAttrs().setSync("return_assembly_process_jbc_pre?");
		template.addTransition(waitInterpretationTrans);
			
		Location jbcFinishLoc = new Location("jbcFinish");
		jbcFinishLoc.setCommited();
		template.addLocation(jbcFinishLoc);
		
		Location cleanupInterpretationLoc = new Location("CleanupInterpretationLoop");
		template.addLocation(cleanupInterpretationLoc);
		
		Transition cleanupInterpretationTrans = new Transition(jbcFinishLoc, cleanupInterpretationLoc);
		cleanupInterpretationTrans.getAttrs().setSync("invoke_assembly_process_jbc_post!");
		template.addTransition(cleanupInterpretationTrans);
		
		Transition cleanupInterpretationWaitTrans = new Transition(cleanupInterpretationLoc, idleLoc);
		cleanupInterpretationWaitTrans.getAttrs().setSync("return_assembly_process_jbc_post?");
		template.addTransition(cleanupInterpretationWaitTrans);
		
		for(String instruction : instructionNames){
			if(!instruction.equals(Constants.ILLEGAL_OPCODE)){
				Location newJbcLoc = new Location("JVM_Execute_"+instruction.toUpperCase());
				Transition newJbcTrans = new Transition(choseJbcImplementationLoc,newJbcLoc);
				newJbcTrans.getAttrs().appendGuard(jbcUniqueIdentifier + "_instruction ==" +jbcUniqueIdentifier.toUpperCase() +"_" +instruction.toUpperCase());
				newJbcTrans.getAttrs().setSync("invoke_" + hwUniqueIdentifier+ "_" + methodId + instruction.toUpperCase() + "_OPCODE!");
				
				Transition jbcReturnTrans = new Transition(newJbcLoc,jbcFinishLoc);
				jbcReturnTrans.getAttrs().setSync("return_" +hwUniqueIdentifier+"_"+methodId + instruction.toUpperCase() + "_OPCODE?");
					
				template.addLocation(newJbcLoc);
				template.addTransition(newJbcTrans);
				template.addTransition(jbcReturnTrans);
			}
		}
		return template;
	}
	public static NTASystem getStaticJvmSystem(String[] instructionNames, String methodId, String jbcUniqueIdentifier, String hwUniqueIdentifier) throws DuplicateKeyException{
		NTASystem system = new NTASystem("StaticJvm");
		
		Template template = getStaticJvmTemplate(instructionNames, methodId, jbcUniqueIdentifier, hwUniqueIdentifier);
		
		system.addTemplate(template);
		
		system.appendDeclaration("urgent chan invoke_initialise_interpretation_loop;\n");
		system.appendDeclaration("urgent chan return_initialise_interpretation_loop;\n");
		system.appendDeclaration("urgent chan invoke_cleanup_interpretation_loop;\n");
		system.appendDeclaration("urgent chan return_cleanup_interpretation_loop;\n");
		
		system.setSystem("system StaticJvm;");
		/*
		for(String instruction : instructionNames){
			if(!instruction.equals(Constants.ILLEGAL_OPCODE)){
				system.appendDeclaration("urgent chan invoke_"+ hwUniqueIdentifier + "_" + methodId + instruction.toUpperCase() +"_OPCODE;");
				system.appendDeclaration("urgent chan return_" + hwUniqueIdentifier+ "_" + methodId + instruction.toUpperCase()+ "_OPCODE;");
			}
		}*/
		
		
		return system;
		
	}
	public static NTASystem getMetamocAVRModel(String uniqueIdentifier) throws DuplicateKeyException{
		NTASystem system = new NTASystem("MetamocAVR");
		
		Template templateFetchStage = getMetamocFetchTemplate();		
		system.addTemplate(templateFetchStage);
		
		Template templateExecuteStage = getMetamocExecuteTemplate(uniqueIdentifier);
		system.addTemplate(templateExecuteStage);
		
		String declarations = getMetamocSystemDeclarations();
		
		system.appendDeclaration(declarations);			
		
		system.setSystem("system FetchStage < ExecuteStage;");

		return system;		
	}
	
	public static Template getMetamocInterfaceModel(String uniqueIdentifier){
		Template template = new Template("METAMOCInterface", Arrays.asList(""));
		
		Location idle = new Location("Idle");
		template.setInitialLocation(idle);
		Location exec = new Location("ExecuteMetamoc");
		template.addLocation(exec);
		
		Transition trans1 = new Transition(idle, exec);
		trans1.getAttrs().setSync(uniqueIdentifier + "_execute?");
		trans1.getAttrs().appendUpdate("instrtype[PIPELINE_FETCH_STAGE] = " + uniqueIdentifier + "_instruction");
		template.addTransition(trans1);
		
		Transition trans2 = new Transition(exec, idle);
		trans2.getAttrs().setSync("fetch!");
		template.addTransition(trans2);
	
		return template;
	}
	
	private static String getMetamocSystemDeclarations() {
		String declarations = "typedef int[-1,131071] address;\n" +
							"typedef int[0,65535] register_mask;\n"+
							"const int INSTR_NONE=-1;\n"+
							"const int PIPELINE_STAGES = 2;\n" +
							"const int PIPELINE_FETCH_STAGE = 0;\n" +
							"const int PIPELINE_EXECUTE_STAGE = 1;\n" +
							"const address INVALID_ADDRESS = -1;\n" +
							"int instrtype[PIPELINE_STAGES];\n" +
							"address instradr[PIPELINE_STAGES];\n" +
							"address dataadr[PIPELINE_STAGES];\n" +
							"register_mask regread[PIPELINE_STAGES];\n" +
							"register_mask regwrite[PIPELINE_STAGES];\n" +
							"urgent chan fetch, execute;\n" +
							"urgent chan main_done, fetch_done, execute_done;\n" +
							"clock cyclecounter;void initialise(){\n" +
							"    int stage;\n" +
							"    for (stage = 0; stage < PIPELINE_STAGES; stage++) {\n" +
							"        instrtype[stage] = INSTR_NONE;\n" +
							"        instradr[stage] = INVALID_ADDRESS;\n" +
							"    }\n" +
							"}\n" +
							"void move(int from, int to){\n" +
							"    instrtype[to] = instrtype[from];\n" +
							"    instradr[to] = instradr[from];\n" +
							"    instrtype[from] = INSTR_NONE;\n" +
							"    instradr[from] = INVALID_ADDRESS;\n" +
							"}\n" +
							"void clear(int stage){\n" +
							"    instrtype[stage] = INSTR_NONE;\n" +
							"    instradr[stage] = INVALID_ADDRESS;\n" +
							"}\n" +
							"register_mask bitwise_neg(register_mask regmask){\n" +
							"    return 65535 - regmask;\n" +
							"}\n";
			/*				
		for(int i = 0; i<Constants.AVR_INSTRUCTIONS.length;i++){
				declarations += "const int AVR_" + Constants.AVR_INSTRUCTIONS[i].getMnemonic().toUpperCase() +" = " +i+ ";\n";
		}*/
		
		return declarations;
	}
	private static Template getMetamocFetchTemplate(){
		Template templateFetchStage = new Template("FetchStage", Arrays.asList(""));
		Location locFetchIdle = new Location("");
		templateFetchStage.setInitialLocation(locFetchIdle);
		Location locFetchWait = new Location("");
		templateFetchStage.addLocation(locFetchWait);
		locFetchWait.setInvariant("x<=1");
		Location locFetchExec = new Location("");
		templateFetchStage.addLocation(locFetchExec);
		Location locFetchMainDone = new Location("");
		templateFetchStage.addLocation(locFetchMainDone);
		Location locFetchDone = new Location("");
		templateFetchStage.addLocation(locFetchDone);
		
		Transition transFetch1 = new Transition(locFetchIdle, locFetchWait);
		transFetch1.getAttrs().setSync("fetch?");
		transFetch1.getAttrs().appendUpdate("x=0");
		templateFetchStage.addTransition(transFetch1);
		Transition transFetch2 = new Transition(locFetchWait,locFetchExec);
		transFetch2.getAttrs().appendGuard("x==1");
		templateFetchStage.addTransition(transFetch2);
		Transition transFetch3 = new Transition(locFetchExec,locFetchIdle);
		transFetch3.getAttrs().setSync("execute!");
		transFetch3.getAttrs().appendUpdate("move(THIS,NEXT)");
		templateFetchStage.addTransition(transFetch3);
		Transition transFetch4 = new Transition(locFetchIdle, locFetchMainDone);
		transFetch4.getAttrs().setSync("main_done?");
		templateFetchStage.addTransition(transFetch4);
		Transition transFetch5 = new Transition(locFetchMainDone, locFetchDone);
		transFetch5.getAttrs().setSync("fetch_done!");
		templateFetchStage.addTransition(transFetch5);
		
		templateFetchStage.appendDeclaration("const int THIS = PIPELINE_FETCH_STAGE;\n" +
												"const int NEXT = PIPELINE_EXECUTE_STAGE;\n" +
												"clock x;");
		
		return templateFetchStage;
	}
	private static Template getMetamocExecuteTemplate(String uniqueIdentifier){
		Template templateExecute = new Template("ExecuteStage", Arrays.asList(""));
		
		Location locIdle = new Location("");
		templateExecute.setInitialLocation(locIdle);
		Location locWait = new Location("");
		locWait.setInvariant("x<=wait");
		templateExecute.addLocation(locWait);
		Location locDone = new Location("");
		locDone.setCommited();
		templateExecute.addLocation(locDone);
		
		Transition trans1 = new Transition(locIdle, locWait);
		trans1.getAttrs().appendUpdate("set_wait()");
		trans1.getAttrs().appendUpdate("x=0");
		trans1.getAttrs().setSync("execute?");
		templateExecute.addTransition(trans1);
		Transition trans2 = new Transition(locWait, locIdle);
		trans2.getAttrs().appendUpdate("clear(THIS)");
		trans2.getAttrs().appendGuard("x==wait");
		templateExecute.addTransition(trans2);
		Transition trans3 = new Transition(locIdle, locDone);
		trans3.getAttrs().setSync("fetch_done?");
		templateExecute.addTransition(trans3);
		
		templateExecute.appendDeclaration("const int THIS = PIPELINE_EXECUTE_STAGE;\n" +
										"int wait;\n" +
										"clock x;\n");
		
		for(Constants.AVRInstruction avrInstr : Constants.AVR_INSTRUCTIONS){
			templateExecute.appendDeclaration("const int CYCLES_"+avrInstr.getMnemonic().toUpperCase()+"="+avrInstr.getClockCyclesWorstCase() + ";");
		}
		
		String setwait = getMetamocSetWaitMethod(uniqueIdentifier);
		templateExecute.appendDeclaration(setwait);
		
		
		return templateExecute;
	}
	private static String getMetamocSetWaitMethod(String uniqueIdentifier) {
		String method = "void set_wait()\n{\n";
		
		for(Constants.AVRInstruction avrInstr : Constants.AVR_INSTRUCTIONS){
				method += "\t if(instrtype[THIS] == " + uniqueIdentifier.toUpperCase()+ "_" + avrInstr.getMnemonic().toUpperCase()+"){\n";
				method += "\t\t wait = CYCLES_" + avrInstr.getMnemonic().toUpperCase() +";\n" +
						"\t\t return;\n" +
						"\t\n}";
		}
		
		method +="}";
		
		return method;
	}
}
