package com.tetaj.modelgenerator.uppaalmodelgeneratorircfg;

import java.util.Arrays;
import java.util.HashMap;
import java.util.HashSet;

import com.tetaj.cfggenerators.constants.Constants;
import com.tetaj.ircfg.*;

import com.jopdesign.wcet.uppaal.model.Location;
import com.jopdesign.wcet.uppaal.model.NTASystem;
import com.jopdesign.wcet.uppaal.model.Template;
import com.jopdesign.wcet.uppaal.model.Transition;

public class UppaalModelGenerator {

	private HashMap<String,ControlFlowGraph> cfgMap;
	private String fqMainMethodName;
	private NTASystem uppaalSystem;
	private HashSet<String> channels;
	private String uniqueIdentifier;
	private String[] instructionMnemonics;
	
	public UppaalModelGenerator(String fqMainMethodName, HashMap<String,ControlFlowGraph> cfgMap, String uniqueIdentifier, String[] instructionMnemonics) throws Exception{
		this.cfgMap = cfgMap;
		this.fqMainMethodName = fqMainMethodName;
		this.channels = new HashSet<String>();
		this.uniqueIdentifier = uniqueIdentifier;
		this.instructionMnemonics = instructionMnemonics;
		
		this.uppaalSystem = generateUppaalSystemMainMethod(fqMainMethodName, cfgMap);
		
	
		
	}
	public String getFqMainMethodName() {
		return fqMainMethodName;
	}

	public NTASystem getUppaalSystem() {
		return uppaalSystem;
	}
	
	private NTASystem generateUppaalSystemMainMethod(String fqMainMethodName, HashMap<String, ControlFlowGraph> cfgMap) throws Exception {
		
		NTASystem system = new NTASystem(fqMainMethodName);
		String systemString = "system "; 
		
		//Create templates for all methods
		for(String fqMethodName : cfgMap.keySet()){
			ControlFlowGraph cfg = cfgMap.get(fqMethodName);
					
			MethodTemplate methodTemplate = new MethodTemplate(fqMethodName, cfg, this.uniqueIdentifier);
			system.addTemplate(methodTemplate.getTemplate());
			channels.add("urgent chan " + "invoke_" + this.uniqueIdentifier + "_" + FQNameHelper.getUppaalFriendlyName(fqMethodName) + ";");
			channels.add("urgent chan " + "return_" + this.uniqueIdentifier + "_" + FQNameHelper.getUppaalFriendlyName(fqMethodName) + ";");
			systemString += "\n\t" + methodTemplate.getUppaalName() + ",";
			
		}
		
		//Adding all method channels
		for (String channel : channels){
			system.appendDeclaration(channel);
		}
		
		for (int i = 0; i<this.instructionMnemonics.length; i++){
			if(this.instructionMnemonics[i] != Constants.ILLEGAL_OPCODE){
				system.appendDeclaration("const int " + this.uniqueIdentifier.toUpperCase() + "_" + this.instructionMnemonics[i].toUpperCase() + " = " + i +";");		
			}
		}
				
		system.appendDeclaration("urgent chan " + this.uniqueIdentifier + "_execute;");
		system.appendDeclaration("int " + this.uniqueIdentifier + "_instruction;");
		
		systemString = systemString.substring(0,systemString.length()-1) + ";";
		system.setSystem(systemString);
				
		return system;
	}
	
	public Template getInitialisationTemplate(String fqMainMethodName) {
		Template initialisationTemplate = new Template("initialisation_T", Arrays.asList(""));
		Location idleLocation = new Location("Idle");
		idleLocation.setCommited();
		initialisationTemplate.setInitialLocation(idleLocation);		
		Location execLocation = new Location("Executing");
		initialisationTemplate.addLocation(execLocation);
		
		Location mainDoneLocation = new Location("Main_Done");
		initialisationTemplate.addLocation(mainDoneLocation);
		
		Location endLocation = new Location("Terminate");
		initialisationTemplate.addLocation(endLocation);
		
		Transition beginMainMethodTransition = new Transition(idleLocation, execLocation);
		beginMainMethodTransition.getAttrs().appendUpdate("initialise()");
		String channelString = FQNameHelper.getUppaalFriendlyName(fqMainMethodName);
		beginMainMethodTransition.getAttrs().setSync("invoke_" + this.uniqueIdentifier + "_"+ channelString + "!");
		channels.add("chan " + channelString + ";");
		initialisationTemplate.addTransition(beginMainMethodTransition);
		
		Transition endMainMethodTransition = new Transition(execLocation, mainDoneLocation);
		endMainMethodTransition.getAttrs().setSync("return_" + this.uniqueIdentifier + "_" + channelString + "?");
		channels.add("chan " + channelString + ";");
		initialisationTemplate.addTransition(endMainMethodTransition);
				
		Transition endMetamocTransition = new Transition(mainDoneLocation, endLocation);
		endMetamocTransition.getAttrs().setSync("main_done!");
		initialisationTemplate.addTransition(endMetamocTransition);
		
		return initialisationTemplate;
	}

}
