package com.tetaj.modelgenerator.annotationextractor.javaparser;

import java.io.BufferedReader;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;
import java.util.LinkedList;

import org.junit.Before;
import org.junit.Test;

import com.tetaj.modelgenerator.annotationextractor.parser.Annotation;
import com.tetaj.modelgenerator.annotationextractor.parser.Annotation.AnnotationType;
import com.tetaj.modelgenerator.annotationextractor.parser.Parser;

import static org.junit.Assert.*;


public class JavaParserTest {

	private static final String PATH_TO_TEST_FILE = "./resources/TesterClassForJavaParser.java";
	
	Parser javaParser;
	BufferedReader bufferedReader;
	
	@Before
	public void setUp() {
		this.javaParser = new Parser();		
		FileReader file = null;
		try {
			file = new FileReader(PATH_TO_TEST_FILE);
		} catch (FileNotFoundException e) {
			e.printStackTrace();
		}
		
		bufferedReader = new BufferedReader(file);
		
		try {
			bufferedReader.mark(1);
		} catch (IOException e) {
			e.printStackTrace();
		}
	}
	
	@Test
	public void testJavaparserSingleAnnotation() {
		/*
		 * line number of the first loop
		 */
		int lineNum = 8;
		LinkedList<Annotation> annoList = null;		
		
		try {
			annoList = javaParser.GetAnnotations(bufferedReader, lineNum);
		} catch (IOException e) {
			e.printStackTrace();
		}
		
		assertNotNull(annoList);
		assertTrue(annoList.size() == 1);
		
		Annotation extractedAnnotation = annoList.getFirst();
		
		assertEquals(AnnotationType.LOOPBOUND, extractedAnnotation.getAnnotationType());
		assertEquals("5", extractedAnnotation.getAnnotationValue());
	}
	
	@Test
	public void testJavaparserDoubleAnnotation() {
		/*
		 * line number of the first loop
		 */
		int lineNum = 15;
		LinkedList<Annotation> annoList = null;		
		
		try {
			annoList = javaParser.GetAnnotations(bufferedReader, lineNum);
		} catch (IOException e) {
			e.printStackTrace();
		}
		
		assertNotNull(annoList);
		assertTrue(annoList.size() == 2);
		
		Annotation extractedAnnotation = annoList.pop();
		assertEquals(AnnotationType.LOOPBOUND, extractedAnnotation.getAnnotationType());
		assertEquals("10", extractedAnnotation.getAnnotationValue());
		
		extractedAnnotation = annoList.pop();
		assertEquals(AnnotationType.LOOPBOUND, extractedAnnotation.getAnnotationType());
		assertEquals("4", extractedAnnotation.getAnnotationValue());
	}
	@Test
	public void testJavaparserNoAnnotation() {
		/*
		 * line number of the first loop
		 */
		int lineNum = 21;
		LinkedList<Annotation> annoList = null;		
		
		try {
			annoList = javaParser.GetAnnotations(bufferedReader, lineNum);
		} catch (IOException e) {
			e.printStackTrace();
		}
		
		assertNotNull(annoList);
		assertTrue(annoList.size() == 0);
	}
	@Test
	public void testJavaparserSingleSpaceAnnotation() {
		/*
		 * line number of the first loop
		 */
		int lineNum = 26;
		LinkedList<Annotation> annoList = null;		
		
		try {
			annoList = javaParser.GetAnnotations(bufferedReader, lineNum);
		} catch (IOException e) {
			e.printStackTrace();
		}
		
		assertNotNull(annoList);
		assertTrue(annoList.size() == 1);
		
		Annotation extractedAnnotation = annoList.pop();
		assertEquals(AnnotationType.LOOPBOUND, extractedAnnotation.getAnnotationType());
		assertEquals("2", extractedAnnotation.getAnnotationValue());
	}
	
	
	@Test
	public void testJavaparserSingleAndNoAnnotation() {
		/*
		 * line number of the first loop
		 */
		int lineNum = 33;
		LinkedList<Annotation> annoList = null;		
		
		try {
			annoList = javaParser.GetAnnotations(bufferedReader, lineNum);
		} catch (IOException e) {
			e.printStackTrace();
		}
		
		assertNotNull(annoList);
		assertTrue(annoList.size() == 1);
		
		Annotation extractedAnnotation = annoList.pop();
		assertEquals(AnnotationType.LOOPBOUND, extractedAnnotation.getAnnotationType());
		assertEquals("10", extractedAnnotation.getAnnotationValue());
	}
}
