package com.tetaj.modelgenerator.avrtoircfg;

import static org.junit.Assert.*;

import java.util.HashMap;
import java.util.List;

import org.junit.Before;
import org.junit.Test;

import com.tetaj.cfggenerators.AVRToIRCFG;
import com.tetaj.ircfg.BasicBlock;
import com.tetaj.ircfg.ControlFlowGraph;
import com.tetaj.ircfg.Edge;
import com.tetaj.prettyprinter.IRCFGPrettyPrinter;

public class AVRToIRCFGTests {

	/*
	 * *basicblocknum has been verified manually
	 */
	private static int simpleMainBasicBlocksNum = 6;
	private static int complexMainBasicBlocksNum = 7;
	private static int mainWithLoopBasicBlocksNum = 7;
	private static int mainWithfuncBasicBlocksNumMAINFUNCTION = 4;
	private static int mainWithfuncBasicBlocksNumFUNCFUNCTION = 2;
	
	@Before
	public void setUp() {
		
	}
	
	@Test
	public void testSimpleMain() throws Exception {
		String sourceFileLocation = "./resources/AVRToIRCFG/simplemain.elf";
		AVRToIRCFG generator = new AVRToIRCFG("main",sourceFileLocation);
		generator.generateIRCFG();
	}
	
	@Test
	public void testSimpleMainSizeOfCFG() throws Exception {
		String sourceFileLocation = "./resources/AVRToIRCFG/simplemain.elf";
		AVRToIRCFG generator = new AVRToIRCFG("main",sourceFileLocation);
		HashMap<String, ControlFlowGraph> CFGs = generator.generateIRCFG();
		assertEquals(1, CFGs.size());
	}
	
	@Test
	public void testSimpleMainNumberOfBasicBlocks() throws Exception {
		String sourceFileLocation = "./resources/AVRToIRCFG/simplemain.elf";
		AVRToIRCFG generator = new AVRToIRCFG("main",sourceFileLocation);
		HashMap<String, ControlFlowGraph> CFGs = generator.generateIRCFG();
		assertEquals(1, CFGs.size());
		
		IRCFGPrettyPrinter prettyPrinter = new IRCFGPrettyPrinter("./prettyprintedCfgs/");
		
		
		for(ControlFlowGraph currentCfg : CFGs.values()) {
			prettyPrinter.createVisualRepresentation(currentCfg);
			
			List<BasicBlock> basicBlocks = currentCfg.getBasicBlocks();
			assertEquals(simpleMainBasicBlocksNum, basicBlocks.size());
		}
	}
	
	@Test
	public void testComplexMain() throws Exception {
		String sourceFileLocation = "./resources/AVRToIRCFG/complexmain.elf";
		AVRToIRCFG generator = new AVRToIRCFG("main",sourceFileLocation);
		generator.generateIRCFG();
	}
	
	@Test
	public void testComplexMainSizeOfCFG() throws Exception {
		String sourceFileLocation = "./resources/AVRToIRCFG/complexmain.elf";
		AVRToIRCFG generator = new AVRToIRCFG("main",sourceFileLocation);
		HashMap<String, ControlFlowGraph> CFGs = generator.generateIRCFG();
		assertEquals(1, CFGs.size());
	}
	
	@Test
	public void testComplexMainNumberOfBasicBlocks() throws Exception {
		String sourceFileLocation = "./resources/AVRToIRCFG/complexmain.elf";
		AVRToIRCFG generator = new AVRToIRCFG("main",sourceFileLocation);
		HashMap<String, ControlFlowGraph> CFGs = generator.generateIRCFG();
		assertEquals(1, CFGs.size());
		
		IRCFGPrettyPrinter prettyPrinter = new IRCFGPrettyPrinter("./prettyprintedCfgs/");
		
		
		for(ControlFlowGraph currentCfg : CFGs.values()) {
			prettyPrinter.createVisualRepresentation(currentCfg);
			
			List<BasicBlock> basicBlocks = currentCfg.getBasicBlocks();
			assertEquals(6, basicBlocks.size());
		}
	}
	@Test
	public void testComplexMainBranchBasicBlocks() throws Exception {
		String sourceFileLocation = "./resources/AVRToIRCFG/complexmain.elf";
		AVRToIRCFG generator = new AVRToIRCFG("main",sourceFileLocation);
		HashMap<String, ControlFlowGraph> CFGs = generator.generateIRCFG();
		assertEquals(1, CFGs.size());
		
		IRCFGPrettyPrinter prettyPrinter = new IRCFGPrettyPrinter("./prettyprintedCfgs/");
		
		
		for(ControlFlowGraph currentCfg : CFGs.values()) {
			prettyPrinter.createVisualRepresentation(currentCfg);
			
			List<BasicBlock> basicBlocks = currentCfg.getBasicBlocks();
			for(BasicBlock bb : basicBlocks) {
				if(bb.getId() == 0) {
					List<Edge> branchEdges = bb.getBranchEdges();
					assertEquals(1, branchEdges.size());
					Edge theEdge = branchEdges.get(0);
					assertEquals(5, theEdge.getTarget().getId());
					
					Edge theNextEdge = bb.getFallthroughEdge();
					assertNotNull(theNextEdge);
					assertEquals(1, theNextEdge.getTarget().getId());
					
				}
				if(bb.getId() == 1) {
					List<Edge> branchEdges = bb.getBranchEdges();
					assertEquals(1, branchEdges.size());
					Edge theEdge = branchEdges.get(0);
					assertEquals(4, theEdge.getTarget().getId());
					
					Edge theNextEdge = bb.getFallthroughEdge();
					assertNotNull(theNextEdge);
					assertEquals(2, theNextEdge.getTarget().getId());
				}
				
			}
		}
	}
	
	
	@Test
	public void testMainWithLoop() throws Exception {
		String sourceFileLocation = "./resources/AVRToIRCFG/simplemainwithloop.elf";
		AVRToIRCFG generator = new AVRToIRCFG("main",sourceFileLocation);
		generator.generateIRCFG();
	}
	
	@Test
	public void testMainWithLoopSizeOfCFG() throws Exception {
		String sourceFileLocation = "./resources/AVRToIRCFG/simplemainwithloop.elf";
		AVRToIRCFG generator = new AVRToIRCFG("main",sourceFileLocation);
		HashMap<String, ControlFlowGraph> CFGs = generator.generateIRCFG();
		assertEquals(1, CFGs.size());
	}
	
	@Test
	public void testMainWithLoopNumberOfBasicBlocks() throws Exception {
		String sourceFileLocation = "./resources/AVRToIRCFG/simplemainwithloop.elf";
		AVRToIRCFG generator = new AVRToIRCFG("main",sourceFileLocation);
		HashMap<String, ControlFlowGraph> CFGs = generator.generateIRCFG();
		assertEquals(1, CFGs.size());
		
		IRCFGPrettyPrinter prettyPrinter = new IRCFGPrettyPrinter("./prettyprintedCfgs/");
		
		
		for(ControlFlowGraph currentCfg : CFGs.values()) {
			prettyPrinter.createVisualRepresentation(currentCfg);
			
			List<BasicBlock> basicBlocks = currentCfg.getBasicBlocks();
			assertEquals(mainWithLoopBasicBlocksNum, basicBlocks.size());
		}
	}
	
	@Test
	public void testMainWithfunc() throws Exception {
		String sourceFileLocation = "./resources/AVRToIRCFG/mainwithfunc.elf";
		AVRToIRCFG generator = new AVRToIRCFG("main",sourceFileLocation);
		generator.generateIRCFG();
	}
	
	@Test
	public void testMainWithfuncSizeOfCFG() throws Exception {
		String sourceFileLocation = "./resources/AVRToIRCFG/mainwithfunc.elf";
		AVRToIRCFG generator = new AVRToIRCFG("main",sourceFileLocation);
		HashMap<String, ControlFlowGraph> CFGs = generator.generateIRCFG();
		/*
		 * there are two functions now
		 */
		assertEquals(2, CFGs.size());
	}
	
	@Test
	public void testMainWithmultiplication() throws Exception {
		String sourceFileLocation = "./resources/AVRToIRCFG/mainwithmultiplication.elf";
		AVRToIRCFG generator = new AVRToIRCFG("main",sourceFileLocation);
		HashMap<String, ControlFlowGraph> CFGs = generator.generateIRCFG();
		
		for(ControlFlowGraph currentCfg : CFGs.values()) {
			List<BasicBlock> basicBlocks = currentCfg.getBasicBlocks();
			if(currentCfg.getMethodName().equals("main")) {
				IRCFGPrettyPrinter prettyPrinter = new IRCFGPrettyPrinter("./prettyprintedCfgs/");
				prettyPrinter.createVisualRepresentation(currentCfg);
				assertEquals(mainWithfuncBasicBlocksNumMAINFUNCTION, basicBlocks.size());
			}
		}
		assertEquals(1, CFGs.size());
	}
	
	@Test
	public void testMainWithFuncBasicBlocks() throws Exception {
		String sourceFileLocation = "./resources/AVRToIRCFG/mainwithfunc.elf";
		AVRToIRCFG generator = new AVRToIRCFG("main",sourceFileLocation);
		HashMap<String, ControlFlowGraph> CFGs = generator.generateIRCFG();
		assertEquals(2, CFGs.size());
		
		for(ControlFlowGraph currentCfg : CFGs.values()) {
			List<BasicBlock> basicBlocks = currentCfg.getBasicBlocks();
			for(BasicBlock bb : basicBlocks) {
				assertNotNull(bb);
			}
		}
	}
	
	@Test
	public void testMainWithfuncNumberOfBasicBlocks() throws Exception {
		String sourceFileLocation = "./resources/AVRToIRCFG/mainwithfunc.elf";
		AVRToIRCFG generator = new AVRToIRCFG("main",sourceFileLocation);
		HashMap<String, ControlFlowGraph> CFGs = generator.generateIRCFG();
		assertEquals(2, CFGs.size());
		
		for(ControlFlowGraph currentCfg : CFGs.values()) {
			List<BasicBlock> basicBlocks = currentCfg.getBasicBlocks();
			if(currentCfg.getMethodName().equals("main")) {
				IRCFGPrettyPrinter prettyPrinter = new IRCFGPrettyPrinter("./prettyprintedCfgs/");
				prettyPrinter.createVisualRepresentation(currentCfg);
				assertEquals(mainWithfuncBasicBlocksNumMAINFUNCTION, basicBlocks.size());
			} else {
				IRCFGPrettyPrinter prettyPrinter = new IRCFGPrettyPrinter("./prettyprintedCfgs/");
				prettyPrinter.createVisualRepresentation(currentCfg);
				assertEquals(mainWithfuncBasicBlocksNumFUNCFUNCTION, basicBlocks.size());
			}
		}
	}
	
	@Test
	public void testMainWithfuncMethodInvokeBlock() throws Exception {
		String sourceFileLocation = "./resources/AVRToIRCFG/mainwithfunc.elf";
		AVRToIRCFG generator = new AVRToIRCFG("main",sourceFileLocation);
		HashMap<String, ControlFlowGraph> CFGs = generator.generateIRCFG();
		assertEquals(2, CFGs.size());
		
		for(ControlFlowGraph currentCfg : CFGs.values()) {
			if(currentCfg.getMethodName().equals("main")) {
				List<BasicBlock> basicBlocks = currentCfg.getBasicBlocks();
				for(BasicBlock bb : basicBlocks) {
					if(bb.getId() == 2) {
						assertEquals(true, bb.isMethodInvoke());
						assertNotNull(bb.getMethodInvokeCfg());
						assertEquals("func", bb.getMethodInvokeCfg().get(bb.getInstructions().size()).getMethodName());
						
					}
				}
			}
		}
	}
	
	@Test
	public void testMainWithexternalfunc() throws Exception {
		String sourceFileLocation = "./resources/AVRToIRCFG/mainwithexternalfunc.elf";
		AVRToIRCFG generator = new AVRToIRCFG("main",sourceFileLocation);
		HashMap<String, ControlFlowGraph> CFGs = generator.generateIRCFG();
		assertEquals(3, CFGs.size());
		
		for(ControlFlowGraph currentCfg : CFGs.values()) {
			IRCFGPrettyPrinter prettyPrinter = new IRCFGPrettyPrinter("./prettyprintedCfgs/");
			prettyPrinter.createVisualRepresentation(currentCfg);
			//List<BasicBlock> basicBlocks = currentCfg.getBasicBlocks();
			
			/*if(currentCfg.getMethodName().equals("main")) {
				IRCFGPrettyPrinter prettyPrinter = new IRCFGPrettyPrinter("./prettyprintedfiles/mainwithexternalfunc.dot");
				prettyPrinter.createVisualRepresentation(currentCfg);
				assertEquals(6, basicBlocks.size());
			}
			else if(currentCfg.getMethodName().equals("externalfunc")) {
				IRCFGPrettyPrinter prettyPrinter = new IRCFGPrettyPrinter("./prettyprintedfiles/externalfunc.dot");
				prettyPrinter.createVisualRepresentation(currentCfg);
				assertEquals(0, basicBlocks.size());
			}*/
		}
	}
	@Test
	public void testFinalTest() throws Exception {
		String sourceFileLocation = "./resources/AVRToIRCFG/finaltest/finaltest.elf";
		AVRToIRCFG generator = new AVRToIRCFG("main",sourceFileLocation);
		HashMap<String, ControlFlowGraph> CFGs = generator.generateIRCFG();
		assertEquals(6, CFGs.size());
		
		for(ControlFlowGraph tempcfg : CFGs.values()) {
			System.out.println("NEW CFG: " + tempcfg.getMethodName());
			for(BasicBlock bb : tempcfg.getBasicBlocks()) {
				System.out.println("basicblock id: " + bb.getId());
			}
			System.out.println("Finished CFG: " + tempcfg.getMethodName());
		}
		
		
		for(ControlFlowGraph currentCfg : CFGs.values()) {
			IRCFGPrettyPrinter prettyPrinter = new IRCFGPrettyPrinter("./prettyprintedCfgs/" + currentCfg.getMethodName() + ".dot");
			prettyPrinter.createVisualRepresentation(currentCfg);
			System.out.println("HER");
			//List<BasicBlock> basicBlocks = currentCfg.getBasicBlocks();
			
			/*if(currentCfg.getMethodName().equals("main")) {
				IRCFGPrettyPrinter prettyPrinter = new IRCFGPrettyPrinter("./prettyprintedfiles/mainwithexternalfunc.dot");
				prettyPrinter.createVisualRepresentation(currentCfg);
				assertEquals(6, basicBlocks.size());
			}
			else if(currentCfg.getMethodName().equals("externalfunc")) {
				IRCFGPrettyPrinter prettyPrinter = new IRCFGPrettyPrinter("./prettyprintedfiles/externalfunc.dot");
				prettyPrinter.createVisualRepresentation(currentCfg);
				assertEquals(0, basicBlocks.size());
			}*/
		}
	}
	
	/*@Test
	public void testHVMGeneration() throws Exception {
		String sourceFileLocation = "./resources/AVRToIRCFG/hvmModified.elf";
		AVRToIRCFG generator = new AVRToIRCFG("main",sourceFileLocation);
		HashMap<String, ControlFlowGraph> CFGs = generator.generateIRCFG();
		//assertEquals(6, CFGs.size());
		
		for(ControlFlowGraph tempcfg : CFGs.values()) {
			System.out.println("NEW CFG: " + tempcfg.getMethodName());
			for(BasicBlock bb : tempcfg.getBasicBlocks()) {
				System.out.println("basicblock id: " + bb.getId());
			}
			System.out.println("Finished CFG: " + tempcfg.getMethodName());
		}
		
		
		for(ControlFlowGraph currentCfg : CFGs.values()) {
			IRCFGPrettyPrinter prettyPrinter = new IRCFGPrettyPrinter("./prettyprintedfiles/" + currentCfg.getMethodName() + ".dot");
			prettyPrinter.createVisualRepresentation(currentCfg);
			
		}
		
	}*/
	
	
	
}
