package com.tetaj.modelgenerator.cfggenerator;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertNotNull;

import org.junit.Test;

import com.tetaj.cfggenerators.jophelper.CFGGenerator;

import com.jopdesign.libgraph.cfg.ControlFlowGraph;
import com.jopdesign.libgraph.cfg.GraphException;
import com.jopdesign.libgraph.struct.TypeException;

public class CFGGeneratorTest {

	@Test (expected=TypeException.class)
	public void testClasspath() throws GraphException, TypeException {
		
		CFGGenerator cfgGenerator = new CFGGenerator("org.outsideclasspath.SomeFile#work()I", ".");
		
		cfgGenerator.getCfgMap();
	}
	
	@Test
	public void testClasspathCorrect() throws GraphException, TypeException {
		
		CFGGenerator cfgGenerator = new CFGGenerator("org.outsideclasspath.SomeFile#work()I", "/tmp/something/:.:./resources/outsideclasspath/");
		
		cfgGenerator.getCfgMap();
	}
	
	@Test
	public void testSingleGraphCFGGenerator() throws GraphException, TypeException {
				
		CFGGenerator cfgGenerator = new CFGGenerator("org.testapp.MethodCall#testSingleMethod(II)I", ".");
		
		//Only one graph is built
		assertEquals(cfgGenerator.getCfgMap().size(), 1);
		
		ControlFlowGraph cfg = cfgGenerator.getCfgMap().get("org.testapp.MethodCall#testSingleMethod(II)I");
		assertNotNull(cfg);
		assertEquals(cfg.getBlockCount(), 3);
		
		assertEquals(cfg.getSignature().toString(),"testSingleMethod(II)I");
	}
	
	@Test
	public void testNestedMethodCFGGenerator() throws GraphException, TypeException {
		CFGGenerator cfgGenerator = new CFGGenerator("org.testapp.MethodCall#testNestedMethod()I", ".");
		
		System.out.println("HashMap Size: " + cfgGenerator.getCfgMap().size());

		ControlFlowGraph cfg = cfgGenerator.getCfgMap().get("org.testapp.MethodCall#testNestedMethod()I");
		assertEquals(cfg.getBlockCount(), 1);

		
	}
	
	@Test
	public void testIncrementerCFGGenerator() throws GraphException, TypeException {
		CFGGenerator cfgGenerator = new CFGGenerator("org.testapp.Incrementer#increment(I)I", ".");
		
		System.out.println("increment()");
		ControlFlowGraph cfg1 = cfgGenerator.getCfgMap().get("org.testapp.Incrementer#increment(I)I");
		System.out.println("addOne()");
		ControlFlowGraph cfg2 = cfgGenerator.getCfgMap().get("org.testapp.Incrementer#addOne(I)I");
		

		
	}
}
