package com.tetaj.modelgenerator.cfggenerator;

import java.util.HashMap;
import java.util.LinkedList;

import junit.framework.Assert;

import org.junit.Test;

import com.tetaj.ircfg.Edge;
import com.tetaj.cfggenerators.AVRToIRCFG;
import com.tetaj.hvmhelper.HvmCfgHelper;
import com.tetaj.ircfg.BasicBlock;
import com.tetaj.ircfg.ControlFlowGraph;
import com.tetaj.modelgenerator.cfganalysis.CfgAnalysis;
import com.tetaj.prettyprinter.AppConfig;
import com.tetaj.prettyprinter.IRCFGPrettyPrinter;


public class HVMRelatedTests {
	
	@Test
	public void testSplitProccessMethod() throws Exception {
		
		AppConfig.appendSourcePaths("/home/semadk/src/eclipse/ModelGenerator/resources/simpleapp/hvm/src");
		
		AVRToIRCFG avrGenerator = new AVRToIRCFG("process_jbc", "/home/semadk/src/eclipse/ModelGenerator/resources/simpleapp/hvm/helloWCET-C-AVR.elf");
		HashMap<String, ControlFlowGraph> cfgs = avrGenerator.generateIRCFG();	
		
		CfgAnalysis analysis = new CfgAnalysis(cfgs);
		analysis.conductCfgAnalyses();
		
		ControlFlowGraph cfg = cfgs.get("process_jbc");
		
		IRCFGPrettyPrinter pp = new IRCFGPrettyPrinter("/home/semadk/Desktop/combined.dot");
		pp.createVisualRepresentation(cfg);
		
		int count = cfgs.size();
		int basicBlockCount = cfg.getBasicBlocks().size();
		
		HvmCfgHelper.handleInterpretationLoopCfg(cfgs);
		
		Assert.assertEquals(count + 1, cfgs.size());
		
		ControlFlowGraph preCfg = cfgs.get("process_jbc__pre__");
		ControlFlowGraph postCfg = cfgs.get("process_jbc__post__");
		
		Assert.assertNotNull(preCfg);
		Assert.assertNotNull(postCfg);
		
		pp = new IRCFGPrettyPrinter("/home/semadk/Desktop/pre.dot");
		pp.createVisualRepresentation(preCfg);
		
		pp = new IRCFGPrettyPrinter("/home/semadk/Desktop/post.dot");
		pp.createVisualRepresentation(postCfg);
		
		Assert.assertEquals(6, preCfg.getBasicBlocks().size() + postCfg.getBasicBlocks().size());
		
		Assert.assertEquals(0, preCfg.getBasicBlock(0).getId());
		
		LinkedList<BasicBlock> foundBasicBlocks = new LinkedList<BasicBlock>();
		LinkedList<BasicBlock> queue = new LinkedList<BasicBlock>();

		queue.push(preCfg.getEntryBlock());
		
		while(queue.size() != 0) {
			BasicBlock current = queue.pop();
			foundBasicBlocks.push(current);
			
			for (Edge edge : current.getOutgoingEdges()) {
				if (!foundBasicBlocks.contains(edge.getTarget())) {
					foundBasicBlocks.push(edge.getTarget());
					queue.push(edge.getTarget());
				}
			}
		}
		
		Assert.assertEquals(preCfg.getBasicBlocks().size(), foundBasicBlocks.size());
		
	}
	

	
}
